/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
/*
 CFDL module skel_drv
*/
#include "fapex.h"
#include "fapex_cfdl.h"
#include "fapex_cfdl_priv.h"
#include "fapex_common_priv.h"
#include "fapex/sys_config.h"
#include "fapex_config_cmn.h"

/*---------------------------------------------------------------------------*/
/* This is the root of the driver context.                                   */
/*---------------------------------------------------------------------------*/
static FAPEXCFDL_StrT *pFAPEXCFDLHdl;

#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXCFDL_StrT FAPIGLD_CfdlStr;
#ifndef __LINUX__
static struct file FAPIGLD_Chrdev[FAPEX_DESCRIPTOR_NUM+1];
#else /* !__LINUX__ */
static struct cdev FAPIGLD_Chrdev[FAPEX_DESCRIPTOR_NUM+1];
#endif /* !__LINUX__ */
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

/* define this name's global data as a system abstraction definitions */
extern FAPEXCFDLSAF_ServicesT FAPEXCFDLSAF_Services_EXT;

/*!
*******************************************************************************
**
** \brief  Get a CFDL context base address.
**
** \param  void.
**
******************************************************************************/
void* FAPEXCFDL_GetCfdlHdl(void)
{
    return (void*)(pFAPEXCFDLHdl);
}

/*!
*******************************************************************************
**
** \brief  Setup character device. (CFDL internal function)
**
** \param  *pCfdlHdl   The pointer to the CFDL context base address.
** \param  *fops       The pointer to the file_operations of CFDL module.
** \param  id          device identifier.
**
******************************************************************************/
static int32_t _fapiCfdl_SetupChrdev(FAPEXCFDL_StrT *pCfdlHdl, struct file_operations *fops,
                                   void *ptr, int32_t id)
{
#ifndef __LINUX__

    /* Allocate a FDL device's internal memory area */
    pCfdlHdl->pOsPrivate[id] = ptr;
//  FAPEXCFDLSAF_CLEAR_MEMORY(pCfdlHdl->pOsPrivate[id],sizeof(struct file));

    return 0;
#else /* !__LINUX__ */
    dev_t   dev;
    int32_t ret;
    struct  cdev *pcdev;
    char    devname[32];

    /* transfer from (major_No,minor_No) to device information */
    dev = MKDEV(FAPEX_MAJOR_DEFAULT, id);

    /* register a device */
    sprintf(devname, "fapi%d", id);
    ret = register_chrdev_region(dev, 1, devname);
    if(ret < 0){
        FAPEXDebugPrint(2, "cannot register: register_chrdev_region %s(%d,%d)\n",
                           "FAPEXFDL", FAPEX_MAJOR_DEFAULT,id);
        return ret;
    }

    /* Allocate an FDL device's internal memory area */
    pcdev = (struct cdev*)ptr;
//  FAPEXCFDLSAF_CLEAR_MEMORY(pcdev,sizeof(struct cdev));

    /* Setup an internal */
    pCfdlHdl->pOsPrivate[id] = (void*)pcdev;

    /* initialize character device (recognize to the kernel) */
    cdev_init(pcdev, fops);

    /* setup parameter and file operations */
#ifdef BUILD_AS_MODULE
    pcdev->owner = THIS_MODULE;
#endif /* BUILD_AS_MODULE */
    pcdev->ops = fops;

    /* recognize to the kernel */
    ret = cdev_add(pcdev, dev, 1);
    if(ret){
        pCfdlHdl->pOsPrivate[id] = NULL;
        FAPEXDebugPrint(2, "cannot register: cdev_add %d(%d,%d)\n",
                           ret, FAPEX_MAJOR_DEFAULT, id);
        return ret;
    }

    return 0;
#endif /* !__LINUX__ */
}

/*!
*******************************************************************************
**
** \brief  Release character device. (CFDL internal function)
**
** \param  id          device identifier.
**
******************************************************************************/
static int32_t _fapiCfdl_ReleaseChrdev(int32_t id)
{
#ifndef __LINUX__
    FAPEXCFDL_StrT *pCfdlHdl;

    /* Get the base address of the CFDL context. */
    pCfdlHdl = (FAPEXCFDL_StrT *)FAPEXCFDL_GetCfdlHdl();
    pCfdlHdl->pOsPrivate[id] = NULL;

    return 0;
#else /* !__LINUX__ */

    FAPEXCFDL_StrT *pCfdlHdl;
    dev_t dev;

    /* remove a device */
    /* transfer from (major_No,minor_No) to device information */
    dev = MKDEV(FAPEX_MAJOR_DEFAULT, id);

    /* unregister a device */
    unregister_chrdev_region(dev,1);

    /* Get the base address of the CFDL context. */
    pCfdlHdl = (FAPEXCFDL_StrT *)FAPEXCFDL_GetCfdlHdl();
    pCfdlHdl->pOsPrivate[id] = NULL;

    return 0;
#endif /* !__LINUX__ */
}

/*!
*******************************************************************************
**
** \brief  return the size of character device. (CFDL internal function)
**
**
******************************************************************************/
int32_t _fapiCfdl_GetChrdevStrSize(void)
{
#ifndef __LINUX__
    return(sizeof(struct file));
#else /* !__LINUX__ */
    return(sizeof(struct cdev));
#endif /* !__LINUX__ */
}

/*!
*******************************************************************************
**
** \brief  Register an FDL.
**
** \param  *fops       The pointer to the file_operations of an FDL module.
** \param  size        A memory size which is allocated by this function. 
**                     Per one device, only include FDL (i.e. exclude HDL).
** \param  *pUtility   The pointer to the CFDL utility structure.
** \param  id          device identifier.
**
******************************************************************************/
int32_t FAPEXCFDL_RegisterFdl(struct file_operations *fops, void *pFdlCtx,
                              FAPEXCFDL_UtilStrT *pUtility, int32_t id)
{
    FAPEXCFDL_StrT *pCfdlHdl;
    int32_t ret;
    uint8_t *ptr;

    /* Get the base address of the CFDL context. */
    pCfdlHdl = FAPEXCFDL_GetCfdlHdl();

    if(pCfdlHdl == NULL) {
        FAPEXDebugPrint(2, "[CFDL/FAPEXCFDL_RegisterFdl] error\n");
        return -1;
    }

    /* Setup a character device */
    ptr = (uint8_t*)pFAPEXCFDLHdl->pCharDevBase;
    ret = _fapiCfdl_SetupChrdev(pCfdlHdl, fops, (void*)(ptr+(_fapiCfdl_GetChrdevStrSize()*id)), id);
    if(ret){
        return ret;
    }

    /* Register a FDL device to the CFDL database. */
    /* operations */
    pCfdlHdl->fop[id] = fops;
    /* function context */
    pCfdlHdl->pFdlContext[id] = pFdlCtx;

    FAPEXDebugPrint(4, "[CFDL/FAPEXCFDL_RegisterFdl] func (id:%d) 0x%08x registered\n",
                       id, (uint32_t)pFdlCtx);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  Unregister an FDL.
**
** \param  id          device identifier.
**
******************************************************************************/
int32_t FAPEXCFDL_UnregisterFdl(int32_t id)
{
    FAPEXCFDL_StrT *pCfdlHdl;
    int32_t ret;

    /* Get the base address of the CFDL context. */
    pCfdlHdl = FAPEXCFDL_GetCfdlHdl();

    if(pCfdlHdl == NULL) {
        FAPEXDebugPrint(2, "[CFDL/FAPEXCFDL_UnregisterFdl] error\n");
        return -1;
    }

    ret = _fapiCfdl_ReleaseChrdev(id);
    if(ret){
        return ret;
    }
    pCfdlHdl->pFdlContext[id] = NULL;

    FAPEXDebugPrint(4, "[CFDL/FAPEXCFDL_UnregisterFdl] func (id:%d) unregistered\n",id);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  Get a base address of a FDL context.
**
** \param  *ptr  The pointer to the CFDL context base.
** \param  id    Device identifier.
**
******************************************************************************/
void* FAPEXCFDL_GetFdlHdl(void *ptr, int32_t id)
{
    if(id < 0) {
        return NULL;
    }

    return (void*)(((FAPEXCFDL_StrT*)(ptr))->pFdlContext[id]);
}

/*!
*******************************************************************************
**
** \brief  Open the CFDL.
**         This function is called with the systemcall "open". And the CFDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXCFDL_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXDebugPrint(4, "[CFDL/FAPEXCFDL_open] CFDL was opened. 0x%08x\n",
                       (uint32_t)FAPEXCFDL_GetCfdlHdl());
    return 0;
}

/*!
*******************************************************************************
**
** \brief  Release the CFDL.
**         This function is called with the systemcall "close". And the CFDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
int32_t FAPEXCFDL_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXDebugPrint(4, "[CFDL/FAPEXCFDL_release]\n");

    return 0;
}

/*!
*******************************************************************************
**
** \brief  General control of the CFDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
int32_t FAPEXCFDL_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                        FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXDebugPrint(4, "[CFDL/FAPEXCFDL_ioctl]\n");

    switch(cmd){
        default:
        break;
    }

    return 0;
}

/*--------------------------------------------------------*/
/* The file operations of the CFDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_CFDL_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the CFDL module.
**         This function is called when the CFDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXCFDL_init(void)
{
    int32_t  ret;
    uint8_t  *ptr;
    uint32_t chrdevStr_size;

    FAPEXDebugPrint(1, "[CFDL/FAPEXCFDL_init]\n");

    /* initialize of FAPIcore module */
    ret = FAPEXCFG_FAPI_Init();
    if(ret) {
        return ret;
    }

    /* set struct file_operations */
    fapex_CFDL_fops.owner   = THIS_MODULE;
    fapex_CFDL_fops.open    = FAPEXCFDL_open;
    fapex_CFDL_fops.ioctl   = FAPEXCFDL_ioctl;
    fapex_CFDL_fops.release = FAPEXCFDL_release;

    /* set SAF */
    FAPEXCFDLSAF_SetServices(&FAPEXCFDLSAF_Services_EXT);

    /* allocate CFDL context */
    chrdevStr_size = _fapiCfdl_GetChrdevStrSize()*FAPEX_DESCRIPTOR_NUM;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    pFAPEXCFDLHdl = (FAPEXCFDL_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXCFDL_StrT));
    FAPEXCFDLSAF_CLEAR_MEMORY(pFAPEXCFDLHdl,sizeof(FAPEXCFDL_StrT));
    pFAPEXCFDLHdl->pCharDevBase = (void*)FAPI_SYS_MALLOC(chrdevStr_size);
#else /* ALLOCATE_DYNAMIC_MEMORY */
    pFAPEXCFDLHdl = (FAPEXCFDL_StrT*)(&FAPIGLD_CfdlStr);
    FAPEXCFDLSAF_CLEAR_MEMORY(pFAPEXCFDLHdl,sizeof(FAPEXCFDL_StrT));
    pFAPEXCFDLHdl->pCharDevBase = (void*)(&(FAPIGLD_Chrdev[0]));
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pFAPEXCFDLHdl->pCharDevBase,chrdevStr_size);

    ptr = (uint8_t*)(pFAPEXCFDLHdl->pCharDevBase);

    ret = _fapiCfdl_SetupChrdev(pFAPEXCFDLHdl, &fapex_CFDL_fops,
                                (void*)(ptr+(_fapiCfdl_GetChrdevStrSize()*FAPEX_CFDL_MINOR)),
                                FAPEX_CFDL_MINOR);
    if(ret) {
        return ret;
    }
    FAPEXDebugPrint(1, "[CFDL/FAPEXCFDL_init] CFDL was loaded.\n");
    return 0;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the CFDL module.
**         This function is called when the CFDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXCFDL_cleanup(void)
{
    FAPEXCFDL_StrT *pCfdlHdl;

    FAPEXDebugPrint(1, "[CFDL/FAPEXCFDL_cleanup]\n");

    pCfdlHdl = FAPEXCFDL_GetCfdlHdl();

    /* CFDL */
    if(pCfdlHdl!=NULL){
        _fapiCfdl_ReleaseChrdev(FAPEX_CFDL_MINOR);

#ifdef ALLOCATE_DYNAMIC_MEMORY
        if(pCfdlHdl->pCharDevBase!=NULL){
            FAPI_SYS_FREE(pCfdlHdl->pCharDevBase);
        }
        FAPI_SYS_FREE(pCfdlHdl);
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    }

    /* cleanup of FAPIcore module */
    FAPEXCFG_FAPI_Cleanup();

    FAPEXDebugPrint(1, "[CFDL/FAPEXCFDL_init] CFDL was removed.\n");
    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXCFDL_init);
module_exit(FAPEXCFDL_cleanup);

EXPORT_SYMBOL(FAPEXCFDL_GetCfdlHdl);
EXPORT_SYMBOL(FAPEXCFDL_GetFdlHdl);
EXPORT_SYMBOL(FAPEXCFDL_RegisterFdl);
EXPORT_SYMBOL(FAPEXCFDL_UnregisterFdl);

MODULE_LICENSE("GPL");
//MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (CFDL object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
