/*!
********************************************************************************
**
** \file        ./boards/fme/mb86h61_dk01/mb86h61_dk01.h
**
** \version     $Id: mb86h61_dk01.h 2579 2010-07-01 15:13:55Z rseng $
**
** \brief       header-file for 'MB86H61 Dev.Kit' install function
**
** This file contains the definitions for the 'install function'
** which will then register all board-contro functions into the BIC-driver
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU SEMICONDUCTOR
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2006-2010 by Fujitsu Semiconductor Europe GmbH
** (C) Copyright 2010 by Fujitsu Semiconductor Europe GmbH
**
********************************************************************************
*/
#ifndef FME_MB86H61_DK01
#define FME_MB86H61_DK01

#ifdef __cplusplus
extern "C" {
#endif

/*
   NOTE1: The 'application/init' should include the relevant 'board-header-file'
   preferabley using 'compiler command-line-define ' like
   -D'BOARD_INCLUDE_FILE="boards/mb86h61_dk01.h"'

   This define is actually setup via the 'MAKE ENVIRONMENT' taking the information from the
   definition of 'BOARDNAME' in the 'config.mak.xxx'-file

   So, then the 'application/init' could contain the following lines:
   #include "boards/boards_common.h"   // general / common defines
   #include BOARD_INCLUDE_FILE         // should be defined via 'Config.mak.user' and 'the MAKE-system'

   and call the 'install-function' using the macro 'BOARDS_INSTALL_PLATFORM_SUPPORT', e.g.
   BOARDS_INSTALL_PLATFORM_SUPPORT(BOARDS_INIT_DEFAULT, 0);

   then changing the 'HW-platform/board' requires only to change the
   header-file specification in 'config.mak.xxx' and to re-compile the 'application/init'

   NOTE: even the 'boards-drivers' are stored in 'dedicaded directories'
         like '..\boards\fme_devkit'  or '..\boards\CustomerSpecific'
         when doing 'MAKE INSTALL', the board-header files like '..\boards\fme_devkit\mb86h61_dk01.h'
         are copied into '..\install\MB86H61\include\boards'
         so the 'dedicaded directory-name' like 'fme_devkit' is not relevant/used !
*/

#define USE_N_SHIFT_CLOCKS 32   // NOTE: we assmue 'ext.32 shift register' (on-board is's just 8-bit)

//-- defines for 'shift-register output-controls' --
#define GPO_SFTRG_CI_IN_ENA_BIT     BIT0
#define GPO_SFTRG_CI_IN_MUX_BIT     BIT1
#define GPO_SFTRG_CI_RESET_BIT      BIT2
#define GPO_SFTRG_SCART_PIN8_12V    BIT3
#define GPO_SFTRG_SCART_PIN8_5V     BIT4
#define GPO_SFTRG_SCART_PIN16       BIT5
#define GPO_SFTRG_USB_CBU0          BIT6
#define GPO_SFTRG_USB_CBU1          BIT7
//-- defines for 'GPIO-pins' --
#define GPIO_OP_FE_RES_A 12
#define GPIO_OP_FE_RES_B 26
#define GPIO_INP_CI_INS0 59
#define GPIO_INP_CI_INS1 73
#define GPIO_UPI_IOWR       60
#define GPIO_UPI_IORD       61
#define GPIO_UPI_BUF_DIR    62
#define GPIO_UPI_CI_WAIT    63
#define GPIO_UPI_CLE        71
#define GPIO_INP_CI_IRQ     72

int32_t BOARD_FME_H61DK01_BoardIfc_Register(uint32_t registerFlags, void *optRegisterDataPtr);
int32_t BOARD_FME_H61DK01_BoardRegister_Mini(uint32_t registerFlags, void *optRegisterDataPtr);
void    BOARD_FME_H61DK01_BoardIfc_Register_CiSlave(void);
//int32_t BOARD_FME_H61DK01_BoardIfc_Regi_DRT(uint32_t registerFlags, void *optRegisterDataPtr);

void    BOARD_FME_H61DK01_BoardIfc_SetAddInitHook(BOARDS_AddInitHookT addInitCbf);
void    BOARD_FME_H61DK01_BoardIfc_InstallSpecialHwTestFuncs(uint32_t flags);
int32_t BOARD_FME_H61DK01_RegisterFrontends(uint32_t registerFlags, void *optInstallDataPtr);
void    BOARD_FME_H61DK01_FrontPanel_Register(void);

#define BOARDS_MAGIC_GPIO_CTRL_DATA 0x12344321
typedef void (*BOARDS_GpioEventCallbackT)(uint32_t value);
typedef struct {
    uint32_t             MagicBoardsGpioCtrlData;
    BOARDS_GpioEventCallbackT  GpioEventCbf;
} BOARDS_GPIO_ControlDataT;

int32_t BOARD_FME_H61DK01_CntrlUserGpios
                 ( int32_t index, uint32_t CtrlCode, void* optPtr );

extern BOARDS_CommonSupportFunctionsT  *FME_H61TB01_CbsfPtr;   //

/*!
** \brief in case a 'further customer specific or extended install-function' exists
** \sa #BOARDS_INSTALL_PLATFORM_SUPPORT_EXT
*/
int32_t BOARD_FME_H61DK01_RegisterBoardIfc_CstExt(uint32_t registerFlags, void *optInstallDataPtr);

void    BOARD_FME_H61DK01_GetTestGpioInput(void *optPtr);
void    BOARD_FME_H61DK01_RegisterGpioPins(BOARDS_FMEDEF_REG_INI_PIN_CBF_T useThisRegistFunction);
void    BOARD_FME_H61DK01_TsAInpSelect(BOARDS_FMEDEF_REG_INI_PIN_CBF_T useThisRegistFunction, uint32_t inpSelFlag);
void    BOARD_FME_H61DK01_ConfigGpiosPAtaAdapter(BOARDS_FMEDEF_REG_INI_PIN_CBF_T useThisRegistFunction);
int32_t BOARD_FME_H61DK01_SetupCiLibHwInterfaceFunctions(void);
int32_t BOARD_FME_H61DK01_GetTsPinsInputLevel(BOARDS_FMEDEF_RD_GPIO_INPREG_CBF  RdGpioInpFnc,
                                              int32_t tsPathIndex,
                                              uint16_t *dstTsInpLevelsPtr);

#define BOARDS_INSTALL_MIN_PLATFORM_SUPPORT  BOARD_FME_H61DK01_BoardRegister_Mini           // special purpose
#define BOARDS_INSTALL_PLATFORM_SUPPORT      BOARD_FME_H61DK01_BoardIfc_Register            // normal case
#define BOARDS_INSTALL_PLATFORM_SUPPORT_CISL BOARD_FME_H61DK01_BoardIfc_Register_CiSlave    // for special test-program
#define BOARDS_INSTALL_PLATFORM_SUPPORT_EXT  BOARD_FME_H61DK01_BoardIfc_Register_CstExt     // special customer-purpose

#define BOARDS_INSTALL_FRONTEND_SUPPORT     BOARD_FME_H61DK01_RegisterFrontends        //!< register supported frontends

#ifdef __cplusplus
}
#endif

/*
********************************************************************************
** End of file
********************************************************************************
*/
#endif
