/*!
********************************************************************************
**
** \file        ./boards/fme/mb86h61_dk01/mb86h61_dk01_ci.c
**
** \version     $Id: mb86h61_dk01_ci.c 2579 2010-07-01 15:13:55Z rseng $
**
** \brief       FME devkit CI-HW specific functions.
**
** This file contains the implementation of the FME devkit specific board
** services functions.
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU SEMICONDUCTOR
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2006-2010 by Fujitsu Semiconductor Europe GmbH
** (C) Copyright 2010 by Fujitsu Semiconductor Europe GmbH
**
********************************************************************************
*/

#define BOARDS_FMEDEF_DIAG_INFO__MACRO_ENABLED

#include <fapi/reg_gpio.h>
#include <fapi/drv_gpio.h>
#include <fapi/board_ifc.h>

#include <boards/boards_common.h>
#include <boards/mb86h61_dk01.h>

#define FBD_ID   BOARDS_FMEDEF_DIAG_INFO_DATA     // simple local re-define to save code-spase

#define BOARD_FME_H61_TB_NUM_CI_SLOTS 1
#define BOARD_FME_H61_TB_CI_FLAGS (BOARDS_FMEDEF_CI_UPI_CS1_IS_USED | BOARDS_FMEDEF_CI_ALLOW_WAIT_SIGNAL)


//#define MEMADDR_CS2_BASE_ADR (0x90000000)    // CS2 Base-Address
#define MEMADDR_CS1_BASE_ADR (0x88000000)    // CS1 Base-Address
//#define MEMADDR_CS0_BASE_ADR (0x80000000)    // CS0 Base-Address

#define CI_SLOT1_CS0_OFFSET (MEMADDR_CS1_BASE_ADR+0x00000000)
#define CI_SLOT1_CS1_OFFSET (MEMADDR_CS1_BASE_ADR+0x00020000)
#define CI_SLOT2_CS0_OFFSET (MEMADDR_CS1_BASE_ADR+0x00040000)
#define CI_SLOT2_CS1_OFFSET (MEMADDR_CS1_BASE_ADR+0x00060000)
#define CI_CTRL_DEV_OFFSET  (MEMADDR_CS1_BASE_ADR+0x00080000)

/*lint -save -e923 -e778 */

//--------------------------------------------------------------------
//-- Info structures on TS-Path structure                           --
//--------------------------------------------------------------------
static const char_t STR_H61TbCiLogic[] = "H61DK01-CiLogic";

static BOARDS_FMEDEF_OneTsPathStructInfoT  AllPathsConfigData[] = {
// MagicSngTsPtInfo,      tsPathInfoStr,
 { BOARDS_SNG_PATH_MAGIC, BRDCOM_BC_STR_TsApath,
//**pMTIM_FeInfos, nMTIM_Inputs, nCIIM_Inputs, nTBPM_Inputs, NotUsedYet
   NULL,            1,            1,            2,            0         }
};

static BOARDS_FMEDEF_AllTsPathsStructureInfoT  BOARD_FME_H61_TB01_AllTsPathInfos = {
//MagicAllTsPathInfos,     tsPathStructInfoStrPtr, allSinglePathMuxInfosPtr, nTsPathInfosAvailable
 BOARDS_TS_PATHMUX_INFO_MAGIC, STR_H61TbCiLogic,   AllPathsConfigData,   BOARD_FME_H61_TB_NUM_CI_SLOTS };

//--------------------------------------------------------------------
//-- Info structures on TS-Path structure                           --
//--------------------------------------------------------------------
static uint32_t  H61_Tb01_CiMemOfsInfos[] = {
  CI_SLOT1_CS0_OFFSET, CI_SLOT1_CS1_OFFSET,         // for 'Slot-0'
  CI_SLOT2_CS0_OFFSET, CI_SLOT2_CS1_OFFSET,         //     'Slot-1' does not exist !!!
  CI_CTRL_DEV_OFFSET                                // optional CTRL-Reg.Latch does not exist
};

// structure with some 'how to access' CI-HW information
static BOARDS_FMEDEF_CiAccessConfigInfoT  H61_Tb01_CiHwAccessInfo = {
    BOARDS_FMEDEF_CI_ACC_INFO_MAGIC,
    BOARD_FME_H61_TB_CI_FLAGS,
    H61_Tb01_CiMemOfsInfos,
    0,
    0
};

static BOARDS_FMEDEF_CiHwAccessInterfaceT  H61_TB01_CiHwInterfaceFunctions;

#ifdef BOARDS_FMEDEF_DIAG_INFO__MACRO_ENABLED
  static const char_t DEBSTR_CiHwCtrl[] = "Ci HW-CTRL ";
  static const char_t DEBSTR_CiPowerON[]   = "Power ON  {on this HW, not controlled by SW}";
  static const char_t DEBSTR_CiPowerOFF[]  = "Power OFF {on this HW, not controlled by SW}";
  static const char_t DEBSTR_CiResetAct[]  = "Reset ACTIVE";
  static const char_t DEBSTR_CiResetOff[]  = "Reset OFF";
  static const char_t DEBSTR_CiInpOff[]    = "TsInpOff";
  static const char_t DEBSTR_CiInpTsG[]    = "TsInpTSG";
  static const char_t DEBSTR_CiInpFeMux[]  = "FE->CiTsInp";
  static const char_t DEBSTR_CiOutOff[]    = "NoTS-Output {not supported by HW}";
  static const char_t DEBSTR_CiOutBypas[]  = "TS-UseBypass (direct from FE)";
  static const char_t DEBSTR_CiOutFroCi[]  = "TS-Bypass-OFF(from Ci)";
  static const char_t DEBSTR_CiIdxErr[] = "CiIdxErr";
#endif

//------------------------------------------------------------
static int32_t BOARD_FME_H61DK01_GetNumOfCiSlots(void)
//------------------------------------------------------------
{
        return BOARD_FME_H61_TB_NUM_CI_SLOTS;
}

//------------------------------------------------------------
static int32_t BOARD_FME_H61DK01_CntrlPowerReset
(uint32_t slotIdx, uint32_t ciControlBitMask, uint32_t ciControlBitValue)
//------------------------------------------------------------
{
    BOARDS_FMEDEF_DIAG_INFO_DATA_DECLARE;
    uint32_t ExtRegBitMask;
    uint32_t ExtRegBitValue;

    ExtRegBitMask  = 0;
    ExtRegBitValue = 0;
    BOARDS_FMEDEF_DIAG_INFO_DATA_SET(slotIdx, FAPI_BOARD_DIAG_BITS_JUST_INFO);
    FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiHwCtrl, (ciControlBitMask<<16)|ciControlBitValue, 0);

    if (slotIdx >= BOARD_FME_H61_TB_NUM_CI_SLOTS)
    {
        BOARDS_FMEDEF_DIAG_INFO_DATA_SET(slotIdx, FAPI_BOARD_DIAG_BITS_ERR_HALT);
        FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiIdxErr, 0, 0);
        return FAPI_ERR_BAD_PARAMETER;
    }

    //-- switch a 'diagnostic LED' if available --
    if (ciControlBitMask&BOARDS_FMEDEF_CI_CNTRL_DIAGLED)
    {
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;
    }

    //-- switch 'POWER' of slot-1 --
    if (ciControlBitMask & BOARDS_FMEDEF_CI_CNTRL_POWER)
    {
        if (ciControlBitValue&BOARDS_FMEDEF_CI_CNTRL_POWER) {
            FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiPowerON, 0, 0); }
        else {
            FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiPowerOFF, 0, 0); }
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;
    }
    if (ciControlBitMask&BOARDS_FMEDEF_CI_CNTRL_RESET)
    {
        ExtRegBitMask |= GPO_SFTRG_CI_RESET_BIT;
        if (ciControlBitValue & BOARDS_FMEDEF_CI_CNTRL_RESET) {
            FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiResetAct, ExtRegBitValue, 0);
            ExtRegBitValue |= GPO_SFTRG_CI_RESET_BIT;            // NOTE: Ci-Reset is Hi-Active
        }
        else {
            FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiResetOff, ExtRegBitValue, 0);
        }
    }

    if (ExtRegBitMask) {
        FME_H61TB01_CbsfPtr->WriteExtLatch(ExtRegBitMask, ExtRegBitValue);
    }
    return FAPI_OK;
}

//------------------------------------------------------------
static int32_t BOARD_FME_H61DK01_ControlTsPathMuxes_CI
                        (int32_t tsPathIndex, uint32_t muxCmd, void *optPtr)
//------------------------------------------------------------
{
    BOARDS_FMEDEF_DIAG_INFO_DATA_DECLARE;
    BOARDS_FMEDEF_RD_GPIO_INPREG_CBF RdGpInpLevels;
    BOARDS_FMEDEF_AllTsPathsStructureInfoT  *dstAllTsPathInfosPtr;
    uint32_t ExtRegBitMask, ExtRegBitValue;
    BOARDS_FMEDEF_TsPathControlCommandsT  contrCmd;
    int32_t  ret_stat;

    BOARDS_FMEDEF_DIAG_INFO_DATA_SET((uint32_t)tsPathIndex, FAPI_BOARD_DIAG_BITS_JUST_INFO);
    contrCmd = (BOARDS_FMEDEF_TsPathControlCommandsT) muxCmd;

    if (tsPathIndex >= BOARD_FME_H61_TB_NUM_CI_SLOTS)
    {
        BOARDS_FMEDEF_DIAG_INFO_DATA_SET((uint32_t)tsPathIndex, FAPI_BOARD_DIAG_BITS_ERR_HALT);
        FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiIdxErr, 0, 0);
    }
    ExtRegBitMask  = 0;         // def. Init
    ExtRegBitValue = 0;
    ret_stat = FAPI_OK;         // default

    //--------------------------------------------
    //-- if getting some Information-Data       --
    //--------------------------------------------
    if (contrCmd==BOARDS_TS_PATH_GET_STRUCTURE_INFOS)
    {
        if(optPtr != NULL)
        {
            dstAllTsPathInfosPtr  = optPtr;
            *dstAllTsPathInfosPtr = BOARD_FME_H61_TB01_AllTsPathInfos;
            return FAPI_OK;
        }
        else return FAPI_ERR_FEATURE_NOT_SUPPORTED;
    }

    else if (contrCmd==BOARDS_TS_PATH_GET_TS_INPUT_LEVEL)
    {
        RdGpInpLevels = FREG_GPIO_GetInput_31_0;
        ret_stat = BOARD_FME_H61DK01_GetTsPinsInputLevel(RdGpInpLevels, tsPathIndex, (uint16_t *)optPtr);
    }

    //--------------------------------------------
    //-- if controling the 'CI-InputMux (CIIM)' --
    //--------------------------------------------
    else if (contrCmd==BOARDS_TS_PATH_SET_CI_INPUT_OFF)
    {
        ExtRegBitMask  = GPO_SFTRG_CI_IN_ENA_BIT;        // control this bit
        ExtRegBitValue = GPO_SFTRG_CI_IN_ENA_BIT;        // and set to '1' -> CI-Input is NOT-active !
        FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiInpOff, 0, 0);
    }
    else if (contrCmd==BOARDS_TS_PATH_SET_CI_INPUT_TS_GEN)
    {
        ExtRegBitMask  = GPO_SFTRG_CI_IN_MUX_BIT;        // control this bit
        ExtRegBitValue = GPO_SFTRG_CI_IN_MUX_BIT;        // and set to '1' -> CI-Input is NOT-active !

        FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiInpTsG, 0, 0);
    }
    else if (contrCmd==BOARDS_TS_PATH_SET_CI_INPUT_MTIM)
    {
        ExtRegBitMask  = GPO_SFTRG_CI_IN_ENA_BIT|GPO_SFTRG_CI_IN_MUX_BIT; // control these bits
        ExtRegBitValue = 0;                 // and set to '0' -> CI-Input is Active (from MTIM) !
        FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiInpFeMux, 0, 0);
    }
    //--------------------------------------------
    //-- if controling the 'TS-BypassMux(TBPM)' --
    //--------------------------------------------
    else if (contrCmd==BOARDS_TS_PATH_SET_TS_BYPASS_NO_TS)
    {
        FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiOutOff, 0, 0);
        ret_stat = FAPI_ERR_FEATURE_NOT_SUPPORTED;
    }
    else if (contrCmd==BOARDS_TS_PATH_SET_TS_BYPASS_USED)
    {
        FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiOutBypas, 0, 0);
        BOARD_FME_H61DK01_TsAInpSelect(BRDCOM_BC_RegisterGpioMode_PreInit, 0);
    }
    else if (contrCmd==BOARDS_TS_PATH_SET_TS_BYPASS_FROM_CI)
    {
        FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_CiOutFroCi, 0, 0);
        BOARD_FME_H61DK01_TsAInpSelect(BRDCOM_BC_RegisterGpioMode_PreInit, 1);
    }

    if (ExtRegBitMask) {
        FME_H61TB01_CbsfPtr->WriteExtLatch(ExtRegBitMask, ExtRegBitValue);
//        BTLIB_IO_PrStrHexStr(8, (const char_t*)"Ts->SerOut:", ExtRegBitValue, (const char_t*)"\n");
    }
    return ret_stat;
}

//------------------------------------------------------------
//--    simple 'straight/direct' implementations     --
//------------------------------------------------------------
static int32_t BOARD_FME_H61DK01_GetInsStatus(uint32_t chkCiSlotIdx, uint32_t * insertedPtr)
{
    uint32_t pin_lev;

//    BTLIB_IO_PrStrHexStr(1, (const char_t*)"\n->CiInsStat[", chkCiSlotIdx, 0);

    if (chkCiSlotIdx >= BOARD_FME_H61_TB_NUM_CI_SLOTS)
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;

    FREG_GPIO_SetDataOut( GPIO_INP_CI_INS0, 0x100);  // make sure this GPIO-is not 'active output'
    pin_lev = (uint32_t) FREG_GPIO_GetInput_63_32();
    pin_lev = (pin_lev >> (GPIO_INP_CI_INS0 - 32)) & 1;  // NOTE:
    *insertedPtr = 1 - pin_lev;
//    BTLIB_IO_PrStrHexStr(2, (const char_t*)"] InsStat:", 1 - pin_lev, (const char_t*)"\n"); BTLIB_IO_RxByte();
    return FAPI_OK;
}

//------------------------------------------------------------
int32_t BOARD_FME_H61DK01_SetupCiLibHwInterfaceFunctions(void)
{
    BOARDS_FMEDEF_CiHwAccessInterfaceT *ci_ptr;
    int32_t ix;
    uint8_t MuxSelIdx;
    void  *voidPtr;

    //-- configure some dedicated GPIO-signals                              --
    //-- NOTE: consider this, if pins are used for 'other HW-test-purpose'  --
    FREG_GPIO_SetDataIn( (FAPI_GPIO_IN_UPI_WAIT - FAPI_GPIO_IN_OFFSET), GPIO_UPI_CI_WAIT+2);
    FREG_GPIO_SetDataOut(GPIO_UPI_BUF_DIR, (FAPI_GPIO_OUT_UPI_BUF_DIR - FAPI_GPIO_OUT_OFFSET));
    FREG_GPIO_SetDataOut(GPIO_UPI_IORD,    (FAPI_GPIO_OUT_UPI_DQM_0   - FAPI_GPIO_OUT_OFFSET));
    FREG_GPIO_SetDataOut(GPIO_UPI_IOWR,    (FAPI_GPIO_OUT_UPI_DQM_1   - FAPI_GPIO_OUT_OFFSET));

    //-- other CI-related setups --
    BRDCOM_CI_InitCiAccessViaUpi(&H61_Tb01_CiHwAccessInfo);
    ci_ptr = &H61_TB01_CiHwInterfaceFunctions;
    ci_ptr->CiHwInterfaceMagicAndVersion = BOARDS_FMEDEF_CI_HW_INTERFACE_MAGIC;
    ci_ptr->pCiHwInfoStr                 = STR_H61TbCiLogic;
    ci_ptr->GetNumberOfCiSlots           = BOARD_FME_H61DK01_GetNumOfCiSlots;
    ci_ptr->InitDriver                   = BRDCOM_CI_DyInit;
    ci_ptr->OpenDriver                   = BRDCOM_CI_DyOpen;
    ci_ptr->CloseDriver                  = BRDCOM_CI_DyClose;
    ci_ptr->ControlPowerReset            = BOARD_FME_H61DK01_CntrlPowerReset;
    ci_ptr->GetInsStatus                 = BOARD_FME_H61DK01_GetInsStatus;
    ci_ptr->ReadIoPort                   = BRDCOM_CI_ReadIoPort;
    ci_ptr->WriteIoPort                  = BRDCOM_CI_WriteIoPort;
    ci_ptr->ReadAttibMem                 = BRDCOM_CI_ReadAttribMem;
    ci_ptr->WriteAttibMem                = BRDCOM_CI_WriteAttribMem;
    ci_ptr->ReadSRam                     = 0;
    ci_ptr->WriteSRam                    = 0;

    FAPI_BOARD_SetCiInterface(&H61_TB01_CiHwInterfaceFunctions);
    FAPI_BOARD_SetControlTsPathHook(BOARD_FME_H61DK01_ControlTsPathMuxes_CI);

    //-- set default TS-Path Muxing --
    for (ix=0;ix<BOARD_FME_H61_TB_NUM_CI_SLOTS;ix++)
    {
        MuxSelIdx = (uint8_t) ix;
        voidPtr   = (void *) &MuxSelIdx;
        (void) FAPI_BOARD_EXT_CONTROL_TS_PATH(ix, (uint32_t) BOARDS_TS_PATH_SET_CI_INPUT_OFF,  voidPtr);
        (void) FAPI_BOARD_EXT_CONTROL_TS_PATH(ix, (uint32_t) BOARDS_TS_PATH_SET_TS_BYPASS_USED, voidPtr);
    }
    return FAPI_OK;
}

