/*!
********************************************************************************
**
** \file        ./boards/fme/mb86h61_dk01/mb86h61_dk01_fp.c
**
** \version     $Id: mb86h61_dk01_fp.c 2579 2010-07-01 15:13:55Z rseng $
**
** \brief       FME devkit 'simple direct front-panel control'
**              mainly for HW-testing.
**              (Fapi-driver could be used as well, but will add code-size)
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU SEMICONDUCTOR
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2006-2010 by Fujitsu Semiconductor Europe GmbH
** (C) Copyright 2010 by Fujitsu Semiconductor Europe GmbH
**
********************************************************************************
*/

#include <fapi/reg_fpc.h>
#include <fapi/board_ifc.h>
#include "fapi/drv_ir.h"

#include <boards/boards_common.h>
#include <boards/boards_com_fp.h>
#include <boards/boards_com_dirx.h>
#include <boards/mb86h61_dk01.h>
//#include "btestlib/btest.h"                       // just for HW-debug

#define FP_CTRL_CONFIG_ACTIVE   0x0005              // default for FP-control register
#define INDICATE_ILLEGAL    0x12345678
#define ALL_DIGIT_SEGS_OFF  0xFF                    //

/*lint -save -e923 -e778 */

static uint32_t prevKeyData = 0;

static BOARDS_FrontPanel_HwNameInfoT  FpHwInfo = {
// HwInfoStrPtr,    SymbolCapabilityBits
   "Fme.H61Dk01Fp", 0  };

#define KEYBIT_ESC      0x40
#define KEYBIT_MENU     0x10
#define KEYBIT_RIGHT    0x02
#define KEYBIT_LEFT     0x80
#define KEYBIT_UP       0x08
#define KEYBIT_DOWN     0x20

typedef struct {
    uint8_t KBit;
    uint8_t FapiCode;
} KEY_BIT_INFOS_T;

static const KEY_BIT_INFOS_T  KeyCodes_NormalMode[] = {
//  KBit,         FapiCode
  { KEYBIT_ESC,   FAPI_IR_KEY_OK   }
 ,{ KEYBIT_MENU,  FAPI_IR_KEY_MENU }
 ,{ KEYBIT_RIGHT, FAPI_IR_KEY_RIGHT}
 ,{ KEYBIT_LEFT,  FAPI_IR_KEY_LEFT }
 ,{ KEYBIT_UP,    FAPI_IR_KEY_UP   }
 ,{ KEYBIT_DOWN,  FAPI_IR_KEY_DOWN }
};
#define MAX_TABLE_SIZE  (sizeof(KeyCodes_NormalMode)/sizeof(KEY_BIT_INFOS_T))


static const uint8_t  SevenSegDigitPatterns[] = {        // NOTE: inverted bits
/*
 *     - BIT7 -
 *    |        |
 *  BIT2      BIT6
 *    |        |
 *    |- BIT1 -|
 *    |        |
 *  BIT3      BIT5
 *    |        |
 *     - BIT4 -   BIT0
 */
  (BIT0|BIT1                              ), // '0'
  (BIT0|BIT1|BIT2|BIT3|BIT4|          BIT7), // '1'
  (BIT0|     BIT2|          BIT5          ), // '2'
  (BIT0|     BIT2|BIT3                    ), // '3'
  (BIT0|          BIT3|BIT4|          BIT7), // '4'
  (BIT0|          BIT3|          BIT6     ), // '5'
  (BIT0|                         BIT6     ), // '6'
  (BIT0|BIT1|BIT2|BIT3|BIT4               ), // '7'
  (BIT0                                   ), // '8'
  (BIT0|          BIT3|BIT4               ), // '9'
  (BIT0|               BIT4               ), // 'A'
  (BIT0|                         BIT6|BIT7), // 'B'
  (BIT0|BIT1|               BIT5|BIT6     ), // 'C'
  (BIT0|     BIT2|                    BIT7), // 'D'
  (BIT0|                    BIT5|BIT6     ), // 'E'
  (BIT0|               BIT4|BIT5|BIT6     ), // 'F'
};

#define FP_7SEG_on_ 0xFFD5C5FF  // _on_
#define FP_7SEG_oFF 0x7171C5FF  // _oFF

/*!
********************************************************************************
** \brief show service info
**
** \param sviPtr   pointer to 'info' structure
********************************************************************************
*/
static int32_t BOARD_FME_H61DK01_ScanFpKeyboard(BOARDS_FrontPanel_KeyPressInfoT *dstKeyPressDataPtr)
{
    volatile CTRLREGS_SERIAL_FPC_T  *fpcPtr;
    const KEY_BIT_INFOS_T           *kycoPtr;
    uint32_t tmpval;
    uint32_t key_code, ix;

    fpcPtr = (volatile CTRLREGS_SERIAL_FPC_T *) FREG_FPC_CONFIG;
    key_code = 0;                           // default: no key pending
    tmpval = fpcPtr->KeyStatus;             // get key-code bit-status
    tmpval = ~tmpval;                       // invert bits
    tmpval &= 0xFF;                         // only 1-byte
    kycoPtr = KeyCodes_NormalMode;          // pointer to actual array (def. 'KeyCodes_NormalMode')
    ix = MAX_TABLE_SIZE;
    while (ix)
    {
        ix--;
        if (tmpval==kycoPtr->KBit)
        {
            key_code = kycoPtr->FapiCode;
            break;
        }
        kycoPtr++;              // check next case
    }

    if (tmpval != prevKeyData)
    {
        prevKeyData = tmpval;
//      BTLIB_IO_PrStrHexStr(8, (const char_t*)" BitInfo: ", tmpval, (const char_t*)"\n");
        tmpval = 0;                                 // clear 'repeated/same' flag
    }
    else {
        tmpval = 1;                                 // set 'repeated/same' flag
    }
    dstKeyPressDataPtr->flgIsRepeatedKey  = (uint8_t) tmpval;   // store flag-info
    dstKeyPressDataPtr->actPendingKeyCode = (uint8_t) key_code;
    return 0;
}

/*!
********************************************************************************
** \brief show service info
**
** \param sviPtr   pointer to 'info' structure
********************************************************************************
*/
static int32_t BOARD_FME_H61DK01_ShowServiceInfo(BOARDS_FrontPanel_ServiceInfoT *sviPtr)
{
    volatile CTRLREGS_SERIAL_FPC_T  *fpcPtr;
    uint32_t  number, divval;
    uint32_t  ix, tmpval;
    uint32_t  dispBits;
    uint32_t  FlgNoLeadingZeros;

    fpcPtr = (volatile CTRLREGS_SERIAL_FPC_T *) FREG_FPC_CONFIG;

//  dispBits  = sviPtr->shortFpMagic;
    //-- this HW-Platfrom can only display channel numbers ... --
    dispBits = sviPtr->bf_UpdateMask;
    if (!(dispBits & (BOARDS_FP_UPDBF_CHAN_NUM|BOARDS_FP_UPDBF_CHAN_NUMNLZ)))
    {
        return 0;
    }
    FlgNoLeadingZeros = dispBits & BOARDS_FP_UPDBF_CHAN_NUMNLZ; // save local flag
    number            = sviPtr->channelNumber;
    dispBits = INDICATE_ILLEGAL;                // default
    divval = 10000;
    if ((number == 0)&&(FlgNoLeadingZeros))
    {
        dispBits = 0xFFFFFFFF;
    }
    else if (number < divval)                            // limit to 'max' value
    {
        dispBits = 0;                               // init
        for (ix=0;ix<4;ix++)
        {
            divval = divval / 10;
            tmpval = number / divval;               // highest digit-position-value
            number -= (tmpval * divval);            // substract for next time
            if ((tmpval==0)&&(FlgNoLeadingZeros)&&(ix!=3)) {
                tmpval = ALL_DIGIT_SEGS_OFF;
            }
            else {
                tmpval = SevenSegDigitPatterns[tmpval]; // get bit-pattern from table
                FlgNoLeadingZeros = 0;              // do not check in future
            }
            tmpval = tmpval << 24;                  // move into 'MS-Byte'
            dispBits  = dispBits >> 8;              // 'shift-down' previous digit-data
            dispBits |= tmpval;
        }
    }
    fpcPtr->SegsData = dispBits;                    // update display
    return 0;
}

/*!
********************************************************************************
** \brief  control interface for front-panel
**
** \param index    not used
** \param command  control-command
** \param optPtr   pointer to specific data/structure, depending on command
********************************************************************************
*/
static int32_t BOARD_FME_H61DK01_FrontPanel_Ctrl(
           int32_t index, BOARDS_FrontPanel_CtrlCodeT command, void*  optPtr )
{
    BOARDS_FrontPanel_HwNameInfoT   **fpHwInfoPP;
    volatile CTRLREGS_SERIAL_FPC_T  *fpcPtr;

    index   = index;
//    BTLIB_IO_PrStr("\n[FPC]\n");
    if      (command == BOARDS_FP_CTRL_GET_FP_KEY)
    {
        (void) BOARD_FME_H61DK01_ScanFpKeyboard((BOARDS_FrontPanel_KeyPressInfoT *) optPtr);
    }

    else if (command == BOARDS_FP_UPDATE_SERVICE_INFO)
    {
        (void) BOARD_FME_H61DK01_ShowServiceInfo((BOARDS_FrontPanel_ServiceInfoT *) optPtr);
    }

    else if (command == BOARDS_FP_SET_PWR_MODE_STANDBY)
    {
        fpcPtr = (volatile CTRLREGS_SERIAL_FPC_T *) FREG_FPC_CONFIG;
        fpcPtr->SegsData = FP_7SEG_oFF;
    }
    else if (command == BOARDS_FP_SET_PWR_MODE_OPERATION)
    {
        fpcPtr = (volatile CTRLREGS_SERIAL_FPC_T *) FREG_FPC_CONFIG;
        fpcPtr->Config   = FP_CTRL_CONFIG_ACTIVE;       // activate
//        fpcPtr->SegsData = 0x12345678;                //
        fpcPtr->SegsData = FP_7SEG_on_;
    }
    else if (command == BOARDS_FP_GET_HW_INFO)
    {
        fpHwInfoPP = optPtr;
        *fpHwInfoPP = &FpHwInfo;
    }
    return 0;
}

/*!
********************************************************************************
** \brief register the 'FME_H61KD01 front-panel' control function
**
** \param none
********************************************************************************
*/
void BOARD_FME_H61DK01_FrontPanel_Register(void)
{
    BRDCOM_BC_DirectAccess_Init( );                            // make sure thi module is 'linked in'
    BRDCOM_SetFrontPanelCtrlHook( BOARD_FME_H61DK01_FrontPanel_Ctrl );
}

