/*!
********************************************************************************
**
** \file        ./boards/fme/mb86h61_dk01/mb86h61_dk01_gpio.c
**
** \version     $Id: mb86h61_dk01_gpio.c 2579 2010-07-01 15:13:55Z rseng $
**
** \brief       function to initialize the GPIO-configuration registers
**
** \note        there exist two methods
**              a) via ..\fapi\sys\src\sys_init_mb86h61_devkit.h
**              b) using this 'baords' driver
**              Actually method b) should be preferred. If the file which includes
**              '<fapi/sys_init.h>' has the define
**              '#define FAPI_SYS_INIT_USE_BOARDS 1', then the GPIO-registration
**              form 'sys_init_mb86h61_devkit.h' will be disabled
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU SEMICONDUCTOR
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2006-2010 by Fujitsu Semiconductor Europe GmbH
** (C) Copyright 2010 by Fujitsu Semiconductor Europe GmbH
**
********************************************************************************
*/

#include <fapi/reg_gpio.h>
#include <fapi/drv_gpio.h>

#include <boards/boards_common.h>          // some defines needed for 'mb86h61_dk01.h'
#include <boards/mb86h61_dk01.h>

#define NO_DIG_RGB_OR_CI_SLAVE_TEST     // comment this out when using 'digital RBG-signals'
                                        // or UPI-Slave,CAM-Emulation Test

//#define USE_P_ATA_ADAPTER             // active when using P-Ata-Adapter

typedef struct {
  uint32_t pinNum;
  uint32_t config;
} GPIO_CONFIG_INFO_T;


static  GPIO_CONFIG_INFO_T  GpioConfig_TsAInput_FromFe[] = {
 { 16, FAPI_GPIO_IN_TS_A_DATA_7  }
,{ 17, FAPI_GPIO_IN_TS_A_DATA_6  }
,{ 18, FAPI_GPIO_IN_TS_A_DATA_5  }
,{ 19, FAPI_GPIO_IN_TS_A_DATA_4  }
,{ 20, FAPI_GPIO_IN_TS_A_DATA_3  }
,{ 21, FAPI_GPIO_IN_TS_A_DATA_2  }
,{ 22, FAPI_GPIO_IN_TS_A_DATA_1  }
,{ 23, FAPI_GPIO_IN_TS_A_DATA_0  }
,{ 24, FAPI_GPIO_IN_TS_A_EN      }
,{ 25, FAPI_GPIO_IN_TS_A_PSTART  }
,{ 15, FAPI_GPIO_IN_TS_A_CLK     }
};

static  GPIO_CONFIG_INFO_T  GpioConfig_TsAInput_FromCi[] = {
 { 88, FAPI_GPIO_IN_TS_A_DATA_7  }
,{ 89, FAPI_GPIO_IN_TS_A_DATA_6  }
,{ 90, FAPI_GPIO_IN_TS_A_DATA_5  }
,{ 91, FAPI_GPIO_IN_TS_A_DATA_4  }
,{ 92, FAPI_GPIO_IN_TS_A_DATA_3  }
,{ 93, FAPI_GPIO_IN_TS_A_DATA_2  }
,{ 94, FAPI_GPIO_IN_TS_A_DATA_1  }
,{ 95, FAPI_GPIO_IN_TS_A_DATA_0  }
,{ 85, FAPI_GPIO_IN_TS_A_EN      }
,{ 86, FAPI_GPIO_IN_TS_A_PSTART  }
,{ 87, FAPI_GPIO_IN_TS_A_CLK     }
};


static  GPIO_CONFIG_INFO_T  BOARD_FME_H61DK01_GpioConfig[] = {
 { 16, FAPI_GPIO_IN_TS_A_DATA_7  }
,{ 17, FAPI_GPIO_IN_TS_A_DATA_6  }
,{ 18, FAPI_GPIO_IN_TS_A_DATA_5  }
,{ 19, FAPI_GPIO_IN_TS_A_DATA_4  }
,{ 20, FAPI_GPIO_IN_TS_A_DATA_3  }
,{ 21, FAPI_GPIO_IN_TS_A_DATA_2  }
,{ 22, FAPI_GPIO_IN_TS_A_DATA_1  }
,{ 23, FAPI_GPIO_IN_TS_A_DATA_0  }
,{ 24, FAPI_GPIO_IN_TS_A_EN      }
,{ 25, FAPI_GPIO_IN_TS_A_PSTART  }
,{ 15, FAPI_GPIO_IN_TS_A_CLK     }
,{ 30, FAPI_GPIO_IN_TS_B_DATA_7  }
,{ 31, FAPI_GPIO_IN_TS_B_DATA_6  }
,{ 32, FAPI_GPIO_IN_TS_B_DATA_5  }
,{ 33, FAPI_GPIO_IN_TS_B_DATA_4  }
,{ 35, FAPI_GPIO_IN_TS_B_DATA_3  }
,{ 34, FAPI_GPIO_IN_TS_B_DATA_2  }
,{ 36, FAPI_GPIO_IN_TS_B_DATA_1  }
,{ 37, FAPI_GPIO_IN_TS_B_DATA_0  }
,{ 29, FAPI_GPIO_IN_TS_B_EN      }
,{ 28, FAPI_GPIO_IN_TS_B_PSTART  }
,{ 27, FAPI_GPIO_IN_TS_B_CLK     }
,{ 13, FAPI_GPIO_BIDI_I2C0_DATA  }
,{ 14, FAPI_GPIO_BIDI_I2C0_CLK   }
,{ 54, FAPI_GPIO_OUT_AD0_BCLK    }
,{ 53, FAPI_GPIO_OUT_AD0_LRCLK   }
,{ 55, FAPI_GPIO_OUT_AD0_CLK     }
,{ 52, FAPI_GPIO_OUT_AD0_DATA    }
,{ 51, FAPI_GPIO_OUT_AD1_DATA    }
,{ 50, FAPI_GPIO_OUT_AD2_DATA    }
,{ 49, FAPI_GPIO_OUT_AD3_DATA    }
,{ 48, FAPI_GPIO_OUT_0           }
,{  2, FAPI_GPIO_IN_UART0_DATA   }
,{  3, FAPI_GPIO_OUT_UART0_DATA  }
,{  9, FAPI_GPIO_OUT_UART1_DATA  }
,{ 10, FAPI_GPIO_IN_UART1_DATA   }
,{ 47, FAPI_GPIO_IN_SFLASH_DATA  }
,{ 44, FAPI_GPIO_OUT_SFLASH_HOLD }
,{ 45, FAPI_GPIO_OUT_SFLASH_CS   }
,{ 46, FAPI_GPIO_OUT_SFLASH_CLK  }
,{ 43, FAPI_GPIO_OUT_SFLASH_DATA }
,{ 75, FAPI_GPIO_OUT_SDIO_CMD    }
,{ 74, FAPI_GPIO_OUT_SDIO_DATA_0 }
,{ 67, FAPI_GPIO_OUT_SDIO_DATA_1 }
,{ 66, FAPI_GPIO_OUT_SDIO_DATA_2 }
,{ 65, FAPI_GPIO_OUT_SDIO_DATA_3 }
#ifdef NO_DIG_RGB_OR_CI_SLAVE_TEST
,{ 76, FAPI_GPIO_OUT_SDIO_CLK    }
,{ 75, FAPI_GPIO_IN_SDIO_CMD     }
,{ 74, FAPI_GPIO_IN_SDIO_DATA_0  }
#endif
,{ 67, FAPI_GPIO_IN_SDIO_DATA_1  }
,{ 66, FAPI_GPIO_IN_SDIO_DATA_2  }
,{ 65, FAPI_GPIO_IN_SDIO_DATA_3  }
,{ 68, FAPI_GPIO_IN_SDIO_CD      }
//,{ 42, FAPI_GPIO_IN_IR_DATA      }
,{ 42, FAPI_GPIO_OUT_SPDIF_OUT   }
#ifdef NO_DIG_RGB_OR_CI_SLAVE_TEST
,{ 84, FAPI_GPIO_OUT_FPC_CLK     }
,{ 78, FAPI_GPIO_IN_FPC_KEY_0    }
,{ 77, FAPI_GPIO_IN_FPC_KEY_1    }
,{ 83, FAPI_GPIO_OUT_FPC_DATA    }
,{ 82, FAPI_GPIO_OUT_FPC_SEG_0   }
,{ 81, FAPI_GPIO_OUT_FPC_SEG_1   }
,{ 80, FAPI_GPIO_OUT_FPC_SEG_2   }
,{ 79, FAPI_GPIO_OUT_FPC_SEG_3   }
#endif
,{ 65, FAPI_GPIO_OUT_UPI_ADDRX0  }
,{ 66, FAPI_GPIO_OUT_UPI_ADDRX1  }
,{ 67, FAPI_GPIO_OUT_UPI_ADDRX2  }
,{ 68, FAPI_GPIO_OUT_UPI_ADDRX3  }
,{ 39, FAPI_GPIO_RESERVED_39     }
,{ 41, FAPI_GPIO_RESERVED_41     }
};

/*!
********************************************************************************
** \brief just a 'security function' should generate a linker-warning (but does not work)
********************************************************************************
#ifdef NO_DIG_RGB_OR_CI_SLAVE_TEST

uint32_t BOARD_FME_H61DK01_GenLinkError_ViolatesWithNormalGpioInit;

void BOARD_FME_H61DK01_GenerateLinkError_ViolatesWithNormalGpioInit(void)
{

}
#endif
*/

/*!
********************************************************************************
** \brief initialize all platform/board specific GPIO pins
**
** this is likely different on all different HW-platforms
**
********************************************************************************
*/
void BOARD_FME_H61DK01_RegisterGpioPins(BOARDS_FMEDEF_REG_INI_PIN_CBF_T useThisRegistFunction)
{
    GPIO_CONFIG_INFO_T *gpc_ptr = BOARD_FME_H61DK01_GpioConfig;
    uint32_t ix, num;
    uint32_t pin, func;

    if (useThisRegistFunction==NULL) return;    // would be severe programming mistake

    num = (sizeof(BOARD_FME_H61DK01_GpioConfig) / sizeof(GPIO_CONFIG_INFO_T));
    for (ix=0;ix<num;ix++)
    {
        pin  = gpc_ptr->pinNum;
        func = gpc_ptr->config;
        useThisRegistFunction( pin, func );
        gpc_ptr++;
    }
}


/*!
********************************************************************************
** \brief (re-)configure GPIO-Pins providing TS-A Input
**
** taking TS either from FrontEnd or from CI-Slot output
**
********************************************************************************
*/
void BOARD_FME_H61DK01_TsAInpSelect(BOARDS_FMEDEF_REG_INI_PIN_CBF_T useThisRegistFunction, uint32_t inpSelFlag)
{
    GPIO_CONFIG_INFO_T *gpc_ptr;
    uint32_t ix, num;
    uint32_t pin, func;

    if (!inpSelFlag)
        gpc_ptr = GpioConfig_TsAInput_FromFe;
    else
        gpc_ptr = GpioConfig_TsAInput_FromCi;

    num = (sizeof(GpioConfig_TsAInput_FromFe) / sizeof(GPIO_CONFIG_INFO_T));
    for (ix=0;ix<num;ix++)
    {
        pin  = gpc_ptr->pinNum;
        func = gpc_ptr->config;
        useThisRegistFunction( pin, func );
        gpc_ptr++;
    }
}

/*!
********************************************************************************
** \brief compose a 'BitPatternValue' corresponding to TS-Input signals
**
** this function can be used for diagnostic purpose to see if 'TS-input' active
**
** \param tsPathIndex        Index of the TS-path
** \param dstTsInpLevelsPtr  pointer to 'U16' where to store value
**
** \return
** - FAPI_OK                if index is OK
** - FAPI_ERR_BAD_PARAMETER if illegal TS-path index
********************************************************************************
*/
int32_t BOARD_FME_H61DK01_GetTsPinsInputLevel(BOARDS_FMEDEF_RD_GPIO_INPREG_CBF  RdGpioInpFnc,
                                              int32_t tsPathIndex,
                                              uint16_t *dstTsInpLevelsPtr)
{
    uint32_t gpio_regdat;
    uint32_t test_data;
    int32_t  ret_stat;

    test_data   = 0;            // just init
    ret_stat    = FAPI_OK;
    tsPathIndex = tsPathIndex;

    if (!RdGpioInpFnc) return FAPI_ERR_BAD_PARAMETER;
//    gpio_regdat = BOARDS_GPIO_GET_INP_31_0();
    gpio_regdat = RdGpioInpFnc();

    if (tsPathIndex==0)
    {
        test_data  = ((gpio_regdat>>5)&0x400);     // GPIO15 -5 -> TS-Data10 (CLK)
        test_data |= ((gpio_regdat>>15)&0x200);    // GPIO24 -15-> TS-Data9  (VAL)
        test_data |= ((gpio_regdat>>17)&0x100);    // GPIO25 -17-> TS-Data8  (SYN)
        test_data |= ((gpio_regdat>>9) & 0x80);    // GPIO16 -9 -> TS-Data7
        test_data |= ((gpio_regdat>>11)& 0x40);    // GPIO17 -11-> TS-Data6
        test_data |= ((gpio_regdat>>13)& 0x20);    // GPIO18 -13-> TS-Data5
        test_data |= ((gpio_regdat>>15)& 0x10);    // GPIO19 -15-> TS-Data4
        test_data |= ((gpio_regdat>>17)& 0x08);    // GPIO20 -17-> TS-Data3
        test_data |= ((gpio_regdat>>19)& 0x04);    // GPIO21 -19-> TS-Data2
        test_data |= ((gpio_regdat>>21)& 0x02);    // GPIO22 -21-> TS-Data1
        test_data |= ((gpio_regdat>>23)& 0x01);    // GPIO23 -23-> TS-Data0
    }
    else if (tsPathIndex==1)
    {
        test_data  = 0;                             // not implemented yet
    }
    else
        ret_stat = FAPI_ERR_BAD_PARAMETER;

    if (ret_stat == FAPI_OK)
        *dstTsInpLevelsPtr = test_data;

    return ret_stat;
}


#ifdef USE_P_ATA_ADAPTER
// NOTE: the P-ATA-Adapter needs to re-configure some GPIO-pins otherwise used for CI-Interface
#define PIN_PATA_IRQ    (GPIO_UPI_CLE)
#define PIN_UPI_WAIT    (GPIO_UPI_CI_WAIT)
#define PIN_UPI_DMARQ   (GPIO_UPI_IORD)         //
#define PIN_UPI_DMAAC   (GPIO_UPI_IOWR)         //
#define PIN_UPI_CS2     (GPIO_UPI_BUF_DIR)      // this will be used as HD-ChipSelect

static  GPIO_CONFIG_INFO_T  GpioConfig_Pins4PAtaTest[] = {
/* { PIN_PATA_IRQ,  FAPI_GPIO_IN_IRQ_IN0   }
,*/{ PIN_UPI_WAIT,  FAPI_GPIO_IN_UPI_WAIT    }
,{ PIN_UPI_DMARQ, FAPI_GPIO_IN_UPI_DMARQ   }
,{ PIN_UPI_DMAAC, FAPI_GPIO_OUT_UPI_CLE    }  // note: actuall 'DMACK'
,{ PIN_UPI_CS2,   FAPI_GPIO_OUT_UPI_CS     }
};

/*!
********************************************************************************
** \brief re-configure relevant GPIO for 'special J900-PAta-Adapter-Test'
**
** this is likely different on all different HW-platforms
**
********************************************************************************
*/
void BOARD_FME_H61DK01_ConfigGpiosPAtaAdapter(BOARDS_FMEDEF_REG_INI_PIN_CBF_T useThisRegistFunction)
{
    GPIO_CONFIG_INFO_T *gpc_ptr = GpioConfig_Pins4PAtaTest;
    uint32_t ix, num;
    uint32_t pin, func;

    if (useThisRegistFunction==NULL) return;    // would be severe programming mistake

//    BOARDS_GPIO_SET_OUTP(GPIO_UPI_BUF_DIR, (FAPI_GPIO_OUT_UPI_CS - FAPI_GPIO_OUT_OFFSET));

    num = (sizeof(GpioConfig_Pins4PAtaTest) / sizeof(GPIO_CONFIG_INFO_T));
    for (ix=0;ix<num;ix++)
    {
        pin  = gpc_ptr->pinNum;
        func = gpc_ptr->config;
        useThisRegistFunction( pin, func );
        gpc_ptr++;
    }
}
#endif
