/*!
********************************************************************************
**
** \file        ./boards/fmc/mb86h61_pdkv1/mb86h61_pdkv1.c
**
** \version     $Id: mb86h61_pdkv1.c  $
**
** \brief       FMC devkit specific board service function.
**
** This file contains the implementation of the FME devkit specific board
** services functions.
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU MICROELECTRONICS
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2006-2010 by Fujitsu Semiconductor Europe GmbH
** (C) Copyright 2010 by Fujitsu Semiconductor Europe GmbH
**
********************************************************************************
*/

#define BOARDS_FMEDEF_DIAG_INFO__MACRO_ENABLED
//#define TEST_PATA_INTERFACE_VIA_J900_ADAPTER

#include <fapi/reg_gpio.h>
#include <fapi/drv_gpio.h>
#include <fapi/board_ifc.h>
//#include <fe/fe.h>

#include <boards/boards_common.h>
#include <boards/boards_com_fp.h>
#include <boards/mb86h61_pdkv1.h>

#ifndef PIX_WLANTUNER

/*lint -save -e923 -e778 */

//#define FBD_ID   BOARDS_FMEDEF_DIAG_INFO_DATA     // simple local re-define to save code-spase
#define FBD_ID   FAPI_BOARD_DIAG_BITS_JUST_INFO

#define GP_XSI 11
#define GP_XSO 38
#define GP_XLD 40
#define GP_XCK 39

#define LATCH_INIT_MASK (0xFF)          // init all bits
//#define LATCH_INIT_BITS (GPO_SFTRG_USB_CBU0|GPO_SFTRG_USB_CBU1|GPO_SFTRG_CI_IN_ENA_BIT)
#define LATCH_INIT_BITS (GPO_SFTRG_CI_IN_ENA_BIT)   // NOTE: by default, USB-VBUS is off

static int32_t BOARD_FMC_H61PDKV1_BoardIfc_InitDetect(uint32_t initDetectFlags, void *optInitDataPtr);
static int32_t BOARD_FMC_H61PDKV1_ControlFeResetPower 
               ( int32_t index, FAPI_BOARD_ControlCommandT command, void* optPtr );

#ifdef BOARDS_FMEDEF_DIAG_INFO__MACRO_ENABLED
  static const char_t DEBSTR_SysPwrActive[] = "SystemPower->ACTIVE";
  static const char_t DEBSTR_SysPwrOff[]    = "SystemPower->OFF";
  static const char_t DEBSTR_UsbPhyCtrl[]   = "USB-PhyCtrl";
#endif


static BOARDS_FMEDEF_SerialIoExtInfoT  H61DkSerIoExtInfos = {
// SIpin,  SOpin,  CLKpin, LDpin,  nShifts,     tDelCnt
   GP_XSI, GP_XSO, GP_XCK, GP_XLD, USE_N_SHIFT_CLOCKS, 20      };

//------------------------------------------------------------
//-- Platform Info/Configuration-Data for Application-SW    --
//------------------------------------------------------------
static const char_t Manufact[] = "Fujitsu(RSD)";
static const char_t Model[]    = "MB86h61 Test-Board";
static const FAPI_BOARD_HwPlatformInfosT  FmcMb86H61PDkv1_HwInfo =
{
    FAPI_BOARD_HWCONFIG_VERSION,//!< MagicAndVersion
    Manufact,                   //!< pHwManufactString
    Model,                      //!< pHwModelString
    NULL,                       //!< pOptionalAdditionalInfo
    0,                          //!< flagUsbHostOnly
    FAPI_BOARD_BITFLAG_SFLASH,  //!< typeOfFlashAvailable
    0,                          //!< AudioConfigurationInfoA
    0,                          //!< AudioConfigurationInfoB
    2,                          //!< numberOfAvlFrontEnds
    1,                          //!< numberOfAvlCiSlots

    {
        { 0,  //!< defaultFeType
          FAPI_BOARD_RSRC_TSIN_FE1, //!< tsPathInputConfig
          1,                        //!< flagCiPathAval
          FAPI_BOARD_RSRC_CIIN_FE1 },   //!< inputOptionsCiPath
        { 0,  //!< defaultFeType
          FAPI_BOARD_RSRC_TSIN_FE2, //!< tsPathInputConfig
          0,                        //!< flagCiPathAval
          0 }                       //!< inputOptionsCiPath
    },
    FAPI_BOARD_HD1_P_ATA,           //!< flagHardDiskInterface
    0,                              //!< flgDdr2MemoryInfos
    FAPI_BOARD_VID_AMP_C0           //!< video/scaler optimize info
};

BOARDS_CommonSupportFunctionsT  *FMC_H61TB01_CbsfPtr;   //

static uint8_t  BefPOff16State, BefPOffPin8State;
static uint8_t  ActPin16State, ActPin8State;

/*!
********************************************************************************
** \brief Controls a LED.
**
** Controls a diagnostic LED and/or a ScopeTrigger-Pin on the hw-platform
**
** \param index   Index of the LED to be controlled.
** \param command LED value to drive (FAPI_BOARD_LedCommandT).
** \param optPtr  Optional parameter pointer depending on 'command'
**                in case of 'FAPI_BOARD_GET_TEST_KEY' points to uint32_t
**
** \sa FAPI_BOARD_LedCommandT
**
** \return
** - FAPI_ERR_FEATURE_NOT_SUPPORTED if there is no function registered
** - any other value is the result of the registered function call
********************************************************************************
*/
static int32_t BOARD_FMC_H61PDKV1_ControlAlifeLed
                 ( int32_t index, FAPI_BOARD_LedCommandT command, void* optPtr )
{
//    uint32_t bit_mask, bit_valu;
    index = index;
    command = command;
    optPtr  = optPtr;

//    if (command == FAPI_BOARD_TEST_LED_ON) { FREG_GPIO_SetDataOut(GPIO_UPI_BUF_DIR, 1); }
//    else                                   { FREG_GPIO_SetDataOut(GPIO_UPI_BUF_DIR, 0); }
    return  FAPI_OK;
}

/*!
********************************************************************************
** \brief Controls the reset-signal and power-supply of the frontend.
**
** This function controls the reset-signal and power-supply of the frontend.
**
** \param index   Index of the frontend to be controlled.
** \param command Type of reset signal (#FAPI_BOARD_ControlCommandT).
** \param optPtr  Optional parameters (not used yet).
**
** \return
** - #FAPI_ERR_FEATURE_NOT_SUPPORTED if there is no function registered
** - any other value is the result of the registered function call
********************************************************************************
*/
static int32_t BOARD_FMC_H61PDKV1_ControlFeResetPower     // todo !
               ( int32_t index, FAPI_BOARD_ControlCommandT command, void* optPtr )
{
    uint32_t gpio_pin;
    uint32_t bit_valu;

    optPtr   = optPtr;
    bit_valu = 0;                       // default
    //-- check supported indexes --
    if      (index==0)
        gpio_pin = GPIO_OP_FE_RES_A;
    else if (index==1)
        gpio_pin = GPIO_OP_FE_RES_B;
    else
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;

    //-- check supported controls --
    if      (command == FAPI_BOARD_RESET_INACTIVE)
        bit_valu = 1;
    else if (command == FAPI_BOARD_RESET_ACTIVE)
        bit_valu = 0;
    else if (command == FAPI_BOARD_RESET_PULSE)
    {
        (void)BOARD_FMC_H61PDKV1_ControlFeResetPower(index, FAPI_BOARD_RESET_ACTIVE, optPtr);
        FMC_H61TB01_CbsfPtr->SimpleTimeDelay(50);
        (void)BOARD_FMC_H61PDKV1_ControlFeResetPower(index, FAPI_BOARD_RESET_INACTIVE, optPtr);
        FMC_H61TB01_CbsfPtr->SimpleTimeDelay(50);
        return FAPI_OK;
    }
    else
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;

    FREG_GPIO_SetDataOut(gpio_pin, bit_valu);

    return  FAPI_OK;
}

/*!
********************************************************************************
** \brief 'USB-Phy-Control', in this case control VBUS On/Off
**
** This function controls relevant modes of 'USB-Phy-Interface'
**
** \param index   Index
** \param command Type of control
** \param optPtr  Optional parameters (not used yet).
**
** \return
** - #FAPI_ERR_FEATURE_NOT_SUPPORTED if there is no function registered
** - FAPI_OK
********************************************************************************
*/
static int32_t BOARD_FMC_H61PDKV1_CtrlUsbPhy
               ( int32_t index, uint32_t ctrl, void* optPtr )
{
    FAPI_BOARD_ControlCommandT command;
    uint32_t bit_mask;
    uint32_t bit_valu;

    optPtr  = optPtr;
    bit_mask = 0;
    bit_valu = 0;
    if (index >= 2) return FAPI_ERR_BAD_PARAMETER;

    command = (FAPI_BOARD_ControlCommandT) ctrl;
    FAPI_BOARD_DIAGNOSTIC_TS_PATH(FBD_ID, (void*) DEBSTR_UsbPhyCtrl, ctrl, 0);
    if     (command == FAPI_BOARD_POWER_ON)
    {
        bit_mask = GPO_SFTRG_USB_CBU0;
        bit_valu = GPO_SFTRG_USB_CBU0;
    }
    else if (command == FAPI_BOARD_POWER_OFF)
    {
        bit_mask = GPO_SFTRG_USB_CBU0;
//        bit_valu = 0;
    }

    if (index) {
        bit_mask = bit_mask << 1;
        bit_valu = bit_valu << 1;
    }
    if (bit_mask)
    {
        FMC_H61TB01_CbsfPtr->WriteExtLatch(bit_mask, bit_valu);
    }


    FREG_GPIO_SetDataOut(GPIO_USB_CBU0, 1);
    FREG_GPIO_SetDataOut(GPIO_USB_CBU1, 1);

    

    return  FAPI_OK;
}

/*!
********************************************************************************
** \brief Controls the reset-signal and power-supply of the harddisk.
**
** This function controls the reset-signal and power-supply of the harddisk.
**
** \param index   Index of the harddisk to be controlled.
** \param Ctrl    Type of reset signal (#FAPI_BOARD_ControlCommandT).
** \param optPtr  Optional parameters, if command equals FAPI_BOARD_POLL_IRQ_STATUS
**                optPtr must point to an uint32_t variable where the resulting
**                status will be written into.
**
** \return
** - #FAPI_ERR_FEATURE_NOT_SUPPORTED if there is no function registered
** - any other value is the result of the registered function call
********************************************************************************
*/
static int32_t BOARD_FMC_H61PDKV1_ControlHdResetPower
               ( int32_t index, FAPI_BOARD_ControlCommandT Ctrl, void* optPtr )
{
    index = index;
    Ctrl  = Ctrl;
    optPtr  = optPtr;
    return FAPI_ERR_FEATURE_NOT_SUPPORTED;
}

/*!
********************************************************************************
** \brief Controls a specific SCART pin.
**
** This function controls a specific SCART pin.
**
** \param index   Index of the Scart connector to be controlled.
** \param command Type of scart pin command (FAPI_BOARD_ScartCommandT).
** \param optPtr  Optional parameters (not used yet).
**
** \return
** - FAPI_ERR_FEATURE_NOT_SUPPORTED if there is no function registered
** - any other value is the result of the registered function call
********************************************************************************
*/

static int32_t BOARD_FMC_H61PDKV1_ControlScartPin8A16
                ( int32_t index, FAPI_BOARD_ScartCommandT command, void* optPtr )
{
    uint32_t bit_mask;
    uint32_t bit_valu;
    index = index;

//        FAPI_BOARD_HW_DIAGNOSTIC(DEBSTR_ControlScart, FAPI_BOARD_DIAG_INFO);
    optPtr = optPtr;
    if (index>0)
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;

    if      (command == FAPI_BOARD_SCART_MODE_OFF)
    {
//        BTLIB_IO_PrStrHexStr(0, (const char_t*)"{off}\n", 0, 0);
        ActPin8State = (uint8_t) command;
        bit_mask = GPO_SFTRG_SCART_PIN8_12V|GPO_SFTRG_SCART_PIN8_5V;  // involved bits
        bit_valu = GPO_SFTRG_SCART_PIN8_12V|GPO_SFTRG_SCART_PIN8_5V;  // set this one
    }
    else if (command == FAPI_BOARD_SCART_MODE_16_9)
    {
//        BTLIB_IO_PrStrHexStr(0, (const char_t*)"{16/9}\n", 0, 0);
        ActPin8State = (uint8_t) command;
        bit_mask = GPO_SFTRG_SCART_PIN8_12V|GPO_SFTRG_SCART_PIN8_5V;   // involved bits
        bit_valu = GPO_SFTRG_SCART_PIN8_5V;                            // set these
    }
    else if (command == FAPI_BOARD_SCART_MODE_4_3)
    {
//        BTLIB_IO_PrStrHexStr(0, (const char_t*)"{4/3}\n", 0, 0);
        ActPin8State = (uint8_t) command;
        bit_mask = GPO_SFTRG_SCART_PIN8_12V|GPO_SFTRG_SCART_PIN8_5V;  // involved bits
        bit_valu = 0;                                                   //
    }

    else if (command == FAPI_BOARD_SCART_VIDEO_OUTPUT_CVBS)
    {
        ActPin16State = (uint8_t) command;
        bit_mask = GPO_SFTRG_SCART_PIN16;                               // involved bits
        bit_valu = GPO_SFTRG_SCART_PIN16;                               // set to '1'
    }
    else if (command == FAPI_BOARD_SCART_VIDEO_OUTPUT_RGB)
    {
        ActPin16State = (uint8_t) command;
        bit_mask = GPO_SFTRG_SCART_PIN16;                               // involved bits
        bit_valu = 0;                                                    // set to '0'
    }
    else
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;

    FMC_H61TB01_CbsfPtr->WriteExtLatch(bit_mask, bit_valu);
    return FAPI_OK;
}

/*!
********************************************************************************
** \brief Controls the 'System Power Mode'
**
** HW-Platform specific controls depending on 'System Power-Mode'
**
** \param index   Optional parameter, not used here
** \param command Enumerate for specific power-mode
** \param optPtr  Optional parameters (not used yet).
**
** \return
** - FAPI_ERR_FEATURE_NOT_SUPPORTED if there is no function registered
** - any other value is the result of the registered function call
********************************************************************************
*/
static int32_t BOARD_FMC_H61PDKV1_ControlSystemPowerMode
                ( int32_t index, FAPI_BOARD_PowerModeCmdT command, void* optPtr )
{
    FAPI_BOARD_ScartCommandT  befcmd;
    optPtr = optPtr;

    if (index>0)
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;

    if      (command == FAPI_BOARD_POWER_FULL_ACTIVE)               //== activ power-mode ==
    {
        FAPI_BOARD_DIAGNOSTIC_GENERALPURPOSE(FBD_ID, (void*) DEBSTR_SysPwrActive, ActPin8State, 0);
        befcmd = (FAPI_BOARD_ScartCommandT) BefPOffPin8State;
        BOARD_FMC_H61PDKV1_ControlScartPin8A16(0,befcmd,0);
        befcmd = (FAPI_BOARD_ScartCommandT) BefPOff16State;
        BOARD_FMC_H61PDKV1_ControlScartPin8A16(0,befcmd,0);
    }
    else if (command == FAPI_BOARD_POWER_MAX_OFF)                   //== lowest power-mode ==
    {
        BefPOffPin8State = ActPin8State;
        BefPOff16State   = ActPin16State;
        FAPI_BOARD_DIAGNOSTIC_GENERALPURPOSE(FBD_ID, (void*) DEBSTR_SysPwrOff, ActPin16State, 0);
        (void) BOARD_FMC_H61PDKV1_ControlScartPin8A16(0,FAPI_BOARD_SCART_MODE_OFF,0);
        (void) BOARD_FMC_H61PDKV1_ControlScartPin8A16(0,FAPI_BOARD_SCART_VIDEO_OUTPUT_CVBS,0);
    }
    else
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;

    return FAPI_OK;
}


/*!
********************************************************************************
** \brief register the 'MB86H61Dk01' HW-platform board-information
**
**  Registers the 'MB86H61Dk01' HW-platform board-information structure
**  (BOARDS_HwPlatformInfosT)
**  the 'init/detect function' hook (FAPI_BOARD_InitDetectHookT)
**  and other 'control function' hooks into the
**  generic 'board interface control' (see fapi/board/board_ifc.c)
**
** \note This function can be called even before FAPI-drivers are initialized.
**  Actually some init-parameters for the initialization of FAPI-drivers later-on,
**  might be obtained be retreiving the registered information/configuration-data
**
** \param registerFlags          set to '0' (for future use)
** \param optRegisterDataPtr      pointer to optional data (future use)
**
** \return  usually FAPI_OK
**
********************************************************************************
*/
int32_t BOARD_FMC_H61PDKV1_BoardIfc_Register(uint32_t registerFlags, void *optRegisterDataPtr)
{
    registerFlags         = registerFlags;          // not used yet
    optRegisterDataPtr    = optRegisterDataPtr;     // not used yet

//    FME_H61TB01_CbsfPtr = BRDCOM_GetIoExtFncsPtr_FapiUpiAta(); // would be like 'old H60'
                                                            // -> should not work, but also not crash !!!
    FMC_H61TB01_CbsfPtr = BRDCOM_GetIoExtFncsPtr_ExtSerShiftRegs();    // we use these functions H61 Testbaord
    FAPI_BOARD_SetBoardInfo( (void*) &FmcMb86H61PDkv1_HwInfo );

    FAPI_BOARD_SetBoardInitDetectHook  ( BOARD_FMC_H61PDKV1_BoardIfc_InitDetect );
    FAPI_BOARD_SetControlTestLedKeyHook( BOARD_FMC_H61PDKV1_ControlAlifeLed     );
    FAPI_BOARD_SetControlFeResetHook   ( BOARD_FMC_H61PDKV1_ControlFeResetPower );
    FAPI_BOARD_SetControlHdResetHook   ( BOARD_FMC_H61PDKV1_ControlHdResetPower );
    FAPI_BOARD_SetControlScartHook     ( BOARD_FMC_H61PDKV1_ControlScartPin8A16 );
    FAPI_BOARD_SetControlPowerMode     ( BOARD_FMC_H61PDKV1_ControlSystemPowerMode );
    FAPI_BOARD_SetControlUsbPhy        ( BOARD_FMC_H61PDKV1_CtrlUsbPhy );
//  FAPI_BOARD_SetControlTsPathHook    (  );    // NOTE: see 'mb86h61_dk01_ci.c'

    BOARD_FMC_H61PDKV1_FrontPanel_Register();

//    // NOTE: following will just-preconfigure GPIOs before 'FAPI_GPIO_Init()' is called
//    BOARD_FMC_H61PDKV1_RegisterGpioPins(BOARDS_GPIO_REGISTER_GPIO_FNC);

    // NOTE: this will call 'FAPI_GPIO_SetPinFunction' but also 'directly configure' the pins
    BOARD_FMC_H61PDKV1_RegisterGpioPins(BRDCOM_BC_RegisterGpioMode_PreInit);

    return FAPI_OK;
}


/*!
********************************************************************************
** \brief initialization of 'MB86H61Dk01' HW-platform specific control functions
**
**  this function will perform eventually necessary initializations
**  before board-control functions can be used
** \note usually this function should be called after FAPI-drivers were initialized
**  so FAPI-drivers (e.g. like I2C-driver) could be used here.
**  Ideally, this function should be able to verify or detect,
**  that the 'present HW' really matches to the board-specific driver.
**  However, for this 'Fujitsu Dev.Kit', this feature is not implemented yet
**  and the return-code always signals that the detected HW is fine.
**  (even the SW is currently accidently running on a different HW-platform)
**
** \param initDetectFlags     enum for special install conditions, not considered here
** \param optInitDataPtr      pointer to optional data, not used here
**
** \return
** - BOARDS_INSTALL_ASSUME_OK, always
**
********************************************************************************
*/
static int32_t BOARD_FMC_H61PDKV1_BoardIfc_InitDetect
                     (uint32_t initDetectFlags, void *optInitDataPtr)
{
    initDetectFlags = initDetectFlags;
    optInitDataPtr  = optInitDataPtr;

    FMC_H61TB01_CbsfPtr->InitAccessExtLatch((void*) &H61DkSerIoExtInfos);
    FMC_H61TB01_CbsfPtr->WriteExtLatch(LATCH_INIT_MASK, LATCH_INIT_BITS);

#ifndef TEST_PATA_INTERFACE_VIA_J900_ADAPTER
    (void) BOARD_FMC_H61PDKV1_SetupCiLibHwInterfaceFunctions();
#endif

    (void) BOARDS_CONTROL_FP(0, BOARDS_FP_SET_PWR_MODE_OPERATION, (void*) 0); // activate FrontPanel

#ifdef TEST_PATA_INTERFACE_VIA_J900_ADAPTER
    // NOTE: this will re-configure different to 'BOARD_FME_H61DK01_SetupCiLibHwInterfaceFunctions'
    BRDCOM_BC_DirectPreInitUpiCs2ViaAtaControler();
    BOARD_FMC_H61PDKV1_ConfigGpiosPAtaAdapter(BRDCOM_BC_DirectConfigureGpioPin);
#endif

    return FAPI_OK;
}
#else

//------------------------------------------------------------
//-- Platform Info/Configuration-Data for Application-SW    --
//------------------------------------------------------------
static const char_t Manufact[] = "*** PIXELA *** (^_^)v";
static const char_t Model[]    = "PIX-WLANTUNER";
static const FAPI_BOARD_HwPlatformInfosT  FmcMb86H61PDkv1_HwInfo =
{
    FAPI_BOARD_HWCONFIG_VERSION,//!< MagicAndVersion
    Manufact,                   //!< pHwManufactString
    Model,                      //!< pHwModelString
    NULL,                       //!< pOptionalAdditionalInfo
    0,                          //!< flagUsbHostOnly
    FAPI_BOARD_BITFLAG_SFLASH,  //!< typeOfFlashAvailable
    0,                          //!< AudioConfigurationInfoA
    0,                          //!< AudioConfigurationInfoB
    1,                          //!< numberOfAvlFrontEnds
    0,                          //!< numberOfAvlCiSlots
	
    {
        { 0,  //!< defaultFeType
			FAPI_BOARD_RSRC_TSIN_FE1 | FAPI_BOARD_RSRC_TSIN_SERIAL, //!< tsPathInputConfig
			0,								//!< flagCiPathAval
			FAPI_BOARD_RSRC_CIIN_NONE },   //!< inputOptionsCiPath
        { 0,  //!< defaultFeType
			FAPI_BOARD_RSRC_TSIN_FE2 | FAPI_BOARD_RSRC_TSIN_SERIAL, //!< tsPathInputConfig
			0,                        //!< flagCiPathAval
			FAPI_BOARD_RSRC_CIIN_NONE }  //!< inputOptionsCiPath
    },
    0,           //!< flagHardDiskInterface
    0,                              //!< flgDdr2MemoryInfos
    FAPI_BOARD_VID_AMP_C0           //!< video/scaler optimize info
};

BOARDS_CommonSupportFunctionsT  *FMC_H61TB01_CbsfPtr;   //


static int32_t BOARD_FMC_H61PDKV1_ControlAlifeLed
( int32_t index, FAPI_BOARD_LedCommandT command, void* optPtr )
{
    return  FAPI_ERR_FEATURE_NOT_SUPPORTED;
}

static int32_t BOARD_FMC_H61PDKV1_ControlFeResetPower
( int32_t index, FAPI_BOARD_ControlCommandT command, void* optPtr )
{
    return  FAPI_ERR_FEATURE_NOT_SUPPORTED;
}

static int32_t BOARD_FMC_H61PDKV1_CtrlUsbPhy
( int32_t index, uint32_t ctrl, void* optPtr )
{
    return  FAPI_ERR_FEATURE_NOT_SUPPORTED;
}

static int32_t BOARD_FMC_H61PDKV1_ControlHdResetPower
( int32_t index, FAPI_BOARD_ControlCommandT Ctrl, void* optPtr )
{
    return FAPI_ERR_FEATURE_NOT_SUPPORTED;
}

static int32_t BOARD_FMC_H61PDKV1_ControlScartPin8A16
( int32_t index, FAPI_BOARD_ScartCommandT command, void* optPtr )
{
    return FAPI_ERR_FEATURE_NOT_SUPPORTED;
}

static int32_t BOARD_FMC_H61PDKV1_ControlSystemPowerMode
( int32_t index, FAPI_BOARD_PowerModeCmdT command, void* optPtr )
{
    return FAPI_ERR_FEATURE_NOT_SUPPORTED;
}

static int32_t BOARD_FMC_H61PDKV1_BoardIfc_InitDetect
(uint32_t initDetectFlags, void *optInitDataPtr)
{
    return FAPI_OK;
}

int32_t BOARD_FMC_H61PDKV1_BoardIfc_Register(uint32_t registerFlags, void *optRegisterDataPtr)
{
    registerFlags         = registerFlags;          // not used yet
    optRegisterDataPtr    = optRegisterDataPtr;     // not used yet
	
	//    FME_H61TB01_CbsfPtr = BRDCOM_GetIoExtFncsPtr_FapiUpiAta(); // would be like 'old H60'
	// -> should not work, but also not crash !!!
	//    FMC_H61TB01_CbsfPtr = BRDCOM_GetIoExtFncsPtr_ExtSerShiftRegs();    // we use these functions H61 Testbaord // not necessary for PIX-WLANTUNER
    FAPI_BOARD_SetBoardInfo( (void*) &FmcMb86H61PDkv1_HwInfo );
	
    FAPI_BOARD_SetBoardInitDetectHook  ( BOARD_FMC_H61PDKV1_BoardIfc_InitDetect );
    FAPI_BOARD_SetControlTestLedKeyHook( BOARD_FMC_H61PDKV1_ControlAlifeLed     ); // not necessary for PIX-WLANTUNER
    FAPI_BOARD_SetControlFeResetHook   ( BOARD_FMC_H61PDKV1_ControlFeResetPower ); // not necessary for PIX-WLANTUNER
    FAPI_BOARD_SetControlHdResetHook   ( BOARD_FMC_H61PDKV1_ControlHdResetPower ); // not necessary for PIX-WLANTUNER
    FAPI_BOARD_SetControlScartHook     ( BOARD_FMC_H61PDKV1_ControlScartPin8A16 ); // not necessary for PIX-WLANTUNER
    FAPI_BOARD_SetControlPowerMode     ( BOARD_FMC_H61PDKV1_ControlSystemPowerMode ); // not necessary for PIX-WLANTUNER
    FAPI_BOARD_SetControlUsbPhy        ( BOARD_FMC_H61PDKV1_CtrlUsbPhy ); // not necessary for PIX-WLANTUNER
	//  FAPI_BOARD_SetControlTsPathHook    (  );    // NOTE: see 'mb86h61_dk01_ci.c'
	
	//    BOARD_FMC_H61PDKV1_FrontPanel_Register(); // not necessary for PIX-WLANTUNER
	
	//    // NOTE: following will just-preconfigure GPIOs before 'FAPI_GPIO_Init()' is called
	//    BOARD_FMC_H61PDKV1_RegisterGpioPins(BOARDS_GPIO_REGISTER_GPIO_FNC);
	
    // NOTE: this will call 'FAPI_GPIO_SetPinFunction' but also 'directly configure' the pins
    BOARD_FMC_H61PDKV1_RegisterGpioPins(BRDCOM_BC_RegisterGpioMode_PreInit);
	
    return FAPI_OK;
}



#endif

