/*!
********************************************************************************
**
** \file        ./boards/fmc/mb86h61_pdkv1/mb86h61_pdkv1.c
**
** \version     $Id: mb86h61_pdkv1.c $
**
** \brief       FME devkit specific board service function.
**
** This file contains the implementation of the FME devkit specific board
** services functions.
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU MICROELECTRONICS
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2006-2010 by Fujitsu Semiconductor Europe GmbH
** (C) Copyright 2010 by Fujitsu Semiconductor Europe GmbH
**
********************************************************************************
*/

#include <fapi/reg_gpio.h>
#include <fapi/drv_gpio.h>
#include <fapi/board_ifc.h>

#include <boards/boards_common.h>
#include <boards/mb86h61_pdkv1.h>

/*lint -save -e322 -e766 -e961 */
#if 0
#include <fe/fe.h>
#include "../../../subsystems/fe/src/fe_priv.h"
#include "../../../subsystems/fe/ce6353/src/ce6353.h"
#include "../../../subsystems/fe/dsf89000/src/dsf89000.h"
#include "../../../subsystems/fe/m88ds3k/src/m88ds3k.h"
#include "../../../subsystems/fe/mb86668/src/mb86668.h"
#include "../../../subsystems/fe/mb86667/src/mb86667.h"
#include "../../../subsystems/fe/fedirect/src/fedirect.h"
#endif

/*!
********************************************************************************
** \brief Detects the various types of frontends
**
** This function detects the various types of frontends supported for this board.
**
** \param index   Index of the frontend to be detected.
** \param optPtr  Frontend handle into which the detected frontend sets its
**                specific function implementations
**
** \return
** - FAPI_ERR_FEATURE_NOT_SUPPORTED if the given index is not supported
** - FAPI_ERR_NOT_OPEN if no frontend was detected
** - FAPI_OK if a frontend was detected
********************************************************************************
*/
static int32_t BOARD_FMC_H61PDKV1_ControlFeDetect( int32_t index,
                                                       void* optPtr )
{
    optPtr = optPtr;

    //-- check supported indices --
    if( (index > 1) || (index < 0) )
        return FAPI_ERR_FEATURE_NOT_SUPPORTED;

 #if 0
    if(CE6353_DTT75404_Detect     ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
        return FAPI_OK;
    //if(CE6353_DTT75404_IDTV_Detect((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
    //    return FAPI_OK;
    if(CE6353_TD911_Detect        ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
        return FAPI_OK;
    //if(CE6353_TD1316_IDTV_Detect  ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
    //    return FAPI_OK;
    if(DSF89000_Detect            ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
        return FAPI_OK;
//    if(M88DS3K_Detect             ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
//        return FAPI_OK;

    if(M88DS3K_HE6306_Detect             ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
        return FAPI_OK;
    //if(MB86668_CD1316_Detect      ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
    //    return FAPI_OK;
    //if(MB86668_CD1600_Detect      ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
    //    return FAPI_OK;
    //if(MB86668_DCT70602_Detect    ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
    //    return FAPI_OK;
    if(MB86668_DCT70701_Detect    ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
        return FAPI_OK;
    if(MB86667_S7HZ1402_Detect    ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
        return FAPI_OK;
    if(FE_DIRECT_Detect           ((uint8_t)index, (feHandleT *)optPtr) == FAPI_OK)
        return FAPI_OK;

#endif
    return FAPI_ERR_NOT_OPEN;
}

/*!
********************************************************************************
** \brief register the 'MB86H61PDkv1' HW-platform frontend support
**
**  Installs the 'MB86H61PDkv1' frontend detect function
**  into the generic 'board control interface (BIC)' (see fapi/board/board_ifc.c)
**
** \param instCondFlag  enum for special install conditions, not considered here
** \param optInstallDataPtr      pointer to optional data, not used here
**
** \return
** - BOARDS_INSTALL_ASSUME_OK, always
**
********************************************************************************
*/
int32_t BOARD_FMC_H61PDKV1_RegisterFrontends
                               (uint32_t instCondFlag, void *optInstallDataPtr)
{
    instCondFlag      = instCondFlag;       // not used yet
    optInstallDataPtr = optInstallDataPtr;
    FAPI_BOARD_SetControlFeDetectHook  ( BOARD_FMC_H61PDKV1_ControlFeDetect );

    return FAPI_OK;
}
