/*!
********************************************************************************
**
** \file        ./boards/fme_devkit/mb86h61_dk01_min.c
**
** \version     $Id: mb86h61_dk01_min.c 2153 2010-06-09 12:23:22Z wlaris $
**
** \brief       FME devkit 'minimalistic initialization'
**
** This file contains the implementation of the FME devkit specific board
** services functions.
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU MICROELECTRONICS
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2006-2010 by Fujitsu Semiconductor Europe GmbH
** (C) Copyright 2010 by Fujitsu Semiconductor Europe GmbH
**
********************************************************************************
*/

#include <fapi/board_ifc.h>

#include <boards/boards_common.h>
#include <boards/boards_com_dirx.h>
#include <boards/mb86h61_pdkv1.h>

/*lint -save -e923 -e778 */

static const BOARDS_FMEDEF_HwShortInfoT  ShortInfo_H61PDKV1 = {
// HwManufacturerStrPtr, HwModelStrPtr
   "Fujitsu Shanghai",            "H61PDKV1" };


/*!
********************************************************************************
** \brief register the 'MB86H61Dk01' HW-platform board-information
**
**  Registers the 'MB86H61PDkV1' HW-platform board-information structure
**  (BOARDS_HwPlatformInfosT)
**  the 'init/detect function' hook (FAPI_BOARD_InitDetectHookT)
**  and other 'control function' hooks into the
**  generic 'board interface control' (see fapi/board/board_ifc.c)
**
** \note This function can be called even before FAPI-drivers are initialized.
**  Actually some init-parameters for the initialization of FAPI-drivers later-on,
**  might be obtained be retreiving the registered information/configuration-data
**
** \param registerFlags          set to '0' (for future use)
** \param optRegisterDataPtr      pointer to optional data (future use)
**
** \return  usually FAPI_OK
**
********************************************************************************
*/
int32_t BOARD_FMC_H61DKV1_BoardRegister_Mini(uint32_t registerFlags, void *optRegisterDataPtr)
{
    registerFlags         = registerFlags;          // not used yet
    optRegisterDataPtr    = optRegisterDataPtr;     // not used yet

    BRDCOM_BC_ShortHwInfoPtr = (BOARDS_FMEDEF_HwShortInfoT*) &ShortInfo_H61PDKV1;
    BOARD_FMC_H61PDKV1_RegisterGpioPins(BRDCOM_BC_DirectConfigureGpioPin);

    return FAPI_OK;
}

