/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FAPI BM configuration code.
*/

#include "fapex_stdhdr.h"
#include "fapi/drv_bm.h"
#include "fapi/drv_intr.h"
#include "fapi/reg_bm.h"
#include "fapex/sys_config.h"
#include "fapex_config_cmn.h"
#include "fapex_bm_func.h"

/*--------------------------------------------------------*/
/* Global Data                                            */
/*--------------------------------------------------------*/
FAPEX_BM_FuncStrT bm_funcStr;

/*--------------------------------------------------------*/
/* Local Data                                             */
/*--------------------------------------------------------*/
#ifdef __LINUX__
static struct semaphore bm_sem;
static uint32_t         result_isr;
#endif

/*--------------------------------------------------------*/
/* ISR.                                                   */
/*--------------------------------------------------------*/
FAPICORE_ISRRET_T FAPEXCFG_BM_Isr(int arg1, void *ptr)
{
#ifndef __LINUX__
    uint32_t result_isr;
    void *pBsr;
#endif

    /* Interrupt service routine for Isr */
    result_isr = bm_funcStr.Isr();

    if(result_isr != 0) {
#ifndef __LINUX__
        /* register bsr */
        pBsr = (void*)FAPEXCFG_INTR_IsBsrFunction(FAPI_INTR_BM);
        if(pBsr != NULL){
            FAPEXCFG_INTR_RegisterBSR(pBsr);
        }
    }
#else
        /* wake up for FAPEXCFG_BM_BsrThread() */
        up(&bm_sem);
    }
#endif
    FAPICORE_ISRRETVAL(result_isr)
}

/*--------------------------------------------------------*/
/* BSR.                                                   */
/*--------------------------------------------------------*/
#ifndef __LINUX__
FAPICORE_BSRRET_T FAPEXCFG_BM_Bsr(unsigned long arg)
{
    bm_funcStr.Bsr(arg);
}

#define BMBSR_FUNCNAME  FAPEXCFG_BM_Bsr

#else
static int FAPEXCFG_BM_BsrThread(void *arg)
{
    int32_t  ret;
    struct sched_param  schedparam = { .sched_priority = 90 };

    sched_setscheduler(current, SCHED_FIFO, &schedparam);

    lock_kernel();
    daemonize("fapex_bm_bsr");
    unlock_kernel();

    while(1) {
        /* wait for FAPEXCFG_DMA_Bsr() */
        ret = down_interruptible(&bm_sem);
        if(ret != 0) {
            break;
        }

        bm_funcStr.Bsr(result_isr);
    }
    return 0;
}
#endif /* !__LINUX__ */

/*--------------------------------------------------------*/
/* INIT.                                                  */
/*--------------------------------------------------------*/
int32_t FAPEXCFG_BM_Init(void)
{
    int32_t ret = 0;
#ifdef __LINUX__
    int32_t bm_pid;
#endif

    bm_funcStr.Init                          = FAPI_BM_Init;
    bm_funcStr.Exit                          = FAPI_BM_Exit;
    bm_funcStr.Isr                           = FAPI_BM_Isr;
    bm_funcStr.Bsr                           = FAPI_BM_Bsr;
    bm_funcStr.Open                          = FAPI_BM_Open;
    bm_funcStr.Close                         = FAPI_BM_Close;
    bm_funcStr.GetBufferFillLevel            = FAPI_BM_GetBufferFillLevel;
    bm_funcStr.GetBufferReadPointerPosition  = FAPI_BM_GetBufferReadPointerPosition;
    bm_funcStr.GetBufferWritePointerPosition = FAPI_BM_GetBufferWritePointerPosition;
    bm_funcStr.GetBufferReadPointer          = FAPI_BM_GetBufferReadPointer;
    bm_funcStr.GetBufferWritePointer         = FAPI_BM_GetBufferWritePointer;
    bm_funcStr.GetHardwareControlledBlock    = FAPI_BM_GetHardwareControlledBlock;
    bm_funcStr.MoveBufferReadPointer         = FAPI_BM_MoveBufferReadPointer;
    bm_funcStr.MoveBufferWritePointer        = FAPI_BM_MoveBufferWritePointer;
    bm_funcStr.SetBufferReadPointer          = FAPI_BM_SetBufferReadPointer;
    bm_funcStr.SetBufferWritePointer         = FAPI_BM_SetBufferWritePointer;
    bm_funcStr.SetHardwareControlledBlock    = FAPI_BM_SetHardwareControlledBlock;
    bm_funcStr.SetReadEnable                 = FAPI_BM_SetReadEnable;
    bm_funcStr.Write                         = FAPI_BM_Write;
    bm_funcStr.WriteMtxCol                   = FAPI_BM_WriteMtxCol;
    bm_funcStr.Read                          = FAPI_BM_Read;
    bm_funcStr.SetAudCallback                = FAPI_BM_SetAudCallback;
    //bm_funcStr.GetAudChannelAssignment       = FAPI_BM_GetAudChannelAssignment;
    bm_funcStr.StartAudFiltering             = FAPI_BM_StartAudFiltering;
    bm_funcStr.StopAudFiltering              = FAPI_BM_StopAudFiltering;
    bm_funcStr.ClearAudFiltering             = FAPI_BM_ClearAudFiltering;
    bm_funcStr.GetInfo                       = FAPI_BM_GetInfo;
    bm_funcStr.SetBufferPurpose              = FAPI_BM_SetBufferPurpose;
    bm_funcStr.GetAddressByBufferPurpose     = FAPI_BM_GetAddressByBufferPurpose;
    bm_funcStr.GetBlockIndexByBufferPurpose  = FAPI_BM_GetBlockIndexByBufferPurpose;
    bm_funcStr.GetSizeByBufferPurpose        = FAPI_BM_GetSizeByBufferPurpose;
    bm_funcStr.SetHighLevelIrq               = FAPI_BM_SetHighLevelIrq;
    bm_funcStr.GetHighLevelIrq               = FAPI_BM_GetHighLevelIrq;
    bm_funcStr.SetLowLevelIrq                = FAPI_BM_SetLowLevelIrq;
    bm_funcStr.GetLowLevelIrq                = FAPI_BM_GetLowLevelIrq;
    bm_funcStr.SetReadSwap                   = FAPI_BM_SetReadSwap;
    bm_funcStr.SetWriteSwap                  = FAPI_BM_SetWriteSwap;
    bm_funcStr.SetOverflowIrq                = FAPI_BM_SetOverflowIrq;
    bm_funcStr.Reconfigure                   = FAPI_BM_Reconfigure;

    /* initialize function of FAPI RAL */
    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_BM] == 0){
        FREG_BM_Init();
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_BM] = 1;
    }

    bm_funcStr.SetHardwareControlledBlock(FAPI_BM0);

    FAPI_SYS_BM_SET_BOARD_PARAMS

    /* setup interrupt */
#ifndef __LINUX__
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_BM,
                               FAPEXCFG_BM_Isr,
                               (FAPICORE_BSRRET_T (*)(unsigned long))BMBSR_FUNCNAME,
                               SA_INTERRUPT, (uint8_t*)"BM",
                               8192, 1024 );
#else
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_BM,
                               FAPEXCFG_BM_Isr, NULL,
                               SA_INTERRUPT, (uint8_t*)"BM",
                               8192, 1024 );

    /* create kernel thread because FAPI_BM_Bsr() use FAPI_SYS_SLEEP(). */
    sema_init(&bm_sem, 0);
    bm_pid = kernel_thread(FAPEXCFG_BM_BsrThread, NULL, SIGCHLD | CLONE_KERNEL);
    if(bm_pid < 0) {
        return bm_pid;
    }
#endif

    /* initialize function of FAPI CDL (bm) */
    ret = bm_funcStr.Init();

    return ret;
}

/*--------------------------------------------------------*/
/* CLEANUP.                                               */
/*--------------------------------------------------------*/
void FAPEXCFG_BM_Cleanup(void)
{
    /* unregister interrupt */
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_BM);
//    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_BM23);

    bm_funcStr.Exit();
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
EXPORT_SYMBOL(bm_funcStr);

EXPORT_SYMBOL(FAPEXCFG_BM_Init);
EXPORT_SYMBOL(FAPEXCFG_BM_Cleanup);
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
