/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FAPI I2C configuration code.
*/

#include "fapex_stdhdr.h"
#include "fapi/drv_i2c.h"
#include "fapi/drv_intr.h"
#include "fapi/reg_i2c.h"
#include "fapex/sys_config.h"
#include "fapex_config_cmn.h"
#include "fapex_i2c_func.h"

/*--------------------------------------------------------*/
/* Global Data                                            */
/*--------------------------------------------------------*/
FAPEX_I2C_FuncStrT i2c_funcStr;

/*--------------------------------------------------------*/
/* ISR.                                                   */
/*--------------------------------------------------------*/
FAPICORE_ISRRET_T FAPEXCFG_I2C_Isr(int arg1, void *ptr)
{
    uint32_t result;
//  void *pBsr;

    /* Interrupt service routine for Isr */
    result = i2c_funcStr.Isr();

    FAPICORE_ISRRETVAL(result)
}

/*--------------------------------------------------------*/
/* INIT.                                                  */
/*--------------------------------------------------------*/
int32_t FAPEXCFG_I2C_Init(void)
{
    int32_t ret = 0;

    i2c_funcStr.Init = FAPI_I2C_Init;
    i2c_funcStr.Exit = FAPI_I2C_Exit;
    i2c_funcStr.Isr = FAPI_I2C_Isr;
    i2c_funcStr.Open = FAPI_I2C_Open;
    i2c_funcStr.Reset = FAPI_I2C_Reset;
    i2c_funcStr.Close = FAPI_I2C_Close;
    i2c_funcStr.CancelProcess = FAPI_I2C_CancelProcess;
    i2c_funcStr.Process = FAPI_I2C_Process;
    i2c_funcStr.Read = FAPI_I2C_Read;
    i2c_funcStr.Write = FAPI_I2C_Write;
    i2c_funcStr.SetOperationMode = FAPI_I2C_SetOperationMode;
    i2c_funcStr.SetProtocol = FAPI_I2C_SetProtocol;
    i2c_funcStr.SetMaxHandleCount = FAPI_I2C_SetMaxHandleCount;

    /* initialize function of FAPI RAL */
    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_I2C] == 0){
        FREG_I2C_Init();
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_I2C] = 1;
    }

    /* setup interrupt */
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_I2C, FAPEXCFG_I2C_Isr, NULL,
                               SA_INTERRUPT, (uint8_t*)"I2C0", 0, 0 );
 
    /* initialize function of FAPI CDL (i2c) */
    ret = i2c_funcStr.Init();
    
    return ret;
}

/*--------------------------------------------------------*/
/* CLEANUP.                                               */
/*--------------------------------------------------------*/
void FAPEXCFG_I2C_Cleanup(void)
{
    /* unregister interrupt */
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_I2C);
//    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_I2C1);
//    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_I2C2);

    i2c_funcStr.Exit();
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
EXPORT_SYMBOL(i2c_funcStr);

EXPORT_SYMBOL(FAPEXCFG_I2C_Init);
EXPORT_SYMBOL(FAPEXCFG_I2C_Cleanup);
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
