/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FAPI ICC configuration code.
*/

#include "fapex_stdhdr.h"
#include "fapi/drv_icc.h"
#include "fapi/drv_intr.h"
#include "fapi/reg_icc.h"
#include "fapex/sys_config.h"
#include "fapex_config_cmn.h"
#include "fapex_icc_func.h"

/*--------------------------------------------------------*/
/* Global Data                                            */
/*--------------------------------------------------------*/
FAPEX_ICC_FuncStrT icc_funcStr;

/*--------------------------------------------------------*/
/* ISR.                                                   */
/*--------------------------------------------------------*/
FAPICORE_ISRRET_T FAPEXCFG_ICC_Isr(int arg1, void *ptr)
{
    uint32_t result;

    /* Interrupt service routine for Isr */
    result = icc_funcStr.Isr();

    FAPICORE_ISRRETVAL(result)
}

/*--------------------------------------------------------*/
/* BSR.                                                   */
/*--------------------------------------------------------*/
/* nothing */

/*--------------------------------------------------------*/
/* INIT.                                                  */
/*--------------------------------------------------------*/
int32_t FAPEXCFG_ICC_Init(void)
{
    int32_t ret = 0;

    icc_funcStr.Init              = FAPI_ICC_Init;
    icc_funcStr.Exit              = FAPI_ICC_Exit;
    icc_funcStr.Isr               = FAPI_ICC_Isr;
    icc_funcStr.Open              = FAPI_ICC_Open;
    icc_funcStr.Close             = FAPI_ICC_Close;
    icc_funcStr.ColdReset         = FAPI_ICC_ColdReset;
    icc_funcStr.WarmReset         = FAPI_ICC_WarmReset;
    icc_funcStr.SetPPS            = FAPI_ICC_SetPPS;
    icc_funcStr.GetAtrParams      = FAPI_ICC_GetAtrParams;
    icc_funcStr.GetBaudRate       = FAPI_ICC_GetBaudRate;
    icc_funcStr.GetClockFrequency = FAPI_ICC_GetClockFrequency;
    icc_funcStr.GetProtocol       = FAPI_ICC_GetProtocol;
    icc_funcStr.GetSlotStatus     = FAPI_ICC_GetSlotStatus;
    icc_funcStr.SetTDA8004TEnable = FAPI_ICC_SetTDA8004TEnable;
    icc_funcStr.SetClkDivider     = FAPI_ICC_SetClkDivider;
    icc_funcStr.SetProtocol       = FAPI_ICC_SetProtocol;
    icc_funcStr.SetBaudRate       = FAPI_ICC_SetBaudRate;
    icc_funcStr.SetCardNotifier   = FAPI_ICC_SetCardNotifier;
    icc_funcStr.SetClockFrequency = FAPI_ICC_SetClockFrequency;
    icc_funcStr.SetGuardTime      = FAPI_ICC_SetGuardTime;
    icc_funcStr.GetGuardTime      = FAPI_ICC_GetGuardTime;
    icc_funcStr.WriteBytes        = FAPI_ICC_WriteBytes;
    icc_funcStr.ReadByte          = FAPI_ICC_ReadByte;
    icc_funcStr.GetData           = FAPI_ICC_GetData;
    icc_funcStr.T0Init            = FAPI_ICC_T0_Init;
    icc_funcStr.T0Open            = FAPI_ICC_T0_Open;
    icc_funcStr.T0GetIccHandle    = FAPI_ICC_TO_GetIccHandle;
    icc_funcStr.T0Mode            = FAPI_ICC_T0_Mode;
    icc_funcStr.T0SendCmd         = FAPI_ICC_T0_SendCmd;
    icc_funcStr.T0SetPps          = FAPI_ICC_TO_SetPps;
    icc_funcStr.T1Init            = FAPI_ICC_T1_Init;
    icc_funcStr.T1Mode            = FAPI_ICC_T1_Mode;

    /* initialize function of FAPI RAL */
    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_ICC] == 0){
        FREG_ICC_Init();
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_ICC] = 1;
    }

    FAPI_SYS_ICC_SET_BOARD_PARAMS

    /* setup interrupt */
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_ICC, FAPEXCFG_ICC_Isr, NULL,
                               SA_INTERRUPT, (uint8_t*)"ICC", 0, 0 );
  
    /* initialize function of FAPI CDL (icc) */
    ret = icc_funcStr.Init();

    return ret;
}

/*--------------------------------------------------------*/
/* CLEANUP.                                               */
/*--------------------------------------------------------*/
void FAPEXCFG_ICC_Cleanup(void)
{
    /* unregister interrupt */
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_ICC);
//    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_ICC1);

    icc_funcStr.Exit();
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
EXPORT_SYMBOL(icc_funcStr);

EXPORT_SYMBOL(FAPEXCFG_ICC_Init);
EXPORT_SYMBOL(FAPEXCFG_ICC_Cleanup);
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
