/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FAPI IR configuration code.
*/

#include "fapex_stdhdr.h"
#include "fapi/drv_ir.h"
#include "fapi/drv_ir_nec.h"
#include "fapi/drv_ir_rc5.h"
#include "fapi/drv_intr.h"
#include "fapi/drv_fpc.h"
#include "fapi/reg_irr.h"
#include "fapi/reg_fpc.h"
#include "fapex/sys_config.h"
#include "fapex_config_cmn.h"
#include "fapex_ir_func.h"

/*--------------------------------------------------------*/
/* Global Data                                            */
/*--------------------------------------------------------*/
FAPEX_IR_FuncStrT ir_funcStr;

/*--------------------------------------------------------*/
/* Local Data                                             */
/*--------------------------------------------------------*/
#ifdef __LINUX__
static uint32_t  result;
#endif

/*--------------------------------------------------------*/
/* ISR.                                                   */
/*--------------------------------------------------------*/
FAPICORE_ISRRET_T FAPEXCFG_IR_Isr(int arg1, void *ptr)
{
#ifndef __LINUX__
    uint32_t result;
#endif
    void *pBsr;

    /* Interrupt service routine for Isr */
    result = ir_funcStr.Isr();

    if(result != 0) {
        /* register bsr */
        pBsr = (void*)FAPEXCFG_INTR_IsBsrFunction(FAPI_INTR_FPC_IR);
        if(pBsr != NULL){
            FAPEXCFG_INTR_RegisterBSR(pBsr);
        }
    }
    FAPICORE_ISRRETVAL(result)
}

/*--------------------------------------------------------*/
/* BSR.                                                   */
/*--------------------------------------------------------*/
#ifndef __LINUX__
FAPICORE_BSRRET_T FAPEXCFG_IR_Bsr(unsigned long arg)
{
    ir_funcStr.Bsr(arg);
}

#define IRBSR_FUNCNAME  FAPEXCFG_IR_Bsr

#else
FAPICORE_BSRRET_T FAPEXCFG_IR_Bsr(unsigned long arg)
{
    ir_funcStr.Bsr(result);
}

#define IRBSR_FUNCNAME (&FAPEXCFG_IR_Tasklet)

static DECLARE_TASKLET( FAPEXCFG_IR_Tasklet, FAPEXCFG_IR_Bsr, (unsigned long)&result );
#endif /* !__LINUX__ */

/*--------------------------------------------------------*/
/* INIT.                                                  */
/*--------------------------------------------------------*/
int32_t FAPEXCFG_IR_Init(void)
{
    int32_t ret = 0;

    ir_funcStr.Init           = FAPI_IR_Init;
    ir_funcStr.Exit           = FAPI_IR_Exit;
    ir_funcStr.Isr            = FAPI_IR_Isr;
    ir_funcStr.SetFrequency   = FAPI_IR_SetFrequency;
    ir_funcStr.Open           = FAPI_IR_Open;
    ir_funcStr.Close          = FAPI_IR_Close;
    ir_funcStr.ReadData       = FAPI_IR_ReadData;
    ir_funcStr.GetStatus      = FAPI_IR_GetStatus;
    ir_funcStr.GetKey         = FAPI_IR_GetKey;
    ir_funcStr.Bsr            = FAPI_IR_Bsr;
    ir_funcStr.SetMapTable    = FAPI_IR_SetMapTable;
    ir_funcStr.RemoveMapTable = FAPI_IR_RemoveMapTable;
    ir_funcStr.SetUserProtocol = FAPI_IR_SetUserProtocol;
    ir_funcStr.SetMaskData    = FAPI_IR_SetMaskData;
    ir_funcStr.SetWakeUpKey   = FAPI_IR_SetWakeUpKey;
    ir_funcStr.NEC_Mode       = FAPI_IR_NEC_Mode;
    ir_funcStr.RC5_Mode       = FAPI_IR_RC5_Mode;

    /*
    (FPC)
     FREG_FPC <----
    (IR)
     FREG_FPC <----
     FREG_IR
    */

    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_FPC] == 0){
        FREG_FPC_Init();
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_FPC] = 1;
    }
    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_IR] == 0){
        FREG_IRR_Init();//FREG_IR_Init(); /*for H61 integated MANGO*/
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_IR] = 1;
    }
    
    /* setup interrupt */
    /* ir */
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_FPC_IR,
                               FAPEXCFG_IR_Isr,
                               (FAPICORE_BSRRET_T (*)(unsigned long))IRBSR_FUNCNAME,
                               SA_INTERRUPT, (uint8_t*)"IR",
                               4096, 32 );

    /* initialize function of FAPI CDL (ir) */
    ret = ir_funcStr.Init();

    return ret;
}

/*--------------------------------------------------------*/
/* CLEANUP.                                               */
/*--------------------------------------------------------*/
void FAPEXCFG_IR_Cleanup(void)
{
    /* unregister interrupt */
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_FPC_IR);

    ir_funcStr.Exit();
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
EXPORT_SYMBOL(ir_funcStr);

EXPORT_SYMBOL(FAPEXCFG_IR_Init);
EXPORT_SYMBOL(FAPEXCFG_IR_Cleanup);
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
