/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FAPI SSP configuration code.
*/

#include "fapex_stdhdr.h"
#include "fapi/drv_intr.h"
#include "fapi/drv_ssp.h"
#include "fapi/reg_ssp.h"
#include "fapex/sys_config.h"
#include "fapex_config_cmn.h"
#include "fapex_ssp_func.h"

/*--------------------------------------------------------*/
/* Global Data                                            */
/*--------------------------------------------------------*/
FAPEX_SSP_FuncStrT ssp_funcStr;

/*--------------------------------------------------------*/
/* Local Data                                             */
/*--------------------------------------------------------*/
#ifdef __LINUX__
static uint32_t  result;
#endif

/*--------------------------------------------------------*/
/* ISR.                                                   */
/*--------------------------------------------------------*/
FAPICORE_ISRRET_T FAPEXCFG_SSP_Isr(int arg1, void *ptr)
{
  #ifndef __LINUX__
    uint32_t result;
  #endif
    void *pBsr;

    /* Interrupt service routine for Isr */
    result = ssp_funcStr.Isr();
    
    if(result != 0) {
        /* register bsr */
        pBsr = (void*)FAPEXCFG_INTR_IsBsrFunction(FAPI_INTR_SSP);
        if(pBsr != NULL){
            FAPEXCFG_INTR_RegisterBSR(pBsr);
        }
    }
    FAPICORE_ISRRETVAL(result)
}

/*--------------------------------------------------------*/
/* BSR.                                                   */
/*--------------------------------------------------------*/
#ifndef __LINUX__
FAPICORE_BSRRET_T FAPEXCFG_SSP_Bsr(unsigned long arg)
{
    ssp_funcStr.Bsr(arg);
}

#define SSPBSR_FUNCNAME  FAPEXCFG_TSD_Bsr

#else
FAPICORE_BSRRET_T FAPEXCFG_SSP_Bsr(unsigned long arg)
{
    result = *(uint32_t *)arg;
    ssp_funcStr.Bsr(result);
}

#define SSPBSR_FUNCNAME (&FAPEXCFG_SSP_Tasklet)

static DECLARE_TASKLET( FAPEXCFG_SSP_Tasklet, FAPEXCFG_SSP_Bsr, (unsigned long)&result );

#endif /* !__LINUX__ */

/*--------------------------------------------------------*/
/* INIT.                                                  */
/*--------------------------------------------------------*/
int32_t FAPEXCFG_SSP_Init(void)
{
    int32_t ret = 0;

    ssp_funcStr.Init           = FAPI_SSP_Init;
    ssp_funcStr.Exit           = FAPI_SSP_Exit;
    ssp_funcStr.Isr            = FAPI_SSP_Isr;
    ssp_funcStr.Bsr            = FAPI_SSP_Bsr;
    ssp_funcStr.Open           = FAPI_SSP_Open;
    ssp_funcStr.Close          = FAPI_SSP_Close;
    ssp_funcStr.SetFrameFormat = FAPI_SSP_SetFrameFormat;
    ssp_funcStr.SetSpeed       = FAPI_SSP_SetSpeed;
    ssp_funcStr.GetData        = FAPI_SSP_GetData;
    ssp_funcStr.Read           = FAPI_SSP_Read;
    ssp_funcStr.Write          = FAPI_SSP_Write;

    /* initialize function of FAPI RAL */
    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_SSP] == 0){
        FREG_SSP_Init();
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_SSP] = 1;
    }
    
    /* setup interrupt */
    /* SSP */
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_SSP,
                               FAPEXCFG_SSP_Isr,
                               (FAPICORE_BSRRET_T (*)(unsigned long))SSPBSR_FUNCNAME,
                               SA_INTERRUPT, (uint8_t*)"SSP",
                               1024, 32 );

    /* initialize function of FAPI CDL (ssp) */
    ret = ssp_funcStr.Init();

    return ret;
}

/*--------------------------------------------------------*/
/* CLEANUP.                                               */
/*--------------------------------------------------------*/
void FAPEXCFG_SSP_Cleanup(void)
{
    ssp_funcStr.Exit();
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
EXPORT_SYMBOL(ssp_funcStr);

EXPORT_SYMBOL(FAPEXCFG_SSP_Init);
EXPORT_SYMBOL(FAPEXCFG_SSP_Cleanup);
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
