/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FAPI TIMER configuration code.
*/

#include "fapex_stdhdr.h"
#include "fapi/drv_timer.h"
#include "fapi/drv_intr.h"
#include "fapi/reg_irq_ctrl.h"
#include "fapex/sys_config.h"
#include "fapex_config_cmn.h"
#include "fapex_timer_func.h"

/*--------------------------------------------------------*/
/* Global Data                                            */
/*--------------------------------------------------------*/
FAPEX_TIMER_FuncStrT timer_funcStr;

/*--------------------------------------------------------*/
/* Local Data                                             */
/*--------------------------------------------------------*/
#ifdef __LINUX__
static uint32_t  result1;
static uint32_t  result2;
#endif

/*--------------------------------------------------------*/
/* ISR.                                                   */
/*--------------------------------------------------------*/
#ifndef __LINUX__
FAPICORE_ISRRET_T FAPEXCFG_TIMER_Isr0(int arg1, void *ptr)
{
    uint32_t result0;
    void *pBsr;

    /* Interrupt service routine for Isr */
    result0 = timer_funcStr.Isr0();

    if(result0 != 0) {
        /* register bsr */
        pBsr = (void*)FAPEXCFG_INTR_IsBsrFunction(FAPI_INTR_TIMER0);
        if(pBsr != NULL){
            FAPEXCFG_INTR_RegisterBSR(pBsr);
        }
    }
    FAPICORE_ISRRETVAL(result0)
}
#endif /* !__LINUX__ */
FAPICORE_ISRRET_T FAPEXCFG_TIMER_Isr1(int arg1, void *ptr)
{
#ifndef __LINUX__
    uint32_t result1;
#endif
    void *pBsr;

    /* Interrupt service routine for Isr */
    result1 = timer_funcStr.Isr1();

    if(result1 != 0) {
        /* register bsr */
        pBsr = (void*)FAPEXCFG_INTR_IsBsrFunction(FAPI_INTR_TIMER1);
        if(pBsr != NULL){
            FAPEXCFG_INTR_RegisterBSR(pBsr);
        }
    }
    FAPICORE_ISRRETVAL(result1)
}
FAPICORE_ISRRET_T FAPEXCFG_TIMER_Isr2(int arg1, void *ptr)
{
#ifndef __LINUX__
    uint32_t result2;
#endif
    void *pBsr;

    /* Interrupt service routine for Isr */
    result2 = timer_funcStr.Isr2();

    if(result2 != 0) {
        /* register bsr */
        pBsr = (void*)FAPEXCFG_INTR_IsBsrFunction(FAPI_INTR_TIMER2);
        if(pBsr != NULL){
            FAPEXCFG_INTR_RegisterBSR(pBsr);
         }
    }
    FAPICORE_ISRRETVAL(result2)
}

/*--------------------------------------------------------*/
/* BSR.                                                   */
/*--------------------------------------------------------*/
#ifndef __LINUX__
FAPICORE_BSRRET_T FAPEXCFG_TIMER_Bsr0(unsigned long arg)
{
    timer_funcStr.Bsr0(arg);
}
FAPICORE_BSRRET_T FAPEXCFG_TIMER_Bsr1(unsigned long arg)
{
    timer_funcStr.Bsr1(arg);
}
FAPICORE_BSRRET_T FAPEXCFG_TIMER_Bsr2(unsigned long arg)
{
    timer_funcStr.Bsr2(arg);
}

#define TIMERBSR_FUNCNAME_0  FAPEXCFG_TIMER_Bsr0
#define TIMERBSR_FUNCNAME_1  FAPEXCFG_TIMER_Bsr1
#define TIMERBSR_FUNCNAME_2  FAPEXCFG_TIMER_Bsr2

#else
FAPICORE_BSRRET_T FAPEXCFG_TIMER_Bsr1(unsigned long arg)
{
    uint32_t  result;

    result = *(uint32_t *)arg;
    timer_funcStr.Bsr1(result);
}
FAPICORE_BSRRET_T FAPEXCFG_TIMER_Bsr2(unsigned long arg)
{
    uint32_t  result;

    result = *(uint32_t *)arg;
    timer_funcStr.Bsr2(result);
}

#define TIMERBSR_FUNCNAME_1 (&FAPEXCFG_TIMER_Tasklet1)
#define TIMERBSR_FUNCNAME_2 (&FAPEXCFG_TIMER_Tasklet2)

static DECLARE_TASKLET( FAPEXCFG_TIMER_Tasklet1, FAPEXCFG_TIMER_Bsr1, (unsigned long)&result1 );
static DECLARE_TASKLET( FAPEXCFG_TIMER_Tasklet2, FAPEXCFG_TIMER_Bsr2, (unsigned long)&result2 );

#endif /* !__LINUX__ */

/*--------------------------------------------------------*/
/* INIT.                                                  */
/*--------------------------------------------------------*/
int32_t FAPEXCFG_TIMER_Init(void)
{
    int32_t ret = 0;

    timer_funcStr.Init              = FAPI_TIMER_Init;
    timer_funcStr.Exit              = FAPI_TIMER_Exit;
#ifndef __LINUX__
    timer_funcStr.Isr0              = FAPI_TIMER_Isr0;
    timer_funcStr.Bsr0              = FAPI_TIMER_Bsr0;
#endif /* !__LINUX__ */
    timer_funcStr.Isr1              = FAPI_TIMER_Isr1;
    timer_funcStr.Isr2              = FAPI_TIMER_Isr2;
    timer_funcStr.Bsr1              = FAPI_TIMER_Bsr1;
    timer_funcStr.Bsr2              = FAPI_TIMER_Bsr2;
    timer_funcStr.SetClockFrequency = FAPI_TIMER_SetClockFrequency;
    timer_funcStr.SetType           = FAPI_TIMER_SetType;
    timer_funcStr.SetResolution     = FAPI_TIMER_SetResolution;
    timer_funcStr.GetIndex          = FAPI_TIMER_GetIndex;
    timer_funcStr.GetResolution     = FAPI_TIMER_GetResolution;
    timer_funcStr.GetTimeStamp      = FAPI_TIMER_GetTimeStamp;
    timer_funcStr.Open              = FAPI_TIMER_Open;
    timer_funcStr.Close             = FAPI_TIMER_Close;
    timer_funcStr.Start             = FAPI_TIMER_Start;
    timer_funcStr.Stop              = FAPI_TIMER_Stop;
    timer_funcStr.ReadValue         = FAPI_TIMER_ReadValue;

    FAPI_SYS_TIMER_SET_BOARD_PARAMS

    /* setup interrupt */
#ifndef __LINUX__
    /* timer 0 */
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_TIMER0,
                               FAPEXCFG_TIMER_Isr0,
                               (FAPICORE_BSRRET_T (*)(unsigned long))TIMERBSR_FUNCNAME_0,
                               SA_INTERRUPT, (uint8_t*)"TIMER0",
                               1024, 4 );
#endif /* !__LINUX__ */
    /* timer 1 */
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_TIMER1,
                               FAPEXCFG_TIMER_Isr1,
                               (FAPICORE_BSRRET_T (*)(unsigned long))TIMERBSR_FUNCNAME_1,
                               SA_INTERRUPT, (uint8_t*)"TIMER1",
                               1024, 4 );
    /* timer 2 */
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_TIMER2,
                               FAPEXCFG_TIMER_Isr2,
                               (FAPICORE_BSRRET_T (*)(unsigned long))TIMERBSR_FUNCNAME_2,
                               SA_INTERRUPT, (uint8_t*)"TIMER2",
                               1024, 4 );

    /* setup software timer resolution for HDMI-CEC */
    timer_funcStr.SetResolution(1, FAPI_TIMER_RESOLUTION_100_USEC);

    /* initialize function of FAPI CDL (timer) */
    ret = timer_funcStr.Init();

    return ret;
}

/*--------------------------------------------------------*/
/* CLEANUP.                                               */
/*--------------------------------------------------------*/
void FAPEXCFG_TIMER_Cleanup(void)
{
    /* unregister interrupt */
#ifndef __LINUX__
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_TIMER0);
#endif /* !__LINUX__ */
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_TIMER1);
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_TIMER2);

    timer_funcStr.Exit();
}


#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
EXPORT_SYMBOL(timer_funcStr);

EXPORT_SYMBOL(FAPEXCFG_TIMER_Init);
EXPORT_SYMBOL(FAPEXCFG_TIMER_Cleanup);
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
