/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FAPI VIDEC configuration code.
*/

#include "fapex_stdhdr.h"
#include "fapi/drv_videc.h"
#include "fapi/drv_intr.h"
#include "fapi/reg_vd.h"
#include "fapi/reg_pint.h"
#include "fapi/reg_pirlo_axi.h"
#include "fapi/reg_vo.h"
#include "fapi/reg_ddr.h"
#include "firmware/firmware_video.h"
#include "fapex/sys_config.h"
#include "fapex_config_cmn.h"
#include "fapex_videc_func.h"

/*--------------------------------------------------------*/
/* Global Data                                            */
/*--------------------------------------------------------*/
FAPEX_VIDEC_FuncStrT videc_funcStr;

/*--------------------------------------------------------*/
/* Local Data                                             */
/*--------------------------------------------------------*/
#ifdef __LINUX__
static struct semaphore disp_sem;
static uint32_t         result_dec;
static uint32_t         result_dsp;
#endif

/*--------------------------------------------------------*/
/* Decode ISR.                                            */
/*--------------------------------------------------------*/
FAPICORE_ISRRET_T FAPEXCFG_VIDEC_IsrDecode(int arg1, void *ptr)
{
#ifndef __LINUX__
    uint32_t result_dec;
#endif
    void *pBsr;

    /* Interrupt service routine for Isr */
    result_dec = videc_funcStr.IsrDecode();

    if(result_dec != 0) {
        /* register bsr */
        pBsr = (void*)FAPEXCFG_INTR_IsBsrFunction(FAPI_INTR_AMCPU_AXP);
        if(pBsr != NULL){
            FAPEXCFG_INTR_RegisterBSR(pBsr);
        }
    }
    FAPICORE_ISRRETVAL(result_dec)
}

/*--------------------------------------------------------*/
/* Display ISR.                                           */
/*--------------------------------------------------------*/
FAPICORE_ISRRET_T FAPEXCFG_VIDEC_IsrDisplay(int arg1, void *ptr)
{
#ifndef __LINUX__
    uint32_t result_dsp;
    void *pBsr;
#endif

    /* Interrupt service routine for Isr */
    result_dsp = videc_funcStr.IsrDisplay();

    if(result_dsp != 0) {
#ifndef __LINUX__
        /* register bsr */
        pBsr = (void*)FAPEXCFG_INTR_IsBsrFunction(FAPI_INTR_VO0);
        if(pBsr != NULL){
            FAPEXCFG_INTR_RegisterBSR(pBsr);
        }
    }
#else
        /* wake up for FAPEXCFG_VIDEC_BsrDisplayThread() */
        up(&disp_sem);
    }
#endif
    FAPICORE_ISRRETVAL(result_dsp)
}

/*--------------------------------------------------------*/
/* Decode BSR.                                            */
/*--------------------------------------------------------*/
#ifndef __LINUX__
FAPICORE_BSRRET_T FAPEXCFG_VIDEC_BsrDecode(unsigned long arg)
{
    videc_funcStr.BsrDecode(arg);
}

#define VIDECBSR_FUNCNAME_DEC  FAPEXCFG_VIDEC_BsrDecode

#else
FAPICORE_BSRRET_T FAPEXCFG_VIDEC_BsrDecode(unsigned long arg)
{
    videc_funcStr.BsrDecode(result_dec);
}

#define VIDECBSR_FUNCNAME_DEC (&FAPEXCFG_VIDEC_TaskletDec)

static DECLARE_TASKLET( FAPEXCFG_VIDEC_TaskletDec, FAPEXCFG_VIDEC_BsrDecode,  (unsigned long)&result_dec );

#endif /* !__LINUX__ */

/*--------------------------------------------------------*/
/* Display BSR.                                           */
/*--------------------------------------------------------*/
#ifndef __LINUX__
FAPICORE_BSRRET_T FAPEXCFG_VIDEC_BsrDisplay(unsigned long arg)
{
    videc_funcStr.BsrDisplay(arg);
}

#define VIDECBSR_FUNCNAME_DSP  FAPEXCFG_VIDEC_BsrDisplay

#else
static int FAPEXCFG_VIDEC_BsrDisplayThread(void *arg)
{
    int32_t  ret;
    struct sched_param  schedparam = { .sched_priority = 99 };

    sched_setscheduler(current, SCHED_FIFO, &schedparam);

    lock_kernel();
    daemonize("fapex_videc_display");
    unlock_kernel();

    while(1) {
        /* wait for FAPEXCFG_VIDEC_BsrDisplay() */
        ret = down_interruptible(&disp_sem);
        if(ret != 0) {
            break;
        }

        videc_funcStr.BsrDisplay(result_dsp);
    }
    return 0;
}
#endif /* !__LINUX__ */

/*--------------------------------------------------------*/
/* INIT.                                                  */
/*--------------------------------------------------------*/
int32_t FAPEXCFG_VIDEC_Init(void)
{
    int32_t ret = 0;
#ifdef __LINUX__
    int32_t disp_pid;
//    struct task_struct *disp_task;
//    struct sched_param schedparam;
#endif

    videc_funcStr.SetBoardMemory                = FAPI_VIDEC_SetBoardMemory;
    videc_funcStr.PreInit                       = FAPI_VIDEC_PreInit;
    videc_funcStr.GetPreInitError               = FAPI_VIDEC_GetPreInitError;
    videc_funcStr.GetFirmwareVersion            = FAPI_VIDEC_GetFirmwareVersion;
    videc_funcStr.Init                          = FAPI_VIDEC_Init;
    videc_funcStr.Exit                          = FAPI_VIDEC_Exit;
    videc_funcStr.IsrDecode                     = FAPI_VIDEC_IsrDecode;
    videc_funcStr.BsrDecode                     = FAPI_VIDEC_BsrDecode;
    videc_funcStr.IsrDisplay                    = FAPI_VIDEC_IsrDisplay;
    videc_funcStr.BsrDisplay                    = FAPI_VIDEC_BsrDisplay;
    videc_funcStr.ReportMemory                  = FAPI_VIDEC_ReportMemory;
    videc_funcStr.Open                          = FAPI_VIDEC_Open;
    videc_funcStr.Close                         = FAPI_VIDEC_Close;
    videc_funcStr.Reset                         = FAPI_VIDEC_Reset;
    videc_funcStr.FreezeContinue                = FAPI_VIDEC_FreezeContinue;
    videc_funcStr.Dump                          = FAPI_VIDEC_Dump;
    videc_funcStr.Start                         = FAPI_VIDEC_Start;
    videc_funcStr.Stop                          = FAPI_VIDEC_Stop;
    videc_funcStr.EnableDisable                 = FAPI_VIDEC_EnableDisable;
    videc_funcStr.SetPlaybackMode               = FAPI_VIDEC_SetPlaybackMode;
    videc_funcStr.SetPlaybackEnvironment        = FAPI_VIDEC_SetPlaybackEnvironment;
    videc_funcStr.GetVesParams                  = FAPI_VIDEC_GetVesParams;
    videc_funcStr.GetVesBufferPosition          = FAPI_VIDEC_GetVesBufferPosition;
    videc_funcStr.GetVesBufferSize              = FAPI_VIDEC_GetVesBufferSize;
    videc_funcStr.GetVesFillLevel               = FAPI_VIDEC_GetVesFillLevel;
    videc_funcStr.GetVesAvailableSpace          = FAPI_VIDEC_GetVesAvailableSpace;
    videc_funcStr.GetState                      = FAPI_VIDEC_GetState;
    videc_funcStr.GetHealth                     = FAPI_VIDEC_GetHealth;
    videc_funcStr.GetDecodingLatency            = FAPI_VIDEC_GetDecodingLatency;
    videc_funcStr.AudCallback                   = FAPI_VIDEC_AudCallback;
    videc_funcStr.AudFeedExternal               = FAPI_VIDEC_AudFeedExternal;
    videc_funcStr.Recover                       = FAPI_VIDEC_Recover;
    videc_funcStr.DecodeSingleDirectIFrame      = FAPI_VIDEC_DecodeSingleDirectIFrame;
    videc_funcStr.DisplaySingleDirectIFrame     = FAPI_VIDEC_DisplaySingleDirectIFrame;
    videc_funcStr.DecodeSequenceOfDirectIFrame  = FAPI_VIDEC_DecodeSequenceOfDirectIFrame;
    videc_funcStr.DisplaySequenceOfDirectIFrame = FAPI_VIDEC_DisplaySequenceOfDirectIFrame;
    videc_funcStr.SetEntropyDecInfoCallback     = FAPI_VIDEC_SetEntropyDecInfoCallback;
    videc_funcStr.SetVideoDecInfoCallback       = FAPI_VIDEC_SetVideoDecInfoCallback;
    videc_funcStr.SetAutoScaleCallback          = FAPI_VIDEC_SetAutoScaleCallback;
    videc_funcStr.SetDecoderToHdEncoder         = FAPI_VIDEC_SetDecoderToHdEncoder;
    videc_funcStr.SetPesParserToDecoder         = FAPI_VIDEC_SetPesParserToDecoder;
    videc_funcStr.SetFrameDecodedCallback       = FAPI_VIDEC_SetFrameDecodedCallback;
    videc_funcStr.SetFrameAvailableCallback     = FAPI_VIDEC_SetFrameAvailableCallback;
    videc_funcStr.SetVesOverflowCallback        = FAPI_VIDEC_SetVesOverflowCallback;
    videc_funcStr.SetDirectIFrameMode           = FAPI_VIDEC_SetDirectIFrameMode;
    videc_funcStr.SetChannelId                  = FAPI_VIDEC_SetChannelId;
    videc_funcStr.SetParamChangeCallback        = FAPI_VIDEC_SetParamChangeCallback;
    videc_funcStr.SetAfdState                   = FAPI_VIDEC_SetAfdState;
    videc_funcStr.SetOutputDelay                = FAPI_VIDEC_SetOutputDelay;
    videc_funcStr.SetEnableDelay                = FAPI_VIDEC_SetEnableDelay;
    videc_funcStr.SetSyncBase                   = FAPI_VIDEC_SetSyncBase;
    videc_funcStr.SetFrameRate                  = FAPI_VIDEC_SetFrameRate;
    videc_funcStr.SetStreamSegmentMode          = FAPI_VIDEC_SetStreamSegmentMode;
    videc_funcStr.ClearEntropyDecInfoCallback   = FAPI_VIDEC_ClearEntropyDecInfoCallback;
    videc_funcStr.ClearVideoDecInfoCallback     = FAPI_VIDEC_ClearVideoDecInfoCallback;
    videc_funcStr.ClearAutoScaleCallback        = FAPI_VIDEC_ClearAutoScaleCallback;
    videc_funcStr.ClearDecoderToHdEncoder       = FAPI_VIDEC_ClearDecoderToHdEncoder;
    videc_funcStr.ClearPesParserToDecoder       = FAPI_VIDEC_ClearPesParserToDecoder;
    videc_funcStr.ClearFrameDecodedCallback     = FAPI_VIDEC_ClearFrameDecodedCallback;
    videc_funcStr.ClearFrameAvailableCallback   = FAPI_VIDEC_ClearFrameAvailableCallback;
    videc_funcStr.ClearVesOverflowCallback      = FAPI_VIDEC_ClearVesOverflowCallback;
    videc_funcStr.ClearDirectIFrameMode         = FAPI_VIDEC_ClearDirectIFrameMode;
    videc_funcStr.ClearChannelId                = FAPI_VIDEC_ClearChannelId;
    videc_funcStr.ClearParamChangeCallback      = FAPI_VIDEC_ClearParamChangeCallback;
    videc_funcStr.ClearOutputDelay              = FAPI_VIDEC_ClearOutputDelay;
    videc_funcStr.ClearEnableDelay              = FAPI_VIDEC_ClearEnableDelay;
    videc_funcStr.ClearSyncBase                 = FAPI_VIDEC_ClearSyncBase;
    videc_funcStr.ClearFrameRate                = FAPI_VIDEC_ClearFrameRate;
    videc_funcStr.ClearStreamSegmentMode        = FAPI_VIDEC_ClearStreamSegmentMode;
    videc_funcStr.firmware_type                 = FIRMWARE_video_embedded;

    /* initialize function of FAPI RAL */
    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_VO] == 0) {
        FREG_VO_Init();
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_VO] = 1;
    }
    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_DDRRAMIF] == 0) {
        FREG_DDR_Init();//FREG_DDRRAMIF_Init();
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_DDRRAMIF] = 1;
    }
    if(FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_VD] == 0) {
        FREG_VD_Init();
        FREG_PINT_Init();
        FREG_PIRLO_AXI_Init();
        FAPEXCFG_RAL_initialized_flag[FAPEX_RAL_MODULEID_VD] = 1;
    }

    FAPI_SYS_VIDEC_SET_BOARD_PARAMS

    /* setup interrupt */
    /* videc 0 decode */
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_AMCPU_AXP,
                               FAPEXCFG_VIDEC_IsrDecode,
                               (FAPICORE_BSRRET_T (*)(unsigned long))VIDECBSR_FUNCNAME_DEC,
                               SA_INTERRUPT, (uint8_t*)"VIDEC0",
                               2048, 256 );
    /* videc 0 display */
#ifndef __LINUX__
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_VO0,
                               FAPEXCFG_VIDEC_IsrDisplay,
                               (FAPICORE_BSRRET_T (*)(unsigned long))VIDECBSR_FUNCNAME_DSP,
                               SA_INTERRUPT, (uint8_t*)"VIDEC1",
                               2048, 256 );
#else
    FAPEXCFG_INTR_RegisterISR( FAPI_INTR_VO0,
                               FAPEXCFG_VIDEC_IsrDisplay, NULL,
                               SA_INTERRUPT, (uint8_t*)"VIDEC1",
                               2048, 256 );

    /* create kernel thread because FAPI_VIDEC_BsrDisplay() use FAPI_SYS_GET_SEMAPHORE(). */
    sema_init(&disp_sem, 0);
    disp_pid = kernel_thread(FAPEXCFG_VIDEC_BsrDisplayThread, NULL, SIGCHLD | CLONE_KERNEL);
    if(disp_pid < 0) {
        return disp_pid;
    }
#endif

    /* initialize function of FAPI CDL (videc) */
    ret = videc_funcStr.Init();

    return ret;
}

/*--------------------------------------------------------*/
/* CLEANUP.                                               */
/*--------------------------------------------------------*/
void FAPEXCFG_VIDEC_Cleanup(void)
{
    /* unregister interrupt */
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_VO0);
    FAPEXCFG_INTR_UnregisterISR(FAPI_INTR_AMCPU_AXP);

    videc_funcStr.Exit();
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
EXPORT_SYMBOL(videc_funcStr);

EXPORT_SYMBOL(FAPEXCFG_VIDEC_Init);
EXPORT_SYMBOL(FAPEXCFG_VIDEC_Cleanup);
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
