/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv
*/
#include "fapex.h"
#include "fapex/abst_libaudec.h"
#include "fapex/sys_config.h"
#include "fapex_fdlaudec_priv.h"
#include "fapex_fdltsd_priv.h"
#include "fapex_fdlauout_priv.h"
#include "fapex_config_cmn.h"

/* for audec_funcStr, auout_funcStr, tsd_funcStr, bm_funcStr, timer_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_AUDEC_StrT  FAPIGLD_FDLAUDECCtx[FAPEXDEV_AUDEC_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

static uint32_t             firmware_boot_flag;     //!< 0:not booted, 1:booted
static FAPI_SYS_SemaphoreT  firmware_boot_sem;

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLAUDEC_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                     FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXDebugPrint(3, "[fapex_fdlaudec/FAPEXFDLAUDEC_read] size:%d\n",size);

    return size;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLAUDEC_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                     FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLAUDEC_write] size:%d\n",size);

    return size;
}
#endif

/*!
*******************************************************************************
**
** \brief  AUDEC driver open
**
** \param  pfunc    pointer to context
**
******************************************************************************/
static int32_t fdlaudec_Open(FAPEXFDL_AUDEC_StrT *pfunc)
{
    FAPI_AUDEC_OpenParamsT params;
    void                   *pCfdlHdl;
    FAPEXFDL_TSD_SubStrT   *pfunc_tsd;
    FAPEXFDL_AUOUT_StrT    *pfunc_auout;
    int32_t ret;

    params.version = FAPI_AUDEC_VERSION;

    /* setup decoder type */
    params.decoder = (FAPI_AUDEC_DecoderEnumT)pfunc->codec_type;

    pCfdlHdl = pfunc->pCfdlHdl;

    /* setup input handler */
    pfunc_tsd = (FAPEXFDL_TSD_SubStrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, pfunc->input_id);
    if(pfunc_tsd != NULL) {
        params.inputStreamHandle = pfunc_tsd->pBMCtxHdl;
    }
    else {
        params.inputStreamHandle = NULL;
    }
    pfunc->pInTSDCtxHdl = pfunc_tsd;
    pfunc->pInBMCtxHdl  = params.inputStreamHandle;

    /* setup output handler */
    pfunc_auout = (FAPEXFDL_AUOUT_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, pfunc->output_id);
    if(pfunc_auout != NULL) {
        params.audioOutHandle = pfunc_auout->pFAPICtxHdl;
    }
    else {
        params.audioOutHandle = NULL;
    }
    pfunc->pAuoutHdl = (void *)pfunc_auout;

    /* setup sample bits */
    params.sampleBits = (FAPI_AUDEC_SampleBitsEnumT)pfunc->sample_bits;

    /* call Open AUDEC */
    pfunc->pFAPICtxHdl = audec_funcStr.Open(&params, &ret);
    FAPEXDebugPrint(3, "[fdlaudec_Open] FAPI context:0x%08x\n",
                       (uint32_t)pfunc->pFAPICtxHdl);
    if(pfunc->pFAPICtxHdl != NULL) {
        if(pfunc_tsd != NULL) {
            if( (pfunc_tsd->pTSDCtxHdl != NULL)
             && !(pfunc->open_flag & FAPEXFDL_AUDEC_SET_PTS_NOTIFIER) ) {
                /* set PTS notifer */
                ret = tsd_funcStr.SetPtsNotifier(pfunc_tsd->pTSDCtxHdl,
                                                 audec_funcStr.PtsNotifier,
                                                 pfunc->pInBMCtxHdl, 1);
                if(ret == FAPI_OK) {
                    pfunc->open_flag |= FAPEXFDL_AUDEC_SET_PTS_NOTIFIER;
                }
            }
            pfunc->status = FAPEX_STATUS_READY;
        }
    }
    else {
        FAPEXDebugPrint(2, "[fdlaudec_Open] error:%d\n", ret);
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  AUDEC driver start
**
** \param  pfunc    pointer to context
**
******************************************************************************/
static int32_t fdlaudec_Start(FAPEXFDL_AUDEC_StrT *pfunc)
{
    uint64_t end_time;
    int64_t  delta_time;
    int32_t  ret;

    if(pfunc->status == FAPEX_STATUS_READY) {
#if (FAPEXFDL_AUDEC_PREBUFFERED_SIZE > 0)
        FAPEXFDL_TSD_SubStrT  *pfunc_tsd;
        uint32_t  wpt;

        pfunc_tsd = pfunc->pInTSDCtxHdl;
        end_time = timer_funcStr.GetTimeStamp(FAPI_TIMER_RESOLUTION_1_MSEC)
                   + FAPEXFDL_AUDEC_PREBUFFERING_WAITING_MAXTIME;
        while(pfunc_tsd->status == FAPEX_STATUS_RUNNING) {
            wpt = bm_funcStr.GetBufferWritePointerPosition(pfunc->pInBMCtxHdl);
            if(wpt > FAPEXFDL_AUDEC_PREBUFFERED_SIZE) {
                break;
            }
            delta_time = timer_funcStr.GetTimeStamp(FAPI_TIMER_RESOLUTION_1_MSEC)
                         - end_time;
            if(delta_time > 0) {
                FAPEXDebugPrint(2, "[fdlaudec_Start] pre-buffering size:%d\n", wpt);
                break;
            }
            FAPI_SYS_SLEEP(1);
        }
#endif
        /* start AUDEC */
        ret = audec_funcStr.Start(pfunc->pFAPICtxHdl);

        if(ret == FAPI_OK) {
            if(pfunc->pAuoutHdl != NULL) {
                FAPEXFDL_AUOUT_StrT  *pfunc_auout;

                pfunc_auout = (FAPEXFDL_AUOUT_StrT*)pfunc->pAuoutHdl;
                if(pfunc_auout->enable_flag & FAPEX_AUOUT_ENABLE_HDMI) {
                    FAPI_AUOUT_StreamInfoT  info;

                    end_time = timer_funcStr.GetTimeStamp(FAPI_TIMER_RESOLUTION_1_MSEC)
                               + FAPEXFDL_AUDEC_FREQ_INFO_WAITING_MAXTIME;
                    while(1) {
                        info = FAPEXCFG_AUOUT_GetStreamInfo(pfunc_auout->pFAPICtxHdl);
                        if(info.sampleFreq != FAPI_AUOUT_UNDEF_HZ) {
                            /* HDMI audio start */
                            ret = auout_funcStr.StartHdmi(pfunc_auout->pFAPICtxHdl,FAPI_AUOUT_I2S_MASK|FAPI_AUOUT_SPDIF_MASK);
                            break;
                        }
                        delta_time = timer_funcStr.GetTimeStamp(FAPI_TIMER_RESOLUTION_1_MSEC)
                                     - end_time;
                        if(delta_time > 0) {
                            break;
                        }
                        FAPI_SYS_SLEEP(1);
                    }
                }
            }
            pfunc->status = FAPEX_STATUS_RUNNING;
        }
        FAPEXDebugPrint(3, "[fdlaudec_Start] FAPI_AUDEC_Start:%d %d\n",
                            ret, bm_funcStr.GetBufferFillLevel(pfunc->pInBMCtxHdl));
    }
    else {
        ret = FAPEX_AUDEC_ERR_INVALID_STATUS;
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  AUDEC driver stop
**
** \param  pfunc    pointer to context
**
******************************************************************************/
static int32_t fdlaudec_Stop(FAPEXFDL_AUDEC_StrT *pfunc)
{
    FAPEXFDL_AUOUT_StrT  *pfunc_auout;

    if(pfunc->status == FAPEX_STATUS_RUNNING) {
        if(pfunc->pAuoutHdl != NULL) {
            pfunc_auout = (FAPEXFDL_AUOUT_StrT*)pfunc->pAuoutHdl;
            if(pfunc_auout->enable_flag & FAPEX_AUOUT_ENABLE_HDMI) {
                /* HDMI audio stop */
                auout_funcStr.StopHdmi();
            }
        }
        /* stop AUDEC */
        audec_funcStr.Stop(pfunc->pFAPICtxHdl);

        pfunc->status = FAPEX_STATUS_READY;
    }
    return 0;
}

/*!
*******************************************************************************
**
** \brief  AUDEC driver set codec type
**
** \param  pfunc    pointer to context
** \param  pArg     pointer to FAPEX_AUDEC_OpenParamStrT
**
******************************************************************************/
static int32_t fdlaudec_SetCodecType(FAPEXFDL_AUDEC_StrT *pfunc,
                                     FAPEX_AUDEC_OpenParamStrT *pArg)
{
    FAPI_AUDEC_CapabilitiesT  cap;
    int32_t  ret = FAPEX_AUDEC_ERR_BAD_PARAMETER;
  
    if(pfunc->status != FAPEX_STATUS_RUNNING) {
        /* copy parameters from an user space to the kernel space */
        FAPEXCFDLSAF_COPY_TO_DRV(&pfunc->codec_type, &(pArg->codecType), sizeof(uint32_t));
        FAPEXCFDLSAF_COPY_TO_DRV(&pfunc->input_id,   &(pArg->inputId),   sizeof(uint32_t));

        /* check codec_type */
        cap = audec_funcStr.GetCapabilities();
        switch(pfunc->codec_type){
            case FAPI_AUDEC_DECODER_AC3:
                if(cap.ac3Decode == 0) {
                    return FAPI_AUDEC_ERR_CODEC_NOT_SUPPORTED;
                }
                break;
            case FAPI_AUDEC_DECODER_DD:
                if(cap.ac3Decode == 0){
                    return FAPI_AUDEC_ERR_CODEC_NOT_SUPPORTED;
                }
                break;
            case FAPI_AUDEC_DECODER_DDPDCV:
                if(cap.ddpdcvDecode == 0){
                    return FAPI_AUDEC_ERR_CODEC_NOT_SUPPORTED;
                }
                break;
            case FAPI_AUDEC_DECODER_MPEG:
                if((cap.mpgDecode == 0) && (cap.mp3Decode == 0)) {
                    return FAPI_AUDEC_ERR_CODEC_NOT_SUPPORTED;
                }
                break;
            case FAPI_AUDEC_DECODER_WMA:
                if(cap.wmaDecode == 0) {
                    return FAPI_AUDEC_ERR_CODEC_NOT_SUPPORTED;
                }
                break;
            case FAPI_AUDEC_DECODER_HEAAC:
                if(cap.heaacDecode == 0) {
                    return FAPI_AUDEC_ERR_CODEC_NOT_SUPPORTED;
                }
                break;
            default:
                return FAPI_AUDEC_ERR_CODEC_NOT_SUPPORTED;
        }

        /* check device id */
        if( (pfunc->input_id >= FAPEXDEV_TSD_SUB_MINOR_FIRST)
         && (pfunc->input_id <= FAPEXDEV_TSD_SUB_MINOR_LAST) ) {
            /* check io setup */
            pfunc->open_flag |= FAPEXFDL_AUDEC_SET_INPUT;
            if(pfunc->open_flag == FAPEXFDL_AUDEC_SET_IODONE) {
                ret = fdlaudec_Open(pfunc);
            }
            else {
                ret = 0;
            }
        }
    }
    else {
        ret = FAPEX_AUDEC_ERR_INVALID_STATUS;
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  AUDEC driver set output
**
** \param  pfunc    pointer to context
** \param  pArg     pointer to FAPEX_AUDEC_OutputParamStrT
**
******************************************************************************/
static int32_t fdlaudec_SetOutput(FAPEXFDL_AUDEC_StrT *pfunc,
                                  FAPEX_AUDEC_OutputParamStrT *pArg)
{
    int32_t  ret = FAPEX_AUDEC_ERR_BAD_PARAMETER;

    if( (pfunc->status != FAPEX_STATUS_LOADED)
     && (pfunc->status != FAPEX_STATUS_RUNNING) ) {
        /* copy parameters from an user space to the kernel space */
        FAPEXCFDLSAF_COPY_TO_DRV(&pfunc->output_id,   &(pArg->outputId),   sizeof(uint32_t));
        FAPEXCFDLSAF_COPY_TO_DRV(&pfunc->sample_bits, &(pArg->sampleBits), sizeof(uint32_t));

        /* check device id */
        if( (pfunc->output_id >= FAPEXDEV_AUOUT_MINOR_FIRST)
         && (pfunc->output_id <= FAPEXDEV_AUOUT_MINOR_LAST) ) {
            /* check io setup */
            pfunc->open_flag |= FAPEXFDL_AUDEC_SET_OUTPUT;
            if(pfunc->open_flag == FAPEXFDL_AUDEC_SET_IODONE) {
                ret = fdlaudec_Open(pfunc);
            }
            else {
                ret = 0;
            }
        }
    }
    else {
        ret = FAPEX_AUDEC_ERR_INVALID_STATUS;
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLAUDEC_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                     FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_AUDEC_StrT *pfunc;
    int32_t  ret = FAPEX_AUDEC_ERR_BAD_PARAMETER;

    FAPEXDebugPrint(3, "[FAPEXFDLAUDEC_ioctl] cmd:%d\n",cmd);

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_AUDEC_StrT*)(file->private_data);

    switch(cmd) {
#ifdef FAPI_AUDEC_CONNECT
        case FAPEX_IOCCOM_START:
            ret = fdlaudec_Start(pfunc);
            break;

        case FAPEX_IOCCOM_STOP:
            ret = fdlaudec_Stop(pfunc);
            break;

        case FAPEX_IOCCOM_AUDEC_OPEN:
            if(pfunc->open_flag == FAPEXFDL_AUDEC_SET_NONE) {
                if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                    ret = FAPI_OK;
#ifndef CONFIG_AUDEC_FIRMWARE_REBOOT
                    if(audec_funcStr.firmware_type != 0)    /* only embedded type */
#endif
                    {
                        /*---------- LOCK ----------*/
                        FAPI_SYS_GET_SEMAPHORE(firmware_boot_sem, FAPI_SYS_SUSPEND);

                        if(firmware_boot_flag == 0) {
                            ret = audec_funcStr.Reboot(0);
                            if(ret == FAPI_OK) {
                                // wait for end of reboot
                                FAPI_SYS_SLEEP(500);
                                firmware_boot_flag = 1;
                            }
                            FAPEXDebugPrint(2, "FAPI_AUDEC_Reboot %d\n", ret);
                        }
                        /*--------- UNLOCK ---------*/
                        FAPI_SYS_SET_SEMAPHORE(firmware_boot_sem, FAPI_SYS_NO_SUSPEND);
                    }
                    if(ret == FAPI_OK) {
                        ret = fdlaudec_SetCodecType(pfunc,
                                                    (FAPEX_AUDEC_OpenParamStrT*)arg);
                    }
                }
            }
            else {
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_RESET:
            ret = audec_funcStr.Reboot(0);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "FAPI_AUDEC_Reboot error:%d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_AUDEC_SET_OUTPUT:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlaudec_SetOutput(pfunc, (FAPEX_AUDEC_OutputParamStrT*)arg);
            }
            break;

        case FAPEX_IOCCOM_AUDEC_GET_CAPABILITIES:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPI_AUDEC_CapabilitiesT          cap;
                FAPEX_AUDEC_CapabilitiesParamStrT *pArg;

                pArg = (FAPEX_AUDEC_CapabilitiesParamStrT *)(arg);

                cap = audec_funcStr.GetCapabilities();

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR(pArg, &cap, sizeof(FAPEX_AUDEC_CapabilitiesParamStrT));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUDEC_GET_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUDEC_StatusStrT  params, *pArg;

                pArg = (FAPEX_AUDEC_StatusStrT *)(arg);

                params.status     = pfunc->status;
                params.firmStatus = audec_funcStr.GetStatus(pfunc->pFAPICtxHdl);

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_AUDEC_StatusStrT));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUDEC_SET_CODEC_PARAMS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUDEC_CodecSpecificParamStrT *pArg, params;

                pArg = (FAPEX_AUDEC_CodecSpecificParamStrT *)arg;

                /* copy parameters from an user space to the kernel space */
                FAPEXCFDLSAF_COPY_TO_DRV(&params.length, &pArg->length, sizeof(uint32_t));

                params.params = (void *)FAPI_SYS_MALLOC(params.length + 64);
                if(params.params != NULL) {
                    void  *param_ptr;

                    param_ptr = (void *)FAPI_SYS_CACHELINE_ALIGNED(params.params);

                    /* copy parameters from an user space to the kernel space */
                    FAPEXCFDLSAF_COPY_TO_DRV(param_ptr, pArg->params, params.length);

                    ret = audec_funcStr.SetCodecParams(pfunc->pFAPICtxHdl,
                                                       param_ptr, params.length);
                    FAPI_SYS_FREE(params.params);
                }
                else {
                    ret = FAPEX_AUDEC_ERR_OUT_OF_MEMORY;
                }
            }
            break;

        case FAPEX_IOCCOM_AUDEC_SET_OUTPUT_DELAY:
            ret = audec_funcStr.SetOutputDelay(pfunc, (int32_t)arg);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "FAPI_AUDEC_SetOutputDelay error:%d\n", ret);
            }
            break;

#endif /* FAPI_AUDEC_CONNECT */

        default:
            ;
    }

    return ret;
}

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLAUDEC_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T mask = 0;

    FAPEXDebugPrint(4, "[FAPEXFDLAUDEC_poll]\n");

    /* actual code which is depend on function is implimented here. */

    return mask;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLAUDEC_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    int32_t ret;

    if (vma->vm_pgoff != 0){
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_IO | VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if (file->f_flags & O_SYNC){
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
                  vma->vm_end - vma->vm_start, vma->vm_page_prot);

    if(ret != 0){
        return -EAGAIN;
    }
    return 0;
#else
    return -1;
#endif /* __LINUX__ */
}
#endif

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLAUDEC_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_AUDEC_StrT *pfunc;
    void     *pCfdlHdl;
    int32_t  id;

    FAPEXDebugPrint(3, "[FAPEXFDLAUDEC_open] file:0x%08x,private_data:0x%08x\n",
                       (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLAUDEC_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLAUDEC_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc = (FAPEXFDL_AUDEC_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc;
        FAPEXDebugPrint(3, "[FAPEXFDLAUDEC_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc));

        /* set parameter */
        pfunc->id       = id;
        pfunc->pCfdlHdl = pCfdlHdl;
        pfunc->status   = FAPEX_STATUS_OPENED;
#ifdef __LINUX__
        kref_init(&pfunc->kref);
#endif
    }
    else {
        pfunc = (FAPEXFDL_AUDEC_StrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLAUDEC_open] fapex_fdlaudec was opened : used_counter %d.\n",
                       pfunc->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlaudec_delete(struct kref *kref)
{
    FAPEXFDL_AUDEC_StrT   *pfunc;
    FAPEXFDL_TSD_SubStrT  *pfunc_tsd;
    int32_t  ret = 0;

    pfunc = container_of(kref, FAPEXFDL_AUDEC_StrT, kref);
#else
static int32_t fdlaudec_delete(FAPEXFDL_AUDEC_StrT *pfunc)
{
    FAPEXFDL_TSD_SubStrT  *pfunc_tsd;
    int32_t  ret = 0;
#endif

#ifdef FAPI_AUDEC_CONNECT
    pfunc_tsd = (FAPEXFDL_TSD_SubStrT*)FAPEXCFDL_GetFdlHdl(pfunc->pCfdlHdl, pfunc->input_id);

    if(pfunc_tsd != NULL) {
        if(pfunc_tsd->pTSDCtxHdl != NULL) {
            /* clear PTS notifer */
            ret = tsd_funcStr.SetPtsNotifier(pfunc_tsd->pTSDCtxHdl, NULL, NULL, 0);
            if(ret != FAPI_OK) {
                goto fdlaudec_delete_end;
            }
        }
    }
    pfunc->input_id  = -1;
    pfunc->output_id = -1;

    if(pfunc->pFAPICtxHdl != NULL) {
        /* call Close FAPI */
        ret = audec_funcStr.Close(pfunc->pFAPICtxHdl);
    }
#endif /* FAPI_AUDEC_CONNECT */

    pfunc->open_flag = FAPEXFDL_AUDEC_SET_NONE;
    pfunc->status    = FAPEX_STATUS_TERMINATED;

fdlaudec_delete_end:
#ifdef __LINUX__
    pfunc->close_result = ret;
#else
    return ret;
#endif
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLAUDEC_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_AUDEC_StrT  *pfunc;
    FAPEXFDLRELRET_T     ret = 0;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_AUDEC_StrT*)(file->private_data);

    if(pfunc != NULL) {
        /* decrement the count on our device */
        pfunc->used_count--;
#ifdef __LINUX__
        ret = kref_put(&pfunc->kref, fdlaudec_delete);
        if(ret != 0) {
            ret = pfunc->close_result;
        }
#else
        ret = fdlaudec_delete(pfunc);
#endif
        FAPEXDebugPrint(3, "[FAPEXFDLAUDEC_release] : used_count %d\n",
                           pfunc->used_count);
    }
    FAPEXFDL_RETVAL(ret);
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlaudec_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLAUDEC_init(void)
{
    int32_t ret;
    int32_t count;
    FAPEXFDL_AUDEC_StrT  *pfunc;
    FAPEXCFDL_UtilStrT   utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_AUDEC_StrT  *pfunc0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLAUDEC_init]\n");

    /* set struct file_operations */
    fapex_fdlaudec_fops.owner   = THIS_MODULE;
    fapex_fdlaudec_fops.open    = FAPEXFDLAUDEC_open;
//  fapex_fdlaudec_fops.read    = FAPEXFDLAUDEC_read;
//  fapex_fdlaudec_fops.write   = FAPEXFDLAUDEC_write;
    fapex_fdlaudec_fops.ioctl   = FAPEXFDLAUDEC_ioctl;
//  fapex_fdlaudec_fops.poll    = FAPEXFDLAUDEC_poll;
//  fapex_fdlaudec_fops.mmap    = FAPEXFDLAUDEC_mmap;
    fapex_fdlaudec_fops.release = FAPEXFDLAUDEC_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc = (FAPEXFDL_AUDEC_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_AUDEC_StrT)*(FAPEXDEV_AUDEC_DESCRIPTOR_NUM));
    if(pfunc == NULL) {
        return FAPEX_AUDEC_ERR_OUT_OF_MEMORY;
    }
    pfunc0 = pfunc;
#else
    pfunc = FAPIGLD_FDLAUDECCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc, sizeof(FAPEXFDL_AUDEC_StrT)*(FAPEXDEV_AUDEC_DESCRIPTOR_NUM));

    for(count=FAPEXDEV_AUDEC_MINOR_FIRST; count<=FAPEXDEV_AUDEC_MINOR_LAST; count++){
        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlaudec_fops, pfunc, &utility, count);
        if(ret) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            FAPEXDebugPrint(1, "[FAPEXFDLAUDEC_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            return ret;
        }
        /* init sequence (common this FDL) */
//      pfunc->status    = FAPEX_STATUS_LOADED;
        pfunc->input_id  = -1;
        pfunc->output_id = -1;

        pfunc ++;
    }

#ifdef FAPI_AUDEC_CONNECT
    FAPEXDebugPrint(1, "call FAPEXCFG_AUDEC_Init\n");

    /* call FAPI (initialize) */
    ret = FAPEXCFG_AUDEC_Init();
#else
    ret = 0;
#endif /* FAPI_AUDEC_CONNECT */

#ifndef CONFIG_AUDEC_FIRMWARE_REBOOT
    if(audec_funcStr.firmware_type != 0)    /* only embedded type */
#endif
    {
        /* setup firmware boot flag */
        firmware_boot_sem = FAPI_SYS_CREATE_SEMAPHORE(1);
        if(firmware_boot_sem == (FAPI_SYS_SemaphoreT)0) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            return FAPEX_AUDEC_ERR_SEMAPHORE_CREATE;
        }
        firmware_boot_flag = 0;
    }

    FAPEXDebugPrint(1, "[FAPEXFDLAUDEC_init] fapex_fdlaudec was loaded. %d\n", ret);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLAUDEC_cleanup(void)
{
    int32_t count;

    FAPEXDebugPrint(1, "[FAPEXFDLAUDEC_cleanup]\n");

#ifndef CONFIG_AUDEC_FIRMWARE_REBOOT
    if(audec_funcStr.firmware_type != 0)    /* only embedded type */
#endif
    {
        /* cleanup sequence (common this FDL) */
        FAPI_SYS_DESTROY_SEMAPHORE(firmware_boot_sem);
        firmware_boot_flag = 0;
    }

#ifdef FAPI_AUDEC_CONNECT
    /* call FAPI API (exit) */
    FAPEXCFG_AUDEC_Cleanup();

    FAPEXDebugPrint(1, "end FAPEXCFG_AUDEC_Cleanup\n");
#endif /* FAPI_AUDEC_CONNECT */

#ifdef ALLOCATE_DYNAMIC_MEMORY
    {
        void                *pCfdlHdl;
        FAPEXFDL_AUDEC_StrT *pfunc;

        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        pfunc    = (FAPEXFDL_AUDEC_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_AUDEC_MINOR_FIRST);

        /* relase the FDL context */
        if(pfunc != NULL) {
            FAPI_SYS_FREE(pfunc);
        }
    }
#endif /* ALLOCATE_DYNAMIC_MEMORY */

    for(count=FAPEXDEV_AUDEC_MINOR_FIRST; count<=FAPEXDEV_AUDEC_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLAUDEC_cleanup] fapex_fdlaudec was unloaded.\n");
    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLAUDEC_init);
module_exit(FAPEXFDLAUDEC_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlaudec object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
