/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv
*/

#include "fapex.h"
#include "fapex/abst_libauout.h"
#include "fapex/sys_config.h"
#include "fapi/drv_intr.h"
#include "fapi/reg_ao.h"
#include "fapex_fdlauout_priv.h"
#include "fapex_cfdlsaf.h"
#include "fapex_config_cmn.h"

/* for auout_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_AUOUT_StrT  FAPIGLD_FDLAUOUTCtx[FAPEXDEV_AUOUT_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

static FAPEXCFDLSAF_SpnLockStrT  auout_isr_lock;

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLAUOUT_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                     FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_read] size:%d\n",size);

    /* actual code which is depend on function is implimented here. */

    return size;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLAUOUT_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                     FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_write] size:%d\n",size);

    /* actual code which is depend on function is implimented here. */

    return size;
}
#endif

/*!
*******************************************************************************
**
** \brief  set stream information
**
** \param  pfunc    pointer to AUOUT context
**
******************************************************************************/
static void fdlauout_SetStreamInfo(FAPEXFDL_AUOUT_StrT *pfunc)
{
    FAPI_AUOUT_StreamInfoT  params;
    uint32_t  flags = 1;

    /* setup parameter */
    params.type       = (FAPI_AUOUT_StreamEnumT)pfunc->stream_type;
    params.sampleFreq = (FAPI_AUOUT_SampleFreqEnumT)pfunc->pcm_info.sampleFreq;
    params.lfeOn      = pfunc->channel_flag & FAPEX_AUOUT_CHANNEL_LFE;
    params.sampleBits = pfunc->pcm_info.sampleBits;

    switch(pfunc->channel_flag) {
        case FAPEX_AUOUT_CHANNEL_CENTER:
            params.channels = FAPI_AUOUT_CHANNEL_C;
            break;
        case FAPEX_AUOUT_CHANNEL_LEFT:
        case FAPEX_AUOUT_CHANNEL_RIGHT:
            params.channels = FAPI_AUOUT_CHANNEL_MONO;
            break;
        case FAPEX_AUOUT_CHANNEL_STEREO:
            if(pfunc->channel_flag == pfunc->dualmono_flag) {
                params.channels = FAPI_AUOUT_CHANNEL_DUALMONO;
            }
            else {
                params.channels = FAPI_AUOUT_CHANNEL_STEREO;
            }
            break;
        case FAPEX_AUOUT_CHANNEL_2P1CH:
            params.channels = FAPI_AUOUT_CHANNEL_L_R_S;
            break;
        case FAPEX_AUOUT_CHANNEL_3CH:
            params.channels = FAPI_AUOUT_CHANNEL_L_C_R;
            break;
        case FAPEX_AUOUT_CHANNEL_3P1CH:
            break;
        case FAPEX_AUOUT_CHANNEL_2S2CH:
            params.channels = FAPI_AUOUT_CHANNEL_L_R_SL_SR;
            break;
        case FAPEX_AUOUT_CHANNEL_4P1CH:
            params.channels = FAPI_AUOUT_CHANNEL_L_C_R_SL_SR;
            break;
        case FAPEX_AUOUT_CHANNEL_3S2CH:
            params.channels = FAPI_AUOUT_CHANNEL_L_C_R_S;
            break;
        case FAPEX_AUOUT_CHANNEL_5P1CH:
        case FAPEX_AUOUT_CHANNEL_7P1CH:
        default:
            params.channels = FAPI_AUOUT_CHANNEL_UNDEF;
    }

    FAPEXCFDLSAF_SPIN_LOCK(&auout_isr_lock, &flags);

    /* call FAPI_AUOUT_SetStreamInfo() */
    auout_funcStr.SetStreamInfo(pfunc->pFAPICtxHdl, &params);

    FAPEXCFDLSAF_SPIN_UNLOCK(&auout_isr_lock, &flags);
}

/*!
*******************************************************************************
**
** \brief  AUOUT pause
**
** \param  pfunc    pointer to context
**
******************************************************************************/
static int32_t fdlauout_Pause(FAPEXFDL_AUOUT_StrT *pfunc)
{
    int32_t             ret = 0;
    FAPI_SYS_HandleT    spdifBufferHandle = NULL;

    /* check param */
    if(NULL == pfunc)
    {
        return -1;
    }

    if(NULL == pfunc->pFAPICtxHdl)
    {
        return -1;
    }

    /*  disable I2S output and reset SPDIF BM write/read pointer */
    auout_funcStr.Stop(pfunc->pFAPICtxHdl);
    FAPEXDebugPrint(1,"[fdlauout_Pause] AUOUT_HANDLE:[0x%08x]\n",pfunc->pFAPICtxHdl);

    /* stop SPDIF correctly */
    spdifBufferHandle = auout_funcStr.GetSpdifBuffer(pfunc->pFAPICtxHdl);
    if ( spdifBufferHandle != NULL )
    {
        (void)bm_funcStr.SetReadEnable(spdifBufferHandle, 0);
    }
    
    return ret;
}

/*!
*******************************************************************************
**
** \brief  AUOUT reset
**
** \param  pfunc    pointer to context
**
******************************************************************************/
static int32_t fdlauout_Reset(FAPEXFDL_AUOUT_StrT *pfunc)
{
    int32_t             ret = 0;
    FAPI_SYS_HandleT    spdifBufferHandle = NULL;

    /* check param */
    if(NULL == pfunc)
    {
        return -1;
    }

    if(NULL == pfunc->pFAPICtxHdl)
    {
        return -1;
    }

    //audec_funcStr.InvalidatePts(pathPtr->handleDec);

    /* enable I2S output and reset SPDIF BM write/read pointer */
    ret = auout_funcStr.Start(pfunc->pFAPICtxHdl);
    FAPEXDebugPrint(1,"[fdlauout_Reset] AUOUT_HANDLE:[0x%08x], ret:[%d]\n",pfunc->pFAPICtxHdl,ret);

    if(FAPI_OK == ret)
    {
        /* start SPDIF correctly */
        spdifBufferHandle = auout_funcStr.GetSpdifBuffer(pfunc->pFAPICtxHdl);
        if ( spdifBufferHandle != NULL )
        {
            (void)bm_funcStr.SetReadEnable(spdifBufferHandle, 1);
        }
    }

    return ret;
}

static int32_t fdlauout_Starthdmi(FAPEXFDL_AUOUT_StrT *pfunc)
{
    int32_t ret = 0;

    /* check param */
    if(NULL == pfunc)
    {
        return -1;
    }

    if(NULL == pfunc->pFAPICtxHdl)
    {
        return -1;
    }

    ret = auout_funcStr.StartHdmi(pfunc->pFAPICtxHdl,
                               FAPI_AUOUT_I2S_MASK | FAPI_AUOUT_SPDIF_MASK);

    return ret;
}

static int32_t fdlauout_Stophdmi(FAPEXFDL_AUOUT_StrT *pfunc)
{
    int32_t ret = 0;

    /* check param */
    if(NULL == pfunc)
    {
        return -1;
    }

    if(NULL == pfunc->pFAPICtxHdl)
    {
        return -1;
    }

    auout_funcStr.StopHdmi();

    return ret;
}
/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLAUOUT_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                     FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_AUOUT_StrT *pfunc;
    int32_t  ret = FAPEX_AUOUT_ERR_BAD_PARAMETER;

    FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_ioctl] cmd:%08x\n",cmd);

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_AUOUT_StrT*)(file->private_data);

    switch(cmd) {
#ifdef FAPI_AUOUT_CONNECT
        /* actual code which is depend on function is implimented here. */
        case FAPEX_IOCCOM_START:
            if((pfunc->enable_flag & FAPEX_AUOUT_ENABLE_TYPE_MASK) == FAPEX_AUOUT_ENABLE_SPDIF) {
                /* call AUOUT start() */
                ret = auout_funcStr.Start(pfunc->pFAPICtxHdl);
                if(ret == FAPI_OK) {
                    pfunc->status = FAPEX_STATUS_RUNNING;
                }
            }
            else {
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_STOP:
            if((pfunc->enable_flag & FAPEX_AUOUT_ENABLE_TYPE_MASK) == FAPEX_AUOUT_ENABLE_SPDIF) {
                /* call AUOUT stop() */
                auout_funcStr.Stop(pfunc->pFAPICtxHdl);

                pfunc->status = FAPEX_STATUS_READY;
            }
            else {
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_PAUSE:
            if( pfunc != NULL ) {
                ret = fdlauout_Pause(pfunc);
            }
            break;

        case FAPEX_IOCCOM_RESET:
            if( pfunc != NULL ) {
                ret = fdlauout_Reset(pfunc);
            }
            break;

        case FAPEX_IOCCOM_AUOUT_OPEN:
            if( (arg != (FAPEXFDLIOCARG4_T)NULL)
             && (pfunc->pFAPICtxHdl == NULL) ) {
                FAPI_AUOUT_OpenParamsT     params;
                FAPEX_AUOUT_OpenParamStrT  *pArg = (FAPEX_AUOUT_OpenParamStrT *)arg;

                /* copy parameters from an user space to the kernel space */
                FAPEXCFDLSAF_COPY_TO_DRV(&(pfunc->enable_flag),  &(pArg->enableFlag),  sizeof(uint32_t));
                FAPEXCFDLSAF_COPY_TO_DRV(&(pfunc->channel_flag), &(pArg->channelFlag), sizeof(uint32_t));

                /* setup open parameter */
                params.version      = FAPI_AUOUT_VERSION;
                params.enableSPDIF  = pfunc->enable_flag & FAPEX_AUOUT_ENABLE_SPDIF;
                params.pcmOutBuffer = NULL;

                if( (pfunc->enable_flag & FAPEX_AUOUT_ENABLE_I2S)
                 && (pfunc->channel_flag != 0) ) {
                    uint32_t  speaker, mask, cnt;

                    cnt     = 0;
                    speaker = FAPEX_AUOUT_SPEAKER_MIN_ID;
                    for(mask=FAPEX_AUOUT_CHANNEL_MIN_ID; mask<=FAPEX_AUOUT_CHANNEL_MAX_ID; mask<<=1) {
                        if(pfunc->channel_flag & mask) {
                            params.speakerArr[cnt++] = (FAPI_AUOUT_SpeakerEnumT)speaker;
                        }
                        speaker ++;
                    }
                    params.speakerArr[cnt] = FAPI_AUOUT_SPEAKER_UNDEF;
                }
                else {
                    params.speakerArr[0] = FAPI_AUOUT_SPEAKER_UNDEF;
                }

                /* call AUOUT open() */
                pfunc->pFAPICtxHdl = auout_funcStr.Open(&params, &ret);
                FAPEXDebugPrint(3, "[fdlauout_Open] FAPI context:0x%08x,ret=%xh \n",
                                   (uint32_t)pfunc->pFAPICtxHdl,ret);
                if(pfunc->pFAPICtxHdl == NULL) {
                    FAPEXDebugPrint(2, "[FAPEXFDLAUOUT_ioctl] FAPI_AUOUT_Open error:%d\n", ret);
                    pfunc->enable_flag  = 0;
                    pfunc->channel_flag = 0;
                }
            }
            else if(pfunc->pFAPICtxHdl != NULL) {
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUOUT_SET_PCM_INFO:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUOUT_PcmInfoStrT *pArg = (FAPEX_AUOUT_PcmInfoStrT *)arg;

                /* copy parameters from an user space to the kernel space */
                FAPEXCFDLSAF_COPY_TO_DRV(&(pfunc->pcm_info), pArg, sizeof(FAPEX_AUOUT_PcmInfoStrT));
                pfunc->enable_sinfo |= FAPEXFDL_AUOUT_SET_PCMINFO;
            }
            if(pfunc->enable_sinfo & FAPEXFDL_AUOUT_SET_STREAMTYPE) {
                fdlauout_SetStreamInfo(pfunc);
            }
            ret = 0;
            break;

        case FAPEX_IOCCOM_AUOUT_SET_STREAM_TYPE:
            pfunc->stream_type   = (int32_t)arg;
            pfunc->enable_sinfo |= FAPEXFDL_AUOUT_SET_STREAMTYPE;

            if(pfunc->enable_sinfo & FAPEXFDL_AUOUT_SET_PCMINFO) {
                fdlauout_SetStreamInfo(pfunc);
            }
            ret = 0;
            break;

        case FAPEX_IOCCOM_AUOUT_SET_MUTE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUOUT_MuteParamStrT  params, *pArg = (FAPEX_AUOUT_MuteParamStrT *)arg;
                uint32_t  speaker;
                uint32_t  mask = FAPEX_AUOUT_CHANNEL_MIN_ID;

                /* copy parameters from an user space to the kernel space */
                FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_AUOUT_MuteParamStrT));

                params.muteFlag &= pfunc->channel_flag;
                for(speaker = FAPEX_AUOUT_SPEAKER_MIN_ID; speaker <= FAPEX_AUOUT_SPEAKER_MAX_ID; speaker++) {
                    if(params.channelFlag & mask) {
                        if(params.muteFlag == 0) {
                            auout_funcStr.Mute(pfunc->pFAPICtxHdl, (FAPI_AUOUT_SpeakerEnumT)speaker);
                        }
                        else {
                            auout_funcStr.UnMute(pfunc->pFAPICtxHdl, (FAPI_AUOUT_SpeakerEnumT)speaker);
                        }
                    }
                    mask <<= 1;
                }
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUOUT_SET_VOLUME:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUOUT_VolumeParamStrT  params, *pArg = (FAPEX_AUOUT_VolumeParamStrT *)arg;
                uint32_t  speaker;
                uint32_t  mask = FAPEX_AUOUT_CHANNEL_MIN_ID;

                /* copy parameters from an user space to the kernel space */
                FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_AUOUT_VolumeParamStrT));

                if(params.channelFlag == FAPEX_AUOUT_CHANNEL_ALL) {
                    auout_funcStr.SetVolume(pfunc->pFAPICtxHdl, FAPI_AUOUT_SPEAKER_ALL, params.volume);
                }
                else {
                    params.channelFlag &= pfunc->channel_flag;
                    for(speaker = FAPEX_AUOUT_SPEAKER_MIN_ID; speaker <= FAPEX_AUOUT_SPEAKER_MAX_ID; speaker++) {
                        if(params.channelFlag & mask) {
                            auout_funcStr.SetVolume(pfunc->pFAPICtxHdl,
                                                    (FAPI_AUOUT_SpeakerEnumT)speaker,
                                                    params.volume);
                        }
                        mask <<= 1;
                    }
                }
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUOUT_SET_DUALMONO:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUOUT_DualMonoParamStrT  params, *pArg = (FAPEX_AUOUT_DualMonoParamStrT *)arg;
                uint32_t  speaker;
                uint32_t  mask = FAPEX_AUOUT_CHANNEL_MIN_ID;

                /* copy parameters from an user space to the kernel space */
                FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_AUOUT_DualMonoParamStrT));

                params.channelFlag &= pfunc->channel_flag;
                if(params.enableFlag != 0) {
                    for(speaker = FAPEX_AUOUT_SPEAKER_MIN_ID; speaker <= FAPEX_AUOUT_SPEAKER_MAX_ID; speaker++) {
                        if(params.channelFlag & mask) {
                            auout_funcStr.EnableDualMono(pfunc->pFAPICtxHdl,
                                                         (FAPI_AUOUT_SpeakerEnumT)speaker);
                            pfunc->dualmono_flag |= mask;
                        }
                    }
                    mask <<= 1;
                }
                else {
                    for(speaker = FAPEX_AUOUT_SPEAKER_MIN_ID; speaker <= FAPEX_AUOUT_SPEAKER_MAX_ID; speaker++) {
                        if(params.channelFlag & mask) {
                            auout_funcStr.DisableDualMono(pfunc->pFAPICtxHdl,
                                                          (FAPI_AUOUT_SpeakerEnumT)speaker);
                            pfunc->dualmono_flag &= ~mask;
                        }
                    }
                    mask <<= 1;
                }
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUOUT_GET_PCM_INFO:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUOUT_PcmInfoStrT params;
                FAPI_AUOUT_StreamInfoT  info;

                info = auout_funcStr.GetStreamInfo(pfunc->pFAPICtxHdl);
                params.sampleFreq   = info.sampleFreq;
                params.sampleBits   = info.sampleBits;
                params.channelCount = info.channels;

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((void *)arg, &params, sizeof(FAPEX_AUOUT_PcmInfoStrT));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUOUT_GET_STREAM_TYPE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                int32_t                 type;
                FAPI_AUOUT_StreamInfoT  info;

                info = auout_funcStr.GetStreamInfo(pfunc->pFAPICtxHdl);
                type = info.type;

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((void *)arg, &type, sizeof(int32_t));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUOUT_GET_VOLUME:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUOUT_VolumeParamStrT params, *pArg = (FAPEX_AUOUT_VolumeParamStrT *)arg;

                /* copy parameters from an user space to the kernel space */
                FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_AUOUT_VolumeParamStrT));

                params.volume       = 0;
                params.channelFlag &= pfunc->channel_flag;
                if(params.channelFlag != 0) {
                    uint32_t  speaker;
                    uint32_t  value;
                    uint32_t  mask = FAPEX_AUOUT_CHANNEL_MIN_ID;
                    int32_t   chk_flag = 0;

                    for(speaker = FAPEX_AUOUT_SPEAKER_MIN_ID; speaker <= FAPEX_AUOUT_SPEAKER_MAX_ID; speaker++) {
                        if(params.channelFlag & mask) {
                            value = auout_funcStr.GetVolume(pfunc->pFAPICtxHdl,
                                                            (FAPI_AUOUT_SpeakerEnumT)speaker);
                            if(chk_flag == 0) {
                                params.volume = value;
                                chk_flag = 1;
                            }
                            else if(value != params.volume) {
                                params.channelFlag &= ~mask;
                            }
                        }
                        mask <<= 1;
                    }
                }

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((void *)arg, &params, sizeof(FAPEX_AUOUT_VolumeParamStrT));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUOUT_GET_SPEAKERS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUOUT_SpeakersParamStrT  params;
                uint32_t  speaker, speakerArr[FAPI_AUOUT_SPEAKER_ALL];

                params.channelCount = auout_funcStr.GetSpeakers(pfunc->pFAPICtxHdl,
                                                                (FAPI_AUOUT_SpeakerEnumT*)speakerArr);
                params.channelFlag = 0;
                for(speaker = FAPEX_AUOUT_SPEAKER_MIN_ID; speaker <= FAPEX_AUOUT_SPEAKER_MAX_ID; speaker++) {
                    if((FAPI_AUOUT_SpeakerEnumT)speakerArr[speaker] == FAPI_AUOUT_SPEAKER_UNDEF) {
                        break;
                    }
                    params.channelFlag |= (1 << speakerArr[speaker]);
                }

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((void *)arg, &params, sizeof(FAPEX_AUOUT_SpeakersParamStrT));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_AUOUT_GET_I2SBUFFERS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_AUOUT_BufferParamStrT  params;
                FAPI_AUOUT_I2sEnumT    buffer;
                FAPI_AUOUT_I2sBufferT  bufferArr[FAPI_AUOUT_I2S_ALL];

                params.bufferCount = auout_funcStr.GetI2sBuffers(pfunc->pFAPICtxHdl, bufferArr);

                for(buffer = (FAPI_AUOUT_I2sEnumT)0; buffer < FAPI_AUOUT_I2S_ALL; buffer++) {
                    params.channelFlag[buffer] = 0;
                    if(bufferArr[buffer].leftCh != FAPI_AUOUT_SPEAKER_UNDEF) {
                        params.channelFlag[buffer] |= (1 << bufferArr[buffer].leftCh);
                    }
                    if(bufferArr[buffer].rightCh != FAPI_AUOUT_SPEAKER_UNDEF) {
                        params.channelFlag[buffer] |= (1 << bufferArr[buffer].rightCh);
                    }
                }

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((void *)arg, &params, sizeof(FAPEX_AUOUT_BufferParamStrT));
                ret = 0;
            }
            break;
        case FAPEX_IOCCOM_AUOUT_START_HDMI:
            if( pfunc != NULL ) {
                ret = fdlauout_Starthdmi(pfunc);
            }
            break;
        case FAPEX_IOCCOM_AUOUT_STOP_HDMI:
            if( pfunc != NULL ) {
                ret = fdlauout_Stophdmi(pfunc);
            }
            break;
#endif /* FAPI_AUOUT_CONNECT */

        default:
            ;
    }

    return ret;
}

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLAUOUT_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T mask = 0;

    FAPEXDebugPrint(4, "[FAPEXFDLAUOUT_poll]\n");

    /* actual code which is depend on function is implimented here. */

    return mask;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLAUOUT_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    int32_t ret;

    if (vma->vm_pgoff != 0){
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_IO | VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if (file->f_flags & O_SYNC){
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
                  vma->vm_end - vma->vm_start, vma->vm_page_prot);

    if(ret != 0){
        return -EAGAIN;
    }
    return 0;
#else
    return -1;
#endif /* __LINUX__ */
}
#endif

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLAUOUT_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_AUOUT_StrT *pfunc;
    void    *pCfdlHdl;
    int32_t id;

    FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_open] file:0x%08x,private_data:0x%08x\n",
                       (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc = (FAPEXFDL_AUOUT_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc;
        FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc));

        /* set parameter */
        pfunc->pCfdlHdl = pCfdlHdl;
        pfunc->id       = id;
        pfunc->status   = FAPEX_STATUS_OPENED;
#ifdef __LINUX__
        kref_init(&pfunc->kref);
#endif
    }
    else {
        pfunc = (FAPEXFDL_AUOUT_StrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_open] fapex_fdlauout was opened : used_counter %d.\n",
                       pfunc->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlauout_delete(struct kref *kref)
{
    FAPEXFDL_AUOUT_StrT  *pfunc;
    int32_t  ret = 0;

    pfunc = container_of(kref, FAPEXFDL_AUOUT_StrT, kref);
#else
static int32_t fdlauout_delete(FAPEXFDL_AUOUT_StrT *pfunc)
{
    int32_t  ret = 0;
#endif

    if(pfunc->pFAPICtxHdl != NULL) {
#ifdef FAPI_AUOUT_CONNECT
        /* close sequence */
        ret = auout_funcStr.Close(pfunc->pFAPICtxHdl);
        if(ret != FAPI_OK) {
            goto fdlauout_delete_end;
        }
#endif /* FAPI_AUOUT_CONNECT */
        pfunc->pFAPICtxHdl = NULL;
    }
    pfunc->status = FAPEX_STATUS_TERMINATED;

fdlauout_delete_end:
#ifdef __LINUX__
    pfunc->close_result = ret;
#else
    return ret;
#endif
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLAUOUT_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_AUOUT_StrT *pfunc;
    FAPEXFDLRELRET_T    ret = 0;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_AUOUT_StrT*)(file->private_data);

    if(pfunc != NULL) {
        /* decrement the count on our device */
        pfunc->used_count--;
#ifdef __LINUX__
        ret = kref_put(&pfunc->kref, fdlauout_delete);
        if(ret != 0) {
            ret = pfunc->close_result;
        }
#else
        ret = fdlauout_delete(pfunc);
#endif
        FAPEXDebugPrint(3, "[FAPEXFDLAUOUT_release] : used_count %d\n",
                           pfunc->used_count);
    }
    FAPEXFDL_RETVAL(ret);
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlauout_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLAUOUT_init(void)
{
    int32_t ret;
    int32_t count;
    FAPEXFDL_AUOUT_StrT  *pfunc;
    FAPEXCFDL_UtilStrT   utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_AUOUT_StrT  *pfunc0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLAUOUT_init]\n");

    /* set struct file_operations */
    fapex_fdlauout_fops.owner   = THIS_MODULE;
    fapex_fdlauout_fops.open    = FAPEXFDLAUOUT_open;
//  fapex_fdlauout_fops.read    = FAPEXFDLAUOUT_read;
//  fapex_fdlauout_fops.write   = FAPEXFDLAUOUT_write;
    fapex_fdlauout_fops.ioctl   = FAPEXFDLAUOUT_ioctl;
//  fapex_fdlauout_fops.poll    = FAPEXFDLAUOUT_poll;
//  fapex_fdlauout_fops.mmap    = FAPEXFDLAUOUT_mmap;
    fapex_fdlauout_fops.release = FAPEXFDLAUOUT_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc = (FAPEXFDL_AUOUT_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_AUOUT_StrT)*(FAPEXDEV_AUOUT_DESCRIPTOR_NUM));
    if(pfunc == NULL) {
        return FAPEX_AUOUT_ERR_OUT_OF_MEMORY;
    }
    pfunc0 = pfunc;
#else
    pfunc = FAPIGLD_FDLAUOUTCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc, sizeof(FAPEXFDL_AUOUT_StrT)*(FAPEXDEV_AUOUT_DESCRIPTOR_NUM));

    for(count=FAPEXDEV_AUOUT_MINOR_FIRST; count<=FAPEXDEV_AUOUT_MINOR_LAST; count++){
        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlauout_fops, pfunc, &utility, count);
        if(ret) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            FAPEXDebugPrint(1, "[FAPEXFDLAUOUT_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            return ret;
        }
        pfunc ++;
    }

#ifdef FAPI_AUOUT_CONNECT
    FAPEXDebugPrint(1, "call FAPEXCFG_AUOUT_Init\n");

    /* call FAPI (initialize) */
    ret = FAPEXCFG_AUOUT_Init();
#else
    ret = 0;
#endif /* FAPI_AUOUT_CONNECT */

    FAPEXDebugPrint(1, "[FAPEXFDLAUOUT_init] fapex_fdlauout was loaded. %d\n", ret);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLAUOUT_cleanup(void)
{
    int32_t count;

    FAPEXDebugPrint(1, "[FAPEXFDLAUOUT_cleanup]\n");

#ifdef FAPI_AUOUT_CONNECT
    /* call FAPI API (cleanup) */
    FAPEXCFG_AUOUT_Cleanup();

    FAPEXDebugPrint(1, "end FAPEXCFG_AUOUT_Cleanup\n");
#endif /* FAPI_AUOUT_CONNECT */

#ifdef ALLOCATE_DYNAMIC_MEMORY
    {
        void                *pCfdlHdl;
        FAPEXFDL_AUOUT_StrT *pfunc;

        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        pfunc    = (FAPEXFDL_AUOUT_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_AUOUT_MINOR_FIRST);

        /* relase the FDL context */
        if(pfunc != NULL) {
            FAPI_SYS_FREE(pfunc);
        }
    }
#endif /* ALLOCATE_DYNAMIC_MEMORY */

    /* cleanup sequence (common this FDL) */
    for(count=FAPEXDEV_AUOUT_MINOR_FIRST; count<=FAPEXDEV_AUOUT_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLAUOUT_cleanup] fapex_fdlauout was unloaded.\n");
    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLAUOUT_init);
module_exit(FAPEXFDLAUOUT_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlauout object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
