/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv (boardifc)
*/

#include "fapex.h"
#include "fapex/abst_libboardifc.h"
#include "fapex/sys_config.h"
#include "fapex_fdlboardifc_priv.h"
#include "fapex_config_cmn.h"

/* for boardifc_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_BOARDIFC_StrT  FAPIGLD_FDLBOARDIFCCtx[FAPEXDEV_BOARDIFC_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */


/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLBOARDIFC_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                        FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_BOARDIFC_StrT *pfunc;
    int32_t  ret = FAPEX_BOARDIFC_ERR_BAD_PARAMETER;
    FAPEX_BOARD_HwPlatformInfosStrT *hwInfoPtr;
    char* tmp1,*tmp2;
    void *tmp3;

    FAPEXDebugPrint(3, "[FAPEXFDLBOARDIFC_ioctl] cmd:%d\n",cmd);

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_BOARDIFC_StrT*)(file->private_data);
    
    switch(cmd) 
    {
        /* actual code which is depend on function is implimented here. */
        /* case ...: */
        case FAPEX_IOCCOM_BOARDIFC_OPEN:
            FAPEXDebugPrint(3, "BOARDIFC_OPEN\n");ret=0;
            break;

        case FAPEX_IOCCOM_BOARDIFC_GETBOARDINFO:
            hwInfoPtr = (void*)board_funcStr.GetBoardInfo();
            ret=0;
            FAPEXDebugPrint(3, "GETBOARDINFO,size 1 = 0x%x, 2 = 0x%x\n",(int32_t)hwInfoPtr->pHwManufactString,(int32_t)hwInfoPtr->pHwModelString);
            tmp1=(char*)((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pHwManufactString;
            tmp2=(char*)((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pHwModelString;
            tmp3=(void*)((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pOptAdditionalInfo;
            FAPEXCFDLSAF_COPY_TO_USR((int32_t*)arg,(int32_t*)hwInfoPtr,sizeof(FAPEX_BOARD_HwPlatformInfosStrT));
            ((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pHwManufactString=tmp1;
            ((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pHwModelString=tmp2;
            ((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pOptAdditionalInfo=tmp3;
            if(hwInfoPtr->pHwManufactString )
            {
                FAPEXCFDLSAF_COPY_TO_USR((int32_t*)(((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pHwManufactString),(int32_t*)(hwInfoPtr->pHwManufactString),25);           
            }
            if(hwInfoPtr->pHwModelString)
            {
                FAPEXCFDLSAF_COPY_TO_USR((int32_t*)(((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pHwModelString),(int32_t*)(hwInfoPtr->pHwModelString),25);
            }
            if(hwInfoPtr->pOptAdditionalInfo)
            {
                FAPEXCFDLSAF_COPY_TO_USR((int32_t*)(((FAPEX_BOARD_HwPlatformInfosStrT*)arg)->pOptAdditionalInfo),(int32_t*)(hwInfoPtr->pOptAdditionalInfo),sizeof(hwInfoPtr->pOptAdditionalInfo));
            }
            break;

        case FAPEX_IOCCOM_BOARDIFC_INIT_DETECT_FUNCTION :
            board_funcStr.BoardServices.controlHook_BrdInitDetect(BOARDS_INIT_DETECT_DEFAULT,0);
            FAPEXDebugPrint(3, "INIT_DETECT_FUNCTION\n");
            ret = 0;
            break;
#if 0
        case FAPEX_IOCCOM_BOARDIFC_CONTROL_TEST_LED_KEY :
            break;

        case FAPEX_IOCCOM_BOARDIFC_CONTROL_FE :
            break;

        case FAPEX_IOCCOM_BOARDIFC_CONTROL_FE_DETECT :
            break;

        case FAPEX_IOCCOM_BOARDIFC_CONTROL_HD :
            break;

        case FAPEX_IOCCOM_BOARDIFC_CONTROL_SCART :
            FAPEXDebugPrint(3, "CONTROL_SCART\n");
            break;

        case FAPEX_IOCCOM_BOARDIFC_EXT_CONTROL_TS_PATH  :
            break;

        case FAPEX_IOCCOM_BOARDIFC_EXT_CONTROL_HDMI_PATH :
            break;

        case FAPEX_IOCCOM_BOARDIFC_DIAGNOSTIC_TS_PATH:
            break;

        case FAPEX_IOCCOM_BOARDIFC_DIAGNOSTIC_HDMI_PATH :
            break;

        case FAPEX_IOCCOM_BOARDIFC_DIAGNOSTIC_CI_CONTROL:
            break;
        case FAPEX_IOCCOM_BOARDIFC_DIAGNOSTIC_GENERALPURPOSE:
            break;
#endif
        case FAPEX_IOCCOM_PLATFORM_SUPPORT:
            boards_funcStr.BoardIfc_Register(BOARDS_REGISTER_DEFAULT,0);
            ret=0;
            FAPEXDebugPrint(3, "PLATFORM_SUPPORT\n");
            break;
        case FAPEX_IOCCOM_FRONTEND_SUPPORT:
            boards_funcStr.RegisterFrontends(BOARDS_REGISTER_DEFAULT,0);
            ret=0;
            FAPEXDebugPrint(3, "FRONTEND_SUPPORT\n");
            break;
        case FAPEX_IOCCOM_PLATFORM_SUPPORT_EXT:
            ret=0;
            FAPEXDebugPrint(3, "PLATFORM_SUPPORT_EXT\n");
            break;

        default:

        break;
    }

    return ret;
}


/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLBOARDIFC_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_BOARDIFC_StrT *pfunc;
    void    *pCfdlHdl;
    int32_t id;

    FAPEXDebugPrint(3, "[FAPEXFDLBOARDIFC_open] file:0x%08x,private_data:0x%08x\n",
                       (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLBOARDIFC_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLBOARDIFC_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc = (FAPEXFDL_BOARDIFC_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc;
        FAPEXDebugPrint(3, "[FAPEXFDLBOARDIFC_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc));

        /* open sequence */
        /* ... */

        /* set parameter */
        pfunc->pCfdlHdl = pCfdlHdl;
        pfunc->id       = id;
#ifdef __LINUX__
        kref_init(&pfunc->kref);
#endif
    }
    else {
        pfunc = (FAPEXFDL_BOARDIFC_StrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLBOARDIFC_open] fapex_fdlboardifc was opened : used_counter %d.\n",
                       pfunc->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlboardifc_delete(struct kref *kref)
{
    FAPEXFDL_BOARDIFC_StrT   *pfunc;

    pfunc = container_of(kref, FAPEXFDL_BOARDIFC_StrT, kref);
#else
static void fdlboardifc_delete(FAPEXFDL_BOARDIFC_StrT *pfunc)
{
#endif
    /* get id */
    //id = pfunc->id;

    /* close sequence */
    /* ... */
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLBOARDIFC_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_BOARDIFC_StrT *pfunc;
    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_BOARDIFC_StrT*)(file->private_data);

    if(pfunc != NULL) {
        /* decrement the count on our device */
        pfunc->used_count--;
#ifdef __LINUX__
        kref_put(&pfunc->kref, fdlboardifc_delete);
#else
        fdlboardifc_delete(pfunc);
#endif

        FAPEXDebugPrint(3, "[FAPEXFDLBOARDIFC_release] : used_count %d\n",
                           pfunc->used_count);
    }
    return 0;
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlboardifc_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLBOARDIFC_init(void)
{
    int32_t ret;
    int32_t count;
    FAPEXFDL_BOARDIFC_StrT  *pfunc;
    FAPEXCFDL_UtilStrT      utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_BOARDIFC_StrT  *pfunc0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLBOARDIFC_init]\n");

    /* set struct file_operations */
    fapex_fdlboardifc_fops.owner   = THIS_MODULE;
    fapex_fdlboardifc_fops.open    = FAPEXFDLBOARDIFC_open;
    //fapex_fdlboardifc_fops.read    = FAPEXFDLBOARDIFC_read;
    //fapex_fdlboardifc_fops.write   = FAPEXFDLBOARDIFC_write;
    fapex_fdlboardifc_fops.ioctl   = FAPEXFDLBOARDIFC_ioctl;
    //fapex_fdlboardifc_fops.poll    = FAPEXFDLBOARDIFC_poll;
    //fapex_fdlboardifc_fops.mmap    = FAPEXFDLBOARDIFC_mmap;
    fapex_fdlboardifc_fops.release = FAPEXFDLBOARDIFC_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc = (FAPEXFDL_BOARDIFC_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_BOARDIFC_StrT)*(FAPEXDEV_BOARDIFC_DESCRIPTOR_NUM));
    if(pfunc == NULL) {
        return FAPEX_BOARDIFC_ERR_OUT_OF_MEMORY;
    }
    pfunc0 = pfunc;
#else
    pfunc = FAPIGLD_FDLBOARDIFCCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc, sizeof(FAPEXFDL_BOARDIFC_StrT)*(FAPEXDEV_BOARDIFC_DESCRIPTOR_NUM));

    for(count=FAPEXDEV_BOARDIFC_MINOR_FIRST; count<=FAPEXDEV_BOARDIFC_MINOR_LAST; count++){
        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlboardifc_fops, pfunc, &utility, count);
        if(ret) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            FAPEXDebugPrint(1, "[FAPEXFDLBOARDIFC_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            return ret;
        }
        /* init sequence (common this FDL) */
        /* ... */

        pfunc ++;
    }

#ifdef FAPI_BOARDIFC_CONNECT
    FAPEXDebugPrint(1, "call FAPEXCFG_BOARD_Init\n");
    /* call FAPI (initialize) */
    ret = FAPEXCFG_BOARD_Init();
#else
    ret = 0;
#endif /* FAPI_BOARDIFC_CONNECT */

    FAPEXDebugPrint(1, "[FAPEXFDLBOARDIFC_init] fapex_fdlboardifc was loaded. %d\n", ret);
    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLBOARDIFC_cleanup(void)
{
    int32_t count;

    FAPEXDebugPrint(1, "[FAPEXFDLBOARDIFC_cleanup]\n");

#ifdef FAPI_BOARDIFC_CONNECT
    /* call FAPI API (cleanup) */
    FAPEXCFG_BOARD_cleanup();

    FAPEXDebugPrint(1, "end FAPEX_BOARDIFC_Cleanup\n");
#endif /* FAPI_BOARDIFC_CONNECT */

    /* cleanup sequence (common this FDL) */
#ifdef ALLOCATE_DYNAMIC_MEMORY
    {
        void                   *pCfdlHdl;
        FAPEXFDL_BOARDIFC_StrT *pfunc;

        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        pfunc    = (FAPEXFDL_BOARDIFC_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_BOARDIFC_MINOR_FIRST);

        /* relase the FDL context */
        if(pfunc != NULL) {
            FAPI_SYS_FREE(pfunc);
        }
    }
#endif /* ALLOCATE_DYNAMIC_MEMORY */

    for(count=FAPEXDEV_BOARDIFC_MINOR_FIRST; count<=FAPEXDEV_BOARDIFC_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLBOARDIFC_cleanup] fapex_fdlboardifc was unloaded.\n");

    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLBOARDIFC_init);
module_exit(FAPEXFDLBOARDIFC_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlboardifc object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
