/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv (clkpwr)
*/

#include "fapex.h"
#include "fapex/abst_libclkpwr.h"
#include "fapex/sys_config.h"
#include "fapex_fdlclkpwr_priv.h"
#include "fapex_config_cmn.h"

/* for clkpwr_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_CLKPWR_StrT  FAPIGLD_FDLCLKPWRCtx[FAPEXDEV_CLKPWR_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLCLKPWR_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                      FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_read] size:%d\n",size);

    return size;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLCLKPWR_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                      FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_write] size:%d\n",size);

    return size;
}
#endif

/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLCLKPWR_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                      FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_CLKPWR_StrT *pfunc;
    int32_t  ret = FAPEX_CLKPWR_ERR_BAD_PARAMETER;

    FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_ioctl] cmd:%08x\n",cmd);

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_CLKPWR_StrT*)(file->private_data);

    switch(cmd) {
#ifdef FAPI_CLKPWR_CONNECT
        /* actual code which is depend on function is implimented here. */
        case FAPEX_IOCCOM_CLKPWR_START:
           if(arg !=(FAPEXFDLIOCARG4_T) NULL) {
                uint64_t component_mask; 
                uint64_t* pArg =(uint64_t*)arg;
                FAPEXCFDLSAF_COPY_TO_DRV(&component_mask,pArg,sizeof(uint64_t));
                ret = clkpwr_funcStr.PowerUp(pfunc->pFAPICtxHdl,component_mask);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                      "FAPI_CLKPWR_Wakeup error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_CLKPWR_STOP:
           if(arg !=(FAPEXFDLIOCARG4_T) NULL) {
                uint64_t  component_mask;
                uint64_t* pArg =(uint64_t*)arg;
                FAPEXCFDLSAF_COPY_TO_DRV(&component_mask,pArg,sizeof(uint64_t));

#ifndef CONFIG_AUDEC_FIRMWARE_REBOOT
                if(audec_funcStr.firmware_type == 0) {  /* non embedded type */
                    /* don't shut down audio processor */
                    component_mask &= ~(FAPEX_CLKPWR_AUDIO_PROCESSOR);
                }
#endif
                ret = clkpwr_funcStr.PowerDown(pfunc->pFAPICtxHdl, component_mask);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                       "FAPI_CLKPWR_Standby error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_RESET:
            if(arg !=(FAPEXFDLIOCARG4_T) NULL) {
                uint64_t  component_mask;
                uint64_t* pArg =(uint64_t*)arg;
                FAPEXCFDLSAF_COPY_TO_DRV(&component_mask,pArg,sizeof(uint64_t));

                ret = clkpwr_funcStr.Reset(pfunc->pFAPICtxHdl, component_mask);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                      "FAPI_CLKPWR_Reset error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_OPEN:
            if(pfunc->pFAPICtxHdl == NULL) {
                FAPI_CLKPWR_OpenParamsT  params;

                /* open sequence */
                params.version = FAPI_CLKPWR_VERSION;
                pfunc->pFAPICtxHdl = clkpwr_funcStr.Open(&params, &ret);
                FAPEXDebugPrint(3, "[fdlclkpwr_Open] FAPI context:0x%08x\n",
                                   (uint32_t)pfunc->pFAPICtxHdl);
                if(pfunc->pFAPICtxHdl == NULL) {
                    FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                       "FAPI_CLKPWR_Open error:%d\n", ret);
                }
            }
            else {
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_CLKPWR_GET_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = clkpwr_funcStr.GetState(pfunc->pFAPICtxHdl);

                if(ret >= 0) {
                    /* copy parameters from the kernel space to an user space */
                    FAPEXCFDLSAF_COPY_TO_USR((int32_t*)(arg), &ret, sizeof(int32_t));
                    ret = 0;
                }
                else {
                    FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                       "FAPI_CLKPWR_GetState error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_CLKPWR_WATCHDOG_START:
            ret = clkpwr_funcStr.StartWatchdog(pfunc->pFAPICtxHdl, (uint32_t)arg);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                   "FAPI_CLKPWR_StartWatchdog error:%d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_CLKPWR_WATCHDOG_STOP:
            ret = clkpwr_funcStr.StopWatchdog(pfunc->pFAPICtxHdl);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                   "FAPI_CLKPWR_StopWatchdog error:%d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_CLKPWR_WATCHDOG_RESET:
            ret = clkpwr_funcStr.ResetWatchdog(pfunc->pFAPICtxHdl);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                   "FAPI_CLKPWR_ResetWatchdog error:%d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_CLKPWR_SET_PWM:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_CLKPWR_PwmParamStrT  params;

                FAPEXCFDLSAF_COPY_TO_DRV(&params,(FAPEX_CLKPWR_PwmParamStrT*)arg,sizeof(FAPEX_CLKPWR_PwmParamStrT)); 
                ret = clkpwr_funcStr.SetPwm(pfunc->pFAPICtxHdl,
                                            params.index,
                                            params.pwmMaximum,
                                            params.pwmSetting);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                   "FAPI_CLKPWR_SetPwm error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_CLKPWR_GET_PWM:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_CLKPWR_PwmParamStrT  params;

                FAPEXCFDLSAF_COPY_TO_DRV(&params,(FAPEX_CLKPWR_PwmParamStrT*)arg,sizeof(FAPEX_CLKPWR_PwmParamStrT)); 
                ret = clkpwr_funcStr.GetPwm(pfunc->pFAPICtxHdl,
                                            params.index,
                                            &params.pwmMaximum,
                                            &params.pwmSetting);

                if(ret == FAPI_OK) {
                    /* copy parameters from the kernel space to an user space */
                    FAPEXCFDLSAF_COPY_TO_USR((FAPEX_CLKPWR_PwmParamStrT*)(arg), &params, sizeof(FAPEX_CLKPWR_PwmParamStrT));
                }
                else {
                    FAPEXDebugPrint(2, "[FAPEXFDLCLKPWR_ioctl] "
                                       "FAPI_CLKPWR_GetPwm error:%d\n", ret);
                }
            }
            break;
#endif /* FAPI_CLKPWR_CONNECT */

        default:
            ;
    }

    return ret;
}

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLCLKPWR_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T mask = 0;

    FAPEXDebugPrint(4, "[FAPEXFDLCLKPWR_poll]\n");

    /* actual code which is depend on function is implimented here. */

    return mask;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLCLKPWR_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    int32_t ret;

    if (vma->vm_pgoff != 0){
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_IO | VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if (file->f_flags & O_SYNC){
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
                  vma->vm_end - vma->vm_start, vma->vm_page_prot);

    if(ret != 0){
        return -EAGAIN;
    }
    return 0;
#else
    return -1;
#endif /* __LINUX__ */
}
#endif

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLCLKPWR_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_CLKPWR_StrT *pfunc;
    void    *pCfdlHdl;
    int32_t id;

    FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_open] file:0x%08x,private_data:0x%08x\n",
                       (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc = (FAPEXFDL_CLKPWR_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc;
        FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc));

        /* set parameter */
        pfunc->pCfdlHdl = pCfdlHdl;
        pfunc->id       = id;
#ifdef __LINUX__
        kref_init(&pfunc->kref);
#endif
    }
    else {
        pfunc = (FAPEXFDL_CLKPWR_StrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_open] fapex_fdlclkpwr was opened : used_counter %d.\n",
                       pfunc->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlclkpwr_delete(struct kref *kref)
{
    FAPEXFDL_CLKPWR_StrT   *pfunc;
    int32_t  ret = 0;

    pfunc = container_of(kref, FAPEXFDL_CLKPWR_StrT, kref);
#else
static int32_t fdlclkpwr_delete(FAPEXFDL_CLKPWR_StrT *pfunc)
{
    int32_t  ret = 0;
#endif

    if(pfunc->pFAPICtxHdl != NULL) {
#ifdef FAPI_CLKPWR_CONNECT
        /* close sequence */
        ret = clkpwr_funcStr.Close(pfunc->pFAPICtxHdl);
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdlclkpwr_delete] FAPI_CLKPWR_Close error:%d\n",
                               ret);
            goto fdlclkpwr_delete_end;
        }
#endif /* FAPI_CLKPWR_CONNECT */
        FAPEXDebugPrint(3, "[fdlclkpwr_delete] FAPI_CLKPWR_Close 0x%08x\n",
                           (uint32_t)pfunc->pFAPICtxHdl);
        pfunc->pFAPICtxHdl = NULL;
    }
fdlclkpwr_delete_end:
#ifdef __LINUX__
    pfunc->close_result = ret;
#else
    return ret;
#endif
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLCLKPWR_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_CLKPWR_StrT  *pfunc;
    FAPEXFDLRELRET_T      ret = 0;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_CLKPWR_StrT*)(file->private_data);

    if(pfunc != NULL) {
        /* decrement the count on our device */
        pfunc->used_count--;
#ifdef __LINUX__
        ret = kref_put(&pfunc->kref, fdlclkpwr_delete);
        if(ret != 0) {
            ret = pfunc->close_result;
        }
#else
        ret = fdlclkpwr_delete(pfunc);
#endif

        FAPEXDebugPrint(3, "[FAPEXFDLCLKPWR_release] : used_count %d\n",
                           pfunc->used_count);
    }
    FAPEXFDL_RETVAL(ret);
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlclkpwr_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLCLKPWR_init(void)
{
    int32_t ret;
    int32_t count;
    FAPEXFDL_CLKPWR_StrT  *pfunc;
    FAPEXCFDL_UtilStrT    utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_CLKPWR_StrT  *pfunc0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLCLKPWR_init]\n");

    /* set struct file_operations */
    fapex_fdlclkpwr_fops.owner   = THIS_MODULE;
    fapex_fdlclkpwr_fops.open    = FAPEXFDLCLKPWR_open;
//  fapex_fdlclkpwr_fops.read    = FAPEXFDLCLKPWR_read;
//  fapex_fdlclkpwr_fops.write   = FAPEXFDLCLKPWR_write;
    fapex_fdlclkpwr_fops.ioctl   = FAPEXFDLCLKPWR_ioctl;
//  fapex_fdlclkpwr_fops.poll    = FAPEXFDLCLKPWR_poll;
//  fapex_fdlclkpwr_fops.mmap    = FAPEXFDLCLKPWR_mmap;
    fapex_fdlclkpwr_fops.release = FAPEXFDLCLKPWR_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc = (FAPEXFDL_CLKPWR_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_CLKPWR_StrT)*(FAPEXDEV_CLKPWR_DESCRIPTOR_NUM));
    if(pfunc == NULL) {
        return FAPEX_CLKPWR_ERR_OUT_OF_MEMORY;
    }
    pfunc0 = pfunc;
#else
    pfunc = FAPIGLD_FDLCLKPWRCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc, sizeof(FAPEXFDL_CLKPWR_StrT)*(FAPEXDEV_CLKPWR_DESCRIPTOR_NUM));

    for(count=FAPEXDEV_CLKPWR_MINOR_FIRST; count<=FAPEXDEV_CLKPWR_MINOR_LAST; count++){
        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlclkpwr_fops, pfunc, &utility, count);
        if(ret) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            FAPEXDebugPrint(1, "[FAPEXFDLCLKPWR_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            return ret;
        }
        /* init sequence (common this FDL) */
        /* ... */

        pfunc ++;
    }

#ifdef FAPI_CLKPWR_CONNECT
    // FAPEXDebugPrint(1, "call FAPEXCFG_CLKPWR_Init\n");

    /* call FAPI (initialize) */
    // ret = FAPEXCFG_CLKPWR_Init(); /* called from CFDL */
#else
    ret = 0;
#endif /* FAPI_CLKPWR_CONNECT */

    FAPEXDebugPrint(1, "[FAPEXFDLCLKPWR_init] fapex_fdlclkpwr was loaded. %d\n", ret);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLCLKPWR_cleanup(void)
{
    int32_t count;

    FAPEXDebugPrint(1, "[FAPEXFDLCLKPWR_cleanup]\n");

#ifdef FAPI_CLKPWR_CONNECT
    /* call FAPI API (cleanup) */
    // FAPEXCFG_CLKPWR_Cleanup(); /* called from CFDL */

    // FAPEXDebugPrint(1, "end FAPEXCFG_CLKPWR_Cleanup\n");
#endif /* FAPI_CLKPWR_CONNECT */

    /* cleanup sequence (common this FDL) */
#ifdef ALLOCATE_DYNAMIC_MEMORY
    {
        void                 *pCfdlHdl;
        FAPEXFDL_CLKPWR_StrT *pfunc;

        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        pfunc    = (FAPEXFDL_CLKPWR_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_CLKPWR_MINOR_FIRST);

        /* relase the FDL context */
        if(pfunc != NULL) {
            FAPI_SYS_FREE(pfunc);
        }
    }
#endif /* ALLOCATE_DYNAMIC_MEMORY */

    for(count=FAPEXDEV_CLKPWR_MINOR_FIRST; count<=FAPEXDEV_CLKPWR_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLCLKPWR_cleanup] fapex_fdlclkpwr was unloaded.\n");

    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLCLKPWR_init);
module_exit(FAPEXFDLCLKPWR_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlclkpwr object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
