/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv (hdmi)
*/

#include "fapex.h"
#include "fapex/abst_libhdmi.h"
#include "fapex/sys_config.h"
#include "fapex_fdlhdmi_priv.h"
#include "fapex_config_cmn.h"

/* for hdmi_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_HDMI_StrT  FAPIGLD_FDLHDMICtx[FAPEXDEV_HDMI_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLHDMI_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                    FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLHDMI_read] size:%d\n",size);

    return size;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLHDMI_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                    FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLHDMI_write] size:%d\n",size);

    return size;
}
#endif

/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLHDMI_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                    FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_HDMI_StrT *pfunc;
    int32_t  status;
    int32_t  ret = FAPEX_HDMI_ERR_BAD_PARAMETER;

    FAPEXDebugPrint(3, "[FAPEXFDLHDMI_ioctl] cmd:%08x\n",cmd);

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_HDMI_StrT*)(file->private_data);

    switch(cmd) {
        /* actual code which is depend on function is implimented here. */
#ifdef FAPI_HDMI_CONNECT
        case FAPEX_IOCCOM_OPEN:
            if(pfunc->pFAPICtxHdl == NULL) {
                FAPI_HDMI_OpenParamsT  params;

                /* open sequence */
                params.version = FAPI_HDMI_VERSION;
                params.block   = FAPI_HDMI_ALL;
                pfunc->pFAPICtxHdl = hdmi_funcStr.Open(&params, &ret);
                FAPEXDebugPrint(3, "[fdlhdmi_Open] FAPI context:0x%08x\n",
                                   (uint32_t)pfunc->pFAPICtxHdl);
                if(pfunc->pFAPICtxHdl == NULL) {
                    FAPEXDebugPrint(2, "[FAPEXFDLHDMI_ioctl] FAPI_HDMI_Open error:%d\n", ret);
                }
            }
            else {
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMI_READ_EDID:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                status = hdmi_funcStr.ReadEDID(pfunc->pFAPICtxHdl);

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((int32_t *)arg, &status, sizeof(int32_t));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMI_GET_EDID_PHYSADDR:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                uint32_t phys_address = 0;

                ret = hdmi_funcStr.GetPhysicalAddressInEDID(pfunc->pFAPICtxHdl, &phys_address);
                if(ret == FAPI_OK) {
                    /* copy parameters from the kernel space to an user space */
                    FAPEXCFDLSAF_COPY_TO_USR((void *)arg, &phys_address, sizeof(uint32_t));
                }
                else {
                    FAPEXDebugPrint(2, "[FAPEXFDLHDMI_ioctl] "
                                       "FAPI_HDMI_GetPhysicalAddressInEDID error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_HDMI_CHECK_HPD:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                status = hdmi_funcStr.CheckHPD(pfunc->pFAPICtxHdl);

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((int32_t *)arg, &status, sizeof(int32_t));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMI_LINK_HDCP:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                status = hdmi_funcStr.LinkHDCP(pfunc->pFAPICtxHdl);

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((int32_t *)arg, &status, sizeof(int32_t));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMI_CHECK_HDCP_STATE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                status = hdmi_funcStr.CheckHDCPState(pfunc->pFAPICtxHdl);

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((int32_t *)arg, &status, sizeof(int32_t));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMI_GET_EDID_INFO:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPI_HDMI_EdidInfoT  info;

                ret = hdmi_funcStr.GetEDID_Info(pfunc->pFAPICtxHdl, &info);
                if(ret == FAPI_OK) {
                    /* copy parameters from the kernel space to an user space */
                    FAPEXCFDLSAF_COPY_TO_USR((void *)arg, &info, sizeof(FAPI_HDMI_EdidInfoT));
                }
                else {
                    FAPEXDebugPrint(2, "[FAPEXFDLHDMI_ioctl] "
                                       "FAPI_HDMI_GetEDID_Info error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_HDMI_SET_DYNAMIC_HDCP:
            ret = hdmi_funcStr.DynamicChangeHDCP(pfunc->pFAPICtxHdl, (uint32_t)arg);
            break;

#endif /* FAPI_HDMI_CONNECT */
        default:
            ;
    }

    return ret;
}

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLHDMI_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T mask = 0;

    FAPEXDebugPrint(4, "[FAPEXFDLHDMI_poll]\n");

    /* actual code which is depend on function is implimented here. */

    return mask;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLHDMI_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    int32_t ret;

    if (vma->vm_pgoff != 0){
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_IO | VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if (file->f_flags & O_SYNC){
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
                  vma->vm_end - vma->vm_start, vma->vm_page_prot);

    if(ret != 0){
        return -EAGAIN;
    }
    return 0;
#else
    return -1;
#endif /* __LINUX__ */
}
#endif

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLHDMI_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_HDMI_StrT  *pfunc;
    void    *pCfdlHdl;
    int32_t id;

    FAPEXDebugPrint(3, "[FAPEXFDLHDMI_open] file:0x%08x,private_data:0x%08x\n",
                       (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLHDMI_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLHDMI_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc = (FAPEXFDL_HDMI_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc;
        FAPEXDebugPrint(3, "[FAPEXFDLHDMI_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc));

        /* set parameter */
        pfunc->pCfdlHdl = pCfdlHdl;
        pfunc->id       = id;
#ifdef __LINUX__
        kref_init(&pfunc->kref);
#endif
    }
    else {
        pfunc = (FAPEXFDL_HDMI_StrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLHDMI_open] fapex_fdlhdmi was opened : used_counter %d.\n",
                       pfunc->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlhdmi_delete(struct kref *kref)
{
    FAPEXFDL_HDMI_StrT  *pfunc;
    int32_t  ret = 0;

    pfunc = container_of(kref, FAPEXFDL_HDMI_StrT, kref);
#else
static int32_t fdlhdmi_delete(FAPEXFDL_HDMI_StrT *pfunc)
{
    int32_t  ret = 0;
#endif

    if(pfunc->pFAPICtxHdl != NULL) {
#ifdef FAPI_HDMI_CONNECT
        /* close sequence */
        ret = hdmi_funcStr.Close(pfunc->pFAPICtxHdl);
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[FAPEXFDLHDMI_release] FAPI_HDMI_Close error:%d\n", ret);
            goto fdlhdmi_delete_end;
        }
#endif /* FAPI_HDMI_CONNECT */
        pfunc->pFAPICtxHdl = NULL;
    }

fdlhdmi_delete_end:
#ifdef __LINUX__
    pfunc->close_result = ret;
#else
    return ret;
#endif
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLHDMI_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_HDMI_StrT  *pfunc;
    FAPEXFDLRELRET_T    ret = 0;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_HDMI_StrT*)(file->private_data);

    if(pfunc != NULL) {
        /* decrement the count on our device */
        pfunc->used_count--;
#ifdef __LINUX__
        ret = kref_put(&pfunc->kref, fdlhdmi_delete);
        if(ret != 0) {
            ret = pfunc->close_result;
        }
#else
        ret = fdlhdmi_delete(pfunc);
#endif

        FAPEXDebugPrint(3, "[FAPEXFDLHDMI_release] : used_count %d\n",
                           pfunc->used_count);
    }
    FAPEXFDL_RETVAL(ret);
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlhdmi_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLHDMI_init(void)
{
    int32_t ret;
    int32_t count;
    FAPEXFDL_HDMI_StrT  *pfunc;
    FAPEXCFDL_UtilStrT  utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_HDMI_StrT  *pfunc0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLHDMI_init]\n");

    /* set struct file_operations */
    fapex_fdlhdmi_fops.owner   = THIS_MODULE;
    fapex_fdlhdmi_fops.open    = FAPEXFDLHDMI_open;
//  fapex_fdlhdmi_fops.read    = FAPEXFDLHDMI_read;
//  fapex_fdlhdmi_fops.write   = FAPEXFDLHDMI_write;
    fapex_fdlhdmi_fops.ioctl   = FAPEXFDLHDMI_ioctl;
//  fapex_fdlhdmi_fops.poll    = FAPEXFDLHDMI_poll;
//  fapex_fdlhdmi_fops.mmap    = FAPEXFDLHDMI_mmap;
    fapex_fdlhdmi_fops.release = FAPEXFDLHDMI_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc = (FAPEXFDL_HDMI_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_HDMI_StrT)*(FAPEXDEV_HDMI_DESCRIPTOR_NUM));
    if(pfunc == NULL) {
        return FAPEX_HDMI_ERR_OUT_OF_MEMORY;
    }
    pfunc0 = pfunc;
#else
    pfunc = FAPIGLD_FDLHDMICtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc, sizeof(FAPEXFDL_HDMI_StrT)*(FAPEXDEV_HDMI_DESCRIPTOR_NUM));

    for(count=FAPEXDEV_HDMI_MINOR_FIRST; count<=FAPEXDEV_HDMI_MINOR_LAST; count++){
        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlhdmi_fops, pfunc, &utility, count);
        if(ret) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            FAPEXDebugPrint(1, "[FAPEXFDLHDMI_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            return ret;
        }
        pfunc ++;
    }

#ifdef FAPI_HDMI_CONNECT
//  FAPEXDebugPrint(1, "call FAPEXCFG_HDMI_Init\n");
//
    /* call FAPI (initialize) -> this function has been called in CFDL */
//  ret = FAPEXCFG_HDMI_Init();
    ret = 0;
#else
    ret = 0;
#endif /* FAPI_HDMI_CONNECT */

    FAPEXDebugPrint(1, "[FAPEXFDLHDMI_init] fapex_fdlhdmi was loaded. %d\n", ret);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLHDMI_cleanup(void)
{
    int32_t count;

    FAPEXDebugPrint(1, "[FAPEXFDLHDMI_cleanup]\n");

#ifdef FAPI_HDMI_CONNECT
    /* call FAPI API (cleanup) -> this function has been called in CFDL */
//  FAPEXCFG_HDMI_Cleanup();
//
//  FAPEXDebugPrint(1, "end FAPEXCFG_HDMI_Cleanup\n");
#endif /* FAPI_HDMI_CONNECT */

    /* cleanup sequence (common this FDL) */
#ifdef ALLOCATE_DYNAMIC_MEMORY
    {
        void *pCfdlHdl;
        FAPEXFDL_HDMI_StrT *pfunc;

        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        pfunc    = (FAPEXFDL_HDMI_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_HDMI_MINOR_FIRST);

        /* relase the FDL context */
        if(pfunc != NULL) {
            FAPI_SYS_FREE(pfunc);
        }
    }
#endif /* ALLOCATE_DYNAMIC_MEMORY */

    for(count=FAPEXDEV_HDMI_MINOR_FIRST; count<=FAPEXDEV_HDMI_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLHDMI_cleanup] fapex_fdlhdmi was unloaded.\n");

    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLHDMI_init);
module_exit(FAPEXFDLHDMI_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlhdmi object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
