/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv (hdmicec)
*/

#include "fapex.h"
#include "fapex/abst_libhdmi.h"
#include "fapex/abst_libhdmicec.h"
#include "fapex/sys_config.h"
#include "fapex_fdlhdmicec_priv.h"
#include "fapex_config_cmn.h"

/* for hdmicec_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_HDMICEC_StrT  FAPIGLD_FDLHDMICECCtx[FAPEXDEV_HDMICEC_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLHDMICEC_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                       FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_read] size:%d\n",size);

    return size;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLHDMICEC_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                       FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_write] size:%d\n",size);

    return size;
}
#endif

/*!
*******************************************************************************
**
** \brief  Event Notifer
**
** \param   value        The current timer value
** \param   optDataPtr   Optional data to be passed to the notify function.
**
******************************************************************************/
static void fdlhdmicec_SendNotifier(uint64_t value, uint32_t optDataPtr)
{
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;

    dt_event_ptr = (FAPEXCFDLSAF_EventStrT *)optDataPtr;

    /* set event */
    FAPEXCFDLSAF_SET_EVENT(dt_event_ptr,
                           FAPEXFDL_HDMICEC_EVENT_SEND,
                           FAPEXCFDLSAF_WAITEVENT_OR);
}

static void fdlhdmicec_ReceiveNotifier(uint64_t value, uint32_t optDataPtr)
{
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;

    dt_event_ptr = (FAPEXCFDLSAF_EventStrT *)optDataPtr;

    /* set event */
    FAPEXCFDLSAF_SET_EVENT(dt_event_ptr,
                           FAPEXFDL_HDMICEC_EVENT_RECEIVE,
                           FAPEXCFDLSAF_WAITEVENT_OR);
}

/*!
*******************************************************************************
**
** \brief  get HDMI-CEC send status
**
** \param   pfunc   pointer to HDMICEC context
** \param   pArg    pointer to status parameter
**
******************************************************************************/
static int32_t fdlhdmicec_GetSendStatus(FAPEXFDL_HDMICEC_StrT *pfunc, uint32_t *pArg)
{
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;
    uint32_t  status;
    int32_t   ret;

    dt_event_ptr = &pfunc->dt_event;
    if(pfunc->status == FAPEX_STATUS_RUNNING) {
        /* wait */
        dt_event_ptr->mode       = FAPEXCFDLSAF_WAITEVENT_SYNC | FAPEXCFDLSAF_WAITEVENT_OR;
        dt_event_ptr->wai_bitptn = FAPEXFDL_HDMICEC_EVENT_SEND;
        dt_event_ptr->tout       = 0;
        ret = FAPEXCFDLSAF_WAIT_EVENT(dt_event_ptr);
    }
    else {
        /* clear read event */
        ret = FAPEXCFDLSAF_SET_EVENT(dt_event_ptr,
                                     ~FAPEXFDL_HDMICEC_EVENT_SEND,
                                     FAPEXCFDLSAF_WAITEVENT_AND);
    }

    if(ret == FAPI_ERR_TIMEOUT) {
        return FAPEX_HDMI_ERR_TIMEOUT;
    }
    else {
        /* get send status */
        status = hdmicec_funcStr.CheckCecSendStatus();

        /* copy parameters from the kernel space to an user space */
        FAPEXCFDLSAF_COPY_TO_USR(pArg, &status, sizeof(uint32_t));
        return 0;
    }
}

/*!
*******************************************************************************
**
** \brief  get HDMI-CEC receive status
**
** \param   pfunc   pointer to HDMICEC context
** \param   pArg    pointer to status parameter
**
******************************************************************************/
static int32_t fdlhdmicec_GetReceiveStatus(FAPEXFDL_HDMICEC_StrT *pfunc, uint32_t *pArg)
{
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;
    uint32_t  status;
    int32_t   ret;

    dt_event_ptr = &pfunc->dt_event;
    if(pfunc->status == FAPEX_STATUS_RUNNING) {
        /* wait */
        dt_event_ptr->mode       = FAPEXCFDLSAF_WAITEVENT_SYNC | FAPEXCFDLSAF_WAITEVENT_OR;
        dt_event_ptr->wai_bitptn = FAPEXFDL_HDMICEC_EVENT_RECEIVE;
        dt_event_ptr->tout       = 0;
        ret = FAPEXCFDLSAF_WAIT_EVENT(dt_event_ptr);
    }
    else {
        /* clear read event */
        ret = FAPEXCFDLSAF_SET_EVENT(dt_event_ptr,
                                     ~FAPEXFDL_HDMICEC_EVENT_RECEIVE,
                                     FAPEXCFDLSAF_WAITEVENT_AND);
    }

    if(ret == FAPI_ERR_TIMEOUT) {
        return FAPEX_HDMI_ERR_TIMEOUT;
    }
    else {
        /* get receive status */
        status = hdmicec_funcStr.CheckCecReceiveStatus();

        /* copy parameters from the kernel space to an user space */
        FAPEXCFDLSAF_COPY_TO_USR(pArg, &status, sizeof(uint32_t));
        return 0;
    }
}

/*!
*******************************************************************************
**
** \brief  check HDMI-CEC send status
**
** \param   pfunc   pointer to HDMICEC context
** \param   pArg    pointer to status parameter
**
******************************************************************************/
static int32_t fdlhdmicec_CheckSendStatus(FAPEXFDL_HDMICEC_StrT *pfunc, uint32_t *pArg)
{
    uint32_t  status;

    /* get send status */
    status = hdmicec_funcStr.CheckCecSendStatus();

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &status, sizeof(uint32_t));

    return 0;
}

/*!
*******************************************************************************
**
** \brief  check HDMI-CEC receive status
**
** \param   pfunc   pointer to HDMICEC context
** \param   pArg    pointer to status parameter
**
******************************************************************************/
static int32_t fdlhdmicec_CheckReceiveStatus(FAPEXFDL_HDMICEC_StrT *pfunc, uint32_t *pArg)
{
    uint32_t  status;

    /* get receive status */
    status = hdmicec_funcStr.CheckCecReceiveStatus();

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &status, sizeof(uint32_t));

    return 0;
}

/*!
*******************************************************************************
**
** \brief  set HDMI-CEC message
**
** \param   pfunc   pointer to HDMICEC context
** \param   pArg    pointer to FAPEX_HDMICEC_MessageParamStrT parameter
**
******************************************************************************/
static int32_t fdlhdmicec_SetMessage(FAPEXFDL_HDMICEC_StrT *pfunc, FAPEX_HDMICEC_MessageParamStrT *pArg)
{
    FAPEX_HDMICEC_MessageParamStrT  params;
    int32_t  ret;

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_HDMICEC_MessageParamStrT));

    ret = hdmicec_funcStr.SetCecMessageParameters(params.initiatorLogicalAddr,
                                                  params.followerLogicalAddr,
                                                  params.opcodeName,
                                                  (FAPI_HDMI_CECMessageParamsT *)&params.message);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  get HDMI-CEC message
**
** \param   pfunc   pointer to HDMICEC context
** \param   pArg    pointer to FAPEX_HDMICEC_MessageParamStrT parameter
**
******************************************************************************/
static int32_t fdlhdmicec_GetMessage(FAPEXFDL_HDMICEC_StrT *pfunc, FAPEX_HDMICEC_MessageParamStrT *pArg)
{
    FAPEX_HDMICEC_MessageParamStrT  params;
    uint32_t  ret;

    ret = hdmicec_funcStr.GetCecMessageParameters(&params.initiatorLogicalAddr,
                                                  &params.opcodeName,
                                                  &params.remainRecMsgNum,
                                                  (FAPI_HDMI_CECMessageParamsT *)&params.message);

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_HDMICEC_MessageParamStrT));

    return ret;
}

/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLHDMICEC_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                       FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_HDMICEC_StrT *pfunc;
    int32_t  status;
    int32_t  ret = FAPEX_HDMI_ERR_BAD_PARAMETER;

    FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_ioctl] cmd:%08x\n",cmd);

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_HDMICEC_StrT*)(file->private_data);

    switch(cmd){
        /* actual code which is depend on function is implimented here. */
#ifdef FAPI_HDMICEC_CONNECT
        case FAPEX_IOCCOM_OPEN:
            if(pfunc->open_flag == 0) {
                /* open sequence */
                ret = hdmicec_funcStr.Open();
                FAPEXDebugPrint(3, "[fdlhdmicec_Open]\n");
                if(ret == FAPI_OK) {
                    FAPI_HDMI_CECConfigParamT  params;
                    FAPEXCFDLSAF_EventStrT     *dt_event_ptr;

                    pfunc->open_flag = 1;
                    dt_event_ptr = &pfunc->dt_event;

                    params.callback1 = fdlhdmicec_SendNotifier;
                    params.optData1  = (uint32_t)dt_event_ptr;
                    params.callback2 = fdlhdmicec_ReceiveNotifier;
                    params.optData2  = (uint32_t)dt_event_ptr;
                    ret = hdmicec_funcStr.SetCecConfigParameters(&params);
                    if(ret == FAPI_OK) {
                        pfunc->status = FAPEX_STATUS_RUNNING;
                    }
                    else {
                        FAPEXDebugPrint(2, "[FAPEXFDLHDMICEC_ioctl] "
                                           "FAPI_HDMI_CEC_SetCecConfigParameters error:%d\n",
                                           ret);
                    }
                }
                else {
                    FAPEXDebugPrint(2, "[FAPEXFDLHDMICEC_ioctl] "
                                       "FAPI_HDMI_CEC_Open error:%d\n", ret);
                }
            }
            else {
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_GET_SEND_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlhdmicec_GetSendStatus(pfunc, (uint32_t*)arg);
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_GET_RECEIVE_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlhdmicec_GetReceiveStatus(pfunc, (uint32_t*)arg);
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_CHECK_SEND_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlhdmicec_CheckSendStatus(pfunc, (uint32_t*)arg);
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_CHECK_RECEIVE_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlhdmicec_CheckReceiveStatus(pfunc, (uint32_t*)arg);
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_SET_MESSAGE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlhdmicec_SetMessage(pfunc, (FAPEX_HDMICEC_MessageParamStrT*)arg);
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_GET_MESSAGE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlhdmicec_GetMessage(pfunc, (FAPEX_HDMICEC_MessageParamStrT*)arg);
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_SEND_MESSAGE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                status = hdmicec_funcStr.SendCecMessage();

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((int32_t *)arg, &status, sizeof(int32_t));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_RECEIVE_MESSAGE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                status = hdmicec_funcStr.ReceivingCecMessage();

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((int32_t *)arg, &status, sizeof(int32_t));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_SET_LOGICAL_ADDRESS:
            {
                uint32_t  address = (uint32_t)arg;

                hdmicec_funcStr.SetCecLogicalAddress(address);
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_HDMICEC_GET_LOGICAL_ADDRESS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                uint32_t  address, *pArg = (uint32_t*)arg;

                address = hdmicec_funcStr.GetCecLogicalAddress();

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR(pArg, &address, sizeof(uint32_t));
                ret = 0;
            }
            break;
#endif /* FAPI_HDMICEC_CONNECT */

        default:
            ;
    }

    return ret;
}

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLHDMICEC_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T mask = 0;

    FAPEXDebugPrint(4, "[FAPEXFDLHDMICEC_poll]\n");

    /* actual code which is depend on function is implimented here. */

    return mask;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLHDMICEC_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    int32_t ret;

    if (vma->vm_pgoff != 0){
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_IO | VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if (file->f_flags & O_SYNC){
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
                  vma->vm_end - vma->vm_start, vma->vm_page_prot);

    if(ret != 0){
        return -EAGAIN;
    }
    return 0;
#else
    return -1;
#endif /* __LINUX__ */
}
#endif

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLHDMICEC_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_HDMICEC_StrT   *pfunc;
    FAPEXCFDLSAF_EventStrT  *dt_event_ptr;
    void    *pCfdlHdl;
    int32_t id;

    FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_open] file:0x%08x,private_data:0x%08x\n",
                       (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc = (FAPEXFDL_HDMICEC_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc;
        FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc));

        dt_event_ptr = &pfunc->dt_event;
        if(dt_event_ptr->initialized == 0) {
            FAPEXCFDLSAF_CREATE_EVENT(dt_event_ptr);
        }

        /* set parameter */
        pfunc->pCfdlHdl = pCfdlHdl;
        pfunc->id       = id;
        pfunc->status   = FAPEX_STATUS_OPENED;
#ifdef __LINUX__
        kref_init(&pfunc->kref);
#endif
    }
    else {
        pfunc = (FAPEXFDL_HDMICEC_StrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_open] fapex_fdlhdmicec was opened : used_counter %d.\n",
                       pfunc->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlhdmicec_delete(struct kref *kref)
{
    FAPEXFDL_HDMICEC_StrT   *pfunc;

    pfunc = container_of(kref, FAPEXFDL_HDMICEC_StrT, kref);
#else
static void fdlhdmicec_delete(FAPEXFDL_HDMICEC_StrT *pfunc)
{
#endif

    if(pfunc->status != FAPEX_STATUS_TERMINATED) {
#ifdef FAPI_HDMICEC_CONNECT
        /* close sequence */
        hdmicec_funcStr.Close();
        pfunc->open_flag = 0;

#endif /* FAPI_HDMICEC_CONNECT */

        /* clear control flags */
        FAPEXCFDLSAF_DELETE_EVENT(&pfunc->dt_event);

        pfunc->status = FAPEX_STATUS_TERMINATED;
    }
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLHDMICEC_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_HDMICEC_StrT *pfunc;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_HDMICEC_StrT*)(file->private_data);

    if(pfunc != NULL) {
        /* decrement the count on our device */
        pfunc->used_count--;
#ifdef __LINUX__
        kref_put(&pfunc->kref, fdlhdmicec_delete);
#else
        fdlhdmicec_delete(pfunc);
#endif

        FAPEXDebugPrint(3, "[FAPEXFDLHDMICEC_release] : used_count %d\n",
                           pfunc->used_count);
    }
    return 0;
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlhdmicec_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLHDMICEC_init(void)
{
    int32_t ret;
    int32_t count;
    FAPEXFDL_HDMICEC_StrT  *pfunc;
    FAPEXCFDL_UtilStrT     utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_HDMICEC_StrT  *pfunc0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLHDMICEC_init]\n");

    /* set struct file_operations */
    fapex_fdlhdmicec_fops.owner   = THIS_MODULE;
    fapex_fdlhdmicec_fops.open    = FAPEXFDLHDMICEC_open;
//  fapex_fdlhdmicec_fops.read    = FAPEXFDLHDMICEC_read;
//  fapex_fdlhdmicec_fops.write   = FAPEXFDLHDMICEC_write;
    fapex_fdlhdmicec_fops.ioctl   = FAPEXFDLHDMICEC_ioctl;
//  fapex_fdlhdmicec_fops.poll    = FAPEXFDLHDMICEC_poll;
//  fapex_fdlhdmicec_fops.mmap    = FAPEXFDLHDMICEC_mmap;
    fapex_fdlhdmicec_fops.release = FAPEXFDLHDMICEC_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc = (FAPEXFDL_HDMICEC_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_HDMICEC_StrT)*(FAPEXDEV_HDMICEC_DESCRIPTOR_NUM));
    if(pfunc == NULL) {
        return FAPEX_HDMI_ERR_OUT_OF_MEMORY;
    }
    pfunc0 = pfunc;
#else
    pfunc = FAPIGLD_FDLHDMICECCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc, sizeof(FAPEXFDL_HDMICEC_StrT)*(FAPEXDEV_HDMICEC_DESCRIPTOR_NUM));

    for(count=FAPEXDEV_HDMICEC_MINOR_FIRST; count<=FAPEXDEV_HDMICEC_MINOR_LAST; count++){
        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlhdmicec_fops, pfunc, &utility, count);
        if(ret) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            FAPEXDebugPrint(1, "[FAPEXFDLHDMICEC_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            return ret;
        }
        /* init sequence (common this FDL) */
//      pfunc->status = FAPEX_STATUS_LOADED;

        pfunc ++;
    }

#ifdef FAPI_HDMICEC_CONNECT
    FAPEXDebugPrint(1, "call FAPEXCFG_HDMICEC_Init\n");

    /* call FAPI (initialize) */
    ret = FAPEXCFG_HDMICEC_Init();
#else
    ret = 0;
#endif /* FAPI_HDMICEC_CONNECT */

    FAPEXDebugPrint(1, "[FAPEXFDLHDMICEC_init] fapex_fdlhdmicec was loaded. %d\n", ret);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLHDMICEC_cleanup(void)
{
    int32_t count;

    FAPEXDebugPrint(1, "[FAPEXFDLHDMICEC_cleanup]\n");

#ifdef FAPI_HDMICEC_CONNECT
    /* call FAPI API (cleanup) */
    FAPEXCFG_HDMICEC_Cleanup();

    FAPEXDebugPrint(1, "end FAPEXCFG_HDMICEC_Cleanup\n");
#endif /* FAPI_HDMICEC_CONNECT */

    /* cleanup sequence (common this FDL) */
#ifdef ALLOCATE_DYNAMIC_MEMORY
    {
        void *pCfdlHdl;
        FAPEXFDL_HDMICEC_StrT *pfunc;

        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        pfunc    = (FAPEXFDL_HDMICEC_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_HDMICEC_MINOR_FIRST);

        /* relase the FDL context */
        if(pfunc != NULL) {
            FAPI_SYS_FREE(pfunc);
        }
    }
#endif /* ALLOCATE_DYNAMIC_MEMORY */

    for(count=FAPEXDEV_HDMICEC_MINOR_FIRST; count<=FAPEXDEV_HDMICEC_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLHDMICEC_cleanup] fapex_fdlhdmicec was unloaded.\n");

    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLHDMICEC_init);
module_exit(FAPEXFDLHDMICEC_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlhdmicec object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
