/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */

#include "fapex.h"
#include "fapex/abst_libtsd.h"
#include "fapex/sys_config.h"
#include "fapex_fdltsd_priv.h"
#include "fapex_fdlvidec_priv.h"
#include "fapex_config_cmn.h"

/* for tsd_funcStr, bm_funcStr, desc_funcStr, sync_funcStr, videc_funcStr, timer_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_TSD_CmnStrT  FAPIGLD_FDLTSDCmnCtx[FAPEXDEV_TSD_CMN_DESCRIPTOR_NUM];
static FAPEXFDL_TSD_SubStrT  FAPIGLD_FDLTSDSubCtx[FAPEXDEV_TSD_SUB_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

//#define TSD_INPUT_MANUAL_MODE_SUPPORTED

/*!
*******************************************************************************
**
** \brief  update section ready index
**
** \param  pfunc_sub    pointer to sub context
** \param  size         update size
**
******************************************************************************/
static void fdltsd_UpdateSctInfo(FAPEXFDL_TSD_SubStrT *pfunc_sub, int32_t size)
{
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;
    FAPEXFDL_TSD_SctInfoT  *sct_data;
    uint32_t  i, r_index, w_index;
    int32_t   updated_size;

    dt_event_ptr = &pfunc_sub->dt_event;

    r_index = dt_event_ptr->read_idx  & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;
    w_index = dt_event_ptr->write_idx & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;
    updated_size = 0;

    i = pfunc_sub->ready_idx;
    while((i != w_index) && (updated_size < size)) {
        sct_data = (FAPEXFDL_TSD_SctInfoT*)pfunc_sub->ntfyFifo[i].pData;
        if(sct_data != NULL) {
            updated_size += sct_data->sct_length + 4;
        }
        if(i == r_index) {
            /* update the read index */
            FAPEXCFDLSAF_SET_EVENT(dt_event_ptr, 0,
                                   FAPEXCFDLSAF_WAITEVENT_QUEUE_READ);
            r_index = (r_index + 1) & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;
        }
        i = (i + 1) & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;
    }
    /* check laset event */
    if( (i != w_index)
     && (pfunc_sub->ntfyFifo[i].notifier_type == FAPEX_TSD_BUFFSTAT_STOPPED) ) {
        if(i == r_index) {
            /* update the read index */
            FAPEXCFDLSAF_SET_EVENT(dt_event_ptr, 0,
                                   FAPEXCFDLSAF_WAITEVENT_QUEUE_READ);
        }
        i = (i + 1) & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;
    }
    pfunc_sub->ready_idx = i;
}

/*!
*******************************************************************************
**
** \brief  Event Notifier common function
**
** \param  pfunc_sub     pointer to sub context
** \param  notifierType  notifier type
**
******************************************************************************/
static void fdltsd_EventNotifier(FAPEXFDL_TSD_SubStrT *pfunc_sub, uint32_t notifierType)
{
    FAPEXCFDLSAF_EventStrT  *dt_event_ptr;
    uint32_t  w_index;

    dt_event_ptr = &pfunc_sub->dt_event;

    w_index = dt_event_ptr->write_idx & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;

    pfunc_sub->ntfyFifo[w_index].notifier_type = notifierType;
    pfunc_sub->ntfyFifo[w_index].pData         = NULL;

    /* set event */
    FAPEXCFDLSAF_SET_EVENT(dt_event_ptr, notifierType,
                           FAPEXCFDLSAF_WAITEVENT_QUEUE_WRITE);
#if 1 /* workaround for #23196 */
	if (dt_event_ptr->write_idx != dt_event_ptr->read_idx &&
		(dt_event_ptr->write_idx & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK) ==
		(dt_event_ptr->read_idx & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK)) {
	  //printk("overflowed\n");
	  FAPEXCFDLSAF_SET_EVENT(dt_event_ptr, 0,
							 FAPEXCFDLSAF_WAITEVENT_QUEUE_READ);
	  pfunc_sub->ready_idx = (pfunc_sub->ready_idx + 1) & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;
	}
#endif
}

/*!
*******************************************************************************
**
** \brief  get buffer fill level
**
** \param  pfunc_sub    pointer to sub context
**
******************************************************************************/
static int32_t fdltsd_GetBufferFillLevel(FAPEXFDL_TSD_SubStrT *pfunc_sub)
{
    FAPEXFDL_VIDEC_StrT  *pfunc_videc;
    uint32_t  half_fill_size;
    uint32_t  fill_size;
    int32_t   ret;

    if( (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_OUT_BOUND)
     && (pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_VIDEO)
     && (pfunc_sub->output_id >= 0) ) {
        /* get Video Stream Buffer fill size */

        pfunc_videc = (FAPEXFDL_VIDEC_StrT*)FAPEXCFDL_GetFdlHdl(pfunc_sub->pCfdlHdl,
                                                                pfunc_sub->output_id);

        ret = videc_funcStr.GetVesFillLevel(pfunc_videc->pFAPICtxHdl, &fill_size);
        if(ret == 0) {
            /* adjustment buffer fill level
               because maximum value of FAPI_VIDEC_GetVesFillLevel() is of 3/4 buffer size */
            half_fill_size = pfunc_sub->buff_size >> 1;
            if(fill_size > half_fill_size) {
                ret = half_fill_size + ((fill_size - half_fill_size) << 1);
            }
            else {
                ret = fill_size;
            }
        }
        else {
            FAPEXDebugPrint(2, "[fdltsd_GetBufferFillLevel] "
                               "FAPI_VIDEC_GetVesFillLevel error:%d\n", ret);
        }
    }
    else if( (pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_TELETEXT)
          && (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_HCB) ) {
        uint32_t  write_pt;

        write_pt = bm_funcStr.GetBufferWritePointerPosition(pfunc_sub->pBMCtxHdl);

        ret = write_pt - pfunc_sub->read_pt;
        if(ret < 0) {
            ret += pfunc_sub->buff_size;
        }
    }
    else {
        /* get Buffer fill size */
        ret = bm_funcStr.GetBufferFillLevel(pfunc_sub->pBMCtxHdl);
        if(ret < 0) {
            FAPEXDebugPrint(2, "[fdltsd_GetBufferFillLevel] "
                               "FAPI_BM_GetBufferFillLevel error:%d\n", ret);
        }
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLTSD_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                   FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXFDL_TSD_CmnStrT   *pfunc_cmn;
    FAPEXFDL_TSD_SubStrT   *pfunc_sub;
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;
    uint32_t  wait_enable;
    uint32_t  wait_time;
    int32_t   fill_size, read_size;
    int32_t   ret;

    FAPEXDebugPrint(3, "[FAPEXFDLTSD_read] size:%d\n", size);

    /* Get the context base of this id's FDL. */
    pfunc_sub = (FAPEXFDL_TSD_SubStrT*)(file->private_data);
    pfunc_cmn = (FAPEXFDL_TSD_CmnStrT*)pfunc_sub->pCmnHdl;

    if( (pfunc_sub->id < FAPEXDEV_TSD_SUB_MINOR_FIRST)
     || (pfunc_sub->pBMCtxHdl == NULL) || (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_OUT_BOUND) ) {
        /* cannot read */
        return 0;
    }
    dt_event_ptr = &pfunc_sub->dt_event;

    /* check size */
    if(size > pfunc_sub->buff_size) {
        size = pfunc_sub->buff_size;
    }

    /* check Buffer Fill Size */
    fill_size = bm_funcStr.GetBufferFillLevel(pfunc_sub->pBMCtxHdl);
    if(fill_size < 0) {
        FAPEXDebugPrint(2, "[FAPEXFDLTSD_read] "
                           "FAPI_BM_GetBufferFillLevel error:%d\n", fill_size);
        return fill_size;
    }

    if(fill_size < size) {
        if(pfunc_sub->status == FAPEX_STATUS_RUNNING) {
            wait_time  = (pfunc_sub->timeout != 0) ?
                         pfunc_sub->timeout : pfunc_cmn->timeout;
            wait_time += FAPEXFDL_TSD_TIMEOUT_INTERRUPT;

            /* wait */
            dt_event_ptr->mode = FAPEXCFDLSAF_WAITEVENT_QUEUE;
            dt_event_ptr->tout = wait_time;

            wait_enable = pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_ALL;

            if( !(wait_enable & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_DATA) ) {
                dt_event_ptr->mode |= FAPEXCFDLSAF_WAITEVENT_ASYNC;
            }

            ret = FAPEXCFDLSAF_WAIT_EVENT(dt_event_ptr);
            if(ret == FAPI_ERR_BAD_PARAMETER) {
                return FAPEX_TSD_ERR_BAD_PARAMETER;
            }
            /* get Buffer Fill Size, again */
            fill_size = bm_funcStr.GetBufferFillLevel(pfunc_sub->pBMCtxHdl);
            if(fill_size < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLTSD_read] "
                                   "FAPI_BM_GetBufferFillLevel error:%d\n", fill_size);
                return fill_size;
            }
        }
    }
    if(fill_size < size) {
        size = fill_size;
    }

    /* read data */
#ifndef __LINUX__
    read_size = bm_funcStr.Read(pfunc_sub->pBMCtxHdl, buf, size, FAPI_SYS_NO_SUSPEND);
    if(read_size <= 0) {
        return read_size;
    }
#else
    read_size = 0;
    if(size > 0) {
        uint32_t    read_pt, tmp_size;
        uint8_t     *buff_top, *read_ptr, *dst_ptr;

        read_pt  = bm_funcStr.GetBufferReadPointerPosition(pfunc_sub->pBMCtxHdl);
        buff_top = (uint8_t*)FAPI_SYS_GET_VIRTUAL_ADDRESS( pfunc_sub->buff_address );
        read_ptr = buff_top + read_pt;
        dst_ptr  = (uint8_t*)buf;
        tmp_size = pfunc_sub->buff_size - read_pt;

        if(tmp_size >= size) {
            /* new read_pointer doesn't reach end of buffer */
            ret = FAPEXCFDLSAF_COPY_TO_USR(dst_ptr, read_ptr, size);
            if(ret == 0) {
                read_size  = size;
                read_pt   += size;
                /* check buffer turned-back point */
                if(read_pt >= pfunc_sub->buff_size) {
                    read_pt -= pfunc_sub->buff_size;
                }

                ret = bm_funcStr.SetBufferReadPointer(pfunc_sub->pBMCtxHdl, read_pt, 0);
            }
        }
        else {
            /* new read_pointer goes round to start of buffer */
            ret = FAPEXCFDLSAF_COPY_TO_USR(dst_ptr, read_ptr, tmp_size);
            if(ret == 0) {
                read_size = tmp_size;
                dst_ptr  += tmp_size;
                size     -= tmp_size;

                ret = FAPEXCFDLSAF_COPY_TO_USR(dst_ptr, buff_top, size);
                if(ret == 0) {
                    read_size += size;
                    ret = bm_funcStr.SetBufferReadPointer(pfunc_sub->pBMCtxHdl, size, 0);
                }
            }
        }
        if(ret != 0) {
            FAPEXDebugPrint(2, "[FAPEXFDLTSD_read] "
                               "FAPI_BM_SetBufferReadPointer error:%d\n", ret);
            return ret;
        }
    }
#endif /* !__LINUX__ */

    if( (pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_UNKNOWN)
     && (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_ALL) ) {
        fdltsd_UpdateSctInfo(pfunc_sub, read_size);
    }
    return read_size;
}

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLTSD_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                   FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXFDL_TSD_CmnStrT   *pfunc_cmn;
    FAPEXFDL_TSD_SubStrT   *pfunc_sub;
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;
    int32_t     fill_size, wrote_size, ret;
    uint32_t    wait_time;

    FAPEXDebugPrint(3, "[FAPEXFDLTSD_write] size:%d\n", size);

    /* Get the context base of this id's FDL. */
    pfunc_sub = (FAPEXFDL_TSD_SubStrT*)(file->private_data);
    pfunc_cmn = (FAPEXFDL_TSD_CmnStrT*)pfunc_sub->pCmnHdl;

    if(pfunc_sub->pBMCtxHdl == NULL) {
        /* cannot write */
        return 0;
    }
    dt_event_ptr = &pfunc_sub->dt_event;

    if(pfunc_sub->status == FAPEX_STATUS_RUNNING) {
        /* check size */
        if(size > pfunc_sub->buff_size) {
            size = pfunc_sub->buff_size;
        }

        wait_time  = (pfunc_sub->timeout != 0) ?
                     pfunc_sub->timeout : pfunc_cmn->timeout;
        wait_time += FAPEXFDL_TSD_TIMEOUT_INTERRUPT;

        do {
            /* wait */
            dt_event_ptr->mode = FAPEXCFDLSAF_WAITEVENT_QUEUE;
            dt_event_ptr->tout = wait_time;

            ret = FAPEXCFDLSAF_WAIT_EVENT(dt_event_ptr);
            if(ret == FAPI_ERR_TIMEOUT) {
                return FAPEX_TSD_ERR_TIMEOUT;
            }

            /* get Buffer Fill Size */
            fill_size = fdltsd_GetBufferFillLevel(pfunc_sub);
            if(fill_size < 0) {
                return -1;
            }
            if(pfunc_sub->buff_size >= (fill_size + size)) {
                break;
            }
        } while(0);
    }
    else {
        /* get Buffer Fill Size */
        fill_size = fdltsd_GetBufferFillLevel(pfunc_sub);
        if(pfunc_sub->buff_size < (fill_size + size)) {
            size = pfunc_sub->buff_size - fill_size;
            /* clear write event */
            ret = FAPEXCFDLSAF_SET_EVENT(dt_event_ptr,
                                         FAPEXFDL_TSD_EVENT_WRITE,
                                         FAPEXCFDLSAF_WAITEVENT_QUEUE_WRITE);
            if(ret < 0) {
                return FAPEX_TSD_ERR_INVALID_STATUS;
            }
        }
    }

    /* write data */
#ifndef __LINUX__
    wrote_size = bm_funcStr.Write(pfunc_sub->pBMCtxHdl, (void *)buf, size, FAPI_SYS_NO_SUSPEND);
#else
    wrote_size = 0;
    if(size > 0) {
        uint32_t    write_pt, tmp_size;
        uint8_t     *buff_top, *write_ptr, *src_ptr;

        write_pt  = bm_funcStr.GetBufferWritePointerPosition(pfunc_sub->pBMCtxHdl);
        buff_top  = (uint8_t*)FAPI_SYS_GET_VIRTUAL_ADDRESS( pfunc_sub->buff_address );
        write_ptr = buff_top + write_pt;
        src_ptr   = (uint8_t*)buf;
        tmp_size  = pfunc_sub->buff_size - write_pt;

        if(tmp_size >= size) {
            /* new write_pointer doesn't reach end of buffer */
            FAPEXCFDLSAF_COPY_TO_DRV(write_ptr, src_ptr, size);

            wrote_size  = size;
            write_pt   += size;
            /* check buffer turned-back point */
            if(write_pt >= pfunc_sub->buff_size) {
                write_pt -= pfunc_sub->buff_size;
            }

            ret = bm_funcStr.SetBufferWritePointer(pfunc_sub->pBMCtxHdl, write_pt, 0);
        }
        else {
            /* new write_pointer goes round to start of buffer */
            FAPEXCFDLSAF_COPY_TO_DRV(write_ptr, src_ptr, tmp_size);

            wrote_size = tmp_size;
            src_ptr   += tmp_size;
            size      -= tmp_size;

            FAPEXCFDLSAF_COPY_TO_DRV(buff_top, src_ptr, size);

            wrote_size += size;
            ret = bm_funcStr.SetBufferWritePointer(pfunc_sub->pBMCtxHdl, size, 0);
        }
        if(ret < 0) {
            FAPEXDebugPrint(2, "[FAPEXFDLTSD_write] "
                               "FAPI_BM_SetBufferWritePointer error:%d\n", ret);
            return -1;
        }
    }
#endif /* !__LINUX__ */

    return wrote_size;
}
#endif

/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLTSD_write_input(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                         FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXFDL_TSD_CmnStrT   *pfunc_cmn;
    uint32_t  valid_flag, *packet_ctrl;
    uint8_t   *write_ptr, *src_ptr;
    int32_t   i, wrote_size;
    uint64_t  end_time;
    int64_t   delta_time;

    FAPEXDebugPrint(3, "[FAPEXFDLTSD_write_input] size:%d\n", size);

    /* Get the context base of this id's FDL. */
    pfunc_cmn = (FAPEXFDL_TSD_CmnStrT*)(file->private_data);

    if((pfunc_cmn->pkt_in_address == 0) || (size < FAPEX_TSD_PACKET_INPUT_BYTES)) {
        /* cannot write */
        return 0;
    }
    packet_ctrl = (uint32_t*)pfunc_cmn->pkt_ctrl_address;

    pfunc_cmn->pkt_in_status = FAPEX_TSD_BUFFSTAT_NONE;
    wrote_size = 0;
    src_ptr    = (uint8_t*)buf;

    while(1) {
        end_time = timer_funcStr.GetTimeStamp(FAPI_TIMER_RESOLUTION_1_MSEC)
                  + FAPEXFDL_TSD_PACKET_IN_MAX_TIMEOUT;

        /* status check */
        valid_flag = *packet_ctrl;

        while(valid_flag == FAPEX_TSD_PACKET_INPUT_STATUS_BUSY_BUFF_ALL) {
            /* wait for buffer empty */
            FAPI_SYS_SLEEP(1);

            delta_time = timer_funcStr.GetTimeStamp(FAPI_TIMER_RESOLUTION_1_MSEC)
                         - end_time;
            if(delta_time > 0) {
                pfunc_cmn->pkt_in_status = FAPEX_TSD_BUFFSTAT_TIMEOUT;
                if(wrote_size > 0) {
                    return wrote_size;
                }
                else {
                    return FAPEX_TSD_ERR_TIMEOUT;
                }
            }
            FAPEXDebugPrint(2, "[FAPEXFDLTSD_write_input] sleep time:%lld\n", - delta_time);

            /* update status */
            valid_flag = *packet_ctrl;
        }

        write_ptr = (uint8_t*)pfunc_cmn->pkt_in_address;
        for(i=0; i<FAPEX_TSD_PACKET_INPUT_BUFF_NUM; i++) {
            /* check valid buffer */
            if(!(valid_flag & FAPEX_TSD_PACKET_INPUT_STATUS_BUSY_FLAG)) {
                /* new write_pointer doesn't reach end of buffer */
                FAPEXCFDLSAF_COPY_TO_DRV(write_ptr, src_ptr, FAPEX_TSD_PACKET_INPUT_BYTES);

#ifdef TSD_INPUT_MANUAL_MODE_SUPPORTED
                if(pfunc_cmn->blk_params.tsSelect != FAPEX_TSD_INPUT_LINEAR_AUTO_MODE) {
                    /* set the notification that packet became valid */
                    *packet_ctrl = (FAPEX_TSD_PACKET_INPUT_STATUS_VALID_FLAG | i)
                                   << (i * FAPEX_TSD_PACKET_INPUT_STATUS_BUFF_SHFT);
                }
#endif
                wrote_size += FAPEX_TSD_PACKET_INPUT_BYTES;
                src_ptr    += FAPEX_TSD_PACKET_INPUT_BYTES;
                if((size - wrote_size) < FAPEX_TSD_PACKET_INPUT_BYTES) {
                    return wrote_size;
                }
            }
            write_ptr   += pfunc_cmn->pkt_in_offset;
            valid_flag >>= FAPEX_TSD_PACKET_INPUT_STATUS_BUFF_SHFT;
        }
    }
}

/*!
*******************************************************************************
**
** \brief  setup TSD block input
**
** \param  pfunc_cmn    pointer to common context
** \param  pArg         pointer to FAPEX_TSD_InputParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetInput(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEX_TSD_InputParamStrT *pArg)
{
    FAPEX_TSD_InputParamStrT     *params;
    FAPI_TSD_TsInputSettingsT   input_params;
    FAPI_TSD_TsOutputSettingsT  output_params, *out_param_ptr;
    FAPEXFDL_TSD_CmnStrT        *pfunc_cmn0, *pfunc_stc, *pfunc_tmp;
    uint32_t  flags, i, seg_mask, blk_mask;
    uint32_t  input_addr0, input_addr1, mapped_size;
    int32_t   ret;

    if(pArg == NULL) {
        return FAPEX_TSD_ERR_BAD_PARAMETER;
    }
    params = &pfunc_cmn->blk_params;

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(params, pArg, sizeof(FAPEX_TSD_InputParamStrT));

    /* setup input parameters */
    flags = params->iflags;
    input_params.version             = FAPI_TSD_VERSION;
    input_params.tsSelect            = params->tsSelect & 0x0F;
    input_params.automaticSync       = (flags & FAPEX_TSD_IFLAGS_AM_SYNC)?      1 : 0;
    input_params.packetStartPolarity = (flags & FAPEX_TSD_IFLAGS_PACKET_START)? 1 : 0;
    input_params.enablePolarity      = (flags & FAPEX_TSD_IFLAGS_EN)?           1 : 0;
    input_params.clockInvert         = (flags & FAPEX_TSD_IFLAGS_INV_CLK)?      1 : 0;
    input_params.serialParallelMode  = (flags & FAPEX_TSD_IFLAGS_SP_MODE_MSK);
    input_params.syncByteDistance    = params->syncByteDistance;
    input_params.writeEndian = 0;

    /* setup output parameters */
    flags = params->oflags;
    if(flags == 0) {
        /* output disable */
        out_param_ptr = NULL;
    }
    else {
        output_params.version     = FAPI_TSD_VERSION;
        output_params.tsOutMode   = 0;
        output_params.clockInvert = (flags & FAPEX_TSD_OFLAGS_OUT_INV_CLK)? 1 : 0;
        output_params.prescaler   = (flags & FAPEX_TSD_OFLAGS_OUT_CLK_PRESCALER_MSK);
        output_params.sbz         = 0;
        out_param_ptr = &output_params;
    }

    /* setup TSD configuration */
    ret = tsd_funcStr.Configure(pfunc_cmn->block_no, &input_params, out_param_ptr);
    if(ret != FAPI_OK) {
        FAPEXDebugPrint(2, "[fdltsd_SetInput] FAPI_TSD_Configure error:%d\n", ret);
        return ret;
    }
    switch(params->tsSelect) {
#ifdef TSD_INPUT_MANUAL_MODE_SUPPORTED
        case FAPEX_TSD_INPUT_DIRECT_MODE:
            input_addr0 = tsd_funcStr.GetPacketRegisterAddress(pfunc_cmn->block_no, 0);
            input_addr1 = tsd_funcStr.GetPacketRegisterAddress(pfunc_cmn->block_no, 1);
            break;
        case FAPEX_TSD_INPUT_LINEAR_MODE:
            input_addr0 = tsd_funcStr.GetLinearPacketRegisterAddress(pfunc_cmn->block_no, 0);
            input_addr1 = tsd_funcStr.GetLinearPacketRegisterAddress(pfunc_cmn->block_no, 1);
            break;
#endif
        case FAPEX_TSD_INPUT_LINEAR_AUTO_MODE:
            input_addr0 = tsd_funcStr.GetLinearPacketAutoRegisterAddress(pfunc_cmn->block_no, 0);
            input_addr1 = tsd_funcStr.GetLinearPacketAutoRegisterAddress(pfunc_cmn->block_no, 1);
            break;
        default:
            input_addr0 = input_addr1 = 0;
    }
    pfunc_cmn->pkt_ctrl_address = tsd_funcStr.GetPacketControlAddress(pfunc_cmn->block_no);
    pfunc_cmn->pkt_in_address   = input_addr0;
    pfunc_cmn->pkt_in_offset    = input_addr1 - input_addr0;

    mapped_size = pfunc_cmn->pkt_in_offset * FAPEX_TSD_PACKET_INPUT_BUFF_NUM;
#ifdef __LINUX__
    mapped_size += FAPI_SYS_GET_PHYSICAL_ADDRESS(input_addr0) & ~PAGE_MASK;
    mapped_size = (mapped_size + ~PAGE_MASK) & PAGE_MASK;
#endif
    pfunc_cmn->pkt_in_area_size = mapped_size;

    pfunc_cmn0 = pfunc_tmp = pfunc_stc = pfunc_cmn - pfunc_cmn->block_no;
    seg_mask   = 0x03 << (pfunc_cmn->block_no & 0xFFFFFFFE);
    blk_mask   = 0x01;

    /*---------- LOCK (stc_channel) ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

    /* update stc_block flag */
    if(params->useStc != 0) {
        /* check the SYNC segment status */
        if(pfunc_cmn0->stc_block & seg_mask) {
            /* SYNC block is shared */
            pfunc_cmn0->stc_block |= 1 << pfunc_cmn->block_no;
        }
        else {
            /* SYNC block is switched */
            pfunc_cmn0->stc_block = 1 << pfunc_cmn->block_no;
        }
    }
    else {
        /* SYNC block is cleared */
        pfunc_cmn0->stc_block &= ~(1 << pfunc_cmn->block_no);
    }

    /* update stc_channel flag */
    for(i=0; i<FAPEXDEV_TSD_CMN_DESCRIPTOR_NUM; i++) {
        if(pfunc_cmn0->stc_block & blk_mask) {
            if(!(pfunc_stc->stc_channel & pfunc_tmp->stc_channel_mask)) {
                pfunc_stc->stc_channel |= pfunc_tmp->stc_channel_mask;
            }
        }
        else {
            pfunc_stc->stc_channel &= ~(pfunc_tmp->stc_channel_mask);
        }
        pfunc_tmp ++;
        if(i & 0x01) {
            pfunc_stc += 2;
        }
        blk_mask <<= 1;
    }

    /*--------- UNLOCK (stc_channel) ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  TSD common descriptor open
**
** \param  pfunc_cmn    pointer to common context
** \param  pArg         pointer to FAPEX_TSD_OpenParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_OpenCmn(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEX_TSD_OpenParamStrT *pArg)
{
    if(pfunc_cmn->pkt_in_area_size != 0) {
        return 0;
    }
    return fdltsd_SetInput(pfunc_cmn, pArg);
}

/*!
*******************************************************************************
**
** \brief  TSD sub descriptor open
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  id           Device identifier
** \param  pArg         pointer to FAPEX_TSD_OpenParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_OpenSub(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                              uint32_t id, FAPEX_TSD_OpenSubIdParamStrT *pArg)
{
    FAPI_TSD_OpenParamsT    params;
    FAPEXCFDLSAF_EventStrT  *dt_event_ptr;
    int32_t  ret;

    if(pfunc_sub->pTSDCtxHdl != NULL) {
        return 0;
    }
    if(pArg == NULL) {
        return FAPEX_TSD_ERR_BAD_PARAMETER;
    }
    /* setup open parameters */
    params.version    = FAPI_TSD_VERSION;
    params.blockIndex = pfunc_cmn->block_no;

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&(params.channelId), &(pArg->channelNo), sizeof(uint32_t));

    /* call Open TSD */
    pfunc_sub->pTSDCtxHdl = tsd_funcStr.Open(&params, &ret);
    FAPEXDebugPrint(3, "[fdltsd_OpenSub] TSD context:0x%08x\n",
                       (uint32_t)pfunc_sub->pTSDCtxHdl);
    if(pfunc_sub->pTSDCtxHdl == NULL) {
        FAPEXDebugPrint(2, "[fdltsd_OpenSub] FAPI_TSD_Open error:%d [%d]\n",
                           ret, pfunc_sub->id);
        return ret;
    }

    dt_event_ptr = &pfunc_sub->dt_event;
    if(dt_event_ptr->initialized == 0) {
        FAPEXCFDLSAF_CREATE_EVENT(dt_event_ptr);
    }

    /*---------- LOCK ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

    /* update status */
    pfunc_cmn->enable_channel |= 1 << pfunc_sub->pid_ch_no;

    /*--------- UNLOCK ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

    pfunc_sub->output_id = -1;
    pfunc_sub->status    = FAPEX_STATUS_OPENED;

    return 0;
}

#ifdef FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
#if (DEBUG >= 2)
/*!
*******************************************************************************
**
** \brief  check stream buffer control list
**
** \param  str          pointer to print message
** \param  pfunc_cmn0   pointer to common-0 context
** \param  address      allocated/freed address value
**
******************************************************************************/
static void fdltsd_CheckBuffList(uint8_t *str, FAPISAF_BuffManageT *pManage,
                                 uint32_t address)
{
    FAPISAF_BuffCtrlT  *alloc, *free;
    uint32_t  alloc_size, free_size;
    uint32_t  alloc_num, free_num, spare_num;

    alloc_size = alloc_num = 0;
    alloc = pManage->alloc_area;
    while(alloc != NULL) {
        alloc_num ++;
        alloc_size += alloc->buff_size;
        alloc = alloc->next_info;
    }

    free_size = free_num = spare_num = 0;
    free  = pManage->free_area;
    while(free != NULL) {
        if(free->buff_top != (uint32_t)-1) {
            free_num ++;
            free_size += free->buff_size;
        }
        else {
            spare_num ++;
        }
        free = free->next_info;
    }

    FAPEXDebugPrint(2, "TSD static allocated stream buffer information [%s] 0x%08x\n"
                       " alloc: 0x%08x [byte], %3d\n"
                       " free : 0x%08x [byte], %3d(+%d)\n",
                       str, address, alloc_size, alloc_num,
                       free_size, free_num, spare_num);
}
#endif /* (DEBUG >= 2) */
#endif /* FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE */

/*!
*******************************************************************************
**
** \brief  allocate stream buffer from dynamic area
**
** \param  size     allocate buffer size
**
******************************************************************************/
static uint32_t fdltsd_AllocDBuffArea(FAPEXFDL_TSD_SubStrT *pfunc_sub, uint32_t size)
{
    uint32_t  ret_address;

#ifndef __LINUX__
    uint32_t  buff_align_rest, allocate_size;

    buff_align_rest = FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE - 1;
    allocate_size   = size + buff_align_rest;

    ret_address = (uint32_t)FAPI_SYS_MALLOC(allocate_size);
    if(ret_address == 0) {
        FAPEXDebugPrint(2, "[fdltsd_AllocDBuffArea] out of memory error\ [%d]n",
                           pfunc_sub->id);
        return 0;
    }
    pfunc_sub->buff_address_temp = ret_address;
    ret_address = FAPI_SYS_GET_PHYSICAL_ADDRESS((ret_address + buff_align_rest) & ~buff_align_rest);
#else
    uint32_t  order;

    order = (PAGE_SHIFT > FAPEXFDL_TSD_STRM_BUFF_SIZE_ORDER_MIN) ?
            PAGE_SHIFT : FAPEXFDL_TSD_STRM_BUFF_SIZE_ORDER_MIN;
    while( (order < 31) && (size > (1 << order)) ) {
        order ++;
    }

    pfunc_sub->buff_pg = alloc_pages(GFP_HIGHUSER, (order - PAGE_SHIFT));
    if(pfunc_sub->buff_pg == NULL) {
        FAPEXDebugPrint(2, "[fdltsd_AllocDBuffArea] out of memory error [%d] (order %d)\n",
                           pfunc_sub->id, order);
        return 0;
    }
    pfunc_sub->order = order - PAGE_SHIFT;
    ret_address = page_to_phys(pfunc_sub->buff_pg);
#endif
    FAPEXDebugPrint(2, "[fdltsd_AllocDBuffArea] %08x\n", ret_address);

    return ret_address;
}

/*!
*******************************************************************************
**
** \brief  free stream buffer from dynamic area
**
** \param  address      free buffer physical address
**
******************************************************************************/
static void fdltsd_FreeDBuffArea(FAPEXFDL_TSD_SubStrT *pfunc_sub)
{
    FAPEXDebugPrint(2, "[fdltsd_FreeDBuffArea] %08x\n", pfunc_sub->buff_address);

#ifndef __LINUX__
    FAPI_SYS_FREE((void*)pfunc_sub->buff_address_temp);
    pfunc_sub->buff_address_temp = 0;
#else
    __free_pages(pfunc_sub->buff_pg, pfunc_sub->order);
    pfunc_sub->buff_pg = NULL;
    pfunc_sub->order   = 0;
#endif
}

/*!
*******************************************************************************
**
** \brief  close TSD's FAPI context
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
**
******************************************************************************/
static int32_t fdltsd_CloseCtx(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub)
{
    int32_t  ret;

    /* call Close TSD */
    if(pfunc_sub->pTSDCtxHdl) {
        ret = tsd_funcStr.Close(pfunc_sub->pTSDCtxHdl);
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdltsd_CloseCtx] FAPI_TSD_Close error:%d [%d]\n",
                               ret, pfunc_sub->id);
            return ret;
        }
        pfunc_sub->pTSDCtxHdl = NULL;
    }

    /* call Close BM */
    if(pfunc_sub->pBMCtxHdl) {
        ret = bm_funcStr.Close(pfunc_sub->pBMCtxHdl);
        if(ret != FAPI_OK) {
            return ret;
        }
        pfunc_sub->pBMCtxHdl = NULL;

        /*---------- LOCK ----------*/
        FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

        pfunc_cmn->bm_channel &= ~(1 << pfunc_sub->bm_ch_no);

        /*--------- UNLOCK ---------*/
        FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

#ifdef FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
        if(pfunc_sub->buff_allocate_type == FAPEXFDL_TSD_STRM_BUFF_TYPE_STATIC) {
            FAPEXFDL_TSD_CmnStrT  *pfunc_cmn0;

            pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;
            FAPISAF_FreeSBuffArea( pfunc_cmn0->sbuff_ctrl,
                                   pfunc_sub->buff_address );
 #if (DEBUG >= 2)
            fdltsd_CheckBuffList("FreeSBuff", pfunc_cmn0->sbuff_ctrl,
                                 pfunc_sub->buff_address);
 #endif
        }
        else
#endif
        if(pfunc_sub->buff_allocate_type == FAPEXFDL_TSD_STRM_BUFF_TYPE_DYNAMIC) {
            fdltsd_FreeDBuffArea( pfunc_sub );
        }
        pfunc_sub->buff_address       = 0;
        pfunc_sub->buff_size          = 0;
        pfunc_sub->buff_allocate_type = FAPEXFDL_TSD_STRM_BUFF_TYPE_UNKNOWN;
    }

#ifdef CONFIG_TSD_USE_INTERNAL_DESC
    /* call Close DESC */
    if(pfunc_sub->pDESCCtxHdl) {
        ret = desc_funcStr.Close(pfunc_sub->pDESCCtxHdl);
        if(ret != FAPI_OK) {
            return ret;
        }
        pfunc_sub->pDESCCtxHdl = NULL;
    }
#endif
    return 0;
}

/*!
*******************************************************************************
**
** \brief  TSD sub descriptor close
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  stc_channel  STC channel flag
**
******************************************************************************/
static int32_t fdltsd_CloseSub(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                               uint32_t stc_channel)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn0, *pfunc_stc;
    uint32_t  ch_mask;
    int32_t   ret;

    if(pfunc_sub->pTSDCtxHdl == NULL) {
        return 0;
    }

    FAPEXDebugPrint(3, "[fdltsd_CloseSub] status %d, flags %08x\n",
                       pfunc_sub->status, pfunc_sub->flags);

    if(stc_channel == (1 << pfunc_sub->pid_ch_no)) {
        if(pfunc_sub->pBMCtxHdl == NULL) {
            /* close STC channel */
            ret = sync_funcStr.SetPcrPid(FAPI_SYNC_INVALIDATE);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_CloseSub] "
                                   "FAPI_SYNC_SetPcrPid error:%d\n", ret);
                return ret;
            }
        }
        pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;
        pfunc_stc  = pfunc_cmn - (pfunc_cmn->block_no & 0x01);

        /*---------- LOCK (stc_channel) ----------*/
        FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

        pfunc_stc->stc_channel |= pfunc_cmn->stc_channel_mask;

        /*--------- UNLOCK (stc_channel) ---------*/
        FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);
    }

    if( (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_ALL)
     && (pfunc_sub->status == FAPEX_STATUS_RUNNING) ) {
        if(pfunc_sub->pTSDCtxHdl != NULL) {
            /* stop channel */
            ret = tsd_funcStr.StopChannel(pfunc_sub->pTSDCtxHdl);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_CloseSub] "
                                   "FAPI_TSD_StopChannel error:%d\n", ret);
                return ret;
            }
        }
        /* set stop event */
        fdltsd_EventNotifier(pfunc_sub, FAPEX_TSD_BUFFSTAT_STOPPED);
    }

    /* close TSD's FAPI context */
    ret = fdltsd_CloseCtx(pfunc_cmn, pfunc_sub);
    if(ret != FAPI_OK) {
        return ret;
    }

    if(pfunc_sub->pSctFifo != NULL) {
        FAPI_SYS_FREE(pfunc_sub->pSctFifo);
        pfunc_sub->pSctFifo = NULL;
        pfunc_sub->sct_max  = 0;
    }

    /* clear control flags */
    FAPEXCFDLSAF_DELETE_EVENT(&pfunc_sub->dt_event);

    pfunc_sub->flags = FAPEXFDL_TSD_CHFLAGS_NONE;

    ch_mask = ~(1 << pfunc_sub->pid_ch_no);

    /*---------- LOCK ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

    /* update status */
    pfunc_cmn->enable_channel  &= ch_mask;
    pfunc_cmn->running_channel &= ch_mask;

    pfunc_sub->status = FAPEX_STATUS_LOADED;
    if(pfunc_cmn->enable_channel == 0) {
        pfunc_cmn->status = FAPEX_STATUS_OPENED;
    }
    else if(pfunc_cmn->running_channel == 0) {
        pfunc_cmn->status = FAPEX_STATUS_READY;
    }

    /*--------- UNLOCK ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  Event Notifier
**
** \param   handle       Valid driver instance handle previously returned by
**                       #FAPI_TSD_Open.
** \param   optDataPtr   Optional data to be passed to the notify function.
**
******************************************************************************/
static void fdltsd_PacketTimeoutNotifier(uint32_t blockIndex, void* optDataPtr)
{
    fdltsd_EventNotifier((FAPEXFDL_TSD_SubStrT *)optDataPtr,
                         FAPEX_TSD_BUFFSTAT_TIMEOUT);
}

static void fdltsd_PacketEndNotifier(FAPI_SYS_HandleT handle, void* optDataPtr)
{
    fdltsd_EventNotifier((FAPEXFDL_TSD_SubStrT *)optDataPtr,
                         FAPEX_TSD_BUFFSTAT_PACKET_END);
}

static void fdltsd_SectionEndNotifier(FAPI_SYS_HandleT handle, uint32_t bufferAddress,
                                      uint32_t bufferLength, uint32_t crcError,
                                      void* optDataPtr)
{
    FAPEXFDL_TSD_SubStrT    *pfunc_sub;
    FAPEXFDL_TSD_SctFifoT   *sct_params;
    FAPEXFDL_TSD_SctInfoT   *sct_data;
    FAPEXCFDLSAF_EventStrT  *dt_event_ptr;
    uint32_t  offset, ntf_type, w_index;

    sct_params = (FAPEXFDL_TSD_SctFifoT*)optDataPtr;
    pfunc_sub  = (FAPEXFDL_TSD_SubStrT*)sct_params->pParentSubHdl;

    sct_data = &sct_params->sct_info[sct_params->write_idx];

    sct_params->write_idx ++;
    if(sct_params->write_idx >= FAPEXFDL_TSD_SECTION_FIFO_MAX) {
        sct_params->write_idx = 0;
    }
    offset = bufferAddress - FAPI_SYS_GET_VIRTUAL_ADDRESS( pfunc_sub->buff_address ) + 4;
    if(offset >= pfunc_sub->buff_size) {
        offset -= pfunc_sub->buff_size;
    }
    sct_data->sct_base   = offset;
    sct_data->sct_length = bufferLength;

    /* setup notifer event data */
    dt_event_ptr = &pfunc_sub->dt_event;

    w_index = dt_event_ptr->write_idx & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;

    ntf_type = FAPEX_TSD_BUFFSTAT_SECTION_END |
               (sct_params->notifier_id << FAPEXFDL_TSD_EVENT_SECTION_NO_SHFT);
    if(crcError != 0) {
        ntf_type |= FAPEX_TSD_BUFFSTAT_CRCERROR;
    }
    pfunc_sub->ntfyFifo[w_index].notifier_type = ntf_type;
    pfunc_sub->ntfyFifo[w_index].pData         = sct_data;

    /* set event */
    FAPEXCFDLSAF_SET_EVENT(dt_event_ptr, ntf_type,
                           FAPEXCFDLSAF_WAITEVENT_QUEUE_WRITE);
}

static void fdltsd_ScrambledPacketNotifier(FAPI_SYS_HandleT handle, uint32_t risingEdge,
                                           void* optDataPtr)
{
    fdltsd_EventNotifier((FAPEXFDL_TSD_SubStrT *)optDataPtr,
                         FAPEX_TSD_BUFFSTAT_SCRAMBLED_PACKET);
}

static void fdltsd_BufferOverflowNotifier(uint32_t blockIndex,
                                          uint32_t channelIndex, void* optDataPtr)
{
    fdltsd_EventNotifier((FAPEXFDL_TSD_SubStrT *)optDataPtr,
                         FAPEX_TSD_BUFFSTAT_OVERFLOW);
}

static void fdltsd_VideoBufferOverflowNotifier(void* optDataPtr)
{
    fdltsd_EventNotifier((FAPEXFDL_TSD_SubStrT *)optDataPtr,
                         FAPEX_TSD_BUFFSTAT_OVERFLOW);
}

static void fdltsd_BufferHiThresholdOverNotifier(uint32_t blockIndex,
                                                 uint32_t channelIndex, void* optDataPtr)
{
    fdltsd_EventNotifier((FAPEXFDL_TSD_SubStrT *)optDataPtr,
                         FAPEX_TSD_BUFFSTAT_THRESHOVER);
}

static void fdltsd_BufferLoThresholdUnderNotifier(uint32_t blockIndex,
                                                  uint32_t channelIndex, void* optDataPtr)
{
    fdltsd_EventNotifier((FAPEXFDL_TSD_SubStrT *)optDataPtr,
                         FAPEX_TSD_BUFFSTAT_THRESHUNDER);
}

/*!
*******************************************************************************
**
** \brief  TSD start
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
**
******************************************************************************/
static int32_t fdltsd_Start(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub)
{
    FAPEXFDL_TSD_CmnStrT    *pfunc_cmn0;
    FAPEXCFDLSAF_EventStrT  *dt_event_ptr;
    uint32_t    enable_cmn, enable_ch, running_ch;
    uint32_t    id, one_ch_flag;
    int32_t     ret;

    enable_ch  = 0x00000001;
    running_ch = 0;

    pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;

    /*---------- LOCK ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

    enable_cmn = pfunc_cmn->enable_channel;

    /*--------- UNLOCK ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);

    /* clear another TSD block flag */

    if(pfunc_sub != NULL) {
        /* sub descriptor */
        enable_ch <<= pfunc_sub->pid_ch_no;
        enable_cmn &= enable_ch;
        one_ch_flag = 1;
    }
    else {
        /* common descriptor */
        id = FAPEXDEV_TSD_SUB_MINOR_FIRST
           + (pfunc_cmn->block_no * FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM);
        pfunc_sub   = (FAPEXFDL_TSD_SubStrT*)FAPEXCFDL_GetFdlHdl(pfunc_cmn->pCfdlHdl, id);
        one_ch_flag = 0;
    }

    /* start channel */
    while(enable_cmn) {
        if(pfunc_sub->status == FAPEX_STATUS_READY) {
            if(pfunc_sub->pBMCtxHdl != NULL) {
                /* reset Buffer Read Pointer */
                if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_OUT_BOUND) {

                    if( (pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_VIDEO)
                     && (pfunc_sub->output_id >= 0) ) {
                        FAPEXFDL_VIDEC_StrT   *pfunc_videc;

                        /* clear overflow callback */
                        ret = bm_funcStr.SetOverflowIrq( pfunc_sub->pBMCtxHdl, NULL, NULL, 0 );

                        if(ret == FAPI_OK) {
                            pfunc_videc = (FAPEXFDL_VIDEC_StrT*)FAPEXCFDL_GetFdlHdl(
                                                    pfunc_sub->pCfdlHdl, pfunc_sub->output_id);

                            ret = videc_funcStr.SetVesOverflowCallback(
                                                    pfunc_videc->pFAPICtxHdl,
                                                    fdltsd_VideoBufferOverflowNotifier,
                                                    pfunc_sub );
                            if(ret != FAPI_OK) {
                                FAPEXDebugPrint(2, "[fdltsd_Start] "
                                                   "FAPI_VIDEC_SetVesOverflowCallback error:%d\n",
                                                   ret);
                            }
                        }
                        else {
                            FAPEXDebugPrint(2, "[fdltsd_Start] "
                                               "FAPI_BM_SetOverflowIrq error:%d\n", ret);
                        }
                        if(ret != FAPI_OK) {
                            return ret;
                        }
                    }
                    ret = bm_funcStr.SetBufferReadPointer(pfunc_sub->pBMCtxHdl,
                                                          (pfunc_sub->buff_size << 1), 1);
                }
                else {
                    pfunc_sub->read_pt = 0;
                    ret = bm_funcStr.SetBufferReadPointer(pfunc_sub->pBMCtxHdl, 0, 0);
                }
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdltsd_Start] "
                                       "FAPI_BM_SetBufferReadPointer error:%d\n", ret);
                    return ret;
                }

                /* reset Buffer Write Pointer */
                ret = bm_funcStr.SetBufferWritePointer(pfunc_sub->pBMCtxHdl, 0, 0);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdltsd_Start] "
                                       "FAPI_BM_SetBufferWritePointer error:%d\n", ret);
                    return ret;
                }

                /* check level IRQ */
                if(bm_funcStr.GetHighLevelIrq(pfunc_sub->pBMCtxHdl) > 0) {
                    ret = bm_funcStr.SetHighLevelIrq(pfunc_sub->pBMCtxHdl,
                                                     NULL, NULL, 0, FAPI_BM_ENABLE_IRQ);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_Start] "
                                           "FAPI_BM_SetHighLevelIrq error:%d\n", ret);
                    }
                }

                /* start AUD filtering */
                if(pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_VIDEO) {
                    ret = bm_funcStr.StartAudFiltering(pfunc_sub->pBMCtxHdl);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_Start] "
                                           "FAPI_BM_StartAudFiltering error:%d\n", ret);
                        return ret;
                    }
                }

                /* check BM Hardware Controlled Block */
                if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_HCB) {
                    ret = bm_funcStr.SetReadEnable(pfunc_sub->pBMCtxHdl, 1);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_Start] "
                                           "FAPI_BM_SetReadEnable error:%d\n", ret);
                        return ret;
                    }
                }

                /* clear parameter */
                if(pfunc_sub->pSctFifo != NULL) {
                    FAPEXFDL_TSD_SctFifoT  *sct_params;
                    uint32_t  i;

                    sct_params = pfunc_sub->pSctFifo;
                    for(i=0; i<pfunc_sub->sct_max; i++) {
                        sct_params->write_idx = 0;
                        sct_params ++;
                    }
                    pfunc_sub->ready_idx = 0;
                }
                dt_event_ptr = &pfunc_sub->dt_event;
                dt_event_ptr->read_idx  = 0;
                dt_event_ptr->write_idx = 0;

                /* start channel */
                ret = tsd_funcStr.StartChannel(pfunc_sub->pTSDCtxHdl);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdltsd_Start] FAPI_TSD_StartChannel error:%d\n",
                                    ret);
                    return ret;
                }
                FAPEXDebugPrint(3, "[fdltsd_Start] id [%d]\n", pfunc_sub->id);
            }
            /* update channel status */
            pfunc_sub->status  = FAPEX_STATUS_RUNNING;
            running_ch        |= enable_ch;
        }
        else if(one_ch_flag != 0) {
            if(pfunc_sub->status != FAPEX_STATUS_RUNNING) {
                return FAPEX_TSD_ERR_INVALID_STATUS;
            }
        }
        enable_cmn &= ~enable_ch;

        /* setup for next channel */
        pfunc_sub ++;
        enable_ch <<= 1;
    }

    /* update block status */
    if(running_ch != 0) {
        /*---------- LOCK ----------*/
        FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

        pfunc_cmn->running_channel |= running_ch;
        pfunc_cmn->status = FAPEX_STATUS_RUNNING;

        /*--------- UNLOCK ---------*/
        FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);
    }

    return 0;
}

/*!
*******************************************************************************
**
** \brief  TSD stop
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
**
******************************************************************************/
static int32_t fdltsd_Stop(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub)
{
    FAPEXFDL_TSD_CmnStrT  *pfunc_cmn0;
    uint32_t    enable_cmn, disable_ch, stop_channel;
    uint32_t    id;
    int32_t     ret = 0;

    disable_ch   = 0x00000001;
    stop_channel = 0;

    pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;

    /*---------- LOCK ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

    enable_cmn = pfunc_cmn->enable_channel;

    /*--------- UNLOCK ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);

    if(pfunc_sub != NULL) {
        /* sub descriptor */
        disable_ch <<= pfunc_sub->pid_ch_no;
        enable_cmn  &= disable_ch;
    }
    else {
        /* common descriptor */
        id = FAPEXDEV_TSD_SUB_MINOR_FIRST
           + (pfunc_cmn->block_no * FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM);
        pfunc_sub = (FAPEXFDL_TSD_SubStrT*)FAPEXCFDL_GetFdlHdl(pfunc_cmn->pCfdlHdl, id);
    }

    /* stop channel */
    while(enable_cmn) {
        if( (enable_cmn & disable_ch)
         && (pfunc_sub->status == FAPEX_STATUS_RUNNING) ) {
            if(pfunc_sub->pBMCtxHdl != NULL) {
                /* stop channel */
                ret = tsd_funcStr.StopChannel(pfunc_sub->pTSDCtxHdl);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdltsd_Stop] "
                                       "FAPI_TSD_StopChannel error:%d\n", ret);
                    break;
                }

                /* check BM Hardware Controlled Block */
                if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_HCB) {
                    ret = bm_funcStr.SetReadEnable(pfunc_sub->pBMCtxHdl, 0);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_Stop] "
                                           "FAPI_BM_SetReadEnable error:%d\n", ret);
                        break;
                    }
                }

                /* stop AUD filtering */
                if(pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_VIDEO) {
                    ret = bm_funcStr.StopAudFiltering(pfunc_sub->pBMCtxHdl);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_Stop] "
                                           "FAPI_BM_StopAudFiltering error:%d\n", ret);
                        return ret;
                    }
                }

                /* check level IRQ */
                if(bm_funcStr.GetHighLevelIrq(pfunc_sub->pBMCtxHdl) > 0) {
                    /* enable the high level IRQ again */
                    ret = bm_funcStr.SetHighLevelIrq(pfunc_sub->pBMCtxHdl,
                                                     NULL, NULL, 0, FAPI_BM_ENABLE_IRQ);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_Stop] "
                                           "FAPI_BM_SetHighLevelIrq error:%d\n", ret);
                        break;
                    }
                }

                /* clear callbacks */
                if( (pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_VIDEO)
                 && (pfunc_sub->output_id >= 0) ) {
                    FAPEXFDL_VIDEC_StrT   *pfunc_videc;

                    pfunc_videc = (FAPEXFDL_VIDEC_StrT*)FAPEXCFDL_GetFdlHdl(
                                            pfunc_sub->pCfdlHdl, pfunc_sub->output_id);

                    ret = videc_funcStr.ClearVesOverflowCallback(pfunc_videc->pFAPICtxHdl);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_Stop] "
                                           "FAPI_VIDEC_ClearVesOverflowCallback error:%d\n",
                                           ret);
                        break;
                    }
                }

                FAPEXDebugPrint(3, "[fdltsd_Stop] id [%d]\n", pfunc_sub->id);

                /* update channel status */
                pfunc_sub->status = FAPEX_STATUS_READY;
                stop_channel     |= disable_ch;

                /* set stop event */
                fdltsd_EventNotifier(pfunc_sub, FAPEX_TSD_BUFFSTAT_STOPPED);
            }
        }
        enable_cmn &= ~disable_ch;

        /* setup for next channel */
        pfunc_sub ++;
        disable_ch <<= 1;
    }

    /* update block status */
    if(stop_channel != 0) {
        /*---------- LOCK ----------*/
        FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

        pfunc_cmn->running_channel &= ~stop_channel;
        if(pfunc_cmn->running_channel == 0) {
            pfunc_cmn->status = FAPEX_STATUS_READY;
        }

        /*--------- UNLOCK ---------*/
        FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);
    }
    return 0;
}

/*!
*******************************************************************************
**
** \brief  set channel status to FAPEX_STATUS_READY
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
**
******************************************************************************/
static void fdltsd_SetChStatusReady(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub)
{
    /* set TSD channel status to READY */
    pfunc_sub->status = FAPEX_STATUS_READY;

    /*---------- LOCK ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

    if(pfunc_cmn->running_channel == 0) {
        pfunc_cmn->status = FAPEX_STATUS_READY;
    }

    /*--------- UNLOCK ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);
}

/*!
*******************************************************************************
**
** \brief  setup BM paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pParams      pointer to FAPEX_TSD_PidParamStrT parameter
** \param  stream_type  stream_type from PMT
**
******************************************************************************/
static int32_t fdltsd_SetupBM(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                              FAPEX_TSD_PidParamStrT *pParams, uint8_t stream_type)
{
    FAPI_BM_OpenParamsT  open_params;
    FAPI_BM_InfoT        bm_info;
    int32_t              channel_id, ret;
    uint32_t             rest_size;
#ifdef FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn0;
#endif

    /* check buffer aligned size */
    rest_size = pParams->bufferSize % FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE;
    if(rest_size != 0) {
        rest_size = FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE - rest_size;
        if(pParams->bufferSize == pParams->bufferThresholdH) {
            pParams->bufferThresholdH += rest_size;
        }
        pParams->bufferSize += rest_size;
    }

    /* check the necessary for BM's setup */
    if(pfunc_sub->pBMCtxHdl != NULL) {
        if( (pfunc_sub->stream_type == stream_type)
         && (pfunc_sub->buff_size >= pParams->bufferSize)
         && ((pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_BUFF_CTRL_MSK)
             == (pParams->flags & FAPEX_TSD_FLTFLAGS_SWAP_MSK)) ) {
            /* already setup BM */
            return FAPI_OK;
        }
        else {
            /* close BM handler */
            ret = bm_funcStr.Close(pfunc_sub->pBMCtxHdl);
            if(ret != FAPI_OK) {
                return ret;
            }
            pfunc_sub->pBMCtxHdl = NULL;
            pfunc_sub->flags    &= ~FAPEXFDL_TSD_CHFLAGS_BUFF_CTRL_MSK;

            /*---------- LOCK ----------*/
            FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

            pfunc_cmn->bm_channel &= ~(1 << pfunc_sub->bm_ch_no);

            /*--------- UNLOCK ---------*/
            FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);
        }
    }

    /* setup BM's open parameters */
    open_params.version    = FAPI_BM_VERSION;
    open_params.blockIndex = pfunc_cmn->block_no;

    /* check hardware controled block buffer */
    open_params.absAddress = 0;
    channel_id          = FAPI_BM_CHANNEL_AUTODETECT;
    pfunc_sub->flags   &= ~(FAPEXFDL_TSD_CHFLAGS_OUT_BOUND | FAPEXFDL_TSD_CHFLAGS_HCB);
    switch(stream_type) {
        case FAPEX_TSD_STREAMTYPE_AUDIO:
            /* audio stream */
            if(pfunc_cmn->pAStreamBuff != NULL) {
                open_params.absAddress = (uint32_t)pfunc_cmn->pAStreamBuff;
            }
            break;

        case FAPEX_TSD_STREAMTYPE_VIDEO:
            /* video stream */
            if(pfunc_cmn->pVStreamBuff != NULL) {
                open_params.absAddress = (uint32_t)pfunc_cmn->pVStreamBuff;
            }
            pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_OUT_BOUND;

            /* disable PTS evaluator & notifer */
            ret = tsd_funcStr.SetPtsEvaluator(pfunc_sub->pTSDCtxHdl, NULL, NULL, 0);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_SetupBM] FAPI_TSD_SetPtsEvaluator error:%d\n",
                                ret);
                return ret;
            }
            break;

        case FAPEX_TSD_STREAMTYPE_TELETEXT:
            /* teletext stream */
            if(pfunc_cmn->pTStreamBuff != NULL) {
                open_params.absAddress = (uint32_t)pfunc_cmn->pTStreamBuff;
            }
            if( (bm_funcStr.GetHardwareControlledBlock() == pfunc_cmn->block_no)
             && !(pfunc_cmn->bm_channel & (1 << FAPI_BM_TELETEXT_BUFFER)) ) {
                channel_id = FAPI_BM_TELETEXT_BUFFER;
                pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_HCB;
            }
            break;
    }
    pfunc_sub->stream_type = stream_type;
    open_params.channelId  = channel_id;

#ifdef FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
    pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;
#endif

    /* setup buffer address */
    if(open_params.absAddress == 0) {
        if(pfunc_sub->buff_allocate_type == FAPEXFDL_TSD_STRM_BUFF_TYPE_RESERVED) {
            pfunc_sub->buff_size = 0;
        }
        if(pfunc_sub->buff_size < pParams->bufferSize) {
#ifdef FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
            if(pfunc_sub->buff_allocate_type == FAPEXFDL_TSD_STRM_BUFF_TYPE_STATIC) {
                FAPISAF_FreeSBuffArea( pfunc_cmn0->sbuff_ctrl,
                                       pfunc_sub->buff_address );
 #if (DEBUG >= 2)
                fdltsd_CheckBuffList("FreeSBuff", pfunc_cmn0->sbuff_ctrl,
                                     pfunc_sub->buff_address);
 #endif
            }
            else
#endif
            if(pfunc_sub->buff_allocate_type == FAPEXFDL_TSD_STRM_BUFF_TYPE_DYNAMIC) {
                fdltsd_FreeDBuffArea( pfunc_sub );
            }

#ifdef FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
            /* 1st step buffer allocation */
            pfunc_sub->buff_address = FAPISAF_AllocSBuffArea( pfunc_cmn0->sbuff_ctrl,
                                                              pParams->bufferSize );
            if(pfunc_sub->buff_address != 0) {
                pfunc_sub->buff_allocate_type = FAPEXFDL_TSD_STRM_BUFF_TYPE_STATIC;
 #if (DEBUG >= 2)
                fdltsd_CheckBuffList("AllocSBuff", pfunc_cmn0->sbuff_ctrl,
                                     pfunc_sub->buff_address);
 #endif
            }
            else
#endif
            {
                /* 2nd step buffer allocation */
                pfunc_sub->buff_address = fdltsd_AllocDBuffArea( pfunc_sub, pParams->bufferSize );
                if(pfunc_sub->buff_address == 0) {
                    pfunc_sub->buff_allocate_type = FAPEXFDL_TSD_STRM_BUFF_TYPE_UNKNOWN;
                    return FAPEX_TSD_ERR_OUT_OF_MEMORY;
                }
                pfunc_sub->buff_allocate_type = FAPEXFDL_TSD_STRM_BUFF_TYPE_DYNAMIC;
            }
        }
        /*for H61 integrated MANGO*/       
        open_params.absAddress = pfunc_sub->buff_address ;
        /*  open_params.absAddress = pfunc_sub->buff_address - bm_funcStr.GetGlobalOffset(pfunc_cmn->block_no);*/
        FAPEXDebugPrint(3, "## [fdltsd_SetupBM] 1.absAddress=%xh\n",open_params.absAddress);
    }
    else {
        /* Audio/Video/TTX stream */
#ifdef FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
        if(pfunc_sub->buff_allocate_type == FAPEXFDL_TSD_STRM_BUFF_TYPE_STATIC) {
            FAPISAF_FreeSBuffArea( pfunc_cmn0->sbuff_ctrl,
                                   pfunc_sub->buff_address );
 #if (DEBUG >= 2)
            fdltsd_CheckBuffList("FreeSBuff", pfunc_cmn0->sbuff_ctrl,
                                 pfunc_sub->buff_address);
 #endif
        }
        else
#endif
        if(pfunc_sub->buff_allocate_type == FAPEXFDL_TSD_STRM_BUFF_TYPE_DYNAMIC) {
            fdltsd_FreeDBuffArea( pfunc_sub );
        }
        pfunc_sub->buff_allocate_type = FAPEXFDL_TSD_STRM_BUFF_TYPE_RESERVED;
        pfunc_sub->buff_address = open_params.absAddress;
    }
    pfunc_sub->buff_size = open_params.size = pParams->bufferSize;

    /* setup buffer swap mode */
    if((pParams->flags & FAPEX_TSD_FLTFLAGS_SWAPWR_MSK)) {
        open_params.swapWrite = FAPI_BM_SWAP;
    }
    else {
        open_params.swapWrite = FAPI_BM_NO_SWAP;
    }
    if((pParams->flags & FAPEX_TSD_FLTFLAGS_SWAPRD_MSK)) {
        open_params.swapRead = FAPI_BM_SWAP;
    }
    else {
        open_params.swapRead = FAPI_BM_NO_SWAP;
    }
    if((pParams->flags & FAPEX_TSD_FLTFLAGS_SWAPMODE_MSK)) {
        open_params.swapMode = FAPI_BM_SWAP_2_BYTES;
    }
    else {
        open_params.swapMode = FAPI_BM_SWAP_4_BYTES;
    }

    /* call Open BM */
    pfunc_sub->pBMCtxHdl = bm_funcStr.Open(&open_params, &ret);
    FAPEXDebugPrint(3, "## [fdltsd_SetupBM] 2.absAddress:%08xh(%08xh),size:%xh ,ret=%xh\n",
                       (uint32_t)open_params.absAddress,FAPI_SYS_GET_VIRTUAL_ADDRESS(open_params.absAddress),open_params.size,ret);
    if(pfunc_sub->pBMCtxHdl == NULL) {
        FAPEXDebugPrint(2, "[fdltsd_SetupBM] FAPI_BM_Open error:%d [%d]\n",
                           ret, pfunc_sub->id);
        return ret;
    }

    /* get BM information */
    ret = bm_funcStr.GetInfo(pfunc_sub->pBMCtxHdl, &bm_info);
    if(ret != FAPI_OK) {
        return ret;
    }
    pfunc_sub->bm_ch_no = bm_info.channelId;

    /*---------- LOCK ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

    pfunc_cmn->bm_channel |= 1 << bm_info.channelId;

    /*--------- UNLOCK ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

    /* set BM information */
    ret = tsd_funcStr.SetBuffer(pfunc_sub->pTSDCtxHdl, bm_info);
    if(ret != FAPI_OK) {
        FAPEXDebugPrint(2, "[fdltsd_SetupBM] FAPI_TSD_SetBuffer error:%d\n", ret);
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  setup BM threshold notifier
**
** \param  pfunc_sub    pointer to sub context
** \param  pParams      pointer to FAPEX_TSD_PidParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetupBMThreshold(FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                       FAPEX_TSD_PidParamStrT *pParams)
{
    int32_t  ret = 0;

    /* low threshold check */
    if(pParams->bufferThresholdL > 0) {
        /* set Low threshold underflow notifier */
        ret = bm_funcStr.SetLowLevelIrq( pfunc_sub->pBMCtxHdl,
                                         fdltsd_BufferLoThresholdUnderNotifier,
                                         pfunc_sub,
                                         pParams->bufferThresholdL,
                                         1 );
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdltsd_SetupBMThreshold] "
                               "FAPI_BM_SetLowLevelIrq error:%d\n", ret);
            return ret;
        }
        pfunc_sub->buff_th_L = pParams->bufferThresholdL;
        pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_NOTIFIER_BUFF_TH;
    }
    else {
        pfunc_sub->buff_th_L = 0;
    }

    /* high threshold check */
    if(pParams->bufferSize > pParams->bufferThresholdH) {
        /* set High threshold overflow notifier */
        ret = bm_funcStr.SetHighLevelIrq( pfunc_sub->pBMCtxHdl,
                                          fdltsd_BufferHiThresholdOverNotifier,
                                          pfunc_sub,
                                          pParams->bufferSize - pParams->bufferThresholdH,
                                          1 );
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdltsd_SetupBMThreshold] "
                               "FAPI_BM_SetHighLevelIrq error:%d\n", ret);
            return ret;
        }
        pfunc_sub->buff_th_H = pParams->bufferThresholdH;
        pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_NOTIFIER_BUFF_TH;
    }
    else {
        pfunc_sub->buff_th_H = pParams->bufferSize;
    }

    /* set buffer overflow notifier */
    if(!(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_OUT_BOUND)) {
        ret = bm_funcStr.SetOverflowIrq( pfunc_sub->pBMCtxHdl,
                                         fdltsd_BufferOverflowNotifier,
                                         pfunc_sub,
                                         1 );
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdltsd_SetupBMThreshold] "
                               "FAPI_BM_SetOverflowIrq error:%d\n", ret);
            return ret;
        }
    }
    pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_NOTIFIER_BUFF_ERR;

    return ret;
}

/*!
*******************************************************************************
**
** \brief  set channel PID value
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pParams      pointer to FAPEX_TSD_PidParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetChannelPid(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                    FAPEX_TSD_PidParamStrT *pParams)
{
    FAPEXFDL_TSD_CmnStrT  *pfunc_cmn0, *pfunc_stc;
    int32_t   old_pid;
    uint32_t  stc_channel;
    uint32_t  stc_undecided_flag;
    uint32_t  blk_shift;
    int32_t   ret = 0;

    old_pid = tsd_funcStr.GetChannelPid(pfunc_sub->pTSDCtxHdl);

    pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;
    pfunc_stc  = pfunc_cmn - (pfunc_cmn->block_no & 0x01);

    stc_undecided_flag = (uint32_t)pfunc_stc->stc_channel_mask;

    /*---------- LOCK (stc_channel) ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

    if(pfunc_cmn == pfunc_stc) {
        stc_channel = (uint32_t)(pfunc_stc->stc_channel & pfunc_cmn->stc_channel_mask);
        blk_shift   = 0;
    }
    else {
        stc_channel = (uint32_t)(pfunc_stc->stc_channel >> FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM);
        blk_shift   = FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM;
    }

    /*--------- UNLOCK (stc_channel) ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);

    if(old_pid == pParams->pid) {
        /****  PID : non-changing  ****/
        if(stc_channel == (1 << pfunc_sub->pid_ch_no)) {
            if(!(pParams->flags & FAPEX_TSD_FLTFLAGS_PCR)) {
                if((pfunc_sub->pBMCtxHdl == NULL) && (pParams->bufferSize == 0)) {
                    /*  PCR : enable -> disable  */
                    ret = sync_funcStr.SetPcrPid(FAPI_SYNC_INVALIDATE);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_SetChannelPid] "
                                           "FAPI_SYNC_SetPcrPid error:%d\n", ret);
                        return ret;
                    }
                }
                /*---------- LOCK (stc_channel) ----------*/
                FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

                /* clear PCR/STC channel flag */
                pfunc_stc->stc_channel |= pfunc_cmn->stc_channel_mask;  /* undecided channel */

                /*--------- UNLOCK (stc_channel) ---------*/
                FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);
            }
        }
        else if(stc_channel != 0) {
            if(pParams->flags & FAPEX_TSD_FLTFLAGS_PCR) {
                /*  PCR : disable -> enable  */
                if((pfunc_sub->pBMCtxHdl == NULL) && (pParams->bufferSize == 0)) {
                    if(stc_channel != stc_undecided_flag) {
                        ret = sync_funcStr.SetPcrPid(FAPI_SYNC_INVALIDATE);
                        if(ret != FAPI_OK) {
                            FAPEXDebugPrint(2, "[fdltsd_SetChannelPid] "
                                               "FAPI_SYNC_SetPcrPid error:%d\n", ret);
                            return ret;
                        }
                    }
                    ret = sync_funcStr.SetPcrPid(pParams->pid);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_SetChannelPid] "
                                           "FAPI_SYNC_SetPcrPid error:%d\n", ret);
                        return ret;
                    }
                }
                /*---------- LOCK (stc_channel) ----------*/
                FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

                /* change PCR/STC channel flag */
                if(blk_shift == 0) {
                    pfunc_stc->stc_channel = (1 << pfunc_sub->pid_ch_no) |
                                             ((uint64_t)stc_undecided_flag << FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM);
                }
                else {
                    pfunc_stc->stc_channel = (1 << (pfunc_sub->pid_ch_no + blk_shift)) |
                                             stc_undecided_flag;
                }

                /*--------- UNLOCK (stc_channel) ---------*/
                FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);
            }
        }
    }
    else {
        /****  PID : changing  ****/
        if(stc_channel != 0) {
            if( (pParams->flags & FAPEX_TSD_FLTFLAGS_PCR)
             || (stc_channel == (1 << pfunc_sub->pid_ch_no)) ) {
                if((pfunc_sub->pBMCtxHdl == NULL) && (pParams->bufferSize == 0) && (old_pid >= 0)) {
                    /*  PCR : disable  */
                    ret = sync_funcStr.SetPcrPid(FAPI_SYNC_INVALIDATE);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_SetChannelPid] "
                                           "FAPI_SYNC_SetPcrPid error:%d\n", ret);
                        return ret;
                    }
                }
                if(!(pParams->flags & FAPEX_TSD_FLTFLAGS_PCR)) {
                    /*---------- LOCK (stc_channel) ----------*/
                    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

                    /* clear PCR/STC channel flag */
                    pfunc_stc->stc_channel |= pfunc_cmn->stc_channel_mask;  /* undecided channel */

                    /*--------- UNLOCK (stc_channel) ---------*/
                    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);
                }
            }
            /* check PCR filtering */
            if(pParams->flags & FAPEX_TSD_FLTFLAGS_PCR) {
                if((pfunc_sub->pBMCtxHdl == NULL) && (pParams->bufferSize == 0)) {
                    /*  PCR : enable  */
                    ret = sync_funcStr.SetPcrPid(pParams->pid);
                    if(ret != FAPI_OK) {
                        FAPEXDebugPrint(2, "[fdltsd_SetChannelPid] "
                                           "FAPI_SYNC_SetPcrPid error:%d\n", ret);
                        return ret;
                    }
                }
                /*---------- LOCK (stc_channel) ----------*/
                FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

                /* change PCR/STC flag */
                if(blk_shift == 0) {
                    pfunc_stc->stc_channel = (1 << pfunc_sub->pid_ch_no) |
                                             ((uint64_t)stc_undecided_flag << FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM);
                }
                else {
                    pfunc_stc->stc_channel = (1 << (pfunc_sub->pid_ch_no + blk_shift)) |
                                             stc_undecided_flag;
                }

                /*--------- UNLOCK (stc_channel) ---------*/
                FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);
            }
        }
        if(pfunc_sub->pTSDCtxHdl != NULL) {
            /* change channel PID value */
            ret = tsd_funcStr.SetChannelPid(
                        pfunc_sub->pTSDCtxHdl, pParams->pid,
                        (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_CHECK_CC),
                        (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_RM_TS_HEADER));
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_SetChannelPid] FAPI_TSD_SetChannelPid error:%d\n",
                                ret);
            }
        }
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  set channel PID value
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_PidParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetPidValue(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                  FAPEX_TSD_PidParamStrT *pArg)
{
    FAPEX_TSD_PidParamStrT  params;
    int32_t  ret;

    /* check channel status */
    if( (pfunc_sub->status != FAPEX_STATUS_READY)
     || (pfunc_sub->pTSDCtxHdl == NULL) ) {
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_TSD_PidParamStrT));

    /* set channel PID */
    ret = fdltsd_SetChannelPid(pfunc_cmn, pfunc_sub, &params);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  set PID filter paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_PidParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetPidFilter(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                   FAPEX_TSD_PidParamStrT *pArg)
{
    FAPEX_TSD_PidParamStrT  params;
    uint8_t  strm_type;
    int32_t  ret = 0;

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_TSD_PidParamStrT));

    /* check continuity counter flag */
    if(params.flags & FAPEX_TSD_FLTFLAGS_CHECK_CC) {
        pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_CHECK_CC;
    }
    else {
        pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_CHECK_CC;
    }
    /* remove TS header flag */
    if(params.flags & FAPEX_TSD_FLTFLAGS_RM_TS_HEADER) {
        pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_RM_TS_HEADER;
    }
    else {
        pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_RM_TS_HEADER;
    }

    /* set channel PID */
    ret = fdltsd_SetChannelPid(pfunc_cmn, pfunc_sub, &params);
    if(ret != FAPI_OK) {
        return ret;
    }

    /* check whether the PCR/STC is only controlled */
    if((params.bufferSize == 0) && (params.flags & FAPEX_TSD_FLTFLAGS_PCR)) {
        ret = fdltsd_CloseCtx(pfunc_cmn, pfunc_sub);
        if(ret != FAPI_OK) {
            return ret;
        }
    }
    else {
        /* setup event parameters */
        if(params.flags & FAPEX_TSD_FLTFLAGS_PACKET_END) {
            if(!(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND)) {
                /* set packet_end notifier */
                ret = tsd_funcStr.SetPacketEndNotifier(
                            pfunc_sub->pTSDCtxHdl,
                            fdltsd_PacketEndNotifier, pfunc_sub, 1);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdltsd_SetPidFilter]"
                                       "FAPI_TSD_SetPacketEndNotifier error:%d\n", ret);
                    return ret;
                }
                pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND;
            }
        }
        else {
            if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND) {
                /* clear packet_end notifier */
                ret = tsd_funcStr.SetPacketEndNotifier(
                            pfunc_sub->pTSDCtxHdl,
                            NULL, NULL, 0);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdltsd_SetPidFilter]"
                                       "FAPI_TSD_SetPacketEndNotifier error:%d\n", ret);
                    return ret;
                }
                pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND;
            }
        }

        /* setup BM's open parameters */
        if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_PES_PARSER) {
            strm_type = pfunc_sub->stream_type;
        }
        else {
            strm_type = FAPEX_TSD_STREAMTYPE_UNKNOWN;
        }

        ret = fdltsd_SetupBM(pfunc_cmn, pfunc_sub, &params, strm_type);
        if(ret != FAPI_OK) {
            return ret;
        }

        /* setup BM threshold notifier */
        ret = fdltsd_SetupBMThreshold(pfunc_sub, &params);
        if(ret != FAPI_OK) {
            return ret;
        }
    }

    /* set TSD channel status to READY */
    fdltsd_SetChStatusReady(pfunc_cmn, pfunc_sub);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  set section filter paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_SctFltParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetSectionFilter(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                       FAPEX_TSD_SctFltParamStrT *pArg)
{
    FAPEX_TSD_SctFltParamStrT    params;
    FAPEXFDL_TSD_SctFifoT        *sct_params;
    FAPI_TSD_FilterDataT         *filter_data;
    uint32_t    size, i;
    int32_t     ret = 0;
    uint32_t    sct_cnt;
    uint32_t    *filter;
    uint32_t    *mask;

    if(pfunc_sub->pTSDCtxHdl == NULL) {
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_TSD_SctFltParamStrT));

    /* setup section filter parameters */
    if(pfunc_sub->table_id_max == 0) {
        pfunc_sub->table_id_max = (params.flags >> FAPEX_TSD_FLTFLAGS_TABLE_NUM_SHIFT) + 1;

        /* allocate filter data */
        size = sizeof(FAPI_TSD_FilterDataT) * pfunc_sub->table_id_max;
        pfunc_sub->pFilterData = (FAPI_TSD_FilterDataT*)FAPI_SYS_MALLOC(size);
        if(pfunc_sub->pFilterData == NULL) {
            return FAPEX_TSD_ERR_OUT_OF_MEMORY;
        }

        /* allocate filter pattern */
        size = sizeof(uint32_t) * FAPEX_TSD_FILTER_SIZE * 2 * pfunc_sub->table_id_max;
        pfunc_sub->pFilter = (uint32_t*)FAPI_SYS_MALLOC(size);
        if(pfunc_sub->pFilter == NULL) {
            FAPI_SYS_FREE(pfunc_sub->pFilterData);
            pfunc_sub->pFilterData = NULL;
            return FAPEX_TSD_ERR_OUT_OF_MEMORY;
        }

        /* allocate section information */
        size = sizeof(FAPEXFDL_TSD_SctFifoT) * pfunc_sub->table_id_max;
        pfunc_sub->pSctFifo = (FAPEXFDL_TSD_SctFifoT*)FAPI_SYS_MALLOC(size);
        if(pfunc_sub->pSctFifo == NULL) {
            FAPI_SYS_FREE(pfunc_sub->pFilter);
            FAPI_SYS_FREE(pfunc_sub->pFilterData);
            pfunc_sub->pFilterData = NULL;
            pfunc_sub->pFilter     = NULL;
            return FAPEX_TSD_ERR_OUT_OF_MEMORY;
        }

        pfunc_sub->pMask = pfunc_sub->pFilter + (FAPEX_TSD_FILTER_SIZE * pfunc_sub->table_id_max);
        i = pfunc_sub->table_id_no = 0;
    }
    else {
        i = pfunc_sub->table_id_no;
    }
    filter_data = pfunc_sub->pFilterData + i;
    filter      = pfunc_sub->pFilter + (FAPEX_TSD_FILTER_SIZE * i);
    mask        = pfunc_sub->pMask   + (FAPEX_TSD_FILTER_SIZE * i);
    sct_params  = pfunc_sub->pSctFifo + i;

    sct_params->notifier_id   = i;
    sct_params->pParentSubHdl = pfunc_sub;
    sct_params->write_idx     = 0;

    sct_cnt = 2;
    for(i=3; i<FAPEX_TSD_FILTER_SIZE; i++) {
        filter[i] = params.filter[i];
        mask  [i] = params.mask  [i];
        if(mask[i] != 0x00) {
            sct_cnt = i;
        }
    }
    mask[2] = params.mask[0];
    filter_data->tableId              = (FAPI_TSD_TableIdT)params.filter[0];
    filter_data->numberOfSectionBytes = sct_cnt - 2;
    filter_data->sectionBytes         = &filter[3];
    filter_data->filterBytes          = &mask[2];
    filter_data->oneShotFiltering     = (params.flags & FAPEX_TSD_FLTFLAGS_ONESHOT_FILTERING);
    filter_data->notifyFunction       = fdltsd_SectionEndNotifier;
    filter_data->optData              = sct_params;

    if(params.flags & FAPEX_TSD_FLTFLAGS_NEW_VERSION) {
        /* filtering only when version number changes */
        filter_data->notifyOnNewVersion = FAPI_TSD_CHECK_VERSION;
    }
    else {
        /* ignore the version when filtering */
        filter_data->notifyOnNewVersion = FAPI_TSD_IGNORE_VERSION;
    }
    pfunc_sub->table_id_no ++;

    if(pfunc_sub->table_id_no < pfunc_sub->table_id_max) {
        return 0;
    }

    /* set section filter */
    ret = tsd_funcStr.SetSectionFilter(
                pfunc_sub->pTSDCtxHdl,
                (FAPI_TSD_NumberOfTablesT)pfunc_sub->table_id_max,
                pfunc_sub->pFilterData);

    FAPI_SYS_FREE(pfunc_sub->pFilter);
    FAPI_SYS_FREE(pfunc_sub->pFilterData);
    pfunc_sub->pFilterData  = NULL;
    pfunc_sub->pFilter      = NULL;
    pfunc_sub->pMask        = NULL;
    pfunc_sub->sct_max      = pfunc_sub->table_id_max;
    pfunc_sub->table_id_max = 0;

    if(ret != FAPI_OK) {
        pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCTEND;
        FAPEXDebugPrint(2, "[fdltsd_SetSectionFilter]"
                           "FAPI_TSD_SetSectionFilter error:%d\n", ret);
        return ret;
    }
    pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCTEND;

    if(params.flags & FAPEX_TSD_FLTFLAGS_PACKET_END) {
        if(!(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND)) {
            /* set packet_end notifier */
            ret = tsd_funcStr.SetPacketEndNotifier(
                        pfunc_sub->pTSDCtxHdl,
                        fdltsd_PacketEndNotifier, pfunc_sub, 1);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_SetSectionFilter]"
                                   "FAPI_TSD_SetPacketEndNotifier error:%d\n", ret);
                return ret;
            }
            pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND;
        }
    }
    else {
        if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND) {
            /* clear packet_end notifier */
            ret = tsd_funcStr.SetPacketEndNotifier(
                        pfunc_sub->pTSDCtxHdl,
                        NULL, NULL, 0);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_SetSectionFilter]"
                                   "FAPI_TSD_SetPacketEndNotifier error:%d\n", ret);
                return ret;
            }
            pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND;
        }
    }

    /* check continuity counter flag */
    if(params.flags & FAPEX_TSD_FLTFLAGS_CHECK_CC) {
        pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_CHECK_CC;
    }
    else {
        pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_CHECK_CC;
    }
    /* remove TS header flag */
    if(params.flags & FAPEX_TSD_FLTFLAGS_RM_TS_HEADER) {
        pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_RM_TS_HEADER;
    }
    else {
        pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_RM_TS_HEADER;
    }

    /* set channel PID */
    ret = tsd_funcStr.SetChannelPid(
                pfunc_sub->pTSDCtxHdl, params.pid,
                (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_CHECK_CC),
                (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_RM_TS_HEADER));
    if(ret != FAPI_OK) {
        FAPEXDebugPrint(2, "[fdltsd_SetSectionFilter]"
                           "FAPI_TSD_SetChannelPid error:%d\n", ret);
        return ret;
    }

    if(params.flags & FAPEX_TSD_FLTFLAGS_TIMEOUT) {
        uint32_t  max_timeout;
        uint32_t  update_flag = 0;

        /* check timeout value */
        if(params.timeout > FAPEXFDL_TSD_TIMEOUT_MAX) {
            params.timeout = FAPEXFDL_TSD_TIMEOUT_MAX;
        }
        pfunc_sub->timeout = params.timeout;

        /*---------- LOCK ----------*/
        FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

        if(pfunc_cmn->timeout == (FAPEXFDL_TSD_TIMEOUT_MAX + 1)) {
            pfunc_cmn->timeout = params.timeout;
            update_flag = 1;
        }
        else if(pfunc_cmn->timeout < params.timeout) {
            pfunc_cmn->timeout = params.timeout;
            update_flag = 1;
        }
        if(pfunc_cmn->timeout > 100) {      /* max 100[msec] */
            pfunc_cmn->timeout = 100;
        }
        max_timeout = pfunc_cmn->timeout;

        if(pfunc_cmn->tout_channel == 0) {
            update_flag = 0;
        }
        pfunc_cmn->tout_channel |= 0x00000001 << pfunc_sub->pid_ch_no;

        /*--------- UNLOCK ---------*/
        FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

        if(!(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_TIMOUT)) {
            /* set timeout notifier */
            if(update_flag == 1) {
                /* clear timeout notifier */
                ret = tsd_funcStr.SetPacketTimeoutNotifier(
                            pfunc_cmn->block_no,
                            NULL, NULL, 0, 0);
            }
            else {
                ret = FAPI_OK;
            }
            if(ret == FAPI_OK) {
                ret = tsd_funcStr.SetPacketTimeoutNotifier(
                            pfunc_cmn->block_no,
                            fdltsd_PacketTimeoutNotifier, pfunc_sub, 1,
                            (max_timeout * FAPEXFDL_TSD_TIMEOUT_TO_CYCLE));
            }
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_SetSectionFilter]"
                                   "FAPI_TSD_SetPacketTimeoutNotifier error:%d\n", ret);
                return ret;
            }
            pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_NOTIFIER_TIMOUT;
        }
    }
    else {
        if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_TIMOUT) {
            uint32_t  tout_channel;

            /*---------- LOCK ----------*/
            FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

            tout_channel = pfunc_cmn->tout_channel
                           & ~(0x00000001 << pfunc_sub->pid_ch_no);
            pfunc_cmn->tout_channel = tout_channel;

            /*--------- UNLOCK ---------*/
            FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

            if(tout_channel == 0) {
                /* clear timeout notifier */
                ret = tsd_funcStr.SetPacketTimeoutNotifier(
                            pfunc_cmn->block_no,
                            NULL, NULL, 0, 0);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdltsd_SetSectionFilter]"
                                       "FAPI_TSD_SetPacketTimeoutNotifier error:%d\n", ret);
                    return ret;
                }
            }
            pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_NOTIFIER_TIMOUT;
        }
    }

    /* setup BM's open parameters */
    ret = fdltsd_SetupBM(pfunc_cmn, pfunc_sub,
                         (FAPEX_TSD_PidParamStrT *)&params,
                         FAPEX_TSD_STREAMTYPE_UNKNOWN);
    if(ret != FAPI_OK) {
        return ret;
    }

    /* setup BM threshold notifier */
    ret = fdltsd_SetupBMThreshold(pfunc_sub,
                                  (FAPEX_TSD_PidParamStrT *)&params);
    if(ret != FAPI_OK) {
        return ret;
    }

    /* set TSD channel status to READY */
    fdltsd_SetChStatusReady(pfunc_cmn, pfunc_sub);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  set PES filter paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_PesParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetPesFilter(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                   FAPEX_TSD_PesParamStrT *pArg)
{
    FAPEX_TSD_PesParamStrT  params;
    int32_t  ret;

    if(pfunc_sub->pTSDCtxHdl == NULL) {
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_TSD_PesParamStrT));

    /* check continuity counter flag */
    if(params.flags & FAPEX_TSD_FLTFLAGS_CHECK_CC) {
        pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_CHECK_CC;
    }
    else {
        pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_CHECK_CC;
    }
    /* remove TS header flag */
    pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_RM_TS_HEADER;

    /* set channel PID */
    ret = fdltsd_SetChannelPid(pfunc_cmn, pfunc_sub, (FAPEX_TSD_PidParamStrT *)&params);
    if(ret != FAPI_OK) {
        return ret;
    }

    if(params.streamType != FAPEX_TSD_STREAMTYPE_TELETEXT) {
        /* setup PES paser */
        if((pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_PES_PARSER)) {
            ret = tsd_funcStr.SetPesParser(pfunc_sub->pTSDCtxHdl, 0, 0, 0);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdltsd_SetPesFilter] "
                                   "FAPI_TSD_SetPesParser error:%d [%d]\n",
                                   ret, pfunc_sub->id);
                return ret;
            }
        }
        ret = tsd_funcStr.SetPesParser(
                    pfunc_sub->pTSDCtxHdl,
                    (params.flags & (FAPEX_TSD_FLTFLAGS_RM_PES_HEADER & ~FAPEX_TSD_FLTFLAGS_RM_TS_HEADER)),
                    (params.flags & FAPEX_TSD_FLTFLAGS_PTS), 1);
        if(ret < 0) {
            FAPEXDebugPrint(2, "[fdltsd_SetPesFilter] "
                               "FAPI_TSD_SetPesParser error:%d [%d]\n",
                               ret, pfunc_sub->id);
            return ret;
        }
        pfunc_sub->pes_paser_no = ret;
        pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_PES_PARSER;
    }

    if(params.flags & FAPEX_TSD_FLTFLAGS_SCRAMBLED_PACKET) {
        if(!(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCRMBLD)) {
            /* set scrambled packet notifier */
            ret = tsd_funcStr.SetScrambledPacketNotifier(
                        pfunc_sub->pTSDCtxHdl,
                        fdltsd_ScrambledPacketNotifier, pfunc_sub, 1,
                        (params.flags & FAPEX_TSD_FLTFLAGS_ONESHOT_FILTERING));
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_SetPesFilter] "
                                   "FAPI_TSD_SetScrambledPacketNotifier error:%d\n", ret);
                return ret;
            }
            pfunc_sub->flags |= FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCRMBLD;
        }
    }
    else {
        if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCRMBLD) {
            /* clear scrambled packet notifier */
            ret = tsd_funcStr.SetScrambledPacketNotifier(
                        pfunc_sub->pTSDCtxHdl,
                        NULL, NULL, 0, 0);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdltsd_SetPesFilter] "
                                   "FAPI_TSD_SetScrambledPacketNotifier error:%d\n", ret);
                return ret;
            }
            pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCRMBLD;
        }
    }

    /* clear section filter parameter */
    if((pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCTEND)) {
        ret = tsd_funcStr.ClearSectionFilter(pfunc_sub->pTSDCtxHdl);
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdltsd_SetPesFilter] "
                               "FAPI_TSD_ClearSectionFilter error:%d\n", ret);
            return ret;
        }

        FAPI_SYS_FREE(pfunc_sub->pSctFifo);
        pfunc_sub->pSctFifo = NULL;
        pfunc_sub->sct_max  = 0;

        pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCTEND;
    }

    /* setup BM's open parameters */
    ret = fdltsd_SetupBM(pfunc_cmn, pfunc_sub,
                         (FAPEX_TSD_PidParamStrT *)&params,
                         params.streamType);
    if(ret != FAPI_OK) {
        return ret;
    }

    /* setup BM threshold notifier */
    ret = fdltsd_SetupBMThreshold(pfunc_sub,
                                  (FAPEX_TSD_PidParamStrT *)&params);
    if( (pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_TELETEXT)
     && (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_HCB) ) {
        /* clear threshold notifier flag */
        pfunc_sub->flags &= ~FAPEXFDL_TSD_CHFLAGS_NOTIFIER_BUFF_TH;
    }
    if(ret != FAPI_OK) {
        return ret;
    }

    /* setup timeout value */
    if(!(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_OUT_BOUND)) {
        pfunc_sub->timeout = FAPEXFDL_TSD_TIMEOUT_PES_FILTER;
    }

    /* set TSD channel status to READY */
    fdltsd_SetChStatusReady(pfunc_cmn, pfunc_sub);

    return ret;
}

#ifdef CONFIG_TSD_USE_INTERNAL_DESC
/*!
*******************************************************************************
**
** \brief  set Descramble Keys paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_DescKeysParamStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetDescramblKeys(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                       FAPEX_TSD_DescKeysParamStrT *pArg)
{
    FAPEX_TSD_DescKeysParamStrT  params;
    FAPI_DESC_OpenParamsT        open_params;
    int32_t     ret;

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_TSD_DescKeysParamStrT));

    /* check the necessary for BM's setup */
    if(pfunc_sub->pDESCCtxHdl == NULL) {
        open_params.version        = FAPI_DESC_VERSION;
        open_params.descBlockIndex = pfunc_cmn->block_no;
        open_params.descChannelId  = FAPI_DESC_CHANNEL_AUTODETECT;
        open_params.tsdHandle = pfunc_sub->pTSDCtxHdl;
        open_params.mode = params.mode;
        open_params.use128BitKey = params.use128BitKey;

        /* call Open DESC */
        pfunc_sub->pDESCCtxHdl = desc_funcStr.Open(&open_params, &ret);
        FAPEXDebugPrint(3, "[fdltsd_SetDescramblKeys] DSEC context:0x%08x\n",
                           (uint32_t)pfunc_sub->pDESCCtxHdl);
      
         if(pfunc_sub->pDESCCtxHdl == NULL) {
            FAPEXDebugPrint(2, "[fdltsd_SetDescramblKeys] FAPI_DESC_Open error:%d\n", ret);
            return ret;
        }
    }

#if 0 /* under consideration for H61 */
    /* set channel mapping */
    ret = desc_funcStr.SetChannelMapping(pfunc_sub->pDESCCtxHdl, pfunc_sub->pid_ch_no);
    if(ret != FAPI_OK) {
        return ret;
    }
#endif
   
    ret = desc_funcStr.Start(pfunc_sub->pDESCCtxHdl);
    if(ret != FAPI_OK) {
        FAPEXDebugPrint(2, "[fdltsd_SetDescramblKeys] FAPI_DESC_Start error:%d\n", ret);
        return ret;
    }
    
    if(params.systemKeyLength > 0) {
        ret = desc_funcStr.SetSystemKey(pfunc_sub->pDESCCtxHdl,
                                        params.systemKey,
                                        params.systemKeyLength);
         if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdltsd_SetDescramblKeys] FAPI_DESC_SetSystemKey error:%d\n", ret);
            return ret;
         }
    }
    
    /* set key parameters */
    if(params.inVectorKeyLength == 0)
        ret =  desc_funcStr.SetKeys(
                pfunc_sub->pDESCCtxHdl,
                params.evenKey, params.evenKeyLength,
                params.oddKey,  params.oddKeyLength,
                NULL ,0);
    else  
        ret = desc_funcStr.SetKeys(
                pfunc_sub->pDESCCtxHdl,
                params.evenKey, params.evenKeyLength,
                params.oddKey,  params.oddKeyLength,
                params.inVectorKey,params.inVectorKeyLength);
    if(ret != FAPI_OK) {
      FAPEXDebugPrint(2, "[fdltsd_SetDescramblKeys] FAPI_DESC_SetKeys error:%d\n", ret);
    }
    return ret;
}
#endif

/*!
*******************************************************************************
**
** \brief  set Buffer Read Size paramter
**
** \param  pfunc_sub    pointer to sub context
** \param  read_size    read size
**
******************************************************************************/
static int32_t fdltsd_SetBufferReadSize(FAPEXFDL_TSD_SubStrT *pfunc_sub, uint32_t read_size)
{
    uint32_t    read_position;
    int32_t     ret = 0;

    if(pfunc_sub->pBMCtxHdl == NULL) {
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }

    if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_OUT_BOUND) {
        /* no management */
        return ret;
    }

    /* update Buffer Read pointer */
    if( (pfunc_sub->stream_type == FAPEX_TSD_STREAMTYPE_TELETEXT)
     && (pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_HCB) ) {
        read_position = pfunc_sub->read_pt + read_size;
        if(read_position >= pfunc_sub->buff_size) {
            read_position -= pfunc_sub->buff_size;
        }
        pfunc_sub->read_pt = read_position;
    }
    else {
        /* get Buffer Read position */
        read_position = bm_funcStr.GetBufferReadPointerPosition(pfunc_sub->pBMCtxHdl);

        /* update Buffer Read position */
        read_position += read_size;
        if(read_position >= pfunc_sub->buff_size) {
            read_position -= pfunc_sub->buff_size;
        }
        ret = bm_funcStr.SetBufferReadPointer(pfunc_sub->pBMCtxHdl, read_position, 0);

        if(ret == FAPI_OK) {
            if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_ALL) {
                fdltsd_UpdateSctInfo(pfunc_sub, read_size);
            }
        }
        else {
            FAPEXDebugPrint(2, "[fdltsd_SetBufferReadSize] "
                               "FAPI_BM_SetBufferReadPointer error:%d\n", ret);
        }
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  set Buffer Write Size paramter
**
** \param  pfunc_sub    pointer to sub context
** \param  read_size    read size
**
******************************************************************************/
static int32_t fdltsd_SetBufferWritePointer(FAPEXFDL_TSD_SubStrT *pfunc_sub, uint32_t write_ptr)
{
    int32_t     ret = 0;

    if(pfunc_sub->pBMCtxHdl == NULL) {
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }

    if(pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_OUT_BOUND) {
        /* no management */
        return ret;
    }

    if(write_ptr >= pfunc_sub->buff_size)
        return FAPEX_TSD_ERR_BAD_PARAMETER;
               
    ret = bm_funcStr.SetBufferWritePointer(pfunc_sub->pBMCtxHdl, write_ptr, 0);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  get Buffer Read Information paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_ReadInfoStrT parameter
**
******************************************************************************/
static int32_t fdltsd_GetBufferReadInfo(FAPEXFDL_TSD_CmnStrT *pfunc_cmn,
                                        FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                        FAPEX_TSD_ReadInfoStrT *pArg)
{
    FAPEX_TSD_ReadInfoStrT  params;
    FAPEXCFDLSAF_EventStrT  *dt_event_ptr;
    FAPEXFDL_TSD_SctInfoT   *sct_data;
    uint32_t  wait_enable;
    uint32_t  wait_time;
    uint32_t  buff_status = FAPEX_TSD_BUFFSTAT_NONE;
    uint32_t  r_index, w_index;
    uint32_t  ntf_type;
    int32_t   ret;
    int32_t   result = 0;

    dt_event_ptr = &pfunc_sub->dt_event;

    if(pfunc_sub->pBMCtxHdl == NULL) {
        ret = FAPEX_TSD_ERR_BAD_PARAMETER;
        goto fdltsd_GetBufferReadInfo_ERR;
    }

    wait_enable = pfunc_sub->flags & FAPEXFDL_TSD_CHFLAGS_NOTIFIER_DATA;

    if(wait_enable != 0) {
        /* check event */
        dt_event_ptr->mode = FAPEXCFDLSAF_WAITEVENT_QUEUE;
#if 0
        if(pfunc_sub->status == FAPEX_STATUS_RUNNING) {
            wait_time  = (pfunc_sub->timeout != 0) ?
                         pfunc_sub->timeout : pfunc_cmn->timeout;
            wait_time += FAPEXFDL_TSD_TIMEOUT_INTERRUPT;
        }
        else {
            wait_time = 1;
        }
#else
        wait_time = 1;
#endif
        dt_event_ptr->tout = wait_time;

        result = FAPEXCFDLSAF_WAIT_EVENT(dt_event_ptr);
#if 0
        if(result == FAPI_ERR_TIMEOUT) {
            if(pfunc_sub->status == FAPEX_STATUS_RUNNING) {
                buff_status = FAPEX_TSD_BUFFSTAT_TIMEOUT;
                ret = FAPEX_TSD_ERR_TIMEOUT;
                FAPEXDebugPrint(2, "[fdltsd_GetBufferReadInfo] "
                                   "timeout error (%d) [%d]\n",
                                   wait_time, pfunc_sub->id);
                goto fdltsd_GetBufferReadInfo_ERR;
            }
        }
#endif
        /* get Buffer fill size */
        ret = fdltsd_GetBufferFillLevel(pfunc_sub);
    }
    else {
        uint32_t  tout, sleep_time;
        uint64_t  end_time;

        /* get Buffer fill size */
        ret = fdltsd_GetBufferFillLevel(pfunc_sub);
        if(ret == 0) {
            tout  = (pfunc_sub->timeout != 0) ? pfunc_sub->timeout : pfunc_cmn->timeout;
            tout += FAPEXFDL_TSD_TIMEOUT_INTERRUPT;

            sleep_time = tout;
            while(sleep_time > 10) {
                sleep_time >>= 1;
            }
            end_time = timer_funcStr.GetTimeStamp(FAPI_TIMER_RESOLUTION_1_MSEC) + tout;

            dt_event_ptr->mode = FAPEXCFDLSAF_WAITEVENT_QUEUE | FAPEXCFDLSAF_WAITEVENT_ASYNC;
            dt_event_ptr->tout = 0;

            while(pfunc_sub->status == FAPEX_STATUS_RUNNING) {
                FAPI_SYS_SLEEP(sleep_time);

                /* check error status with asynchronous mode */
                result = FAPEXCFDLSAF_WAIT_EVENT(dt_event_ptr);
                if(result == 0) {
                    buff_status = dt_event_ptr->evt_bitptn
                                & (FAPEXFDL_TSD_EVENT_WARNING | FAPEXFDL_TSD_EVENT_ERROR);
                    break;
                }
                result = 0;

                /* get Buffer fill size */
                ret = fdltsd_GetBufferFillLevel(pfunc_sub);
                if(ret != 0) {
                    break;
                }

                /* check timeout */
                if(timer_funcStr.GetTimeStamp(FAPI_TIMER_RESOLUTION_1_MSEC) >= end_time) {
                    buff_status = FAPEX_TSD_BUFFSTAT_TIMEOUT;
                    ret = FAPEX_TSD_ERR_TIMEOUT;
                    break;
                }
            }
        }
    }
    if(ret < 0) {
        goto fdltsd_GetBufferReadInfo_ERR;
    }
    params.fillSize = ret;

    /* get Buffer Read pointer offset */
    if(!(pfunc_sub->flags & (FAPEXFDL_TSD_CHFLAGS_OUT_BOUND | FAPEXFDL_TSD_CHFLAGS_HCB))) {
        params.readOffset = bm_funcStr.GetBufferReadPointerPosition(pfunc_sub->pBMCtxHdl);
    }
    else {
        params.readOffset = pfunc_sub->read_pt;
    }
    FAPEXDebugPrint(2, "## [fdltsd_GetBufferReadInfo] params.readOffset = %xh\n",params.readOffset);
    if((wait_enable != 0) && (result == 0)) {
        r_index = dt_event_ptr->read_idx & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;

        ntf_type = pfunc_sub->ntfyFifo[r_index].notifier_type;
        params.bufferStatus = ntf_type & FAPEXFDL_TSD_EVENT_ALL;

        sct_data = (FAPEXFDL_TSD_SctInfoT*)pfunc_sub->ntfyFifo[r_index].pData;

        if(sct_data != NULL) {
            /* get offset/length/number of current section */
            params.sectionOffset = sct_data->sct_base;
            params.sectionLength = sct_data->sct_length;
            params.sectionNo     = ntf_type >> FAPEXFDL_TSD_EVENT_SECTION_NO_SHFT;
        }
        else {
            params.sectionOffset = 0;
            params.sectionLength = 0;
            params.sectionNo     = -1;

            if(r_index == pfunc_sub->ready_idx) {
                /* skip ready index */
                pfunc_sub->ready_idx = (pfunc_sub->ready_idx + 1) & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;
            }
        }
        /* update the read index */
        FAPEXCFDLSAF_SET_EVENT(dt_event_ptr, 0,
                               FAPEXCFDLSAF_WAITEVENT_QUEUE_READ);
        r_index = (r_index + 1) & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;

        /* check next event */
        if(pfunc_sub->ntfyFifo[r_index].pData == NULL) {
            w_index = dt_event_ptr->write_idx & FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK;
            if(r_index != w_index) {
                ntf_type = pfunc_sub->ntfyFifo[r_index].notifier_type;
                params.bufferStatus |= ntf_type & FAPEXFDL_TSD_EVENT_ALL;

                /* update the read index */
                FAPEXCFDLSAF_SET_EVENT(dt_event_ptr, 0,
                                       FAPEXCFDLSAF_WAITEVENT_QUEUE_READ);
            }
        }
    }
    else {
        params.bufferStatus  = buff_status;
        params.sectionOffset = 0;
        params.sectionLength = 0;
        params.sectionNo     = -1;
    }

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_TSD_ReadInfoStrT));

    return 0;

/* for error */
fdltsd_GetBufferReadInfo_ERR:
    params.fillSize      = 0;
    params.readOffset    = 0;
    params.sectionOffset = 0;
    params.sectionLength = 0;
    params.sectionNo     = -1;
    params.bufferStatus  = buff_status;

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_TSD_ReadInfoStrT));

    return ret;
}

/*!
*******************************************************************************
**
** \brief  get Buffer Write Information paramter
**
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_WriteInfoStrT parameter
**
******************************************************************************/
static int32_t fdltsd_GetBufferWriteInfo(FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                         FAPEX_TSD_WriteInfoStrT *pArg)
{
    FAPEX_TSD_WriteInfoStrT  params;
    FAPEXCFDLSAF_EventStrT   *dt_event_ptr;
    int32_t  ret;

    params.bufferStatus = FAPEX_TSD_BUFFSTAT_NONE;

    if(pfunc_sub->pBMCtxHdl == NULL) {
        ret = FAPEX_TSD_ERR_BAD_PARAMETER;
        goto fdltsd_GetBufferWriteInfo_ERR;
    }

    /* check event */
    dt_event_ptr = &pfunc_sub->dt_event;

    params.bufferStatus |= dt_event_ptr->evt_bitptn
                         & (FAPEXFDL_TSD_EVENT_WARNING | FAPEXFDL_TSD_EVENT_ERROR);

    /* get Buffer fill size */
    ret = fdltsd_GetBufferFillLevel(pfunc_sub);
    if(ret < 0) {
        goto fdltsd_GetBufferWriteInfo_ERR;
    }
    params.fillSize = ret;

    /* get Buffer Write pointer offset */
    params.writeOffset = bm_funcStr.GetBufferWritePointerPosition(pfunc_sub->pBMCtxHdl);

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_TSD_WriteInfoStrT));

    return 0;

/* for error */
fdltsd_GetBufferWriteInfo_ERR:
    params.fillSize    = 0;
    params.writeOffset = 0;

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_TSD_WriteInfoStrT));

    return ret;
}

/*!
*******************************************************************************
**
** \brief  get TSD block information paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pArg         pointer to FAPEX_TSD_BlockInfoStrT parameter
**
******************************************************************************/
static int32_t fdltsd_GetBlockInfo(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEX_TSD_BlockInfoStrT *pArg)
{
    uint32_t    count, tmp_flag, enable_flag, running_flag;

    /*---------- LOCK ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

    /* count the enable channels */
    enable_flag  = pfunc_cmn->enable_channel;
    running_flag = pfunc_cmn->running_channel;

    /*--------- UNLOCK ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

    count    = 0;
    tmp_flag = enable_flag;
    while(tmp_flag != 0) {
        if((tmp_flag & 0x00000001)) {
            count ++;
        }
        tmp_flag >>= 1;
    }

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->usedChannelNum,     &count,                 sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->enableChannelFlag,  &enable_flag,           sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->runningChannelFlag, &running_flag,          sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->blockParams,        &pfunc_cmn->blk_params, sizeof(FAPEX_TSD_InputParamStrT));

    return 0;
}

/*!
*******************************************************************************
**
** \brief  get TSD STC channel flag
**
** \param  pfunc_cmn    pointer to common context
**
******************************************************************************/
static uint32_t fdltsd_GetStcChannelFlag(FAPEXFDL_TSD_CmnStrT *pfunc_cmn)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn0, *pfunc_stc;
    uint32_t  stc_channel;

    pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;
    pfunc_stc  = pfunc_cmn - (pfunc_cmn->block_no & 0x01);

    /*---------- LOCK (stc_channel) ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

    if(pfunc_cmn == pfunc_stc) {
        stc_channel = (uint32_t)(pfunc_stc->stc_channel & pfunc_cmn->stc_channel_mask);
    }
    else {
        stc_channel = (uint32_t)(pfunc_stc->stc_channel >> FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM);
    }

    /*--------- UNLOCK (stc_channel) ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);

    return stc_channel;
}

/*!
*******************************************************************************
**
** \brief  get TSD channel information paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_ChannelInfoStrT parameter
**
******************************************************************************/
static int32_t fdltsd_GetChannelInfo(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                     FAPEX_TSD_ChannelInfoStrT *pArg)
{
    FAPEX_TSD_ChannelInfoStrT  params;
    FAPI_BM_InfoT              bm_info;
    int32_t  pid;
    int32_t  ret;

    params.channelNo = pfunc_sub->pid_ch_no;

    if(pfunc_sub->pTSDCtxHdl == NULL) {
        uint32_t  stc_channel;

        stc_channel = fdltsd_GetStcChannelFlag(pfunc_cmn);

        /* check PCR only channel */
        if(stc_channel & (1 << pfunc_sub->pid_ch_no)) {
            params.pid = sync_funcStr.GetPcrPid();
            ret = 0;
            goto fdltsd_GetChannelInfo_ERR2;
        }
        else {
            ret = FAPEX_TSD_ERR_INVALID_STATUS;
            goto fdltsd_GetChannelInfo_ERR0;
        }
    }

    /* get PID value */
    pid = tsd_funcStr.GetChannelPid(pfunc_sub->pTSDCtxHdl);
    if(pid < 0) {
        ret = pid;
        FAPEXDebugPrint(2, "[fdltsd_GetChannelInfo] "
                           "FAPI_TSD_GetChannelPid error:%d\n", ret);
        goto fdltsd_GetChannelInfo_ERR1;
    }
    params.pid = pid;

    /* get Buffer information value */
    if(pfunc_sub->pBMCtxHdl) {
        ret = bm_funcStr.GetInfo(pfunc_sub->pBMCtxHdl, &bm_info);
        if(ret != FAPI_OK) {
            goto fdltsd_GetChannelInfo_ERR2;
        }
        params.bufferSize = bm_info.size;
    }
    else {
        params.bufferSize = 0;
    }

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_TSD_ChannelInfoStrT));

    return 0;

/* for error */
fdltsd_GetChannelInfo_ERR0:
    params.channelNo  = 0;

fdltsd_GetChannelInfo_ERR1:
    params.pid        = (uint32_t)-1;

fdltsd_GetChannelInfo_ERR2:
    params.bufferSize = 0;

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_TSD_ChannelInfoStrT));

    return ret;
}

/*!
*******************************************************************************
**
** \brief  get TSD block/channel status information paramter
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  pArg         pointer to FAPEX_TSD_StatusStrT parameter
**
******************************************************************************/
static int32_t fdltsd_GetStatus(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, FAPEXFDL_TSD_SubStrT *pfunc_sub,
                                FAPEX_TSD_StatusStrT *pArg)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn0, *pfunc_stc;
    FAPEX_TSD_StatusStrT params;
    uint32_t  stc_channel;

    params.syncStatus = FAPI_SYNC_INITIAL_STATUS;

    /* check TSD status */
    if(pfunc_sub != NULL) {
        params.status = pfunc_sub->status;
    }
    else {
        params.status = pfunc_cmn->status;
    }

    /* check STC status */
    pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;
    pfunc_stc  = pfunc_cmn - (pfunc_cmn->block_no & 0x01);

    if(pfunc_cmn0->stc_block & (0x03 << (pfunc_cmn->block_no & 0xFFFFFFFE))) {
        /* TSD block uses the STC */
        if(pfunc_sub == NULL) {
            params.syncStatus = sync_funcStr.GetStatus();
        }
        else {
            /*---------- LOCK (stc_channel) ----------*/
            FAPI_SYS_GET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_SUSPEND);

            if(pfunc_cmn == pfunc_stc) {
                stc_channel = (uint32_t)(pfunc_stc->stc_channel & pfunc_cmn->stc_channel_mask);
            }
            else {
                stc_channel = (uint32_t)(pfunc_stc->stc_channel >> FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM);
            }

            /*--------- UNLOCK (stc_channel) ---------*/
            FAPI_SYS_SET_SEMAPHORE(pfunc_cmn0->cmn_sem, FAPI_SYS_NO_SUSPEND);

            if(stc_channel & (1 << pfunc_sub->pid_ch_no)) {
                params.syncStatus = sync_funcStr.GetStatus();
            }
        }
    }

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_TSD_StatusStrT));

    return 0;
}

/*!
*******************************************************************************
**
** \brief  get STC value (unit of 90kHz)
**
** \param  pfunc_cmn    pointer to common context
** \param  pArg         pointer to STC value
**
******************************************************************************/
static int32_t fdltsd_GetStc(FAPEXFDL_TSD_CmnStrT *pfunc_cmn, uint64_t *pArg)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn0;
    int64_t  stc_value;

    /* check STC block */
    pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;

    if(pfunc_cmn0->stc_block & (0x03 << (pfunc_cmn->block_no & 0xFFFFFFFE))) {
        /* TSD block uses the STC */
        sync_funcStr.GetStc((uint64_t *)&stc_value);

        stc_value += sync_funcStr.GetDecoderOffset();
        if(stc_value < 0) {
            stc_value += FAPI_SYNC_STC_MAX;
        }
        else if(stc_value >= FAPI_SYNC_STC_MAX) {
            stc_value -= FAPI_SYNC_STC_MAX;
        }
        stc_value = FAPI_DIV64(stc_value, 300);
    }
    else {
        /* TSD block doesn't use the STC */
        stc_value = -1;
    }

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &stc_value, sizeof(uint64_t));

    return 0;
}

/*!
*******************************************************************************
**
** \brief  get TS packet input control mapping information
**
** \param  pfunc_cmn    pointer to common context
** \param  pMapInfo     pointer to FAPEX_TSD_PacketInMapInfoStrT parameter
**
******************************************************************************/
static int32_t fdltsd_GetPacketInMapInfo(FAPEXFDL_TSD_CmnStrT *pfunc_cmn,
                                         FAPEX_TSD_PacketInMapInfoStrT *pArg)
{
    FAPEX_TSD_PacketInMapInfoStrT  map_info;

    map_info.mapSize = pfunc_cmn->pkt_in_area_size;
#ifndef __LINUX__
    map_info.mapOffset = 0;
#else
    map_info.mapOffset = FAPI_SYS_GET_PHYSICAL_ADDRESS(pfunc_cmn->pkt_in_address) & ~PAGE_MASK;
#endif
    map_info.packetOffset = pfunc_cmn->pkt_in_offset;

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &map_info, sizeof(FAPEX_TSD_PacketInMapInfoStrT));

    return 0;
}

/*!
*******************************************************************************
**
** \brief  get TS packet input information parameters
**
** \param  pfunc_cmn    pointer to common context
** \param  pPktInfo     pointer to FAPEX_TSD_PacketInInfoStrT parameter
**
******************************************************************************/
static int32_t fdltsd_SetPacketInInfo(FAPEXFDL_TSD_CmnStrT *pfunc_cmn,
                                      FAPEX_TSD_PacketInInfoStrT *pArg)
{
    uint32_t  valid_flag, *packet_ctrl;

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&valid_flag, &pArg->inputStatus, sizeof(uint32_t));

    if((valid_flag & FAPEXFDL_TSD_PACKET_IN_INVALID_MASK)) {
        return FAPEX_TSD_ERR_BAD_PARAMETER;
    }
    packet_ctrl  = (uint32_t*)pfunc_cmn->pkt_ctrl_address;
    *packet_ctrl = valid_flag;

    return 0;
}

/*!
*******************************************************************************
**
** \brief  get TS packet input information parameters
**
** \param  pfunc_cmn    pointer to common context
** \param  pPktInfo     pointer to FAPEX_TSD_PacketInInfoStrT parameter
**
******************************************************************************/
static int32_t fdltsd_GetPacketInInfo(FAPEXFDL_TSD_CmnStrT *pfunc_cmn,
                                      FAPEX_TSD_PacketInInfoStrT *pArg)
{
    FAPEX_TSD_PacketInInfoStrT  pkt_info;
    uint32_t  *packet_ctrl;

    packet_ctrl = (uint32_t*)pfunc_cmn->pkt_ctrl_address;

    pkt_info.inputStatus  = *packet_ctrl & FAPEXFDL_TSD_PACKET_IN_BUSY_MASK;
    pkt_info.bufferStatus = pfunc_cmn->pkt_in_status;

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &pkt_info, sizeof(FAPEX_TSD_PacketInInfoStrT));

    return 0;
}

/*!
*******************************************************************************
**
** \brief  resets the STC regulation.
**
** \param  pfunc_cmn    pointer to common context
**
******************************************************************************/
static int32_t fdltsd_ResetStcRegulation(FAPEXFDL_TSD_CmnStrT *pfunc_cmn)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn0;

    /* check STC block */
    pfunc_cmn0 = pfunc_cmn - pfunc_cmn->block_no;

    if(pfunc_cmn0->stc_block & (0x03 << (pfunc_cmn->block_no & 0xFFFFFFFE))) {
        /* TSD block uses the STC */
        sync_funcStr.ResetRegulation();
        return 0;
    }
    else {
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }
}

static int32_t fdltsd_BmWrite(FAPEXFDL_TSD_SubStrT *pfunc_sub, FAPEX_TSD_BmWriteParmStrT *pArg)
{
    FAPEX_TSD_BmWriteParmStrT   bmParams;
    uint8_t                     *buffer = NULL;
    int32_t                     ret = 0;

    /* check param */
    if(NULL == pfunc_sub)
    {
        return -1;
    }

    if(NULL == pfunc_sub->pBMCtxHdl)
    {
        return -1;
    }

    memset(&bmParams, 0, sizeof(FAPEX_TSD_BmWriteParmStrT));
    
    FAPEXCFDLSAF_COPY_TO_DRV(&bmParams, pArg, sizeof(FAPEX_TSD_BmWriteParmStrT));
    
    buffer = FAPI_SYS_MALLOC_UNCACHED(sizeof(uint8_t)*(bmParams.bufferLen));
    if(NULL == buffer)
    {
        FAPEXDebugPrint(2,"[fdltsd_BmWrite] ERROR:no memory\n");
        return -1;
    }
    memset(buffer, 0, sizeof(uint8_t)*(bmParams.bufferLen));

    FAPEXCFDLSAF_COPY_TO_DRV(buffer, (uint8_t *)pArg->bufferPtr, sizeof(uint8_t)*(bmParams.bufferLen));

    ret = bm_funcStr.Write(
                        pfunc_sub->pBMCtxHdl,
                        (void*)buffer,
                        (uint32_t)bmParams.bufferLen,
                        bmParams.timeout);
    
    if(ret != bmParams.bufferLen)
    {
        FAPEXDebugPrint(2,"[fdltsd_BmWrite] ERROR: BM write error(%d)\n",ret);
    }

    if(NULL != buffer)
    {
        FAPI_SYS_FREE(buffer);
    }

    FAPEXDebugPrint(1,"[fdltsd_BmWrite] handle:[%d], ret:[%d]\n",pfunc_sub->pBMCtxHdl,ret);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLTSD_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                   FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn;
    FAPEXFDL_TSD_SubStrT *pfunc_sub;
    uint32_t id;
    int32_t  ret = FAPEX_TSD_ERR_BAD_PARAMETER;

    id = FAPEXCFDLSAF_GETID(inode);

    FAPEXDebugPrint(3, "[FAPEXFDLTSD_ioctl] %3d cmd:%08x\n", id, cmd);

    /* Get the context base of this id's FDL. */
    if(id < FAPEXDEV_TSD_SUB_MINOR_FIRST) {
        pfunc_cmn = (FAPEXFDL_TSD_CmnStrT *)(file->private_data);
        pfunc_sub = NULL;
    }
    else {
        pfunc_sub = (FAPEXFDL_TSD_SubStrT *)(file->private_data);
        if(pfunc_sub != NULL) {
            pfunc_cmn = (FAPEXFDL_TSD_CmnStrT *)pfunc_sub->pCmnHdl;
        }
        else {
            pfunc_cmn = NULL;
        }
    }

    switch(cmd){
        /* actual code which is depend on function is implimented here. */
        /* case ...: */
#ifdef FAPI_TSD_CONNECT
        case FAPEX_IOCCOM_START:
            ret = fdltsd_Start(pfunc_cmn, pfunc_sub);
            break;

        case FAPEX_IOCCOM_STOP:
            ret = fdltsd_Stop(pfunc_cmn, pfunc_sub);
            break;

        case FAPEX_IOCCOM_TSD_OPEN_SUBID:
            if(pfunc_sub != NULL) {
                ret = fdltsd_OpenSub(pfunc_cmn, pfunc_sub, id,
                                     (FAPEX_TSD_OpenSubIdParamStrT*)arg);
            }
            break;

        case FAPEX_IOCCOM_CLOSE_SUBID:
            if((pfunc_sub != NULL) && (pfunc_cmn != NULL)) {
                uint32_t  stc_channel;

                stc_channel = fdltsd_GetStcChannelFlag(pfunc_cmn);

                ret = fdltsd_CloseSub(pfunc_cmn, pfunc_sub, stc_channel);
            }
            break;

        case FAPEX_IOCCOM_TSD_SET_PID_FILTER:
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_SetPidFilter(pfunc_cmn, pfunc_sub,
                                          (FAPEX_TSD_PidParamStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_SET_SECTION_FILTER:
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_SetSectionFilter(pfunc_cmn, pfunc_sub,
                                              (FAPEX_TSD_SctFltParamStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_SET_PES_FILTER:
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_SetPesFilter(pfunc_cmn, pfunc_sub,
                                          (FAPEX_TSD_PesParamStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_SET_KEYS:
#ifdef CONFIG_TSD_USE_INTERNAL_DESC
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_SetDescramblKeys(pfunc_cmn, pfunc_sub,
                                              (FAPEX_TSD_DescKeysParamStrT *)arg);
            }
#endif
            break;

        case FAPEX_IOCCOM_TSD_SET_BUFFER_READSIZE:
            if(pfunc_sub != NULL) {
                ret = fdltsd_SetBufferReadSize(pfunc_sub, (uint32_t)arg);
            }
            break;
           
        //added for bugzilla #804
        case FAPEX_IOCCOM_TSD_SET_BUFFER_WRITEPOINTEROFFSET:
           if(pfunc_sub != NULL) {
                ret = fdltsd_SetBufferWritePointer(pfunc_sub, (uint32_t)arg);
            }
            break; 	
        
        case FAPEX_IOCCOM_TSD_GET_BUFFER_READINFO:
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_GetBufferReadInfo(pfunc_cmn, pfunc_sub,
                                               (FAPEX_TSD_ReadInfoStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_GET_BUFFER_WRITEINFO:
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_GetBufferWriteInfo(pfunc_sub,
                                                (FAPEX_TSD_WriteInfoStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_GET_BUFFER_LEVEL:
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_GetBufferFillLevel(pfunc_sub);
                if(ret >= 0) {
                    /* copy parameters from the kernel space to an user space */
                    FAPEXCFDLSAF_COPY_TO_USR((uint32_t*)arg, &ret, sizeof(uint32_t));
                    ret = 0;
                }
            }
            break;

        case FAPEX_IOCCOM_TSD_SET_PID_VALUE:
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_SetPidValue(pfunc_cmn, pfunc_sub, (FAPEX_TSD_PidParamStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_OPEN:
            if( (pfunc_cmn != NULL) && (pfunc_sub == NULL) ) {
                ret = fdltsd_OpenCmn(pfunc_cmn, (FAPEX_TSD_OpenParamStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_SET_INPUT:
            if( (pfunc_cmn != NULL) && (pfunc_sub == NULL) ) {
                ret = fdltsd_SetInput(pfunc_cmn, (FAPEX_TSD_InputParamStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_GET_BLOCK_INFO:
            if((pfunc_cmn != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL)) {
                ret = fdltsd_GetBlockInfo(pfunc_cmn, (FAPEX_TSD_BlockInfoStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_GET_CHANNEL_INFO:
            if((pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL)) {
                ret = fdltsd_GetChannelInfo(pfunc_cmn, pfunc_sub, (FAPEX_TSD_ChannelInfoStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_GET_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdltsd_GetStatus(pfunc_cmn, pfunc_sub, (FAPEX_TSD_StatusStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_GET_STC_VALUE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdltsd_GetStc(pfunc_cmn, (uint64_t *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_GET_INPUT_MAP_INFO:
            if( (pfunc_sub == NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_GetPacketInMapInfo(pfunc_cmn, (FAPEX_TSD_PacketInMapInfoStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_SET_INPUT_PACKETINFO:
            if( (pfunc_sub == NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_SetPacketInInfo(pfunc_cmn, (FAPEX_TSD_PacketInInfoStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_TSD_GET_INPUT_PACKETINFO:
            if( (pfunc_sub == NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_GetPacketInInfo(pfunc_cmn, (FAPEX_TSD_PacketInInfoStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_RESET:
            if(pfunc_sub == NULL) {
                ret = fdltsd_ResetStcRegulation(pfunc_cmn);
            }
            break;
          
        case FAPEX_IOCCOM_TSD_BM_WRITE:
            if( (pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL) ) {
                ret = fdltsd_BmWrite(pfunc_sub, (FAPEX_TSD_BmWriteParmStrT *)arg);
            }
            break;
            
#endif /* FAPI_TSD_CONNECT */
        default:
            ;
    }

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLTSD_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T        mask = 0;
    FAPEXFDL_TSD_SubStrT   *pfunc_sub;
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;

    /* Get the context base of this id's FDL. */
    pfunc_sub = (FAPEXFDL_TSD_SubStrT*)(file->private_data);
    dt_event_ptr = &pfunc_sub->dt_event;

    FAPEXDebugPrint(4, "[FAPEXFDLTSD_poll]\n");

    FAPEXCFDLSAF_POLL_EVENT(file, dt_event_ptr, NULL, ptbl);

    /* actual code which is depend on function is implimented here. */

    if(dt_event_ptr->read_idx != dt_event_ptr->write_idx) {
        mask |= POLLIN | POLLRDNORM;
    }
    if(dt_event_ptr->evt_bitptn & FAPEXFDL_TSD_EVENT_MESSAGE) {
        mask |= POLLMSG;
    }
    if(dt_event_ptr->evt_bitptn & FAPEXFDL_TSD_EVENT_WARNING) {
        mask |= POLLPRI;
    }
    if(dt_event_ptr->evt_bitptn & FAPEXFDL_TSD_EVENT_ERROR) {
        mask |= POLLERR;
    }
    return mask;
}

/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLTSD_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    FAPEXFDL_TSD_SubStrT *pfunc_sub;
    int32_t  ret;

    /* Get the context base of this id's FDL. */
    pfunc_sub = (FAPEXFDL_TSD_SubStrT *)(file->private_data);

    if(pfunc_sub == NULL) {
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }
    if(pfunc_sub->id < FAPEXDEV_TSD_SUB_MINOR_FIRST) {
        return FAPEX_TSD_ERR_BAD_PARAMETER;
    }

    if( (vma->vm_pgoff != 0) || ((vma->vm_end - vma->vm_start) > pfunc_sub->buff_size) ) {
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if(file->f_flags & O_SYNC) {
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma,
                          vma->vm_start,
                          pfunc_sub->buff_address >> PAGE_SHIFT,    /* physical page */
                          vma->vm_end - vma->vm_start,
                          vma->vm_page_prot);
    if(ret != 0) {
        return -EAGAIN;
    }
    return 0;
#else
    FAPEXFDL_TSD_SubStrT *pfunc_sub;

    /* Get the context base of this id's FDL. */
    pfunc_sub = (FAPEXFDL_TSD_SubStrT *)(file->private_data);

    if( (vma->vm_length > pfunc_sub->buff_size)
     || (vma->vm_offset != 0) ) {
        return FAPEX_TSD_ERR_BAD_PARAMETER;
    }
    vma->vm_start = (void *)pfunc_sub->buff_address;

    return 0;
#endif /* __LINUX__ */
}

/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLTSD_mmap_input(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn;
#ifdef __LINUX__
    int32_t   ret;
    uint32_t  input_address;

    FAPEXDebugPrint(3, "[FAPEXFDLTSD_mmap_input]\n");

    /* Get the context base of this id's FDL. */
    pfunc_cmn = (FAPEXFDL_TSD_CmnStrT *)(file->private_data);

    if(pfunc_cmn == NULL) {
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }

    if(pfunc_cmn->pkt_in_address == 0) {
        /* cannot mmap */
        return FAPEX_TSD_ERR_INVALID_STATUS;
    }

    if( (vma->vm_pgoff != 0) || ((vma->vm_end - vma->vm_start) > pfunc_cmn->pkt_in_area_size) ) {
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if(file->f_flags & O_SYNC) {
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    input_address = FAPI_SYS_GET_PHYSICAL_ADDRESS(pfunc_cmn->pkt_in_address);

    ret = remap_pfn_range(vma,
                          vma->vm_start,
                          input_address >> PAGE_SHIFT,  /* physical page */
                          vma->vm_end - vma->vm_start,
                          vma->vm_page_prot);
    if(ret != 0) {
        return -EAGAIN;
    }
    return 0;
#else
    /* Get the context base of this id's FDL. */
    pfunc_cmn = (FAPEXFDL_TSD_CmnStrT *)(file->private_data);

    if( (vma->vm_length > pfunc_cmn->pkt_in_area_size)
     || (vma->vm_offset != 0) ) {
        return FAPEX_TSD_ERR_BAD_PARAMETER;
    }
    vma->vm_start = (void *)pfunc_cmn->pkt_in_address;

    return 0;
#endif /* __LINUX__ */
}

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLTSD_open(FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn;
    void    *pCfdlHdl;
    int32_t id;

    FAPEXDebugPrint(3, "[FAPEXFDLTSD_open] file:0x%08x,private_data:0x%08x\n",
                       (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLTSD_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLTSD_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc_cmn = (FAPEXFDL_TSD_CmnStrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc_cmn;
        FAPEXDebugPrint(3, "[FAPEXFDLTSD_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc_cmn));

        /* set parameter */
        pfunc_cmn->pCfdlHdl = pCfdlHdl;
        pfunc_cmn->id       = id;
        pfunc_cmn->status   = FAPEX_STATUS_OPENED;
#ifdef __LINUX__
        kref_init(&pfunc_cmn->kref);
#endif
    }
    else {
        pfunc_cmn = (FAPEXFDL_TSD_CmnStrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc_cmn->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc_cmn->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLTSD_open] id [%d], used_count %d.\n",
                       pfunc_cmn->id, pfunc_cmn->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdltsd_delete(struct kref *kref)
{
    FAPEXFDL_TSD_CmnStrT  *pfunc_cmn;

    pfunc_cmn = container_of(kref, FAPEXFDL_TSD_CmnStrT, kref);
#else
static void fdltsd_delete(FAPEXFDL_TSD_CmnStrT *pfunc_cmn)
{
#endif

    pfunc_cmn->status = FAPEX_STATUS_TERMINATED;
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLTSD_release(FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn;

    pfunc_cmn = (FAPEXFDL_TSD_CmnStrT*)(file->private_data);

    if(pfunc_cmn != NULL) {
        if(pfunc_cmn->status > FAPEX_STATUS_OPENED) {
            FAPEXFDL_RETVAL( FAPEX_TSD_ERR_INVALID_STATUS );
        }

        /* decrement the count on our device */
        pfunc_cmn->used_count--;
#ifdef __LINUX__
        kref_put(&pfunc_cmn->kref, fdltsd_delete);
#else
        fdltsd_delete(pfunc_cmn);
#endif

        FAPEXDebugPrint(3, "[FAPEXFDLTSD_release] id [%d], used_count %d\n",
                           pfunc_cmn->id, pfunc_cmn->used_count);
    }
    return 0;
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdltsd_cmn_fops;
struct file_operations fapex_fdltsd_sub_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLTSD_init(void)
{
    int32_t   ret;
    int32_t   count_cmn, count_sub, i, j;
    void      *pCfdlHdl;
    uint32_t  size_0, size_1;
    FAPEXFDL_TSD_CmnStrT    *pfunc_cmn, *pfunc_cmn0;
    FAPEXFDL_TSD_SubStrT    *pfunc_sub;
    FAPEXCFDL_UtilStrT      utility;
    uint32_t                buff_align_size;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_TSD_SubStrT    *pfunc_sub0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLTSD_init]\n");

    /* set struct file_operations */
    fapex_fdltsd_cmn_fops.owner   = THIS_MODULE;
    fapex_fdltsd_cmn_fops.open    = FAPEXFDLTSD_open;
    fapex_fdltsd_cmn_fops.write   = FAPEXFDLTSD_write_input;
    fapex_fdltsd_cmn_fops.ioctl   = FAPEXFDLTSD_ioctl;
    fapex_fdltsd_cmn_fops.mmap    = FAPEXFDLTSD_mmap_input;
    fapex_fdltsd_cmn_fops.release = FAPEXFDLTSD_release;

    fapex_fdltsd_sub_fops.owner   = THIS_MODULE;
    fapex_fdltsd_sub_fops.open    = FAPEXFDLTSD_open;
    fapex_fdltsd_sub_fops.read    = FAPEXFDLTSD_read;
    fapex_fdltsd_sub_fops.ioctl   = FAPEXFDLTSD_ioctl;
    fapex_fdltsd_sub_fops.poll    = FAPEXFDLTSD_poll;
    fapex_fdltsd_sub_fops.mmap    = FAPEXFDLTSD_mmap;
    fapex_fdltsd_sub_fops.release = FAPEXFDLTSD_release;

    size_0 = sizeof(FAPEXFDL_TSD_CmnStrT) * (FAPEXDEV_TSD_CMN_DESCRIPTOR_NUM);
    size_1 = sizeof(FAPEXFDL_TSD_SubStrT) * (FAPEXDEV_TSD_SUB_DESCRIPTOR_NUM);

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc_cmn = (FAPEXFDL_TSD_CmnStrT*)FAPI_SYS_MALLOC(size_0);
    if(pfunc_cmn == NULL) {
        return FAPEX_TSD_ERR_OUT_OF_MEMORY;
    }
    pfunc_sub = (FAPEXFDL_TSD_SubStrT*)FAPI_SYS_MALLOC(size_1);
    if(pfunc_sub == NULL) {
        FAPI_SYS_FREE(pfunc_cmn);
        return FAPEX_TSD_ERR_OUT_OF_MEMORY;
    }
    pfunc_sub0 = pfunc_sub;
#else
    pfunc_cmn = FAPIGLD_FDLTSDCmnCtx;
    pfunc_sub = FAPIGLD_FDLTSDSubCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc_cmn, size_0);
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc_sub, size_1);

    pfunc_cmn0 = pfunc_cmn;

    count_cmn = FAPEXDEV_TSD_CMN_MINOR_FIRST;
    count_sub = FAPEXDEV_TSD_SUB_MINOR_FIRST;
    pCfdlHdl  = (void*)FAPEXCFDL_GetCfdlHdl();

    for(i=0; i<FAPEXDEV_TSD_CMN_DESCRIPTOR_NUM; i++) {
        pfunc_cmn->pCfdlHdl = pCfdlHdl;
        pfunc_cmn->block_no = i;
//      pfunc_cmn->status   = FAPEX_STATUS_LOADED;
        pfunc_cmn->timeout  = FAPEXFDL_TSD_TIMEOUT_MAX + 1;
        pfunc_cmn->cmn_sem  = FAPI_SYS_CREATE_SEMAPHORE(1);
        if(pfunc_cmn->cmn_sem == (FAPI_SYS_SemaphoreT)0) {
            ret = FAPEX_TSD_ERR_SEMAPHORE_CREATE;
            FAPEXDebugPrint(2, "[FAPEXFDLTSD_init] create semaphore error\n");
            goto FAPEXFDLTSD_init_error;
        }
        pfunc_cmn->stc_channel_mask = ((0x01ULL << FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM) - 1)
                                      << (FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM * (i & 0x01));

        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdltsd_cmn_fops, pfunc_cmn, &utility, count_cmn);
        if(ret) {
            FAPEXDebugPrint(2, "[FAPEXFDLTSD_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            goto FAPEXFDLTSD_init_error;
        }
        count_cmn ++;

        for(j=0; j<FAPEXDEV_TSD_CH_SUB_DESCRIPTOR_NUM; j++) {
            pfunc_sub->pCfdlHdl = pCfdlHdl;
            pfunc_sub->pid_ch_no = j;
            pfunc_sub->pCmnHdl   = pfunc_cmn;
//          pfunc_sub->status    = FAPEX_STATUS_LOADED;

            /* Register this id's FDL device to CFDL. */
            ret = FAPEXCFDL_RegisterFdl(&fapex_fdltsd_sub_fops, pfunc_sub, &utility, count_sub);
            if(ret) {
                FAPEXDebugPrint(2, "[FAPEXFDLTSD_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
                goto FAPEXFDLTSD_init_error;
            }
            pfunc_sub ++;
            count_sub ++;
        }
        pfunc_cmn ++;
    }

#ifdef FAPI_TSD_CONNECT
    /* call FAPI (initialize) */
    FAPEXDebugPrint(1, "call FAPEXCFG_SYNC_Init\n");

    ret = FAPEXCFG_SYNC_Init();
    if( ret != FAPI_OK ) {
        FAPEXDebugPrint(2, "FAPEXCFG_SYNC_Init:error %d\n", ret);
        goto FAPEXFDLTSD_init_error;
    }

    FAPEXDebugPrint(1, "call FAPEXCFG_TSD_Init\n");

    ret = FAPEXCFG_TSD_Init();
    if( ret != FAPI_OK ) {
        FAPEXDebugPrint(2, "FAPEXCFG_TSD_Init:error %d\n", ret);
        goto FAPEXFDLTSD_init_error;
    }

    FAPEXDebugPrint(1, "call FAPEXCFG_BM_Init\n");

    ret = FAPEXCFG_BM_Init();
    if( ret != FAPI_OK ) {
        FAPEXDebugPrint(2, "FAPEXCFG_BM_Init:error %d\n", ret);
        goto FAPEXFDLTSD_init_error;
    }

    FAPEXDebugPrint(1, "call FAPEXCFG_DESC_Init\n");

    ret = FAPEXCFG_DESC_Init();
    if( ret != FAPI_OK ) {
        FAPEXDebugPrint(2, "FAPEXCFG_DESC_Init:error %d\n", ret);
        goto FAPEXFDLTSD_init_error;
    }

    /* setup Audio/Video/Teletext stream buffer area */
    pfunc_cmn = pfunc_cmn0;
    for(i=0; i<FAPEXDEV_TSD_CMN_DESCRIPTOR_NUM; i++) {

        /* setup Audio Stream BM area top address */
        if(bm_funcStr.GetBlockIndexByBufferPurpose(FAPI_BM_AUDIO_STREAM_BUFFER) == i) {
            pfunc_cmn->pAStreamBuff = (uint8_t*)bm_funcStr.GetAddressByBufferPurpose(
                                                                FAPI_BM_AUDIO_STREAM_BUFFER);
        }
        else {
            pfunc_cmn->pAStreamBuff = NULL;
        }

        /* setup Video Stream BM area top address */
        if(bm_funcStr.GetBlockIndexByBufferPurpose(FAPI_BM_VIDEO_STREAM_BUFFER) == i) {
            pfunc_cmn->pVStreamBuff = (uint8_t*)bm_funcStr.GetAddressByBufferPurpose(
                                                                FAPI_BM_VIDEO_STREAM_BUFFER);
        }
        else {
            pfunc_cmn->pVStreamBuff = NULL;
        }

        /* setup Teletext Stream BM area top address */
        if(bm_funcStr.GetBlockIndexByBufferPurpose(FAPI_BM_TELETEXT_BUFFER) == i) {
            pfunc_cmn->pTStreamBuff = (uint8_t*)bm_funcStr.GetAddressByBufferPurpose(
                                                                FAPI_BM_TELETEXT_BUFFER);
        }
        else {
            pfunc_cmn->pTStreamBuff = NULL;
        }
        pfunc_cmn ++;
    }
#endif /* FAPI_TSD_CONNECT */

    /* setup general stream buffer area information */
#ifndef __LINUX__
    buff_align_size = FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE;
#else
    buff_align_size = (FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE > PAGE_SIZE) ?
                       FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE : PAGE_SIZE;
#endif

#ifdef FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
    if(FAPEX_SYS_MMAP_BM_SECTION_BYTES > 0) {
        pfunc_cmn0->sbuff_ctrl = FAPISAF_InitSBuffArea(FAPEX_SYS_MMAP_BM_SECTION_START,
                                                       FAPEX_SYS_MMAP_BM_SECTION_BYTES,
                                                       buff_align_size,
                                                       FAPEXDEV_TSD_SUB_DESCRIPTOR_NUM);
 #if (DEBUG >= 2)
        fdltsd_CheckBuffList("InitSBuff", pfunc_cmn0->sbuff_ctrl,
                             FAPEX_SYS_MMAP_BM_SECTION_START);
 #endif
    }
    else
#endif
    {
        pfunc_cmn0->sbuff_ctrl = FAPISAF_InitSBuffArea(0, 0,
                                                       buff_align_size,
                                                       FAPEXDEV_TSD_SUB_DESCRIPTOR_NUM);
    }
    if(pfunc_cmn0->sbuff_ctrl == NULL) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
        FAPI_SYS_FREE(pfunc_sub);
        FAPI_SYS_FREE(pfunc_cmn);
#endif
        return FAPEX_TSD_ERR_OUT_OF_MEMORY;
    }

    FAPEXDebugPrint(1, "[FAPEXFDLTSD_init] fapex_fdltsd was loaded. %d\n", ret);

    return 0;

FAPEXFDLTSD_init_error:
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPI_SYS_FREE(pfunc_cmn0);
    FAPI_SYS_FREE(pfunc_sub0);
#endif
    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLTSD_cleanup(void)
{
    int32_t count;
    void    *pCfdlHdl;
    FAPEXFDL_TSD_CmnStrT *pfunc_cmn, *pfunc_cmn0;

    FAPEXDebugPrint(1, "[FAPEXFDLTSD_cleanup]\n");

#ifdef FAPI_TSD_CONNECT
    /* call FAPI API (cleanup) */
    FAPEXCFG_TSD_Cleanup();
    FAPEXDebugPrint(1, "end FAPEXCFG_TSD_Cleanup\n");

    FAPEXCFG_BM_Cleanup();
    FAPEXDebugPrint(1, "end FAPEXCFG_BM_Cleanup\n");

    FAPEXCFG_DESC_Cleanup();
    FAPEXDebugPrint(1, "end FAPEXCFG_DESC_Cleanup\n");

    FAPEXCFG_SYNC_Cleanup();
    FAPEXDebugPrint(1, "end FAPEXCFG_SYNC_Cleanup\n");

#endif /* FAPI_TSD_CONNECT */

    /* cleanup sequence (common this FDL) */
    pCfdlHdl  = (void*)FAPEXCFDL_GetCfdlHdl();
    pfunc_cmn = (FAPEXFDL_TSD_CmnStrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_TSD_CMN_MINOR_FIRST);

    if(pfunc_cmn != NULL) {
        pfunc_cmn0 = pfunc_cmn;

        FAPISAF_ExitSBuffArea(pfunc_cmn0->sbuff_ctrl);
        pfunc_cmn0->sbuff_ctrl = NULL;

        for(count=0; count<FAPEXDEV_TSD_CMN_DESCRIPTOR_NUM; count++) {
            FAPI_SYS_DESTROY_SEMAPHORE(pfunc_cmn->cmn_sem);
            pfunc_cmn->cmn_sem = (FAPI_SYS_SemaphoreT)0;
            pfunc_cmn ++;
        }

#ifdef ALLOCATE_DYNAMIC_MEMORY
        FAPI_SYS_FREE(pfunc_cmn0);
    }
    {
        FAPEXFDL_TSD_SubStrT *pfunc_sub;

        pfunc_sub = (FAPEXFDL_TSD_SubStrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_TSD_SUB_MINOR_FIRST);
        if(pfunc_sub != NULL) {
            FAPI_SYS_FREE(pfunc_sub);
        }
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    }

    for(count=FAPEXDEV_TSD_CMN_MINOR_FIRST; count<=FAPEXDEV_TSD_CMN_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }
    for(count=FAPEXDEV_TSD_SUB_MINOR_FIRST; count<=FAPEXDEV_TSD_SUB_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLTSD_cleanup] fapex_fdltsd was unloaded.\n");

    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLTSD_init);
module_exit(FAPEXFDLTSD_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdltsd object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
