/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL TSD header file (private)
*/

#ifndef _FAPEX_FDLTSD_PRIV_H
#define _FAPEX_FDLTSD_PRIV_H

#include "fapex/abst_libtsd.h"
#include "fapi/drv_tsd.h"
#include "fapex/sys_config.h"

/*!
*************************************************************************
** \brief FDL TSD stream buffer statically allocation enable flag
*************************************************************************/
#if defined(FAPEX_SYS_MMAP_BM_SECTION_START) && defined(FAPEX_SYS_MMAP_BM_SECTION_END)
#if FAPEX_SYS_MMAP_BM_SECTION_START < FAPEX_SYS_MMAP_BM_SECTION_END
#define FAPEXFDL_TSD_STATIC_STREAM_BUFFER_ENABLE
#endif
#endif

/*!
*************************************************************************
** \brief FDL TSD stream buffer management control structure
*************************************************************************/
typedef struct _FAPEXFDL_TSD_BuffCtrlT      FAPEXFDL_TSD_BuffCtrlT;

struct _FAPEXFDL_TSD_BuffCtrlT {
    uint32_t   buff_top;    //!< buffer area physical address
    uint32_t   buff_size;   //!< buffer area size;

    FAPEXFDL_TSD_BuffCtrlT  *prev_info;     //!< previous buffer information
    FAPEXFDL_TSD_BuffCtrlT  *next_info;     //!< next buffer information

};

/*!
*************************************************************************
** \brief FDL TSD section data information structure
*************************************************************************/
typedef struct {
    uint32_t    sct_base;           //!< offset address where the section has been stored.
    uint32_t    sct_length;         //!< length of current section */

} FAPEXFDL_TSD_SctInfoT;

/*!
*************************************************************************
** \brief FDL TSD section information fifo structure
*************************************************************************/
#define FAPEXFDL_TSD_SECTION_FIFO_MAX   64

typedef struct {
    uint32_t    notifier_id;        //!< notifier identifier
    void        *pParentSubHdl;     //!< The pointer to FAPEXFDL_TSD_SubStrT context

    FAPEXFDL_TSD_SctInfoT   sct_info[FAPEXFDL_TSD_SECTION_FIFO_MAX];
                                    //!< array for FAPEXFDL_TSD_SctInfoT data
    uint32_t    write_idx;          //!< write index of sct_info[]

} FAPEXFDL_TSD_SctFifoT;

/*!
*************************************************************************
** \brief FDL TSD Event Notifier information fifo structure
*************************************************************************/
typedef struct {
    /*!
        notifier type (basically type)<br>
        - FAPEX_TSD_BUFFSTAT_PACKET_END
        - FAPEX_TSD_BUFFSTAT_SECTION_END
        - FAPEX_TSD_BUFFSTAT_SCRAMBLED_PACKET
        - FAPEX_TSD_BUFFSTAT_THRESHUNDER
        - FAPEX_TSD_BUFFSTAT_THRESHOVER
        - FAPEX_TSD_BUFFSTAT_OVERFLOW
        - FAPEX_TSD_BUFFSTAT_TIMEOUT
        - FAPEX_TSD_BUFFSTAT_STOPPED
        .
        optional bit
        - FAPEX_TSD_BUFFSTAT_CRCERROR
    */
    uint32_t    notifier_type;
    /*
        notifier data
    */
    void*       pData;

} FAPEXFDL_TSD_NotifierFifoT;

#if 1 /* workaround for #23196 */
#define FAPEXFDL_TSD_NOTIFIER_FIFO_MAX          2048
#else
#define FAPEXFDL_TSD_NOTIFIER_FIFO_MAX          64
#endif
#define FAPEXFDL_TSD_NOTIFIER_FIFO_MAX_MASK     (FAPEXFDL_TSD_NOTIFIER_FIFO_MAX - 1)

/*!
*************************************************************************
** \brief FDL TSD common context structure
*************************************************************************/
typedef struct {
    int32_t     id;             //!< Device identifier
    uint32_t    used_count;     //!< used count
#ifdef __LINUX__
    struct kref kref;           //!< kref struct
#endif

    void        *pCfdlHdl;      //!< The pointer to the CFDL context
    int32_t     status;         //!< TSD block status

    uint32_t    block_no;       //!< TSD block number
    uint32_t    stc_block;      //!< STC enable block flag

    FAPEX_TSD_InputParamStrT  blk_params;   //!< TS block parameters

    uint32_t    pkt_ctrl_address;   //!< TS Packet Control Address (virtual)
    uint32_t    pkt_in_address;     //!< TS Packet Register Address (virtual)
    uint32_t    pkt_in_offset;      //!< each offset of TS Packet Register Address
    uint32_t    pkt_in_area_size;   //!< TS Packet input area size
    int32_t     pkt_in_status;      //!< TS Packet input status

    uint32_t    enable_channel;     //!< enable TSD channel flag
    uint32_t    running_channel;    //!< running channel flag
    uint32_t    bm_channel;         //!< enable BM channel flag
    uint32_t    tout_channel;       //!< enable Timeout channel flag
    uint64_t    stc_channel;        //!< PCR/STC channel flag
    uint64_t    stc_channel_mask;   //!< PCR/STC channel mask for TSD block

    uint8_t     *pAStreamBuff;      //!< Audio Stream Buffer Top offset address
    uint8_t     *pVStreamBuff;      //!< Video Stream Buffer Top offset address
    uint8_t     *pTStreamBuff;      //!< Teletext Stream Buffer Top offset address

    FAPISAF_BuffManageT *sbuff_ctrl;    //!< static buffer management parameter pointer

    uint32_t    timeout;            //!< timeout for packet receive [msec]

    FAPI_SYS_SemaphoreT  cmn_sem;   //!< Semaphore for common TSD block control

} FAPEXFDL_TSD_CmnStrT;

#define FAPEXFDL_TSD_PACKET_IN_BUSY_MASK    FAPEX_TSD_PACKET_INPUT_STATUS_BUSY_BUFF_ALL
#define FAPEXFDL_TSD_PACKET_IN_INVALID_MASK ~(FAPEX_TSD_PACKET_INPUT_STATUS_VALID_BUFF_ALL)
#define FAPEXFDL_TSD_PACKET_IN_MAX_TIMEOUT  (100)

/*!
*************************************************************************
** \brief FDL TSD sub context structure
*************************************************************************/
typedef struct {
    int32_t     id;             //!< Device identifier
    uint32_t    used_count;     //!< used count
#ifdef __LINUX__
    struct kref kref;           //!< kref struct
#endif

    void        *pCfdlHdl;      //!< The pointer to the CFDL context
    int32_t     status;         //!< TSD channel status

    uint32_t    pid_ch_no;      //!< PID channel number
    uint32_t    bm_ch_no;       //!< BM channel number

    void        *pCmnHdl;       //!< the pointer to the TSD common context
    void        *pTSDCtxHdl;    //!< FAPI TSD driver's context
    void        *pBMCtxHdl;     //!< FAPI BM driver's context
    void        *pDESCCtxHdl;   //!< FAPI DESC driver's context

    uint32_t    flags;          //!< TSD channel structure flags (see #FAPEXFDL_TSD_CHFLAGS_xxx)
    uint8_t     stream_type;    //!< stream_type (see #FAPEX_TSD_STREAMTYPE_xxx)
    uint32_t    pes_paser_no;   //!< PES Parser No.
    uint32_t    read_pt;        //!< buffer read position for HWCB

    uint32_t    table_id_max;   //!< max number of table_id filter setting
    uint32_t    table_id_no;    //!< the table_id filter setting number
    FAPI_TSD_FilterDataT    *pFilterData;   //!< pointer for filter data
    uint32_t                *pFilter;       //!< pointer for filter pattern
    uint32_t                *pMask;         //!< pointer for mask pattern

    uint32_t    buff_address;       //!< buffer top physical address
    uint32_t    buff_size;          //!< buffer size
    uint32_t    buff_allocate_type; //!< buffer allocation type (see #FAPEXFDL_TSD_STRM_BUFF_TYPE_xxx)
#ifndef __LINUX__
    uint32_t    buff_address_temp;  //!< buffer top address (non-alignment)
#else
    struct page *buff_pg;           //!< buffer area page
    uint32_t    order;              //!< order of buff_size
#endif
    uint32_t    buff_th_H;          //!< buffer threshold heigh level
    uint32_t    buff_th_L;          //!< buffer threshold low level
    uint32_t    timeout;            //!< timeout for packet receive [msec]

    FAPEXCFDLSAF_EventStrT     dt_event;   //!< data event control

    FAPEXFDL_TSD_NotifierFifoT ntfyFifo[FAPEXFDL_TSD_NOTIFIER_FIFO_MAX];
                                           //!< The FAPEXFDL_TSD_NotifierFifoT array

    uint32_t                sct_max;       //!< max number of section filter setting
    FAPEXFDL_TSD_SctFifoT   *pSctFifo;     //!< The pointer to the FAPEXFDL_TSD_SctFifoT array
    int32_t                 ready_idx;     //!< ready index of event queue

    int32_t     output_id;          //!< the identifier for output device

} FAPEXFDL_TSD_SubStrT;

/*! \brief TSD channel structure flags */
#define FAPEXFDL_TSD_CHFLAGS_NONE               (0x0000)
#define FAPEXFDL_TSD_CHFLAGS_BUFF_CTRL_MSK      (0x0007)  //!< mask for buffer control flag (see #FAPEX_TSD_FLTFLAGS_SWAPxxx)
#define FAPEXFDL_TSD_CHFLAGS_PES_PARSER         (0x0020)  //!< using PES Parser
#define FAPEXFDL_TSD_CHFLAGS_HCB                (0x0040)  //!< using Hardware Controlled Block
#define FAPEXFDL_TSD_CHFLAGS_OUT_BOUND          (0x0080)  //!< buffer read pointer outside flag
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_TIMOUT    (0x0100)  //!< packet timeout notifer already assigned flag
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND    (0x0200)  //!< packet end notifer already assigned flag
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCTEND    (0x0400)  //!< section end notifer already assigned flag
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCRMBLD   (0x0800)  //!< scrambled packet notifer already assigned flag
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_BUFF_TH   (0x1000)  //!< buffer threshold notifer already assigned flag
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_BUFF_ERR  (0x2000)  //!< buffer error notifer already assigned flag
#define FAPEXFDL_TSD_CHFLAGS_RM_TS_HEADER       (0x4000)  //!< remove TS header flag
#define FAPEXFDL_TSD_CHFLAGS_CHECK_CC           (0x8000)  //!< check continuity counter flag
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_DATA      \
 (FAPEXFDL_TSD_CHFLAGS_NOTIFIER_PKTEND | FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCTEND | \
  FAPEXFDL_TSD_CHFLAGS_NOTIFIER_SCRMBLD | FAPEXFDL_TSD_CHFLAGS_NOTIFIER_BUFF_TH)
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_EXCEPT    \
 (FAPEXFDL_TSD_CHFLAGS_NOTIFIER_TIMOUT | FAPEXFDL_TSD_CHFLAGS_NOTIFIER_BUFF_ERR)
#define FAPEXFDL_TSD_CHFLAGS_NOTIFIER_ALL       \
 (FAPEXFDL_TSD_CHFLAGS_NOTIFIER_DATA | FAPEXFDL_TSD_CHFLAGS_NOTIFIER_EXCEPT)

/*! \brief TSD timeout parameter */
#define FAPEXFDL_TSD_TIMEOUT_TO_CYCLE   (200000000 / 1000)  //!< TSD clock cycle : 200MHz
#define FAPEXFDL_TSD_TIMEOUT_MAX        (uint32_t)(0xFFFFFFFFUL / FAPEXFDL_TSD_TIMEOUT_TO_CYCLE) /* [msec] */
#define FAPEXFDL_TSD_TIMEOUT_INTERRUPT  (100)
#define FAPEXFDL_TSD_TIMEOUT_PES_FILTER (10000)

/*! \brief TSD event type */
#define FAPEXFDL_TSD_EVENT_SECTION_NO_SHFT  20
#define FAPEXFDL_TSD_EVENT_SECTION_MASK     \
 ((0x0000001F << FAPEXFDL_TSD_EVENT_SECTION_NO_SHFT) | FAPEX_TSD_BUFFSTAT_SECTION_END)

#define FAPEXFDL_TSD_EVENT_READ             (FAPEX_TSD_BUFFSTAT_PACKET_END | FAPEXFDL_TSD_EVENT_SECTION_MASK)
#define FAPEXFDL_TSD_EVENT_WRITE            (FAPEX_TSD_BUFFSTAT_NONE)
#define FAPEXFDL_TSD_EVENT_MESSAGE          (FAPEX_TSD_BUFFSTAT_SCRAMBLED_PACKET)
#define FAPEXFDL_TSD_EVENT_WARNING          (FAPEX_TSD_BUFFSTAT_THRESHOVER | FAPEX_TSD_BUFFSTAT_THRESHUNDER)
#define FAPEXFDL_TSD_EVENT_ERROR            (FAPEX_TSD_BUFFSTAT_TIMEOUT | FAPEX_TSD_BUFFSTAT_OVERFLOW)
#define FAPEXFDL_TSD_EVENT_ALL              \
 (FAPEX_TSD_BUFFSTAT_PACKET_END  | FAPEX_TSD_BUFFSTAT_SECTION_END | FAPEX_TSD_BUFFSTAT_SCRAMBLED_PACKET | \
  FAPEX_TSD_BUFFSTAT_THRESHUNDER | FAPEX_TSD_BUFFSTAT_THRESHOVER  | FAPEX_TSD_BUFFSTAT_OVERFLOW | \
  FAPEX_TSD_BUFFSTAT_CRCERROR    | FAPEX_TSD_BUFFSTAT_TIMEOUT | FAPEX_TSD_BUFFSTAT_STOPPED)

/*! \brief stream buffer allocation type */
#define FAPEXFDL_TSD_STRM_BUFF_TYPE_UNKNOWN     (0)     //!< not allocated
#define FAPEXFDL_TSD_STRM_BUFF_TYPE_STATIC      (1)     //!< allocated from static common area
#define FAPEXFDL_TSD_STRM_BUFF_TYPE_RESERVED    (2)     //!< allocated from statically reserved area
#define FAPEXFDL_TSD_STRM_BUFF_TYPE_DYNAMIC     (3)     //!< allocated from dynamic area

/*! \brief stream buffer align size (MB86H60 hardware specification) */
#define FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE         (1024)
#define FAPEXFDL_TSD_STRM_BUFF_SIZE_ORDER_MIN       (10)
#define FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE_MASK    (FAPEXFDL_TSD_STRM_BUFF_ALIGNED_SIZE - 1)


#endif /* !_FAPEX_FDLTSD_PRIV_H */
