/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 *
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv (ttx)
*/

#include "fapex.h"
#include "fapex/abst_libttx.h"
#include "fapex/abst_libtsd.h"
#include "fapex/sys_config.h"
#include "fapex_fdlttx_priv.h"
#include "fapex_fdltsd_priv.h"
#include "fapex_config_cmn.h"

/* for ttx_funcStr, bm_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_TTX_StrT  FAPIGLD_FDLTTXCtx[FAPEXDEV_TTX_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLTTX_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                   FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLTTX_read] size:%d\n", size);

    return size;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLTTX_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                   FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLTTX_write] size:%d\n", size);

    return size;
}
#endif

/*!
*******************************************************************************
**
** \brief  start Teletext
**
** \param  pfunc    pointer to TTX context
**
******************************************************************************/
static int32_t fdlttx_Start(FAPEXFDL_TTX_StrT *pfunc)
{
    /* check VBI-Teletext status */
    if(pfunc->status == FAPEX_STATUS_READY) {
        ttx_funcStr.Enable();

        pfunc->status = FAPEX_STATUS_RUNNING;
        return 0;
    }
    else {
        return FAPEX_TTX_ERR_INVALID_STATUS;
    }
}

/*!
*******************************************************************************
**
** \brief  stop Teletext
**
** \param  pfunc    pointer to TTX context
**
******************************************************************************/
static int32_t fdlttx_Stop(FAPEXFDL_TTX_StrT *pfunc)
{
    /* check VBI-Teletext status */
    if(pfunc->status == FAPEX_STATUS_RUNNING) {
        ttx_funcStr.Disable();

        pfunc->status = FAPEX_STATUS_READY;
    }
    return 0;
}

/*!
*******************************************************************************
**
** \brief  callback function for Teletext
**
******************************************************************************/
static void fdlttx_OverflowIrqHandler(uint32_t blockIndex, uint32_t channelIndex, void* optData)
{
    FAPI_SYS_HandleT bm_handle = (FAPI_SYS_HandleT)optData;

    ttx_funcStr.EnableReinsertion(FAPI_TTX_HOR_START_PAL_DISABLE);

    /* reset buffer pointers before starting teletext */
    bm_funcStr.SetBufferWritePointer( bm_handle, 0, 0 );
    bm_funcStr.SetBufferReadPointer ( bm_handle, 0, 0 );

    /* enable the high level IRQ until high level IRQ is received */
    bm_funcStr.SetHighLevelIrq(bm_handle, NULL, NULL, 0, FAPI_BM_ENABLE_IRQ);
}

static void fdlttx_HighLevelIrqHandler(uint32_t blockIndex, uint32_t channelIndex, void* optData)
{
    FAPI_SYS_HandleT bm_handle = (FAPI_SYS_HandleT)optData;

    /* disable the high level IRQ until TTX is stopped or low level IRQ is received */
    bm_funcStr.SetHighLevelIrq(bm_handle, NULL, NULL, 0, FAPI_BM_DISABLE_BUT_KEEP_IRQ);

    ttx_funcStr.EnableReinsertion(FAPI_TTX_HOR_START_PAL_ENABLE);
}

static void fdlttx_LowLevelIrqHandler(uint32_t blockIndex, uint32_t channelIndex, void* optData)
{
    FAPI_SYS_HandleT bm_handle = (FAPI_SYS_HandleT)optData;

    ttx_funcStr.EnableReinsertion(FAPI_TTX_HOR_START_PAL_DISABLE);

    /* enable the high level IRQ until high level IRQ is received */
    bm_funcStr.SetHighLevelIrq(bm_handle, NULL, NULL, 0, FAPI_BM_ENABLE_IRQ);
}

/*!
*******************************************************************************
**
** \brief  setup input for Teletext
**
** \param  pfunc    pointer to TTX context
** \param  pArg     pointer to FAPEX_TTX_OpenParamStrT parameter
**
******************************************************************************/
static int32_t fdlttx_SetupInput(FAPEXFDL_TTX_StrT *pfunc, FAPEX_TTX_OpenParamStrT *pArg)
{
    FAPEXFDL_TSD_SubStrT  *pfunc_tsd;
    FAPI_BM_InfoT         bm_info;
    void     *pCfdlHdl;
    int32_t  input_id;
    int32_t  ret = FAPEX_TTX_ERR_BAD_PARAMETER;

    if(pfunc->open_flag != 0) {
        return 0;
    }
    if(pArg == NULL) {
        return ret;
    }
    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&input_id, &(pArg->inputId), sizeof(int32_t));

    /* check device id */
    if( (input_id >= FAPEXDEV_TSD_SUB_MINOR_FIRST)
     && (input_id <= FAPEXDEV_TSD_SUB_MINOR_LAST) ) {

        pCfdlHdl = pfunc->pCfdlHdl;

        /* setup input handler */
        pfunc_tsd = (FAPEXFDL_TSD_SubStrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, input_id);
        if(pfunc_tsd != NULL) {

            /* check VBI-Teletext stream */
            if(pfunc_tsd->stream_type == FAPEX_TSD_STREAMTYPE_TELETEXT) {
                uint32_t  high_level, low_level;

                /* clear callback */
                ret = bm_funcStr.SetHighLevelIrq(pfunc_tsd->pBMCtxHdl,
                                                 NULL, NULL, 0, FAPI_BM_DISABLE_IRQ);
                if((ret != FAPI_OK) && (ret != FAPI_BM_ERR_HIGH_LEVEL_IRQ_NOT_ASSIGNED)) {
                    FAPEXDebugPrint(2, "[fdlttx_SetupInput] FAPI_BM_SetHighLevelIrq error %d\n",
                                       ret);
                    return ret;
                }

                /* get Buffer information */
                ret = bm_funcStr.GetInfo(pfunc_tsd->pBMCtxHdl, &bm_info);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdlttx_SetupInput] FAPI_BM_GetInfo error %d\n",
                                       ret);
                    return ret;
                }

                /* assign callback (overflow) */
                ret = bm_funcStr.SetOverflowIrq(pfunc_tsd->pBMCtxHdl,
                                                fdlttx_OverflowIrqHandler,
                                                pfunc_tsd->pBMCtxHdl, 1);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdlttx_SetupInput] FAPI_BM_SetOverflowIrq error %d\n",
                                       ret);
                    return ret;
                }

                /* assign callback (HighLevel) */
                if(pfunc_tsd->buff_th_H == bm_info.size) {
                    high_level = bm_info.size >> 1;
                }
                else {
                    high_level = bm_info.size - pfunc_tsd->buff_th_H;
                }
                ret = bm_funcStr.SetHighLevelIrq(pfunc_tsd->pBMCtxHdl,
                                                 fdlttx_HighLevelIrqHandler,
                                                 pfunc_tsd->pBMCtxHdl,
                                                 high_level, FAPI_BM_ENABLE_IRQ);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdlttx_SetupInput] FAPI_BM_SetHighLevelIrq error %d\n",
                                       ret);
                    return ret;
                }

                /* assign callback (LowLevel) */
                if(pfunc_tsd->buff_th_L == 0) {
                    low_level = 0x1000; /* 4KB */
                    while(low_level >= high_level) {
                        low_level >>= 1;
                    }
                }
                else {
                    low_level = pfunc_tsd->buff_th_L;
                }
                ret = bm_funcStr.SetLowLevelIrq(pfunc_tsd->pBMCtxHdl,
                                                fdlttx_LowLevelIrqHandler,
                                                pfunc_tsd->pBMCtxHdl,
                                                low_level, FAPI_BM_ENABLE_IRQ);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdlttx_SetupInput] FAPI_BM_SetLowLevelIrq error %d\n",
                                       ret);
                    return ret;
                }

                /* reset BM from reading */
                ret = bm_funcStr.SetReadEnable(pfunc_tsd->pBMCtxHdl, 0);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdlttx_SetupInput] FAPI_BM_SetReadEnable error %d\n",
                                       ret);
                    return ret;
                }
                pfunc->open_flag = 1;
                pfunc->status    = FAPEX_STATUS_READY;
            }
        }
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLTTX_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                   FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_TTX_StrT *pfunc;
    int32_t  ret = FAPEX_TTX_ERR_BAD_PARAMETER;

    FAPEXDebugPrint(3, "[FAPEXFDLTTX_ioctl] cmd:%08x\n",cmd);

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_TTX_StrT*)(file->private_data);

    switch(cmd) {
        case FAPEX_IOCCOM_START:
            ret = fdlttx_Start(pfunc);
            break;

        case FAPEX_IOCCOM_STOP:
            ret = fdlttx_Stop(pfunc);
            break;

        case FAPEX_IOCCOM_TTX_OPEN:
            ret = fdlttx_SetupInput(pfunc, (FAPEX_TTX_OpenParamStrT *)arg);
            break;

        default:
            ;
    }

    return ret;
}

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLTTX_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T mask = 0;

    FAPEXDebugPrint(4, "[FAPEXFDLTTX_poll]\n");

    /* actual code which is depend on function is implimented here. */

    return mask;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLTTX_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    int32_t ret;

    if (vma->vm_pgoff != 0){
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_IO | VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if (file->f_flags & O_SYNC){
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
                  vma->vm_end - vma->vm_start, vma->vm_page_prot);

    if(ret != 0){
        return -EAGAIN;
    }
    return 0;
#else
    return -1;
#endif /* __LINUX__ */
}
#endif

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLTTX_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_TTX_StrT *pfunc;
    int32_t id;
    void    *pCfdlHdl;

    FAPEXDebugPrint(3, "[FAPEXFDLTTX_open] file:0x%08x,private_data:0x%08x\n",
                       (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLTTX_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLTTX_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc = (FAPEXFDL_TTX_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc;
        FAPEXDebugPrint(3, "[FAPEXFDLTTX_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc));

        /* set parameter */
        pfunc->pCfdlHdl = pCfdlHdl;
        pfunc->status   = FAPEX_STATUS_OPENED;
#ifdef __LINUX__
        kref_init(&pfunc->kref);
#endif
    }
    else {
        pfunc = (FAPEXFDL_TTX_StrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLTTX_open] fapex_fdlttx was opened : used_counter %d.\n",
                       pfunc->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlttx_delete(struct kref *kref)
{
    FAPEXFDL_TTX_StrT  *pfunc;

    pfunc = container_of(kref, FAPEXFDL_TTX_StrT, kref);
#else
static void fdlttx_delete(FAPEXFDL_TTX_StrT *pfunc)
{
#endif
    pfunc->open_flag = 0;
    pfunc->status    = FAPEX_STATUS_TERMINATED;
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLTTX_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_TTX_StrT  *pfunc;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_TTX_StrT*)(file->private_data);

    if(pfunc != NULL) {
        /* decrement the count on our device */
        pfunc->used_count--;
#ifdef __LINUX__
        kref_put(&pfunc->kref, fdlttx_delete);
#else
        fdlttx_delete(pfunc);
#endif

        FAPEXDebugPrint(3, "[FAPEXFDLTTX_release] : used_count %d\n",
                           pfunc->used_count);
    }
    return 0;
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlttx_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLTTX_init(void)
{
    int32_t ret;
    int32_t count;
    FAPEXFDL_TTX_StrT   *pfunc;
    FAPEXCFDL_UtilStrT  utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_TTX_StrT   *pfunc0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLTTX_init]\n");

    /* set struct file_operations */
    fapex_fdlttx_fops.owner   = THIS_MODULE;
    fapex_fdlttx_fops.open    = FAPEXFDLTTX_open;
//  fapex_fdlttx_fops.read    = FAPEXFDLTTX_read;
//  fapex_fdlttx_fops.write   = FAPEXFDLTTX_write;
    fapex_fdlttx_fops.ioctl   = FAPEXFDLTTX_ioctl;
//  fapex_fdlttx_fops.poll    = FAPEXFDLTTX_poll;
//  fapex_fdlttx_fops.mmap    = FAPEXFDLTTX_mmap;
    fapex_fdlttx_fops.release = FAPEXFDLTTX_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc = (FAPEXFDL_TTX_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_TTX_StrT)*(FAPEXDEV_TTX_DESCRIPTOR_NUM));
    if(pfunc == NULL) {
        return FAPEX_TTX_ERR_OUT_OF_MEMORY;
    }
    pfunc0 = pfunc;
#else
    pfunc = FAPIGLD_FDLTTXCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc, sizeof(FAPEXFDL_TTX_StrT)*(FAPEXDEV_TTX_DESCRIPTOR_NUM));

    for(count=FAPEXDEV_TTX_MINOR_FIRST; count<=FAPEXDEV_TTX_MINOR_LAST; count++){
//      pfunc_sub->status = FAPEX_STATUS_LOADED;

        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlttx_fops, pfunc, &utility, count);
        if(ret) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            FAPEXDebugPrint(1, "[FAPEXFDLTTX_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            return ret;
        }
        pfunc ++;
    }

#ifdef FAPI_TTX_CONNECT
    FAPEXDebugPrint(1, "call FAPEXCFG_TTX_Init\n");

    /* call FAPI (initialize) */
    ret = FAPEXCFG_TTX_Init();
#else
    ret = 0;
#endif /* FAPI_TTX_CONNECT */

    FAPEXDebugPrint(1, "[FAPEXFDLTTX_init] fapex_fdlttx was loaded. %d\n", ret);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLTTX_cleanup(void)
{
    int32_t count;

    FAPEXDebugPrint(1, "[FAPEXFDLTTX_cleanup]\n");

#ifdef FAPI_TTX_CONNECT
    /* call FAPI API (cleanup) */
    FAPEXCFG_TTX_Cleanup();

    FAPEXDebugPrint(1, "end FAPEXCFG_TTX_Cleanup\n");
#endif /* FAPI_TTX_CONNECT */

    /* cleanup sequence (common this FDL) */
#ifdef ALLOCATE_DYNAMIC_MEMORY
    {
        void              *pCfdlHdl;
        FAPEXFDL_TTX_StrT *pfunc;

        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        pfunc    = (FAPEXFDL_TTX_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_TTX_MINOR_FIRST);

        /* relase the FDL context */
        if(pfunc != NULL) {
            FAPI_SYS_FREE(pfunc);
        }
    }
#endif /* ALLOCATE_DYNAMIC_MEMORY */

    for(count=FAPEXDEV_TTX_MINOR_FIRST; count<=FAPEXDEV_TTX_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLTTX_cleanup] fapex_fdlttx was unloaded.\n");
    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLTTX_init);
module_exit(FAPEXFDLTTX_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlttx object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
