/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 * 
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv
*/

#include "fapex.h"
#include "fapex/abst_lib.h"
#include "fapex/abst_libvidec.h"
#include "fapex/sys_config.h"
#include "fapex_fdltsd_priv.h"
#include "fapex_fdlvidec_priv.h"
#include "fapex_fdlviout_priv.h"
#include "fapex_fdlvienc_priv.h"
#include "fapex_config_cmn.h"

/* for videc_funcStr, viscale_funcStr, tsd_funcStr, bm_funcStr, sync_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_VIDEC_StrT  FAPIGLD_FDLVIDECCtx[FAPEXDEV_VIDEC_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

/*! Bank offsets on ramif_v */
static uint32_t FAPEXFDLVIDEC_BANK_OFFSET_1920HD[22] = {
    0x0000,0x0340,0x0680,0x09C0,0x0D00,0x1040,0x1380,0x16C0,0x1A00,
    0x2000,0x2340,0x2680,0x29C0,0x2D00,0x3040,0x3380,0x36C0,0x3A00,
    0x4000,0x4340,0x4680,0x49C0
};
static uint32_t FAPEXFDLVIDEC_BANK_OFFSET_1440HD[22] = {
    0x0000,0x0270,0x04E0,0x0750,0x09C0,0x0C30,0x0EA0,0x1110,0x1380,
    0x15F0,0x1860,0x1AD0,0x1D40,0x2000,0x2270,0x24E0,0x2750,0x29C0,
    0x2C30,0x2EA0,0x3110,0x3380
};
static uint32_t FAPEXFDLVIDEC_BANK_OFFSET_1280HD[22] = {
    0x0000,0x0168,0x02D0,0x0438,0x05A0,0x0708,0x0870,0x09D8,0x0B40,
    0x0CA8,0x0E10,0x0F78,0x10E0,0x1248,0x13B0,0x1518,0x1680,0x17E8,
    0x1950,0x1AB8,0x1C20,0x1D88
};
static uint32_t FAPEXFDLVIDEC_BANK_OFFSET_SD[22] = {
    0x0000,0x00A8,0x0150,0x01F8,0x02A0,0x0348,0x03F0,0x0498,0x0540,
    0x05E8,0x0690,0x0738,0x07E0,0x0888,0x0930,0x09D8,0x0A80,0x0B28,
    0x0BD0,0x0C78,0x0D20,0x0DC8
};

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLVIDEC_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                     FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_read] size:%d\n",size);

    /* actual code which is depend on function is implimented here. */

    return size;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLVIDEC_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                     FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_write] size:%d\n",size);

    /* actual code which is depend on function is implimented here. */

    return size;
}
#endif

/*!
*******************************************************************************
**
** \brief  Event Notifer
**
** \param   optDataPtr   Optional data to be passed to the notify function.
**
******************************************************************************/
static void fdlvidec_DecAvailableNotifier(void* optDataPtr)
{
    FAPEX_VIDEC_DecFrameInfoStrT  *pFrameInfo;
    FAPEXFDL_VIDEC_StrT *pfunc = (FAPEXFDL_VIDEC_StrT *)optDataPtr;
    uint32_t  flags;

    FAPEXCFDLSAF_SPIN_LOCK(&pfunc->lock, &flags);

    pFrameInfo = &pfunc->frameInfo;
    pFrameInfo->enableFlag |= FAPEX_VIDEC_INFO_DEC_AVAILABLE;

    FAPEXCFDLSAF_SPIN_UNLOCK(&pfunc->lock, &flags);

    /* set event */
    FAPEXCFDLSAF_SET_EVENT(
            &pfunc->dt_event,
            FAPEXFDL_VIDEC_EVENT_DEC_AVAILABLE,
            FAPEXCFDLSAF_WAITEVENT_OR);

    FAPEXDebugPrint(1, "dec start\n");
}

static void fdlvidec_DispAvailableNotifier(void* optDataPtr)
{
    FAPEX_VIDEC_DecFrameInfoStrT  *pFrameInfo;
    FAPEXFDL_VIDEC_StrT *pfunc = (FAPEXFDL_VIDEC_StrT *)optDataPtr;
    uint32_t  flags;

    FAPEXCFDLSAF_SPIN_LOCK(&pfunc->lock, &flags);

    pFrameInfo = &pfunc->frameInfo;
#if 0
    pFrameInfo->enableFlag |= FAPEX_VIDEC_INFO_DISP_AVAILABLE;
#else
    pFrameInfo->enableFlag |= FAPEX_VIDEC_INFO_DISP_AVAILABLE | FAPEX_VIDEC_INFO_DISP_CHANGED;
#endif

    FAPEXCFDLSAF_SPIN_UNLOCK(&pfunc->lock, &flags);

    /* set event */
    FAPEXCFDLSAF_SET_EVENT(
            &pfunc->dt_event,
            FAPEXFDL_VIDEC_EVENT_DISP_AVAILABLE,
            FAPEXCFDLSAF_WAITEVENT_OR);

    FAPEXDebugPrint(1, "disp start\n");
}

static void fdlvidec_FrameInfoNotifier(FAPI_VIDEC_VideoDecInfoStateT* videcInfoPtr,
                                       void* optDataPtr)
{
    FAPEX_VIDEC_DecFrameInfoStrT  *pFrameInfo;
    FAPEXFDL_VIDEC_StrT *pfunc = (FAPEXFDL_VIDEC_StrT *)optDataPtr;
    uint32_t  flags;


    pFrameInfo = &pfunc->frameInfo;

    FAPEXCFDLSAF_SPIN_LOCK(&pfunc->lock, &flags);

    pFrameInfo->enableFlag   |= FAPEX_VIDEC_INFO_DISP_ENABLE;
    pFrameInfo->decodedWidth  = videcInfoPtr->decodedWidth;
    pFrameInfo->decodedHeight = videcInfoPtr->decodedHeight;
    pFrameInfo->displayWidth  = videcInfoPtr->displayWidth;
    pFrameInfo->displayHeight = videcInfoPtr->displayHeight;
    pFrameInfo->offset_y      = pfunc->pOffsetLut[videcInfoPtr->frameBank] << 12;
    pFrameInfo->offset_c      = pFrameInfo->offset_y + pfunc->c_offset;

    FAPEXCFDLSAF_SPIN_UNLOCK(&pfunc->lock, &flags);

    /* set event */
    FAPEXCFDLSAF_SET_EVENT(
            &pfunc->dt_event,
            FAPEXFDL_VIDEC_EVENT_DISP_INFO,
            FAPEXCFDLSAF_WAITEVENT_OR);
}

static void fdlvidec_ParamChangeNotifier(FAPI_VIDEC_FrameInfoStateT* nextInfoPtr,
                                         FAPI_VIDEC_FrameInfoStateT* lastInfoPtr,
                                         void* optDataPtr)
{
    FAPEX_VIDEC_DecFrameInfoStrT  *pFrameInfo;
    FAPEXFDL_VIDEC_StrT *pfunc = (FAPEXFDL_VIDEC_StrT *)optDataPtr;
    uint32_t  flags, chg_flag;

    pFrameInfo = &pfunc->frameInfo;

    if(pfunc->codec_type == FAPI_VIDEC_H264_HP41) {
        /* check bank type */
        if( (nextInfoPtr->decodedWidth  != lastInfoPtr->decodedWidth)
         || (nextInfoPtr->decodedHeight != lastInfoPtr->decodedHeight) ) {
            FAPI_VIDEC_GetStateT  params;
            int32_t  ret;

            ret = videc_funcStr.GetState(pfunc->pFAPICtxHdl, &params);
            if(ret == 0) {
                switch(params.decodeFrameBankSizeType) {
                    case FAPI_VIDEC_FRAME_BANK_SIZE_1920HD:
                        pfunc->pOffsetLut = FAPEXFDLVIDEC_BANK_OFFSET_1920HD;
                        pfunc->c_offset   = 0x00000230 << 12;
                        FAPEXDebugPrint(1, "bank type : 1920\n");
                        break;
                    case FAPI_VIDEC_FRAME_BANK_SIZE_1440HD:
                        pfunc->pOffsetLut = FAPEXFDLVIDEC_BANK_OFFSET_1440HD;
                        pfunc->c_offset   = 0x000001A4 << 12;
                        FAPEXDebugPrint(1, "bank type : 1440\n");
                        break;
                    case FAPI_VIDEC_FRAME_BANK_SIZE_1280HD:
                        pfunc->pOffsetLut = FAPEXFDLVIDEC_BANK_OFFSET_1280HD;
                        pfunc->c_offset   = 0x000000E6 << 12;
                        FAPEXDebugPrint(1, "bank type : 1280\n");
                        break;
                    case FAPI_VIDEC_FRAME_BANK_SIZE_SD:
                        pfunc->pOffsetLut = FAPEXFDLVIDEC_BANK_OFFSET_SD;
                        pfunc->c_offset   = 0x00000072 << 12;
                        FAPEXDebugPrint(1, "bank type : SD\n");
                        break;
                    default:
                        pfunc->pOffsetLut = FAPEXFDLVIDEC_BANK_OFFSET_1920HD;
                        pfunc->c_offset   = 0x00000230 << 12;
                        FAPEXDebugPrint(1, "bank type : 1920\n");
                }
            }
        }
    }
    if(!(pfunc->event_flags & FAPEX_VIDEC_NOTIFY_FLAGS_PARAM_CHANGE)) {
        if((lastInfoPtr->decodedWidth == 0) || (lastInfoPtr->decodedHeight == 0)) {
            /* 1st detected parameters */
            FAPEXCFDLSAF_SPIN_LOCK(&pfunc->lock, &flags);

            pFrameInfo->aspectRatio    = nextInfoPtr->aspectRatio;
            pFrameInfo->frameField     = nextInfoPtr->frameField;
            pFrameInfo->fieldInversion = nextInfoPtr->fieldInversion;
            pFrameInfo->decodedWidth   = nextInfoPtr->decodedWidth;
            pFrameInfo->decodedHeight  = nextInfoPtr->decodedHeight;

            FAPEXCFDLSAF_SPIN_UNLOCK(&pfunc->lock, &flags);
        }
        return;
    }

    chg_flag = 0;
    FAPEXCFDLSAF_SPIN_LOCK(&pfunc->lock, &flags);

    if(nextInfoPtr->aspectRatio != lastInfoPtr->aspectRatio) {
        chg_flag |= FAPEX_VIDEC_INFO_DISP_CHANGED_AR;
    }
    pFrameInfo->aspectRatio = nextInfoPtr->aspectRatio;

    if(nextInfoPtr->frameField  != lastInfoPtr->frameField) {
        chg_flag |= FAPEX_VIDEC_INFO_DISP_CHANGED_FRM_FLD;
    }
    pFrameInfo->frameField = nextInfoPtr->frameField;

    if(nextInfoPtr->fieldInversion  != lastInfoPtr->fieldInversion) {
        chg_flag |= FAPEX_VIDEC_INFO_DISP_CHANGED_FLD_INV;
    }
    pFrameInfo->fieldInversion = nextInfoPtr->fieldInversion;

    if( (nextInfoPtr->decodedWidth  != lastInfoPtr->decodedWidth)
     || (nextInfoPtr->decodedHeight != lastInfoPtr->decodedHeight) ) {
        chg_flag |= FAPEX_VIDEC_INFO_DISP_CHANGED_DEC_SIZE;
    }
    pFrameInfo->decodedWidth  = nextInfoPtr->decodedWidth;
    pFrameInfo->decodedHeight = nextInfoPtr->decodedHeight;

    pFrameInfo->enableFlag |= chg_flag;

    FAPEXCFDLSAF_SPIN_UNLOCK(&pfunc->lock, &flags);

    /* set event */
    FAPEXCFDLSAF_SET_EVENT(
            &pfunc->dt_event,
            FAPEXFDL_VIDEC_EVENT_DISP_CHANGE,
            FAPEXCFDLSAF_WAITEVENT_OR);

    FAPEXDebugPrint(1, "param changed %04x\n", chg_flag);
}

/*!
*******************************************************************************
**
** \brief  start VIDEC
**
** \param  pfunc       pointer to context
** \param  start_flag  #FAPEX_VIDEC_StartParamStrT parameter value
**
******************************************************************************/
static int32_t fdlvidec_Start(FAPEXFDL_VIDEC_StrT *pfunc, FAPEX_VIDEC_StartParamStrT *pArg)
{
    FAPEX_VIDEC_StartParamStrT  params;
    int32_t  ret = FAPI_OK;

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pArg, sizeof(FAPEX_VIDEC_StartParamStrT));

    /* setup output */
    if(params.outputMode == FAPI_VIDEC_ENABLE_FORCE_OFF) {
        ret = videc_funcStr.EnableDisable(
                    pfunc->pFAPICtxHdl,
                    (FAPI_VIDEC_EnableDisableStateEnumT)params.outputMode );
        if(ret < 0) {
            FAPEXDebugPrint(2, "[fdlvidec_Start] FAPI_VIDEC_EnableDisable error %d\n", ret);
        }
    }

    if(ret == FAPI_OK) {
        /* set PES parser */
        ret = videc_funcStr.SetPesParserToDecoder(
                    pfunc->pFAPICtxHdl,
                    (FAPI_VIDEC_PesParserEnumT)pfunc->pes_parser_no );
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdlvidec_Start] "
                               "FAPI_VIDEC_SetPesParserToDecoder error %d\n", ret);
            return ret;
        }

        /* set Playback environment */
        ret = videc_funcStr.SetPlaybackEnvironment(
                    pfunc->pFAPICtxHdl,
                    (FAPI_VIDEC_PlaybackEnvironmentEnumT)pfunc->playback_env);
        if(ret < 0) {
            FAPEXDebugPrint(2, "[fdlvidec_Start] "
                               "FAPI_VIDEC_SetPlaybackEnvironment error %d\n", ret);
            return ret;
        }
        if(pfunc->playback_env == FAPI_VIDEC_LIVE_PLAYBACK_ENVIRONMENT) {
            ret = videc_funcStr.SetSyncBase(pfunc->pFAPICtxHdl,
                                            FAPI_VIDEC_SYNC_PCR_BASED);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_VIDEC_SetSyncBase error %d\n", ret);
                return ret;
            }

            /* enable STC synchronization */
            sync_funcStr.ResetStatus();

            ret = sync_funcStr.SetEnvironment(FAPI_SYNC_PCR_BASED);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_SYNC_SetEnvironment error %d\n", ret);
                return ret;
            }
            ret = sync_funcStr.SetStcRegulation(1);
        }
        else if(pfunc->playback_env == FAPI_VIDEC_STREAMING_PLAYBACK_ENVIRONMENT) {
            ret = videc_funcStr.SetSyncBase(pfunc->pFAPICtxHdl,
                                            FAPI_VIDEC_SYNC_STREAMING_PCR_BASED);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_VIDEC_SetSyncBase error %d\n", ret);
                return ret;
            }

            /* enable STC synchronization */
            sync_funcStr.ResetStatus();

            ret = sync_funcStr.SetEnvironment(FAPI_SYNC_STREAMING_PTS_BASED);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_SYNC_SetEnvironment error %d\n", ret);
                return ret;
            }
            ret = sync_funcStr.SetStcRegulation(1);
        }
 
        else {
            uint32_t  sync_env;

            /* disable STC synchronization */
            if(pfunc->playback_env == FAPI_VIDEC_PVR_PLAYBACK_ENVIRONMENT) {
                sync_env = FAPI_SYNC_VIDEO_PTS_BASED;
            }
            else {
                sync_env = FAPI_SYNC_AUDIO_PTS_BASED;
            }
            ret = sync_funcStr.SetEnvironment(sync_env);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_SYNC_SetEnvironment error %d\n", ret);
                return ret;
            }
            ret = sync_funcStr.SetStcRegulation(0);
        }
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdlvidec_Start] "
                               "FAPI_SYNC_SetStcRegulation error %d\n", ret);
            return ret;
        }

        if(params.outputMode != FAPI_VIDEC_ENABLE_FORCE_OFF) {
            /* set Playback Mode */
            ret = videc_funcStr.SetPlaybackMode(
                        pfunc->pFAPICtxHdl,
                        (FAPI_VIDEC_PlaybackModeEnumT)params.playbackMode );
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_VIDEC_SetPlaybackMode error %d\n", ret);
                return ret;
            }
            pfunc->playback_mode = params.playbackMode;
        }

        if(pfunc->codec_type == FAPI_VIDEC_MPEG2_MPHL) {
            /* stream is input from TSD buffer */
            ret = videc_funcStr.SetDirectIFrameMode(
                        pfunc->pFAPICtxHdl,
                        FAPI_VIDEC_NON_DIRECT_IFRAME_MODE);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_VIDEC_SetDirectIFrameMode error %d\n", ret);
                return ret;
            }
        }

        /* clear the decoded video frame information */
        memset(&pfunc->frameInfo, 0, sizeof(FAPEX_VIDEC_DecFrameInfoStrT));
        pfunc->frameInfo.aspectRatio = FAPEX_VIDEC_AR_UNKNOWN;

        /* setup callbacks */
        pfunc->event_flags = 0;

        if(params.flags & FAPEX_VIDEC_NOTIFY_FLAGS_DEC_AVAILABLE) {
            ret = videc_funcStr.SetFrameDecodedCallback(pfunc->pFAPICtxHdl,
                                                        fdlvidec_DecAvailableNotifier, pfunc);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_VIDEC_SetFrameDecodedCallback error %d\n", ret);
                return ret;
            }
            pfunc->event_flags |= FAPEX_VIDEC_NOTIFY_FLAGS_DEC_AVAILABLE;
        }
        if(params.flags & FAPEX_VIDEC_NOTIFY_FLAGS_DISP_AVAILABLE) {
            ret = videc_funcStr.SetFrameAvailableCallback(pfunc->pFAPICtxHdl,
                                                          fdlvidec_DispAvailableNotifier, pfunc);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_VIDEC_SetFrameAvailableCallback error %d\n", ret);
                return ret;
            }
            pfunc->event_flags |= FAPEX_VIDEC_NOTIFY_FLAGS_DISP_AVAILABLE;
        }
        if(params.flags & FAPEX_VIDEC_NOTIFY_FLAGS_FRAME_INFO) {
            ret = videc_funcStr.SetVideoDecInfoCallback(pfunc->pFAPICtxHdl,
                                                        fdlvidec_FrameInfoNotifier, pfunc);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                   "FAPI_VIDEC_SetVideoDecInfoCallback error %d\n", ret);
                return ret;
            }
            pfunc->event_flags |= FAPEX_VIDEC_NOTIFY_FLAGS_FRAME_INFO;
        }
        ret = videc_funcStr.SetParamChangeCallback(pfunc->pFAPICtxHdl,
                                                   fdlvidec_ParamChangeNotifier, pfunc);
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdlvidec_Start] "
                               "FAPI_VIDEC_SetParamChangeCallback error %d\n", ret);
            return ret;
        }
        pfunc->event_flags |= (params.flags & FAPEX_VIDEC_NOTIFY_FLAGS_PARAM_CHANGE);

        /* Video Decoder start */
        ret = videc_funcStr.Start(pfunc->pFAPICtxHdl);
        if(ret == FAPI_OK) {
            pfunc->status = FAPEX_STATUS_RUNNING;

            if(params.outputMode != FAPI_VIDEC_ENABLE_FORCE_OFF) {
                ret = videc_funcStr.EnableDisable(
                            pfunc->pFAPICtxHdl,
                            (FAPI_VIDEC_EnableDisableStateEnumT)params.outputMode );
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[fdlvidec_Start] "
                                       "FAPI_VIDEC_EnableDisable error %d\n", ret);
                }
            }
        }
        else {
            FAPEXDebugPrint(2, "[fdlvidec_Start] FAPI_VIDEC_Start error %d\n", ret);
        }
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  stop VIDEC
**
** \param  pfunc       pointer to context
**
******************************************************************************/
static int32_t fdlvidec_Stop(FAPEXFDL_VIDEC_StrT *pfunc)
{
    int32_t  ret;

    /* disable output */
    ret = videc_funcStr.EnableDisable(pfunc->pFAPICtxHdl, FAPI_VIDEC_ENABLE_FORCE_OFF);
    if(ret < 0) {
        return ret;
    }
    /* disable auto scaler */
    ret = videc_funcStr.ClearAutoScaleCallback(pfunc->pFAPICtxHdl);
    if(ret < 0) {
        FAPEXDebugPrint(2, "[fdlvidec_Stop] "
                           "FAPI_VIDEC_ClearAutoScaleCallback error %d\n", ret);
        return ret;
    }

    /* stop Video decoder */
    ret = videc_funcStr.Stop(pfunc->pFAPICtxHdl);
    if(ret == FAPI_OK) {
        pfunc->status = FAPEX_STATUS_READY;
    }
    else {
        FAPEXDebugPrint(2, "[fdlvidec_Stop] FAPI_VIDEC_Stop error %d\n", ret);
        return ret;
    }

    if(pfunc->event_flags != 0) {
        /* set end event */
        pfunc->frameInfo.enableFlag |= FAPEX_VIDEC_INFO_STOPPED;

        ret = FAPEXCFDLSAF_SET_EVENT(&pfunc->dt_event,
                                     FAPEXFDL_VIDEC_EVENT_END,
                                     FAPEXCFDLSAF_WAITEVENT_OR);
        if(ret < 0) {
            return FAPEX_VIDEC_ERR_INVALID_STATUS;
        }

        /* clear callbacks */
        ret = videc_funcStr.ClearFrameDecodedCallback(pfunc->pFAPICtxHdl);
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdlvidec_Stop] "
                               "FAPI_VIDEC_ClearFrameDecodedCallback error %d\n", ret);
            return ret;
        }

        ret = videc_funcStr.ClearFrameAvailableCallback(pfunc->pFAPICtxHdl);
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdlvidec_Stop] "
                               "FAPI_VIDEC_ClearFrameAvailableCallback error %d\n", ret);
            return ret;
        }

        ret = videc_funcStr.ClearVideoDecInfoCallback(pfunc->pFAPICtxHdl);
        if(ret < 0) {
            FAPEXDebugPrint(2, "[fdlvidec_Stop] "
                               "FAPI_VIDEC_ClearVideoDecInfoCallback error %d\n", ret);
            return ret;
        }

        ret = videc_funcStr.ClearParamChangeCallback(pfunc->pFAPICtxHdl);
        if(ret != FAPI_OK) {
            FAPEXDebugPrint(2, "[fdlvidec_Stop] "
                               "FAPI_VIDEC_ClearParamChangeCallback error %d\n", ret);
        }
        pfunc->event_flags = 0;
    }

    /* This is necessary for PTS-based sync */
    sync_funcStr.ResetStatus();

    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIDEC descriptor open
**
** \param  pfunc    pointer to context
** \param  pArg     pointer to #FAPEX_VIDEC_OpenParamStrT parameter
**
******************************************************************************/
static int32_t fdlvidec_Open(FAPEXFDL_VIDEC_StrT *pfunc, FAPEX_VIDEC_OpenParamStrT *pArg)
{
    FAPI_VIDEC_OpenParamsT     params;
    FAPI_VIDEC_PesParserEnumT  parser_no;
    FAPEXFDL_TSD_SubStrT       *pfunc_tsd;
    FAPI_BM_InfoT              bm_info;
    FAPEXCFDLSAF_EventStrT     *dt_event_ptr;
    int32_t  ret = 0;

    if(pfunc->pFAPICtxHdl != NULL) {
        return 0;
    }
    if(pArg == NULL) {
        return FAPEX_VIDEC_ERR_BAD_PARAMETER;
    }
#if (PROFILE > 0)
    videc_funcStr.ReportMemory();
#endif

    /* copy parameters from an user space to the kernel space */
    params.version       = FAPI_VIDEC_VERSION;
    params.decoderNumber = FAPI_VIDEC_DECODER_NUMBER_0;
    FAPEXCFDLSAF_COPY_TO_DRV(&(pfunc->input_id),
                             &(pArg->inputId),
                             sizeof(int32_t));
    FAPEXCFDLSAF_COPY_TO_DRV(&(params.decoderType),
                             &(pArg->codecType),
                             sizeof(FAPI_VIDEC_DecoderEnumT));
    FAPEXCFDLSAF_COPY_TO_DRV(&(params.decoderPlaybackEnvironment),
                             &(pArg->playbackEnv),
                             sizeof(FAPI_VIDEC_PlaybackEnvironmentEnumT));
    FAPEXCFDLSAF_COPY_TO_DRV(&(params.decoderAfdState),
                             &(pArg->afdFlag),
                             sizeof(FAPI_VIDEC_AfdStateEnumT));
    FAPEXCFDLSAF_COPY_TO_DRV(&(params.decoderDelimiterType),
                             &(pArg->delimiterType),
                             sizeof(FAPI_VIDEC_DelimiterEnumT));
    params.decoderSecurityMode     = FAPI_VIDEC_SECURITY_MODE_UNSECURED; /* tentative */
    params.decodeFrameBankSizeType = FAPI_VIDEC_FRAME_BANK_SIZE_1920HD;
    params.decoderPlaybackMode     = FAPI_VIDEC_NORMAL_PLAYBACK_MODE;

    /* call Open FAPI */
    pfunc->pFAPICtxHdl = (void*)videc_funcStr.Open(&params, &ret);
    FAPEXDebugPrint(3, "[fdlvidec_Open] FAPI context:0x%08x\n",
                       (uint32_t)pfunc->pFAPICtxHdl);
    if(pfunc->pFAPICtxHdl == NULL) {
        FAPEXDebugPrint(2, "[fdlvidec_Open] FAPI_VIDEC_Open error:%d\n", ret);
        return ret;
    }
    pfunc->codec_type   = params.decoderType;
    pfunc->playback_env = params.decoderPlaybackEnvironment;

    /* default bank information */
    pfunc->pOffsetLut = FAPEXFDLVIDEC_BANK_OFFSET_1920HD;
    pfunc->c_offset   = 0x00000230 << 12;
    pfunc->bank_size = (pfunc->pOffsetLut[21] + pfunc->pOffsetLut[1]) << 12;
#ifdef __LINUX__
    pfunc->bank_size = (pfunc->bank_size + ~PAGE_MASK) & PAGE_MASK;
#endif

    ret = FAPEX_VIDEC_ERR_INVALID_STATUS;

    /* set input stream buffer id */
    pfunc_tsd = (FAPEXFDL_TSD_SubStrT*)FAPEXCFDL_GetFdlHdl(pfunc->pCfdlHdl, pfunc->input_id);

    if(pfunc_tsd == NULL) {
        FAPEXDebugPrint(2, "[fdlvidec_Open] tsd channel for videc is not opened yet %d\n", ret);
        return ret;
    }
    pfunc_tsd->output_id = pfunc->id;

    if(pfunc_tsd->pBMCtxHdl == NULL) {
        FAPEXDebugPrint(2, "[fdlvidec_Open] bm channel for videc is not opened yet %d\n", ret);
        return ret;
    }

    /* get BM information */
    ret = bm_funcStr.GetInfo(pfunc_tsd->pBMCtxHdl, &bm_info);
    if(ret != FAPI_OK) {
        FAPEXDebugPrint(2, "[fdlvidec_Open] FAPI_BM_GetInfo error %d\n", ret);
        return ret;
    }

    /* set BM channel */
    ret = videc_funcStr.SetChannelId(pfunc->pFAPICtxHdl,
                                     bm_info.blockIndex, bm_info.channelId);
    if(ret != FAPI_OK) {
        FAPEXDebugPrint(2, "[fdlvidec_Open] FAPI_VIDEC_SetChannelId error %d\n", ret);
        return ret;
    }

    /* set PES parser */
    switch(pfunc->codec_type) {
        case FAPI_VIDEC_MPEG2_MPHL:
            ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                            videc_funcStr.AudCallback,
                                            pfunc->pFAPICtxHdl,
                                            0x000000B3, 0x000000FF);
            if(ret != FAPI_OK) {
                break;
            }
            ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                            videc_funcStr.AudCallback,
                                            pfunc->pFAPICtxHdl,
                                            0x00000000, 0x000000FF);
            break;

        case FAPI_VIDEC_H264_HP41:
            if(params.decoderDelimiterType == FAPI_VIDEC_DELIMITER_PES_HEADER) {
                ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                                videc_funcStr.AudCallback,
                                                pfunc->pFAPICtxHdl,
                                                0x000000E0, 0x000000F0);
            }
            else {
                ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                                videc_funcStr.AudCallback,
                                                pfunc->pFAPICtxHdl,
                                                0x00000009, 0x000000FF);
            }
            break;

        case FAPEX_VIDEC_CODEC_TYPE_AVS:
            ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                            videc_funcStr.AudCallback,
                                            pfunc->pFAPICtxHdl,
                                            0x000000B0, 0x000000FF);
            if(ret != FAPI_OK) {
                break;
            }
            ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                            videc_funcStr.AudCallback,
                                            pfunc->pFAPICtxHdl,
                                            0x000000B3, 0x000000FF);
            if(ret != FAPI_OK) {
                break;
            }
            ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                            videc_funcStr.AudCallback,
                                            pfunc->pFAPICtxHdl,
                                            0x000000B6, 0x000000FF);
            break;
        
        case FAPEX_VIDEC_CODEC_TYPE_VC1:
            ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                            videc_funcStr.AudCallback,
                                            pfunc->pFAPICtxHdl,                                
                                            0x0000000F,0x000000FF);
            if(ret != FAPI_OK) {
                break;
            }
            ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                            videc_funcStr.AudCallback,
                                            pfunc->pFAPICtxHdl,
                                            0x0000000D,0x000000FF);                     
            if(ret != FAPI_OK) {
                break;
            }
            ret = bm_funcStr.SetAudCallback(pfunc_tsd->pBMCtxHdl,
                                            videc_funcStr.AudCallback,
                                            pfunc->pFAPICtxHdl,
                                            0x0000000C,0x000000FF);                     
            if(ret != FAPI_OK) {
                break;
            }
  
    }
    if(ret != FAPI_OK) {
        FAPEXDebugPrint(2, "[fdlvidec_Open] FAPI_BM_SetAudCallback error %d\n", ret);
        return ret;
    }

    if(pfunc_tsd->id < FAPEXDEV_TSD2_SUB_MINOR_FIRST) {
        parser_no = FAPI_VIDEC_TSD_PESPARSER_0; /*for H61 integrated MANGO*/
    }
    else {
        parser_no = FAPI_VIDEC_TSD_PESPARSER_0; /*for H61 integrated MANGO*/
    }
    parser_no += pfunc_tsd->pes_paser_no;
    pfunc->pes_parser_no = parser_no;

    /* setup event parameter */
    dt_event_ptr = &pfunc->dt_event;
    if(dt_event_ptr->initialized == 0) {
        FAPEXCFDLSAF_CREATE_EVENT(dt_event_ptr);
    }
    pfunc->frameInfo.enableFlag = 0;

    FAPEXCFDLSAF_SPIN_LOCK_CREATE(&pfunc->lock);

    FAPEXDebugPrint(3, "[fdlvidec_Open] Setup OK\n");
    pfunc->status = FAPEX_STATUS_OPENED;

    return 0;
}

/*!
*******************************************************************************
**
** \brief  VIDEC get status information
**
** \param  pfunc    pointer to context
** \param  pArg     pointer to FAPEX_VIDEC_StatusStrT parameter
**
******************************************************************************/
static int32_t fdlvidec_getStatus(FAPEXFDL_VIDEC_StrT *pfunc, FAPEX_VIDEC_StatusStrT *pArg)
{
    FAPEX_VIDEC_StatusStrT  status;
    FAPI_VIDEC_GetStateT    params;
    int32_t  ret;

    ret = videc_funcStr.GetState(pfunc->pFAPICtxHdl, &params);
    if(ret == 0) {
        status.firmwareVersionInFlash   = params.firmwareVersionInFlash;
        status.firmwareVersionSupported = params.firmwareVersionSupported;
        status.decoderType              = params.decoderType;
        status.decodeFrameBankSizeType  = params.decodeFrameBankSizeType;
        status.displayWidth             = params.displayWidth;
        status.displayHeight            = params.displayHeight;
        status.decodedWidth             = params.decodedWidth;
        status.decodedHeight            = params.decodedHeight;
        status.aspectRatio              = params.aspectRatio;
        status.averageBitRate           = params.vesAverageBitRate;
        status.currBitRate              = params.vesCurrBitRate;
        status.frameRate                = params.vesFrameRate;
        status.bootState                = params.bootState;
        status.resetState               = params.resetState;
        status.enableState              = params.enableState;
        status.startState               = params.startState;
        status.enableState              = params.enableState;
        status.freezeState              = params.freezeState;
        status.healthState              = params.healthState;

        /* copy status from the kernel space to an user space */
        FAPEXCFDLSAF_COPY_TO_USR(pArg, &status, sizeof(FAPEX_VIDEC_StatusStrT));
    }
    else {
        FAPEXDebugPrint(2, "[fdlvidec_getStatus] FAPI_VIDEC_GetState error %d\n", ret);
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIDEC set output id
**
** \param  pfunc    pointer to context
** \param  pArg     pointer to output_id parameter
**
******************************************************************************/
static int32_t fdlvidec_setOutput(FAPEXFDL_VIDEC_StrT *pfunc, int32_t *pArg)
{
    FAPEXFDL_VIOUT_SubStrT  *pfunc_viout;
    FAPEXFDL_VIENC_SubStrT  *pfunc_vienc;
    int32_t  output_id;
    int32_t  i, setup_max_num;
    int32_t  vienc_setup_flag, viout_setup_flag;
    int32_t  ret;

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&output_id, pArg, sizeof(uint32_t));

    /* check output_id */
    if(output_id == FAPEXDEV_VIOUT_CMN_MINOR) {
        output_id     = FAPEXDEV_VIOUT_SUB_MINOR_FIRST;
        setup_max_num = FAPEXDEV_VIOUT_SUB_DESCRIPTOR_NUM;
    }
    else if( (output_id >= FAPEXDEV_VIOUT_SUB_MINOR_FIRST)
          && (output_id <= FAPEXDEV_VIOUT_SUB_MINOR_LAST) ) {
        setup_max_num = 1;
    }
    else {
        FAPEXDebugPrint(2, "[fdlvidec_setOutput] id error\n");
        return FAPEX_VIDEC_ERR_BAD_PARAMETER;
    }

    /* setup output */
    vienc_setup_flag = viout_setup_flag = 0;
    for(i=0; i<setup_max_num; i++) {
        pfunc_viout = (FAPEXFDL_VIOUT_SubStrT*)FAPEXCFDL_GetFdlHdl(pfunc->pCfdlHdl,
                                                                   output_id);
        if(pfunc_viout != NULL) {
            if( (pfunc_viout->scaler_type == FAPI_VISCALE_SCALER_VIDEO)
             || (pfunc_viout->scaler_type == FAPI_VISCALE_SCALER_SD) ) {

                if(vienc_setup_flag == 0) {
                    /* set link to VIENC(HD) */
                    if(pfunc_viout->pViencHdl != NULL) {
                        pfunc_vienc = (FAPEXFDL_VIENC_SubStrT*)pfunc_viout->pViencHdl;

                        /* link to Vienc */
                        ret = videc_funcStr.SetDecoderToHdEncoder(pfunc->pFAPICtxHdl,
                                                                  pfunc_vienc->pFAPICtxHdl);
                        if(ret == FAPI_OK) {
                            vienc_setup_flag = 1;
                        }
                        else {
                            FAPEXDebugPrint(2, "[fdlvidec_setOutput] "
                                               "FAPI_VIDEC_SetDecoderToHdEncoder error %d\n", ret);
                        }
                    }
                }

                /* set VIOUT auto scaler */
                ret = videc_funcStr.SetAutoScaleCallback(pfunc->pFAPICtxHdl,
                                                         viscale_funcStr.AutoScaler,
                                                         pfunc_viout->pFAPICtxHdl);
                if(ret == FAPI_OK) {
                    viout_setup_flag ++;
                }
                else {
                    FAPEXDebugPrint(2, "[fdlvidec_setOutput] "
                                       "FAPI_VIDEC_SetAutoScaleCallback error %d %d\n",
                                       ret, output_id);
                }
            }
        }
        output_id ++;
    }
    /* check setup status */
    if(viout_setup_flag > 0) {
        FAPEXDebugPrint(3, "[fdlvidec_setOutput] Setup OK %d %d\n", viout_setup_flag, vienc_setup_flag);
        pfunc->status = FAPEX_STATUS_READY;
        return 0;
    }
    else {
        FAPEXDebugPrint(2, "[fdlvidec_setOutput] Setup failed.\n");
        return FAPEX_VIDEC_ERR_INVALID_STATUS;
    }
}

/*!
*******************************************************************************
**
** \brief  VIDEC get Video ES parameter
**
** \param  pfunc    pointer to context
** \param  pArg     pointer to #FAPEX_VIDEC_VESParamStrT parameter
**
******************************************************************************/
static int32_t fdlvidec_getVesParams(FAPEXFDL_VIDEC_StrT *pfunc, FAPEX_VIDEC_VESParamStrT *pArg)
{
    FAPEX_VIDEC_VESParamStrT  params;
    FAPI_VIDEC_VESParamsT     ves_params;
    int32_t  ret;

    /* get Video ES parameters */
    ret = videc_funcStr.GetVesParams(pfunc->pFAPICtxHdl, (FAPI_VIDEC_VESParamsT *)&ves_params,
                                     FAPI_SYS_SUSPEND);
    if(ret == FAPI_OK) {
        params.profile        = ves_params.vesProfile;
        params.level          = ves_params.vesLevel;
        params.averageBitRate = ves_params.vesAverageBitRate;
        params.currBitRate    = ves_params.vesCurrBitRate;
        params.aspectRatio    = ves_params.vesAspectRatio;
        params.frameRate      = ves_params.vesFrameRate;
        params.decoderLatency = ves_params.vesDecoderLatency;
        params.frameField     = ves_params.proggInter;
        params.displayWidth   = ves_params.displayWidth;
        params.displayHeight  = ves_params.displayHeight;
        params.decodedWidth   = ves_params.decodedWidth;
        params.decodedHeight  = ves_params.decodedHeight;
        params.sarWidth       = ves_params.sarWidth;
        params.sarHeight      = ves_params.sarHeight;
    }
    else {
        params.profile        = FAPEX_VIDEC_PROFILE_UNKNOWN;
        params.level          = FAPEX_VIDEC_LEVEL_UNKNOWN;
        params.averageBitRate = 0;
        params.currBitRate    = 0;
        params.aspectRatio    = FAPEX_VIDEC_AR_UNKNOWN;
        params.frameRate      = 0;
        params.decoderLatency = 0;
        params.frameField     = 0;
        params.displayWidth   = 0;
        params.displayHeight  = 0;
        params.decodedWidth   = 0;
        params.decodedHeight  = 0;
        params.sarWidth       = 0;
        params.sarHeight      = 0;

        FAPEXDebugPrint(2, "[fdlvidec_getVesParams] FAPI_VIDEC_GetVesParams error %d\n", ret);
    }
    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(pArg, &params, sizeof(FAPEX_VIDEC_VESParamStrT));

    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIDEC get display frame information
**
** \param  pfunc    pointer to context
** \param  pArg     pointer to #FAPEX_VIDEC_DecFrameInfoStrT parameter
**
******************************************************************************/
static int32_t fdlvidec_getDispFrameInfo(FAPEXFDL_VIDEC_StrT *pfunc, FAPEX_VIDEC_DecFrameInfoStrT *pArg)
{
    FAPEX_VIDEC_DecFrameInfoStrT  *pFrameInfo;
    FAPI_VIDEC_GetStateT          params;
    FAPEXCFDLSAF_EventStrT        *dt_event_ptr;
    uint32_t  temp_data;
    int32_t   ret, result;

    pFrameInfo = &pfunc->frameInfo;

    if(pfunc->event_flags != 0) {
        if(pFrameInfo->enableFlag == 0) {
            /* check event */
            dt_event_ptr = &pfunc->dt_event;
            dt_event_ptr->mode       = FAPEXCFDLSAF_WAITEVENT_SYNC | FAPEXCFDLSAF_WAITEVENT_OR;
            dt_event_ptr->wai_bitptn = FAPEXFDL_VIDEC_EVENT_FRAME_INFO | FAPEXFDL_VIDEC_EVENT_END;
            dt_event_ptr->tout       = 0;

            result = FAPEXCFDLSAF_WAIT_EVENT(dt_event_ptr);
        }
        else {
            result = 0;
        }

        if(result != FAPI_ERR_TIMEOUT) {
            uint32_t  flags;

            FAPEXCFDLSAF_SPIN_LOCK(&pfunc->lock, &flags);

            /* copy parameters from the kernel space to an user space */
            FAPEXCFDLSAF_COPY_TO_USR(pArg, pFrameInfo, sizeof(FAPEX_VIDEC_DecFrameInfoStrT));

            pFrameInfo->enableFlag = 0;

            FAPEXCFDLSAF_SPIN_UNLOCK(&pfunc->lock, &flags);

            return 0;
        }
        else {
            ret = FAPEX_VIDEC_ERR_TIMEOUT;
        }
    }
    else {
        ret = FAPEX_VIDEC_ERR_INVALID_STATUS;
    }

    result = videc_funcStr.GetState(pfunc->pFAPICtxHdl, &params);
    if(result < 0) {
        FAPEXDebugPrint(2, "[fdlvidec_getDispFrameInfo] FAPI_VIDEC_GetState error %d\n", result);
        params.aspectRatio   = FAPI_VIDEC_AR_NONE;
        params.decodedWidth  = 0;
        params.decodedHeight = 0;
        params.displayWidth  = 0;
        params.displayHeight = 0;
    }
    temp_data = 0;

    /* copy parameters from the kernel space to an user space */
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->enableFlag,     &temp_data,            sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->aspectRatio,    &params.aspectRatio,   sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->frameField    , &temp_data,            sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->fieldInversion, &temp_data,            sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->decodedWidth,   &params.decodedWidth,  sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->decodedHeight,  &params.decodedHeight, sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->displayWidth,   &params.displayWidth,  sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->displayHeight,  &params.displayHeight, sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->offset_y,       &temp_data,            sizeof(uint32_t));
    FAPEXCFDLSAF_COPY_TO_USR(&pArg->offset_c,       &temp_data,            sizeof(uint32_t));

    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIDEC set Playback environment
**
** \param  pfunc    pointer to context
** \param  mode     value of #FAPI_VIDEC_PlaybackEnvironmentEnumT
**
******************************************************************************/
static int32_t fdlvidec_setPlaybackEnv(FAPEXFDL_VIDEC_StrT *pfunc,
                                       FAPI_VIDEC_PlaybackEnvironmentEnumT mode)
{
    int32_t  ret;

    if(pfunc->status == FAPEX_STATUS_RUNNING) {
        return FAPEX_VIDEC_ERR_INVALID_STATUS;
    }

    if( (mode != FAPI_VIDEC_LIVE_PLAYBACK_ENVIRONMENT)
     && (mode != FAPI_VIDEC_PVR_PLAYBACK_ENVIRONMENT)
     && (mode != FAPI_VIDEC_VOD_PLAYBACK_ENVIRONMENT) 
     && (mode != FAPI_VIDEC_STREAMING_PLAYBACK_ENVIRONMENT)) {
        FAPEXDebugPrint(2, "[fdlvidec_setPlaybackEnv] mode error\n");
        return FAPEX_VIDEC_ERR_BAD_PARAMETER;
    }

    if(pfunc->playback_env == mode) {
        return 0;
    }

    /* clear link to Pes parser */
    ret = videc_funcStr.ClearPesParserToDecoder(pfunc->pFAPICtxHdl);
    if(ret < 0) {
        FAPEXDebugPrint(2, "[fdlvidec_setPlaybackEnv] "
                           "FAPI_VIDEC_ClearPesParserToDecoder error %d\n", ret);
        return ret;
    }
    pfunc->playback_env = mode;

    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIDEC set Playback mode
**
** \param  pfunc    pointer to context
** \param  mode     value of #FAPI_VIDEC_PlaybackModeEnumT
**
******************************************************************************/
static int32_t fdlvidec_setPlaybackMode(FAPEXFDL_VIDEC_StrT *pfunc,
                                        FAPI_VIDEC_PlaybackModeEnumT mode)
{
    int32_t  ret = 0;

    switch(mode) {
        case FAPI_VIDEC_NORMAL_PLAYBACK_MODE:
            /* clear link to Pes parser */
            ret = videc_funcStr.ClearPesParserToDecoder(pfunc->pFAPICtxHdl);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_VIDEC_ClearPesParserToDecoder error %d\n", ret);
                return ret;
            }

            if(pfunc->playback_env != FAPI_VIDEC_LIVE_PLAYBACK_ENVIRONMENT) {
                ret = sync_funcStr.SetStcRegulation(0);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                       "FAPI_SYNC_SetStcRegulation error %d\n", ret);
                    return ret;
                }
            }

            /* set Playback environment */
            ret = videc_funcStr.SetPlaybackEnvironment(
                        pfunc->pFAPICtxHdl,
                        (FAPI_VIDEC_PlaybackEnvironmentEnumT)pfunc->playback_env);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_VIDEC_SetPlaybackEnvironment error %d\n", ret);
                return ret;
            }

            ret = videc_funcStr.SetPesParserToDecoder(
                        pfunc->pFAPICtxHdl,
                        (FAPI_VIDEC_PesParserEnumT)pfunc->pes_parser_no);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_VIDEC_SetPesParserToDecoder error %d\n", ret);
                return ret;
            }

            if(pfunc->playback_env == FAPI_VIDEC_LIVE_PLAYBACK_ENVIRONMENT) {
                sync_funcStr.ResetStatus();
                ret = sync_funcStr.SetStcRegulation(1);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                       "FAPI_SYNC_SetStcRegulation error %d\n", ret);
                }
            }
            break;

        case FAPI_VIDEC_PLAYBACK_PAUSE:
            /* clear link to Pes parser */
            ret = videc_funcStr.ClearPesParserToDecoder(pfunc->pFAPICtxHdl);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_VIDEC_ClearPesParserToDecoder error %d\n", ret);
                return ret;
            }

            ret = sync_funcStr.SetStcRegulation(0);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_SYNC_SetStcRegulation error %d\n", ret);
                return ret;
            }

            ret = videc_funcStr.SetPlaybackEnvironment(pfunc->pFAPICtxHdl,
                                                       FAPI_VIDEC_PVR_PLAYBACK_ENVIRONMENT);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_VIDEC_SetPlaybackEnvironment error %d\n", ret);
            }
            break;

        case FAPI_VIDEC_FIRST_I_ONLY_PLAYBACK_MODE:
        case FAPI_VIDEC_I_ONLY_PLAYBACK_MODE:
        case FAPI_VIDEC_I_AND_P_ONLY_PLAYBACK_MODE:
            break;

        case FAPI_VIDEC_SLOW_FORWARD_2X_PLAYBACK_MODE:
        case FAPI_VIDEC_SLOW_FORWARD_4X_PLAYBACK_MODE:
        case FAPI_VIDEC_SLOW_FORWARD_8X_PLAYBACK_MODE:
            /* clear link to Pes parser */
            ret = videc_funcStr.ClearPesParserToDecoder(pfunc->pFAPICtxHdl);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_VIDEC_ClearPesParserToDecoder error %d\n", ret);
                return ret;
            }

            ret = sync_funcStr.SetStcRegulation(0);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_SYNC_SetStcRegulation error %d\n", ret);
                return ret;
            }

            ret = videc_funcStr.SetPlaybackEnvironment(pfunc->pFAPICtxHdl,
                                                       FAPI_VIDEC_PVR_PLAYBACK_ENVIRONMENT);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_VIDEC_SetPlaybackEnvironment error %d\n", ret);
            }
            break;

        default:
            FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] mode error\n");
            return FAPEX_VIDEC_ERR_BAD_PARAMETER;
    }

    if(ret == 0) {
        ret = videc_funcStr.SetPlaybackMode(pfunc->pFAPICtxHdl, mode);
        if(ret == FAPI_OK) {
            pfunc->playback_mode = mode;
        }
        else {
            FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                               "FAPI_VIDEC_SetPlaybackMode error %d\n", ret);
            return ret;
        }

        if(pfunc->codec_type == FAPI_VIDEC_MPEG2_MPHL) {
            /* stream is input from TSD buffer */
            ret = videc_funcStr.SetDirectIFrameMode(pfunc->pFAPICtxHdl,
                                                  FAPI_VIDEC_NON_DIRECT_IFRAME_MODE);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[fdlvidec_setPlaybackMode] "
                                   "FAPI_VIDEC_SetDirectIFrameMode error %d\n", ret);
            }
        }
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIDEC decode single I-frame
**
** \param  pfunc    pointer to context
** \param  pEsData  pointer to #FAPEX_VIDEC_ESdataParamStrT parameter
**
******************************************************************************/
static int32_t fdlvidec_decodeSingleIframe(FAPEXFDL_VIDEC_StrT *pfunc,
                                           FAPEX_VIDEC_ESdataParamStrT *pEsData)
{
    FAPEX_VIDEC_ESdataParamStrT params;
    uint32_t  *data;
    int32_t   ret;
#ifdef __LINUX__
    uint32_t  order;
    struct page *buff_pg;
#endif

    /* stream is input from non TSD buffer */
    ret = videc_funcStr.SetDirectIFrameMode(pfunc->pFAPICtxHdl,
                                            FAPI_VIDEC_DIRECT_IFRAME_MODE);
    if(ret < 0) {
        FAPEXDebugPrint(2, "[fdlvidec_decodeSingleIframe] "
                           "FAPI_VIDEC_SetDirectIFrameMode error %d\n", ret);
        return ret;
    }

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pEsData, sizeof(FAPEX_VIDEC_ESdataParamStrT));

#ifdef __LINUX__
    order = PAGE_SHIFT;
    while( (order < 31) && (params.size > (1 << order)) ) {
        order ++;
    }

    buff_pg = alloc_pages(GFP_HIGHUSER, (order - PAGE_SHIFT));

    if(buff_pg == NULL) {
        return FAPEX_VIDEC_ERR_OUT_OF_MEMORY;
    }
    else {
        data = (uint32_t *)page_address(buff_pg);

        FAPEXCFDLSAF_COPY_TO_DRV(data, params.data, params.size);
#else
        data = params.data;
#endif
        ret = videc_funcStr.DecodeSingleDirectIFrame(pfunc->pFAPICtxHdl,
                                                     data, params.size);
        if(ret < 0) {
            FAPEXDebugPrint(2, "[fdlvidec_decodeSingleIframe] "
                               "FAPI_VIDEC_DecodeSingleDirectIFrame error %d\n", ret);
        }
#ifdef __LINUX__
        __free_pages(buff_pg, (order - PAGE_SHIFT));
    }
#endif
    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIDEC decode sequence I-frame
**
** \param  pfunc    pointer to context
** \param  pEsData  pointer to #FAPEX_VIDEC_ESdataParamStrT parameter
**
******************************************************************************/
static int32_t fdlvidec_decodeSequenceIframe(FAPEXFDL_VIDEC_StrT *pfunc,
                                             FAPEX_VIDEC_ESdataParamStrT *pEsData)
{
    FAPEX_VIDEC_ESdataParamStrT params;
    uint32_t  *data;
    int32_t   ret;

    /* stream is input from non TSD buffer */
    ret = videc_funcStr.SetDirectIFrameMode(pfunc->pFAPICtxHdl,
                                            FAPI_VIDEC_DIRECT_IFRAME_MODE);
    if(ret < 0) {
        FAPEXDebugPrint(2, "[fdlvidec_decodeSequenceIframe] "
                           "FAPI_VIDEC_SetDirectIFrameMode error %d\n", ret);
        return ret;
    }

    /* copy parameters from an user space to the kernel space */
    FAPEXCFDLSAF_COPY_TO_DRV(&params, pEsData, sizeof(FAPEX_VIDEC_ESdataParamStrT));

#ifdef __LINUX__
    data = (uint32_t *)FAPI_SYS_MALLOC(params.size);
    if(data == NULL) {
        return FAPEX_VIDEC_ERR_OUT_OF_MEMORY;
    }
    else {
        FAPEXCFDLSAF_COPY_TO_DRV(data, params.data, params.size);
#else
        data = params.data;
#endif
        ret = videc_funcStr.DecodeSequenceOfDirectIFrame(pfunc->pFAPICtxHdl,
                                                         data, params.size);
        if(ret < 0) {
            FAPEXDebugPrint(2, "[fdlvidec_decodeSingleIframe] "
                               "FAPI_VIDEC_DecodeSequenceOfDirectIFrame error %d\n",
                               ret);
        }
#ifdef __LINUX__
        FAPI_SYS_FREE(data);
    }
#endif
    return ret;
}

/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLVIDEC_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                     FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_VIDEC_StrT *pfunc;
    int32_t  ret = FAPEX_VIDEC_ERR_BAD_PARAMETER;

    FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_ioctl] cmd:%08x\n",cmd);

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_VIDEC_StrT*)(file->private_data);

    switch(cmd){
        /* actual code which is depend on function is implimented here. */
        /* case ...: */
#ifdef FAPI_VIDEC_CONNECT
        case FAPEX_IOCCOM_VIDEC_START:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlvidec_Start(pfunc, (FAPEX_VIDEC_StartParamStrT*)arg);
            }
            break;

        case FAPEX_IOCCOM_STOP:
            ret = fdlvidec_Stop(pfunc);
            break;

        case FAPEX_IOCCOM_PAUSE:
            ret = fdlvidec_setPlaybackMode(pfunc, FAPI_VIDEC_PLAYBACK_PAUSE);
            break;

        case FAPEX_IOCCOM_RESET:
            ret = videc_funcStr.Reset(pfunc->pFAPICtxHdl, (FAPI_VIDEC_ResetEnumT)arg);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                   "FAPI_VIDEC_Reset error %d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_OPEN:
            ret = fdlvidec_Open(pfunc, (FAPEX_VIDEC_OpenParamStrT*)arg);
            break;

        case FAPEX_IOCCOM_VIDEC_GET_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlvidec_getStatus(pfunc, (FAPEX_VIDEC_StatusStrT *)arg);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_SET_OUTPUT:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlvidec_setOutput(pfunc, (int32_t *)arg);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_SET_SYNC:
            if((uint32_t)arg == 0) {
                ret = videc_funcStr.ClearPesParserToDecoder(pfunc->pFAPICtxHdl);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                       "FAPI_VIDEC_ClearPesParserToDecoder error %d\n", ret);
                }
            }
            else {
                ret = videc_funcStr.SetPesParserToDecoder(
                            pfunc->pFAPICtxHdl,
                            (FAPI_VIDEC_PesParserEnumT)pfunc->pes_parser_no);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                       "FAPI_VIDEC_SetPesParserToDecoder error %d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_VIDEC_SET_DISPON:
            if((FAPI_VIDEC_EnableDisableStateEnumT)arg != FAPI_VIDEC_ENABLE_FORCE_OFF) {
                ret = videc_funcStr.EnableDisable(pfunc->pFAPICtxHdl,
                                                  (FAPI_VIDEC_EnableDisableStateEnumT)arg);
                if(ret < 0) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                       "FAPI_VIDEC_EnableDisable error %d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_VIDEC_SET_DISPOFF:
            ret = videc_funcStr.EnableDisable(pfunc->pFAPICtxHdl, FAPI_VIDEC_ENABLE_FORCE_OFF);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                   "FAPI_VIDEC_EnableDisable error %d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_GET_VES_PARAMS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlvidec_getVesParams(pfunc, (FAPEX_VIDEC_VESParamStrT*)arg);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_GET_FRAME_BANK_SIZE:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((uint32_t*)arg, &pfunc->bank_size, sizeof(uint32_t));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_VIDEC_GET_DEC_FRAME_INFO:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                ret = fdlvidec_getDispFrameInfo(pfunc, (FAPEX_VIDEC_DecFrameInfoStrT*)arg);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_SET_PLAYBACK_ENV:
            ret = fdlvidec_setPlaybackEnv(pfunc, (FAPI_VIDEC_PlaybackEnvironmentEnumT)arg);
            break;

        case FAPEX_IOCCOM_VIDEC_SET_PLAYBACK_MODE:
            ret = fdlvidec_setPlaybackMode(pfunc, (FAPI_VIDEC_PlaybackModeEnumT)arg);
            break;

        case FAPEX_IOCCOM_VIDEC_SET_FREEZE_MODE:
            ret = videc_funcStr.FreezeContinue(pfunc->pFAPICtxHdl,
                                               (FAPI_VIDEC_FreezeContinueStateEnumT)arg);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                   "FAPI_VIDEC_FreezeContinue error %d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_RECOVER:
            ret = videc_funcStr.Recover(pfunc->pFAPICtxHdl);
            if(ret == FAPI_OK) {
                if(pfunc->status == FAPEX_STATUS_RUNNING) {
                    ret = videc_funcStr.Start(pfunc->pFAPICtxHdl);
                    if(ret < 0) {
                        FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                           "FAPI_VIDEC_Start error %d\n", ret);
                    }
                }
            }
            else {
                FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                   "FAPI_VIDEC_Recover error %d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_DEC_SINGLE_I_FRAME:
            ret = fdlvidec_decodeSingleIframe(pfunc, (FAPEX_VIDEC_ESdataParamStrT*)arg);
            break;

        case FAPEX_IOCCOM_VIDEC_DISP_SINGLE_I_FRAME:
            ret = videc_funcStr.DisplaySingleDirectIFrame(pfunc->pFAPICtxHdl);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                   "FAPI_VIDEC_DisplaySingleDirectIFrame error %d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_DEC_SEQUENCE_I_FRAME:
            ret = fdlvidec_decodeSequenceIframe(pfunc, (FAPEX_VIDEC_ESdataParamStrT*)arg);
            break;

        case FAPEX_IOCCOM_VIDEC_DISP_SEQUENCE_I_FRAME:
            ret = videc_funcStr.DisplaySequenceOfDirectIFrame(pfunc->pFAPICtxHdl);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                   "FAPI_VIDEC_DisplaySequenceOfDirectIFrame error %d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_SET_OUTPUT_DELAY:
            ret = videc_funcStr.SetOutputDelay(pfunc->pFAPICtxHdl, (int32_t)arg);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                   "FAPI_VIDEC_SetOutputDelay error %d\n", ret);
            }
            break;

        case FAPEX_IOCCOM_VIDEC_SET_STREAMSEGMENTMODE:
            ret = videc_funcStr.SetStreamSegmentMode(pfunc->pFAPICtxHdl,(uint32_t)arg);
            if(ret < 0) {
                FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_ioctl] "
                                   "FAPI_VIDEC_SetStreamSegmentMode error %d\n", ret);
            }
            break;
#endif /* FAPI_VIDEC_CONNECT*/

        default:
            ;
    }

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLVIDEC_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T        mask = 0;
    FAPEXFDL_VIDEC_StrT    *pfunc;
    FAPEXCFDLSAF_EventStrT *dt_event_ptr;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_VIDEC_StrT*)(file->private_data);
    dt_event_ptr = &pfunc->dt_event;

    FAPEXDebugPrint(4, "[FAPEXFDLVIDEC_poll]\n");

    FAPEXCFDLSAF_POLL_EVENT(file, dt_event_ptr, NULL, ptbl);

    /* actual code which is depend on function is implimented here. */

    if(dt_event_ptr->evt_bitptn & FAPEXFDL_VIDEC_EVENT_WRITE) {
        mask |= POLLOUT | POLLWRNORM;
    }
    if(dt_event_ptr->evt_bitptn & FAPEXFDL_VIDEC_EVENT_READ) {
        mask |= POLLIN | POLLRDNORM;
    }
    if(dt_event_ptr->evt_bitptn & FAPEXFDL_VIDEC_EVENT_MESSAGE) {
        mask |= POLLMSG;
    }

    return mask;
}

/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLVIDEC_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    FAPEXFDL_VIDEC_StrT *pfunc;
    int32_t ret;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_VIDEC_StrT *)(file->private_data);

    if(pfunc == NULL) {
        return FAPEX_VIDEC_ERR_INVALID_STATUS;
    }

    if( (vma->vm_pgoff != 0) || ((vma->vm_end - vma->vm_start) > pfunc->bank_size) ) {
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if(file->f_flags & O_SYNC) {
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma,
                          vma->vm_start,
                          FAPEXFDL_VIDEC_RAMIFV_BASE_ADDR >> PAGE_SHIFT,    /* physical page */
                          vma->vm_end - vma->vm_start,
                          vma->vm_page_prot);
    if(ret != 0) {
        return -EAGAIN;
    }
    return 0;
#else
    if(vma->vm_offset != 0) {
        return FAPEX_VIDEC_ERR_BAD_PARAMETER;
    }
    vma->vm_start = (void *)FAPEXFDL_VIDEC_RAMIFV_BASE_ADDR;

    return 0;
#endif /* __LINUX__ */
}

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLVIDEC_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_VIDEC_StrT *pfunc;
    void    *pCfdlHdl;
    int32_t id;

    FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_open] file:0x%08x,private_data:0x%08x\n",
        (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc = (FAPEXFDL_VIDEC_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc;
        FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc));

        /* set parameter */
        pfunc->pCfdlHdl = pCfdlHdl;
        pfunc->id       = id;
#ifdef __LINUX__
        kref_init(&pfunc->kref);
#endif
    }
    else {
        pfunc = (FAPEXFDL_VIDEC_StrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_open] fapex_fdlvidec was opened : used_counter %d.\n",
                       pfunc->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlvidec_delete(struct kref *kref)
{
    FAPEXFDL_VIDEC_StrT  *pfunc;
    int32_t  ret = 0;

    pfunc = container_of(kref, FAPEXFDL_VIDEC_StrT, kref);
#else
static int32_t fdlvidec_delete(FAPEXFDL_VIDEC_StrT *pfunc)
{
    int32_t  ret = 0;
#endif

    if(pfunc->event_flags != 0) {
        /* set end event */
        pfunc->frameInfo.enableFlag |= FAPEX_VIDEC_INFO_STOPPED;

        ret = FAPEXCFDLSAF_SET_EVENT(&pfunc->dt_event,
                                     FAPEXFDL_VIDEC_EVENT_END,
                                     FAPEXCFDLSAF_WAITEVENT_OR);
        if(ret < 0) {
            ret = FAPEX_VIDEC_ERR_INVALID_STATUS;
            goto fdlvidec_delete_end;
        }
        pfunc->event_flags = 0;
    }
    pfunc->bank_size = 0;

#ifdef FAPI_VIDEC_CONNECT
    {
        FAPEXFDL_TSD_SubStrT  *pfunc_tsd;

        pfunc_tsd = (FAPEXFDL_TSD_SubStrT*)FAPEXCFDL_GetFdlHdl(pfunc->pCfdlHdl, pfunc->input_id);

        if(pfunc_tsd != NULL) {
            if(pfunc_tsd->pTSDCtxHdl != NULL) {
                /* clear PTS notifer */
                ret = tsd_funcStr.SetPtsNotifier(pfunc_tsd->pTSDCtxHdl, NULL, NULL, 0);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdlvidec_delete] "
                                       "FAPI_TSD_SetPtsNotifier error %d\n", ret);
                    goto fdlvidec_delete_end;
                }
                pfunc_tsd->output_id = -1;
            }
        }

        /* close sequence */
        if(pfunc->pFAPICtxHdl != NULL) {
            ret = videc_funcStr.EnableDisable(pfunc->pFAPICtxHdl, FAPI_VIDEC_ENABLE_FORCE_OFF);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_delete] "
                                   "FAPI_VIDEC_EnableDisable error %d\n", ret);
                goto fdlvidec_delete_end;
            }

            /* call Close FAPI */
            ret = videc_funcStr.Close(pfunc->pFAPICtxHdl);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvidec_delete] "
                                   "FAPI_VIDEC_Close error %d\n", ret);
                goto fdlvidec_delete_end;
            }
            pfunc->pFAPICtxHdl = NULL;
        }
    }
#endif /* FAPI_VIDEC_CONNECT */

    /* clear control flags */
    FAPEXCFDLSAF_SPIN_LOCK_DELETE(&pfunc->lock);

    FAPEXCFDLSAF_DELETE_EVENT(&pfunc->dt_event);

    pfunc->status = FAPEX_STATUS_TERMINATED;

fdlvidec_delete_end:
#ifdef __LINUX__
    pfunc->close_result = ret;
#else
    return ret;
#endif
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLVIDEC_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_VIDEC_StrT  *pfunc;
    FAPEXFDLRELRET_T     ret = 0;

    /* Get the context base of this id's FDL. */
    pfunc = (FAPEXFDL_VIDEC_StrT*)(file->private_data);

    if(pfunc != NULL) {
        /* decrement the count on our device */
        pfunc->used_count--;
#ifdef __LINUX__
        ret = kref_put(&pfunc->kref, fdlvidec_delete);
        if(ret != 0) {
            ret = pfunc->close_result;
        }
#else
        ret = fdlvidec_delete(pfunc);
#endif

        FAPEXDebugPrint(3, "[FAPEXFDLVIDEC_release] : used_count %d\n",
                           pfunc->used_count);
    }
    FAPEXFDL_RETVAL(ret);
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlvidec_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLVIDEC_init(void)
{
    int32_t ret;
    int32_t count;
    FAPEXFDL_VIDEC_StrT  *pfunc;
    FAPEXCFDL_UtilStrT   utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_VIDEC_StrT  *pfunc0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLVIDEC_init]\n");

    /* set struct file_operations */
    fapex_fdlvidec_fops.owner   = THIS_MODULE;
    fapex_fdlvidec_fops.open    = FAPEXFDLVIDEC_open;
//  fapex_fdlvidec_fops.read    = FAPEXFDLVIDEC_read;
//  fapex_fdlvidec_fops.write   = FAPEXFDLVIDEC_write;
    fapex_fdlvidec_fops.ioctl   = FAPEXFDLVIDEC_ioctl;
    fapex_fdlvidec_fops.poll    = FAPEXFDLVIDEC_poll;
    fapex_fdlvidec_fops.mmap    = FAPEXFDLVIDEC_mmap;
    fapex_fdlvidec_fops.release = FAPEXFDLVIDEC_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc = (FAPEXFDL_VIDEC_StrT*)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_VIDEC_StrT)*(FAPEXDEV_VIDEC_DESCRIPTOR_NUM));
    if(pfunc == NULL) {
        return FAPEX_VIDEC_ERR_OUT_OF_MEMORY;
    }
    pfunc0 = pfunc;
#else
    pfunc = FAPIGLD_FDLVIDECCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc, sizeof(FAPEXFDL_VIDEC_StrT)*(FAPEXDEV_VIDEC_DESCRIPTOR_NUM));

    for(count=FAPEXDEV_VIDEC_MINOR_FIRST; count<=FAPEXDEV_VIDEC_MINOR_LAST; count++){
        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlvidec_fops, pfunc, &utility, count);
        if(ret) {
#ifdef ALLOCATE_DYNAMIC_MEMORY
            FAPI_SYS_FREE(pfunc0);
#endif
            FAPEXDebugPrint(2, "[FAPEXFDLVIDEC_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
            return ret;
        }
//      pfunc->status = FAPEX_STATUS_LOADED;
        pfunc ++;
    }

#ifdef FAPI_VIDEC_CONNECT
    FAPEXDebugPrint(1, "call FAPEXCFG_VIDEC_Init\n");

    /* call FAPI (initialize) */
    ret = FAPEXCFG_VIDEC_Init();
#else
    ret = 0;
#endif /* FAPI_VIDEC_CONNECT */

    FAPEXDebugPrint(1, "[FAPEXFDLVIDEC_init] fapex_fdlvidec was loaded. %d\n", ret);

    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLVIDEC_cleanup(void)
{
    int32_t count;

    FAPEXDebugPrint(1, "[FAPEXFDLVIDEC_cleanup]\n");

#ifdef FAPI_VIDEC_CONNECT
    /* call FAPI API (cleanup) */
    FAPEXCFG_VIDEC_Cleanup();

    FAPEXDebugPrint(1, "end FAPEXCFG_VIDEC_Cleanup\n");
#endif /* FAPI_VIDEC_CONNECT */

    /* cleanup sequence (common this FDL) */
#ifdef ALLOCATE_DYNAMIC_MEMORY
    {
        void                *pCfdlHdl;
        FAPEXFDL_VIDEC_StrT *pfunc;

        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        pfunc    = (FAPEXFDL_VIDEC_StrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_VIDEC_MINOR_FIRST);

        /* relase the FDL context */
        if(pfunc != NULL) {
            FAPI_SYS_FREE(pfunc);
        }
    }
#endif /* ALLOCATE_DYNAMIC_MEMORY */

    for(count=FAPEXDEV_VIDEC_MINOR_FIRST; count<=FAPEXDEV_VIDEC_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLVIDEC_cleanup] fapex_fdlvidec was unloaded.\n");
    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLVIDEC_init);
module_exit(FAPEXFDLVIDEC_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlvidec object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
