/*
 *
 * \attention  This program is provided as is. You can redistribute it and/or
 * modify it. Fujitsu Semiconductor Limited accepts no responsibility or
 * liability for any errors or omissions.
 * 
 * Copyright (C) 2008-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
 */
/*
 FDL module skel_drv
*/

#include "fapex.h"
#include "fapex/abst_lib.h"
#include "fapex/abst_libvienc.h"
#include "fapex/sys_config.h"
#include "fapex_fdlvienc_priv.h"
#include "fapex_config_cmn.h"

/* for vienc_funcStr */
#include "fapex_func_ext.h"

/*
*******************************************************************************
*******************************************************************************
** Global Data
*******************************************************************************
*******************************************************************************/

/*
*******************************************************************************
*******************************************************************************
** Local Data
*******************************************************************************
*******************************************************************************/
#ifndef ALLOCATE_DYNAMIC_MEMORY
static FAPEXFDL_VIENC_CmnStrT  FAPIGLD_FDLVIENCCmnCtx[FAPEXDEV_VIENC_CMN_DESCRIPTOR_NUM];
static FAPEXFDL_VIENC_SubStrT  FAPIGLD_FDLVIENCSubCtx[FAPEXDEV_VIENC_SUB_DESCRIPTOR_NUM];
#endif /* !ALLOCATE_DYNAMIC_MEMORY */

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Read method of this id's FDL.
**         This function is called with the systemcall "read".
**
** If copy data to the argument's buffer, use FAPI_SYS_copyFromUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLREADRET_T FAPEXFDLVIENC_read(FAPEXFDLREADARG1_T file, FAPEXFDLREADARG2_T buf,
                                     FAPEXFDLREADARG3_T size, FAPEXFDLREADARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLVIENC_read] size:%d\n",size);

    return size;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Write method of this id's FDL.
**         This function is called with the systemcall "write".
**
** If copy data from the argument's buffer, use FAPI_SYS_copyToUser().
**
** \param  file    The pointer to the struct file.
** \param  buf     The pointer to the base address of the buffer.
** \param  size    The buffer size.
** \param  offset  The offset (now, not supported).
**
******************************************************************************/
FAPEXFDLWRTRET_T FAPEXFDLVIENC_write(FAPEXFDLWRTARG1_T file, FAPEXFDLWRTARG2_T buf,
                                     FAPEXFDLWRTARG3_T size, FAPEXFDLWRTARG4_T offset)
{
    FAPEXDebugPrint(3, "[FAPEXFDLVIENC_write] size:%d\n",size);

    return size;
}
#endif

/*!
*******************************************************************************
**
** \brief  VIENC start
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  id           Device identifier
**
******************************************************************************/
static int32_t fdlvienc_Start(FAPEXFDL_VIENC_CmnStrT *pfunc_cmn, FAPEXFDL_VIENC_SubStrT *pfunc_sub,
                              uint32_t id)
{
    uint32_t    enable_cmn, enable_ch, running_ch;
    uint32_t    one_ch_flag;
    int32_t     ret;

    enable_cmn = pfunc_cmn->enable_channel;
    enable_ch  = 0x00000001;
    running_ch = 0;

    if(pfunc_sub != NULL) {
        /* sub descriptor */
        enable_ch <<= id - FAPEXDEV_VIENC_SUB_MINOR_FIRST;
        enable_cmn &= enable_ch;
        one_ch_flag = 1;
    }
    else {
        /* common descriptor */
        id ++;
        pfunc_sub   = (FAPEXFDL_VIENC_SubStrT *)FAPEXCFDL_GetFdlHdl(pfunc_cmn->pCfdlHdl, id);
        one_ch_flag = 0;
    }

    /* start channel */
    while(enable_cmn) {
        if(pfunc_sub->status == FAPEX_STATUS_READY) {

#ifdef FAPI_VIENC_CONNECT
            /* call FAPI (start) */
            ret = vienc_funcStr.Start(pfunc_sub->pFAPICtxHdl);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvienc_Start] FAPI_VIENC_Start error:%d\n", ret);
                return ret;
            }
            if(pfunc_sub->enc_type == FAPI_VIENC_ENCODER_SD) {
                ret = vienc_funcStr.EnableDisableWss(pfunc_sub->pFAPICtxHdl,
                                                     FAPI_VIENC_PARAMETER_ON);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[fdlvienc_Start] FAPI_VIENC_EnableDisableWss error:%d\n",
                                       ret);
                    return ret;
                }
            }
#endif /* FAPI_VIENC_CONNECT */

            /* update channel status */
            pfunc_sub->status  = FAPEX_STATUS_RUNNING;
            running_ch        |= enable_ch;
        }
        else if(one_ch_flag != 0) {
            if(pfunc_sub->status != FAPEX_STATUS_RUNNING) {
                return FAPEX_VIENC_ERR_INVALID_STATUS;
            }
        }
        enable_cmn &= ~enable_ch;

        /* setup for next channel */
        pfunc_sub ++;
        enable_ch <<= 1;
    }

    /* update block status */
    if(running_ch != 0) {
        /*---------- LOCK ----------*/
        FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

        pfunc_cmn->running_channel |= running_ch;
        pfunc_cmn->status = FAPEX_STATUS_RUNNING;

        /*--------- UNLOCK ---------*/
        FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);
    }
    return 0;
}

/*!
*******************************************************************************
**
** \brief  VIENC stop
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  id           Device identifier
**
******************************************************************************/
static int32_t fdlvienc_Stop(FAPEXFDL_VIENC_CmnStrT *pfunc_cmn, FAPEXFDL_VIENC_SubStrT *pfunc_sub,
                             uint32_t id)
{
    uint32_t    enable_cmn, disable_ch, stop_channel;
    int32_t     ret = 0;

    enable_cmn = pfunc_cmn->enable_channel;
    disable_ch   = 0x00000001;
    stop_channel = 0;

    if(pfunc_sub != NULL) {
        /* sub descriptor */
        disable_ch <<= id - FAPEXDEV_VIENC_SUB_MINOR_FIRST;
        enable_cmn  &= disable_ch;
    }
    else {
        /* common descriptor */
        id ++;
        pfunc_sub = (FAPEXFDL_VIENC_SubStrT *)FAPEXCFDL_GetFdlHdl(pfunc_cmn->pCfdlHdl, id);
    }

    /* stop channel */
    while(enable_cmn) {
        if(pfunc_sub->status == FAPEX_STATUS_RUNNING) {

#ifdef FAPI_VIENC_CONNECT
            /* call FAPI (stop) */
            ret = vienc_funcStr.Stop(pfunc_sub->pFAPICtxHdl);
            if(ret != FAPI_OK) {
                FAPEXDebugPrint(2, "[fdlvienc_Stop] FAPI_VIENC_Stop error:%d\n", ret);
                break;
            }
#endif /* FAPI_VIENC_CONNECT */

            /* update channel status */
            pfunc_sub->status = FAPEX_STATUS_READY;
            stop_channel     |= disable_ch;
        }
        enable_cmn &= ~disable_ch;

        /* setup for next channel */
        pfunc_sub ++;
        disable_ch <<= 1;
    }

    /* update block status */
    if(stop_channel != 0) {
        /*---------- LOCK ----------*/
        FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

        pfunc_cmn->running_channel &= ~stop_channel;
        if(pfunc_cmn->running_channel == 0) {
            pfunc_cmn->status = FAPEX_STATUS_READY;
        }

        /*--------- UNLOCK ---------*/
        FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);
    }

    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIENC common descriptor open
**
** \param  pfunc_cmn    pointer to common context
** \param  sys_mode     system mode
**
******************************************************************************/
static int32_t fdlvienc_OpenCmn(FAPEXFDL_VIENC_CmnStrT *pfunc_cmn, int32_t sys_mode)
{
    FAPI_VIENC_InitClocksEnumT  clk_type;
    int32_t  ret;

    if(pfunc_cmn->system_mode != FAPEX_VIENC_SYSTEM_MODE_UNKNOWN) {
        return 0;
    }

    switch(sys_mode) {
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x2397P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x2397P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x24P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x24P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x25P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x25P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x2997P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x2997P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x30P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x30P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x50P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x50P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x5994P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x5994P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x60P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x60P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x50I:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x50I;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x5994I:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x5994I;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1920x1080x60I:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1920x1080x60I;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1280x720x2397P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1280x720x2397P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1280x720x24P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1280x720x24P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1280x720x25P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1280x720x25P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1280x720x2997P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1280x720x2997P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1280x720x30P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1280x720x30P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1280x720x50P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1280x720x50P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1280x720x5994P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1280x720x5994P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1280x720x60P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_1280x720x60P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_720x576x50I:
            clk_type = FAPI_VIENC_INIT_CLOCKS_720x576x50I;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_720x576x50P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_720x576x50P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_720x480x5994P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_720x480x5994P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_720x480x5994I:
            clk_type = FAPI_VIENC_INIT_CLOCKS_720x480x5994I;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_640x480x5994P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_VGA_640x480x5994P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_800x600x5994P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_VGA_800x600x5994P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1024x768x5994P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_VGA_1024x768x5994P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_1360x768x5994P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_VGA_1360x768x5994P;
            break;
        case FAPEX_VIENC_SYSTEM_MODE_LCD_640x480x50P:
        case FAPEX_VIENC_SYSTEM_MODE_LCD_1024x768x50P:
            clk_type = FAPI_VIENC_INIT_CLOCKS_LCD;
            break;
        default:
            clk_type = FAPI_VIENC_INIT_CLOCKS_UNKNOWN;
            break;
    }
    ret = vienc_funcStr.InitClocks(clk_type);

    if(ret == FAPI_OK) {
        pfunc_cmn->system_mode = sys_mode;
    }
    else {
        FAPEXDebugPrint(2, "[FAPEXFDLVIENC_ioctl] "
                           "FAPI_VIENC_InitClocks error:%d\n", ret);
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIENC sub descriptor open
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  id           Device identifier
** \param  pArg         pointer to FAPEX_VIENC_OpenSubIDParamStrT parameter
**
******************************************************************************/
static int32_t fdlvienc_OpenSub(FAPEXFDL_VIENC_CmnStrT *pfunc_cmn, FAPEXFDL_VIENC_SubStrT *pfunc_sub,
                                uint32_t id, FAPEX_VIENC_OpenSubIDParamStrT *pArg)
{
    FAPI_VIENC_OpenParamsT  params;
    int32_t  ret;

    if(pfunc_sub->pFAPICtxHdl != NULL) {
        return 0;
    }
    if(pArg == NULL) {
        return FAPEX_VIENC_ERR_BAD_PARAMETER;
    }

    /* copy parameters from an user space to the kernel space */
    params.version = FAPI_VIENC_VERSION;
    FAPEXCFDLSAF_COPY_TO_DRV(&params.encoderType,
                             &(pArg->encoderType),
                             sizeof(int32_t));
    FAPEXCFDLSAF_COPY_TO_DRV(&params.interlaceProgressive,
                             &(pArg->interlaceProgressive),
                             sizeof(int32_t));
    FAPEXCFDLSAF_COPY_TO_DRV(&params.colourSystem,
                             &(pArg->colourSystem),
                             sizeof(int32_t));
    FAPEXCFDLSAF_COPY_TO_DRV(&params.colourComponent,
                             &(pArg->colourComponent),
                             sizeof(int32_t));
    FAPEXCFDLSAF_COPY_TO_DRV(&params.crossChrominanceFilter,
                             &(pArg->crossChrominanceFilter),
                             sizeof(int32_t));

#ifdef FAPI_VIENC_CONNECT
    /* call FAPI (open) */
    pfunc_sub->pFAPICtxHdl = vienc_funcStr.Open(&params, &ret);

    FAPEXDebugPrint(3, "[FAPEXFDLVIENC_ioctl] VIENC context:0x%08x (%d),interlaceProgressive=%xh\n",
                       (uint32_t)pfunc_sub->pFAPICtxHdl, pfunc_sub->id,params.interlaceProgressive);

    if(pfunc_sub->pFAPICtxHdl != NULL)
#endif /* FAPI_VIENC_CONNECT */
    {
        /*---------- LOCK ----------*/
        FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

        pfunc_cmn->enable_channel |= 1 << (id - FAPEXDEV_VIENC_SUB_MINOR_FIRST);
        if(pfunc_cmn->status == FAPEX_STATUS_OPENED) {
            pfunc_cmn->status = FAPEX_STATUS_READY;
        }

        /*--------- UNLOCK ---------*/
        FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

        pfunc_sub->enc_type = params.encoderType;
        pfunc_sub->status   = FAPEX_STATUS_READY;
    }
    else {
        FAPEXDebugPrint(2, "[fdlvienc_OpenSub] FAPI_VIENC_Open error:%d\n", ret);
    }
    return ret;
}

/*!
*******************************************************************************
**
** \brief  VIENC sub descriptor close
**
** \param  pfunc_cmn    pointer to common context
** \param  pfunc_sub    pointer to sub context
** \param  id           Device identifier
**
******************************************************************************/
static int32_t fdlvienc_CloseSub(FAPEXFDL_VIENC_CmnStrT *pfunc_cmn,
                                 FAPEXFDL_VIENC_SubStrT *pfunc_sub, uint32_t id)
{
    uint32_t  ch_mask;
    int32_t   ret;

    if(pfunc_sub->pFAPICtxHdl == NULL) {
        return 0;
    }

#ifdef FAPI_VIENC_CONNECT
    /* call FAPI (close) */
    ret = vienc_funcStr.Close(pfunc_sub->pFAPICtxHdl);
    if(ret != FAPI_OK) {
        FAPEXDebugPrint(2, "[fdlvienc_CloseSub] FAPI_VIENC_Close error:%d\n", ret);
        return ret;
    }
#endif /* FAPI_VIENC_CONNECT */
    pfunc_sub->pFAPICtxHdl = NULL;

    ch_mask = ~(1 << (id - FAPEXDEV_VIENC_SUB_MINOR_FIRST));

    /*---------- LOCK ----------*/
    FAPI_SYS_GET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_SUSPEND);

    /* update status */
    pfunc_cmn->enable_channel  &= ch_mask;
    pfunc_cmn->running_channel &= ch_mask;

    pfunc_sub->status = FAPEX_STATUS_LOADED;
    if(pfunc_cmn->enable_channel == 0) {
        pfunc_cmn->status = FAPEX_STATUS_OPENED;
    }
    else if(pfunc_cmn->running_channel == 0) {
        pfunc_cmn->status = FAPEX_STATUS_READY;
    }

    /*--------- UNLOCK ---------*/
    FAPI_SYS_SET_SEMAPHORE(pfunc_cmn->cmn_sem, FAPI_SYS_NO_SUSPEND);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  General control of this id's FDL.
**         This function is called with systemcall "ioctl".
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
** \param  cmd     The ioctl command. This is set by user application. On Linux,
**                 the kernel copy this data to the kernel space. On RTOS, this
**                 memory area is shared with the user.
**                 memory of the user
** \param  arg     The argument of the ioctl command. This is set by user 
**                 application. On Linux, the kernel copy this data to the kernel
**                 space. On RTOS, thismemory area is shared with the user.
**
******************************************************************************/
FAPEXFDLIOCRET_T FAPEXFDLVIENC_ioctl(FAPEXFDLIOCARG1_T inode, FAPEXFDLIOCARG2_T file,
                                     FAPEXFDLIOCARG3_T cmd, FAPEXFDLIOCARG4_T arg)
{
    FAPEXFDL_VIENC_CmnStrT *pfunc_cmn;
    FAPEXFDL_VIENC_SubStrT *pfunc_sub;
    uint32_t  id;
    int32_t   ret = FAPEX_VIENC_ERR_BAD_PARAMETER;

    FAPEXDebugPrint(3, "[FAPEXFDLVIENC_ioctl] cmd:%08x\n",cmd);

    id = FAPEXCFDLSAF_GETID(inode);

    /* Get the context base of this id's FDL. */
    if(id < FAPEXDEV_VIENC_SUB_MINOR_FIRST) {
        pfunc_cmn = (FAPEXFDL_VIENC_CmnStrT *)(file->private_data);
        pfunc_sub = NULL;
    }
    else {
        pfunc_sub = (FAPEXFDL_VIENC_SubStrT *)(file->private_data);
        if(pfunc_sub != NULL) {
            pfunc_cmn = (FAPEXFDL_VIENC_CmnStrT *)pfunc_sub->pCmnHdl;
        }
        else {
            pfunc_cmn = NULL;
        }
    }

    switch(cmd){
        /* actual code which is depend on function is implimented here. */
        /* case ...: */
        case FAPEX_IOCCOM_START:
            ret = fdlvienc_Start(pfunc_cmn, pfunc_sub, id);
            break;

        case FAPEX_IOCCOM_STOP:
            ret = fdlvienc_Stop(pfunc_cmn, pfunc_sub, id);
            break;

        case FAPEX_IOCCOM_VIENC_OPEN:
            if(pfunc_cmn != NULL) {
                ret = fdlvienc_OpenCmn(pfunc_cmn, (int32_t)arg);
            }
            break;

        case FAPEX_IOCCOM_VIENC_OPEN_SUBID:
            if(pfunc_sub != NULL) {
                ret = fdlvienc_OpenSub(pfunc_cmn, pfunc_sub, id,
                                       (FAPEX_VIENC_OpenSubIDParamStrT*)arg);
            }
            break;

        case FAPEX_IOCCOM_CLOSE_SUBID:
            if(pfunc_sub != NULL) {
                ret = fdlvienc_CloseSub(pfunc_cmn, pfunc_sub, id);
            }
            break;

        case FAPEX_IOCCOM_VIENC_GET_STATUS:
            if(arg != (FAPEXFDLIOCARG4_T)NULL) {
                FAPEX_VIENC_StatusStrT params;

                if(pfunc_sub == NULL) {
                    if(pfunc_cmn != NULL) {
                        params.status = pfunc_cmn->status;
                    }
                    else {
                        params.status = FAPEX_STATUS_TERMINATED;
                    }
                }
                else {
                    params.status = pfunc_sub->status;
                }

                /* copy parameters from the kernel space to an user space */
                FAPEXCFDLSAF_COPY_TO_USR((void *)arg, &params, sizeof(FAPEX_VIENC_StatusStrT));
                ret = 0;
            }
            break;

        case FAPEX_IOCCOM_VIENC_GET_WSSDATA:
            if((pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL)) {
                /* call FAPI_VIENC_GetWssData */
                ret = vienc_funcStr.GetWssData(pfunc_sub->pFAPICtxHdl, (uint32_t*)arg);

                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIENC_ioctl] FAPI_VIENC_GetWssData error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_VIENC_SET_WSSDATA:
            if((pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL)) {
                /* call FAPI_VIENC_SetWssData */
                ret = vienc_funcStr.SetWssData(pfunc_sub->pFAPICtxHdl, (uint32_t)arg);

                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIENC_ioctl] FAPI_VIENC_SetWssData error:%d\n", ret);
                }
            }
            break;
        
         /* Martin added at 20100513 for SMARTGO requires*/
        case FAPEX_IOCCOM_VIENC_SET_CCOMPONENT:
            if((pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL)) {
                /* call FAPI_VIENC_SetColourComponent */
                FAPEXDebugPrint(1, "[FAPEXFDLVIENC_ioctl] FAPEX_IOCCOM_VIENC_SET_CCOMPONENT:arg: %08x\n",arg);
                ret = vienc_funcStr.SetColourComponent(pfunc_sub->pFAPICtxHdl, (uint32_t)arg);

                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIENC_ioctl] FAPEX_IOCCOM_VIENC_SET_CCOMPONENT error:%d\n", ret);
                }
            }
            break;
            
        case FAPEX_IOCCOM_VIENC_GET_CCOMPONENT:
            if((pfunc_sub != NULL) && (arg != (FAPEXFDLIOCARG4_T)NULL)) {
                FAPI_VIENC_ComponentEnumT  value;
                /* call FAPI_VIENC_getColourComponent */
                ret = vienc_funcStr.GetColourComponent(pfunc_sub->pFAPICtxHdl, &value);
                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIENC_ioctl] FAPEX_IOCCOM_VIENC_GET_CCOMPONENT error:%d\n", ret);
                }
                else{
                    FAPEXCFDLSAF_COPY_TO_DRV((void*)arg, &value, sizeof(FAPI_VIENC_ComponentEnumT));
                }
            }
            break;
            
        case FAPEX_IOCCOM_VIENC_ENABLE_WSS:
            if((pfunc_sub != NULL)) {
                /* call FAPI_VIENC_EnableDisableWss */
                ret = vienc_funcStr.EnableDisableWss(pfunc_sub->pFAPICtxHdl, FAPI_VIENC_PARAMETER_ON);

                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIENC_ioctl] FAPEX_IOCCOM_VIENC_SET_CCOMPONENT error:%d\n", ret);
                }
            }
            break;

        case FAPEX_IOCCOM_VIENC_DISABLE_WSS:
            if((pfunc_sub != NULL)) {
                /* call FAPI_VIENC_EnableDisableWss */
                ret = vienc_funcStr.EnableDisableWss(pfunc_sub->pFAPICtxHdl, FAPI_VIENC_PARAMETER_OFF);

                if(ret != FAPI_OK) {
                    FAPEXDebugPrint(2, "[FAPEXFDLVIENC_ioctl] FAPEX_IOCCOM_VIENC_SET_CCOMPONENT error:%d\n", ret);
                }
            }
            break;
        /* Martin added end */
       
        case FAPEX_IOCCOM_VIENC_SET_DAC:
        if(pfunc_sub != NULL)
            ret = vienc_funcStr.SetDacAssignment(pfunc_sub->pFAPICtxHdl,arg);
            if(ret != FAPI_OK){
                FAPEXDebugPrint(2, "[FAPEXFDLVIENC_ioctl] FAPI_VIENC_SetDacAssignment error:%d\n", ret);
            }
        default:
            break;
    }

    return ret;
}

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Poll method of this id's FDL.
**         This function is called with the systemcall "select".
**
** \param  file    The pointer to the struct file.
** \param  ptbl    An internal data.
**
******************************************************************************/
FAPEXFDLPLRET_T FAPEXFDLVIENC_poll(FAPEXFDLPLARG1_T file, FAPEXFDLPLARG2_T ptbl)
{
    FAPEXFDLPLRET_T mask = 0;

    FAPEXDebugPrint(4, "[FAPEXFDLVIENC_poll]\n");

    /* actual code which is depend on function is implimented here. */

    return mask;
}
#endif

#if 0 /* not supported */
/*!
*******************************************************************************
**
** \brief  Mmap method of this id's FDL.
**         This function is called with the systemcall "mmap".
**
** \param  file    The pointer to the struct file.
** \param  vma     system depend
**
******************************************************************************/
FAPEXFDLMMPRET_T FAPEXFDLVIENC_mmap(FAPEXFDLMMPARG1_T file, FAPEXFDLMMPARG2_T vma)
{
#ifdef __LINUX__
    int32_t ret;

    if (vma->vm_pgoff != 0){
        return -EINVAL;
    }

    /* This area doen't swapout */
    vma->vm_flags |= VM_IO | VM_RESERVED;

    /* This area is IO area. Set noncache, and it doesn't dump to a core
       file of a process */
    if (file->f_flags & O_SYNC){
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        vma->vm_flags |= VM_IO;
    }

    ret = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
                  vma->vm_end - vma->vm_start, vma->vm_page_prot);

    if(ret != 0){
        return -EAGAIN;
    }
    return 0;
#else
    return -1;
#endif /* __LINUX__ */
}
#endif

/*!
*******************************************************************************
**
** \brief  Open this id's FDL.
**         This function is called with the systemcall "open". And this id's FDL 
**         are opened.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLOPENRET_T FAPEXFDLVIENC_open (FAPEXFDLOPENARG1_T inode, FAPEXFDLOPENARG2_T file)
{
    FAPEXFDL_VIENC_CmnStrT *pfunc_cmn;
    void    *pCfdlHdl;
    int32_t id;

    FAPEXDebugPrint(3, "[FAPEXFDLVIENC_open] file:0x%08x,private_data:0x%08x\n",
        (uint32_t)file, (uint32_t)file->private_data);

    if(file->private_data == NULL){
        id = FAPEXCFDLSAF_GETID(inode);
        FAPEXDebugPrint(3, "[FAPEXFDLVIENC_open] minor_no:%d\n", id);

        /* Get the context base of the CFDL. */
        pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();
        FAPEXDebugPrint(3, "[FAPEXFDLVIENC_open] pCfdlHdl:0x%08x\n", (uint32_t)(pCfdlHdl));

        /* Get the context base of this id's FDL. */
        pfunc_cmn = (FAPEXFDL_VIENC_CmnStrT *)FAPEXCFDL_GetFdlHdl(pCfdlHdl, id);
        file->private_data = pfunc_cmn;
        FAPEXDebugPrint(3, "[FAPEXFDLVIENC_open] pFdlHdl:0x%08x\n", (uint32_t)(pfunc_cmn));

        /* set parameter */
        pfunc_cmn->pCfdlHdl = pCfdlHdl;
        pfunc_cmn->id       = id;
        pfunc_cmn->status   = FAPEX_STATUS_OPENED;
#ifdef __LINUX__
        kref_init(&pfunc_cmn->kref);
#endif
    }
    else {
        pfunc_cmn = (FAPEXFDL_VIENC_CmnStrT *)file->private_data;
#ifdef __LINUX__
        kref_get(&pfunc_cmn->kref);
#endif
    }

    /* increment our usage count for the device */
    pfunc_cmn->used_count++;

    FAPEXDebugPrint(3, "[FAPEXFDLVIENC_open] fapex_fdlvienc was opened : used_counter %d.\n",
                       pfunc_cmn->used_count);

    return 0;
}

/*!
*******************************************************************************
**
** \brief  delete this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  kref    The pointer to the kref structure.
**
******************************************************************************/
#ifdef __LINUX__
static void fdlvienc_delete(struct kref *kref)
{
    FAPEXFDL_VIENC_CmnStrT   *pfunc_cmn;

    pfunc_cmn = container_of(kref, FAPEXFDL_VIENC_CmnStrT, kref);
#else
static void fdlvienc_delete(FAPEXFDL_VIENC_CmnStrT *pfunc_cmn)
{
#endif

    pfunc_cmn->status = FAPEX_STATUS_TERMINATED;
}

/*!
*******************************************************************************
**
** \brief  Release this FDL id's module.
**         This function is called with the systemcall "close". And this id's FDL
**         are released.
**
** \param  inode   The pointer to the struct inode.
** \param  file    The pointer to the struct file.
**
******************************************************************************/
FAPEXFDLRELRET_T FAPEXFDLVIENC_release (FAPEXFDLRELARG1_T inode, FAPEXFDLRELARG2_T file)
{
    FAPEXFDL_VIENC_CmnStrT *pfunc_cmn;

    pfunc_cmn = (FAPEXFDL_VIENC_CmnStrT *)(file->private_data);

    if(pfunc_cmn != NULL) {
        if(pfunc_cmn->status > FAPEX_STATUS_OPENED) {
            FAPEXFDL_RETVAL( FAPEX_VIENC_ERR_INVALID_STATUS );
        }

        /* decrement the count on our device */
        pfunc_cmn->used_count--;
#ifdef __LINUX__
        kref_put(&pfunc_cmn->kref, fdlvienc_delete);
#else
        fdlvienc_delete(pfunc_cmn);
#endif

        FAPEXDebugPrint(3, "[FAPEXFDLVIENC_release] id %d, used_count %d\n",
                           pfunc_cmn->id, pfunc_cmn->used_count);
    }
    return 0;
}

/*--------------------------------------------------------*/
/* The file operations of this FDL module.                */
/*--------------------------------------------------------*/
struct file_operations fapex_fdlvienc_fops;

/*!
*******************************************************************************
**
** \brief  Initialize process of the FDL module.
**         This function is called when the FDL module is loaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_init".
**         And user application doesn't call this function. If do "insmod" command,
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
int32_t FAPEXFDLVIENC_init(void)
{
    int32_t ret;
    int32_t count;
    void    *pCfdlHdl;
    FAPEXFDL_VIENC_CmnStrT  *pfunc_cmn;
    FAPEXFDL_VIENC_SubStrT  *pfunc_sub;
    FAPEXCFDL_UtilStrT      utility;
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPEXFDL_VIENC_SubStrT  *pfunc_sub0;
#endif

    FAPEXDebugPrint(1, "[FAPEXFDLVIENC_init]\n");

    /* set struct file_operations */
    fapex_fdlvienc_fops.owner   = THIS_MODULE;
    fapex_fdlvienc_fops.open    = FAPEXFDLVIENC_open;
//  fapex_fdlvienc_fops.read    = FAPEXFDLVIENC_read;
//  fapex_fdlvienc_fops.write   = FAPEXFDLVIENC_write;
    fapex_fdlvienc_fops.ioctl   = FAPEXFDLVIENC_ioctl;
//  fapex_fdlvienc_fops.poll    = FAPEXFDLVIENC_poll;
//  fapex_fdlvienc_fops.mmap    = FAPEXFDLVIENC_mmap;
    fapex_fdlvienc_fops.release = FAPEXFDLVIENC_release;

#ifdef ALLOCATE_DYNAMIC_MEMORY
    pfunc_cmn = (FAPEXFDL_VIENC_CmnStrT *)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_VIENC_CmnStrT));
    if(pfunc_cmn == NULL) {
        return FAPEX_VIENC_ERR_OUT_OF_MEMORY;
    }
    pfunc_sub = (FAPEXFDL_VIENC_SubStrT *)FAPI_SYS_MALLOC(sizeof(FAPEXFDL_VIENC_SubStrT)*FAPEXDEV_VIENC_SUB_DESCRIPTOR_NUM);
    if(pfunc_sub == NULL) {
        FAPI_SYS_FREE(pfunc_cmn);
        return FAPEX_VIENC_ERR_OUT_OF_MEMORY;
    }
    pfunc_sub0 = pfunc_sub;
#else
    pfunc_cmn = FAPIGLD_FDLVIENCCmnCtx;
    pfunc_sub = FAPIGLD_FDLVIENCSubCtx;
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc_cmn, sizeof(FAPEXFDL_VIENC_CmnStrT));
    FAPEXCFDLSAF_CLEAR_MEMORY(pfunc_sub, sizeof(FAPEXFDL_VIENC_SubStrT)*FAPEXDEV_VIENC_SUB_DESCRIPTOR_NUM);

    pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();

    pfunc_cmn->pCfdlHdl    = pCfdlHdl;
//  pfunc_cmn->status      = FAPEX_STATUS_LOADED;
    pfunc_cmn->system_mode = FAPEX_VIENC_SYSTEM_MODE_UNKNOWN;
    pfunc_cmn->cmn_sem     = FAPI_SYS_CREATE_SEMAPHORE(1);
    if(pfunc_cmn->cmn_sem == (FAPI_SYS_SemaphoreT)0) {
        ret = FAPEX_VIENC_ERR_SEMAPHORE_CREATE;
        goto FAPEXFDLVIENC_init_error;
    }

    /* Register this id's FDL device to CFDL. */
    ret = FAPEXCFDL_RegisterFdl(&fapex_fdlvienc_fops, pfunc_cmn, &utility, FAPEXDEV_VIENC_CMN_MINOR);
    if(ret) {
        goto FAPEXFDLVIENC_init_error;
    }

    for(count=FAPEXDEV_VIENC_SUB_MINOR_FIRST; count<=FAPEXDEV_VIENC_SUB_MINOR_LAST; count++) {
        pfunc_sub->pCfdlHdl = pCfdlHdl;
        pfunc_sub->pCmnHdl  = pfunc_cmn;
//      pfunc_sub->status   = FAPEX_STATUS_LOADED;

        /* Register this id's FDL device to CFDL. */
        ret = FAPEXCFDL_RegisterFdl(&fapex_fdlvienc_fops, pfunc_sub, &utility, count);
        if(ret) {
            goto FAPEXFDLVIENC_init_error;
        }
        pfunc_sub ++;
    }

#ifdef FAPI_VIENC_CONNECT
    FAPEXDebugPrint(1, "call FAPEXCFG_VIENC_Init\n");

    /* call FAPI (initialize) */
    ret = FAPEXCFG_VIENC_Init();
#else
    ret = 0;
#endif /* FAPI_VIENC_CONNECT */

    FAPEXDebugPrint(1, "[FAPEXFDLVIENC_init] fapex_fdlvienc was loaded. %d\n", ret);

    return ret;

FAPEXFDLVIENC_init_error:
#ifdef ALLOCATE_DYNAMIC_MEMORY
    FAPI_SYS_FREE(pfunc_cmn);
    FAPI_SYS_FREE(pfunc_sub0);
#endif
    FAPEXDebugPrint(2, "[FAPEXFDLVIENC_init] FAPEXCFDL_RegisterFdl:error %d\n", ret);
    return ret;
}

/*!
*******************************************************************************
**
** \brief  Cleanup process of the FDL module.
**         This function is called when the FDL module is unloaded. It is executed
**         only one times.
**         On Linux, this function is registered to the kernel using "module_exit".
**         And user application doesn't call this function. If do "rmmod" command, 
**         this is called by the kernel.
**         On RTOS, user application has to call this function.
**
** \param  void (can't set any argument)
**
******************************************************************************/
void FAPEXFDLVIENC_cleanup(void)
{
    int32_t count;
    void    *pCfdlHdl;
    FAPEXFDL_VIENC_CmnStrT *pfunc_cmn;

    FAPEXDebugPrint(1, "[FAPEXFDLVIENC_init]\n");

#ifdef FAPI_VIENC_CONNECT
    /* call FAPI API (cleanup) */
    FAPEXCFG_VIENC_Cleanup();

    FAPEXDebugPrint(1, "end FAPEXCFG_VIENC_Cleanup\n");
#endif /* FAPI_VIENC_CONNECT */

    /* cleanup sequence (common this FDL) */
    pCfdlHdl = (void*)FAPEXCFDL_GetCfdlHdl();

    /* relase the FDL context */
    pfunc_cmn = (FAPEXFDL_VIENC_CmnStrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_VIENC_CMN_MINOR);
    if(pfunc_cmn != NULL) {
        FAPI_SYS_DESTROY_SEMAPHORE(pfunc_cmn->cmn_sem);
        pfunc_cmn->cmn_sem = (FAPI_SYS_SemaphoreT)0;

#ifdef ALLOCATE_DYNAMIC_MEMORY
        FAPI_SYS_FREE(pfunc_cmn);
    }
    {
        FAPEXFDL_VIENC_SubStrT *pfunc_sub;

        pfunc_sub = (FAPEXFDL_VIENC_SubStrT*)FAPEXCFDL_GetFdlHdl(pCfdlHdl, FAPEXDEV_VIENC_SUB_MINOR_FIRST);
        if(pfunc_sub != NULL) {
            FAPI_SYS_FREE(pfunc_sub);
        }
#endif /* ALLOCATE_DYNAMIC_MEMORY */
    }

    /* Unregister this id's FDL device to CFDL. */
    FAPEXCFDL_UnregisterFdl(FAPEXDEV_VIENC_CMN_MINOR);

    for(count=FAPEXDEV_VIENC_SUB_MINOR_FIRST; count<=FAPEXDEV_VIENC_SUB_MINOR_LAST; count++) {
        /* Unregister this id's FDL device to CFDL. */
        FAPEXCFDL_UnregisterFdl(count);
    }

    FAPEXDebugPrint(1, "[FAPEXFDLVIENC_cleanup] fapex_fdlvienc was unloaded.\n");

    return;
}

#ifdef __LINUX__
/*******************************
 MACRO (Linux dep.)
********************************/
#ifdef BUILD_AS_MODULE
module_init(FAPEXFDLVIENC_init);
module_exit(FAPEXFDLVIENC_cleanup);

MODULE_LICENSE("Proprietary. Fujitsu Semiconductor.");
MODULE_DESCRIPTION("FAPI driver (fapex_fdlvienc object)");
#endif /* BUILD_AS_MODULE */
#endif /* __LINUX__ */
