/*!
*************************************************************************
**
** \file    fapex_auout_func.h
**
** \brief   function pointer definiton header file 
**
** \attention  This program is provided as is. You can redistribute it and/or
** modify it. Fujitsu Semiconductor Limited accepts no responsibility or
** liability for any errors or omissions.
** 
** Copyright (C) 2009-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
**
*************************************************************************
*/
#ifndef _FAPEX_AUOUT_FUNC_H
#define _FAPEX_AUOUT_FUNC_H

#include "fapi/drv_auout.h"
#include "fapi/drv_auout_apt_11db.h"
#include "fapi/drv_auout_apt_ssrc.h"
#include "fapi/drv_auout_apt_vol.h"
typedef struct {
    int32_t                (*Init)( void );
    void                   (*Exit)( void );
    uint32_t               (*Isr)( void );
    void                   (*SetInternalDac)( FAPI_AUOUT_I2sEnumT i2sId );
    void                   (*SetSpeakers)( FAPI_AUOUT_I2sEnumT i2sId,
                                           FAPI_AUOUT_SpeakerEnumT left,
                                           FAPI_AUOUT_SpeakerEnumT right);
    void                   (*SetI2sParams)( FAPI_AUOUT_I2sEnumT i2sId,
                                            FAPI_AUOUT_FormatEnumT format,
                                            FAPI_AUOUT_ClkEnumT aoClk,
                                            FAPI_AUOUT_BClkEnumT bitClk,
                                            uint32_t lrClkInv,
                                            uint32_t switchChannel,
                                            uint32_t dmonoBypass );
    int32_t                (*GetI2sParams)( FAPI_SYS_HandleT handle,
                                            FAPI_AUOT_I2sParamsT paramsArr[] );                                  
    void                   (*SetSpdifParams)( FAPI_AUOUT_ClkEnumT amcClk,
                                              uint32_t switchChannel,
                                              uint32_t parity );
    FAPI_SYS_HandleT       (*Open)(const FAPI_AUOUT_OpenParamsT* paramsPtr,
                                         int32_t* errorCodePtr );
    int32_t                (*Close)( FAPI_SYS_HandleT handle );
    int32_t                (*Start)( FAPI_SYS_HandleT handle );
    void                   (*Stop)( FAPI_SYS_HandleT handle );
    void                   (*SetStreamInfo)( FAPI_SYS_HandleT handle,
                                             FAPI_AUOUT_StreamInfoT* infoPtr );
    FAPI_AUOUT_StreamInfoT (*GetStreamInfo)( FAPI_SYS_HandleT handle );
    int32_t                (*StartHdmi)( FAPI_SYS_HandleT handle,uint8_t ifcMask );
    void                   (*StopHdmi)( void );
    void                   (*Mute)( FAPI_SYS_HandleT handle,
                                           FAPI_AUOUT_SpeakerEnumT speaker );
    void                   (*UnMute)( FAPI_SYS_HandleT handle,
                                           FAPI_AUOUT_SpeakerEnumT speaker );
    void                   (*SetVolume)( FAPI_SYS_HandleT handle,
                                           FAPI_AUOUT_SpeakerEnumT speaker,
                                           uint32_t volume );
    uint32_t               (*GetVolume)( FAPI_SYS_HandleT handle,
                                           FAPI_AUOUT_SpeakerEnumT speaker );
    void                   (*EnableDualMono)( FAPI_SYS_HandleT handle,
                                           FAPI_AUOUT_SpeakerEnumT speaker );
    void                   (*DisableDualMono)( FAPI_SYS_HandleT handle,
                                           FAPI_AUOUT_SpeakerEnumT speaker );
    int32_t                (*GetSpeakers)( FAPI_SYS_HandleT handle,
                                                   FAPI_AUOUT_SpeakerEnumT speakerArr[]);
    int32_t                (*GetI2sBuffers)( FAPI_SYS_HandleT handle,
                                                     FAPI_AUOUT_I2sBufferT bufferArr[]);
    FAPI_SYS_HandleT       (*GetSpdifBuffer)(FAPI_SYS_HandleT handle);
    FAPI_SYS_HandleT       (*GetPcmBuffer)(FAPI_SYS_HandleT handle);
    int32_t                (*RestartIfEmptied)(FAPI_SYS_HandleT handle);
    int32_t                (*SetSpdifCopyProtect)( FAPI_SYS_HandleT handle, 
                                                   uint8_t cpFlag);
    int32_t                (*GetSpdifCopyProtect)(FAPI_SYS_HandleT handle);
    void                   (*SetSpdifStatus)(FAPI_SYS_HandleT handle);
    int32_t                (*WritePcm)( FAPI_SYS_HandleT  auoutHandle,
                                        uint64_t  pts,
                                        FAPI_AUOUT_SampleBufDescrT* sampleBufDescr,
                                        uint8_t   ifcMask,
                                        uint32_t  timeout );
    /* Audio Processing Tool API */
     void                 (*SetApt)( FAPI_AUOUT_AptEnumT      aptId, 
                                     FAPI_AUOUT_AptExecFunc   execFunc,
                                     FAPI_AUOUT_AptAddFunc    addFunc,
                                     FAPI_AUOUT_AptRemoveFunc removeFunc,
                                     FAPI_AUOUT_AptResetFunc  resetFunc);

    FAPI_SYS_HandleT      (*AddApt)( FAPI_SYS_HandleT    auoutHandle,
                                     FAPI_AUOUT_AptEnumT aptId,
                                     void*               paramsPtr,
                                     uint32_t            paramLength,
                                     uint8_t             ifcMask,
                                     int32_t*            errorCodePtr);

    int32_t               (*RemoveApt)( FAPI_SYS_HandleT aptHandle );

    int32_t               (*SetAptParams)( FAPI_SYS_HandleT aptHandle,
                                           void*            paramsPtr,
                                           uint32_t         paramLength);
                                   
    int32_t               (*GetAptLastError)( FAPI_SYS_HandleT aptHandle );
    int32_t               (*ResetAptChain)( FAPI_SYS_HandleT auoutHandle );

    // SYNC API
    int32_t                (*CalculateSkipStallTime)( FAPI_SYS_HandleT handle,
                                                              uint64_t pts);
    int32_t                (*ExecuteSkipStall)( FAPI_SYS_HandleT handle,
                                                        uint64_t pts);
    void                   (*ExecuteSkipStallTime)( FAPI_SYS_HandleT handle,
                                                            int32_t skipStallTime);
    FAPI_AUOUT_SyncStateEnumT (*GetSyncState)( FAPI_SYS_HandleT handle);
    FAPI_AUOUT_SampleBufDescrT*  (*AptVolExecFunc)( 
                                             FAPI_AUOUT_SampleBufDescrT* inpDescrPtr,
                                             void*    paramsPtr,
                                             uint32_t paramLength,
                                             uint8_t  ifcMask,
                                             int32_t* errorCodePtr);
    FAPI_AUOUT_SampleBufDescrT*  (*Apt11DBExecFunc) (
                                             FAPI_AUOUT_SampleBufDescrT* inpDescrPtr,
                                             void*    paramsPtr,
                                             uint32_t paramLength,
                                             uint8_t  ifcMask,
                                             int32_t* errorCodePtr);
   FAPI_AUOUT_SampleBufDescrT*  (*AptSsrcExecFunc)(
                                             FAPI_AUOUT_SampleBufDescrT* inpDescrPtr,
                                             void*    paramsPtr,
                                             uint32_t paramLength,
                                             uint8_t  ifcMask,
                                             int32_t* errorCodePtr);
                                             
   int32_t                   (*AptSsrcAddFunc)(  void*    paramsPtr,
                                                uint32_t paramLength,
                                                uint8_t  ifcMask);      
   int32_t                   (*AptSsrcRemoveFunc)( void*    paramsPtr,
                                                   uint32_t paramLength,
                                                   uint8_t  ifcMask);
   int32_t                   (*AptSsrcResetFunc)( void*    paramsPtr,
                                                  uint32_t paramLength,
                                                  uint8_t  ifcMask);

} FAPEX_AUOUT_FuncStrT;

#endif /* _FAPEX_AUOUT_FUNC_H */
