/*!
*************************************************************************
**
** \file    fapex_dma_func.h
**
** \brief   function pointer definiton header file 
**
** \attention  This program is provided as is. You can redistribute it and/or
** modify it. Fujitsu Semiconductor Limited accepts no responsibility or
** liability for any errors or omissions.
**
** Copyright (C) 2009-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
**
*************************************************************************
*/
#ifndef _FAPEX_DMA_FUNC_H
#define _FAPEX_DMA_FUNC_H

#include "fapi/drv_dma.h"

typedef struct {
    int32_t            (*Init)( void );
    void               (*Exit)( void );
    uint32_t           (*Isr)( void );
    void               (*Bsr)(uint32_t privData);
    FAPI_DMA_RequestT* (*AllocateRequest)( uint32_t feature, uint32_t selfDestroy );
    void               (*ReleaseRequest)( FAPI_DMA_RequestT* dmaRequestPtr );
    uint32_t           (*GetRequestList)( uint32_t feature, FAPI_DMA_RequestS* dmaRequestPtrArray[], uint32_t requestArraySize );
    int32_t            (*CancelRequest)( FAPI_DMA_RequestT* dmaRequestPtr, uint32_t feature );
    int32_t            (*SendRequest)( FAPI_DMA_RequestT* dmaRequestP );
    int32_t            (*SendRequest2)( FAPI_DMA_RequestT* dmaRequestP, FAPI_DMA_RequestT* dmaRequest2Ptr );
    uint32_t           (*GetRequestStatus)( FAPI_DMA_RequestT* dmaRequestP );

    FAPI_DMA_RequestT* (*ATA_ReadData)(
                           FAPI_SYS_HandleT   ideHandle,
                           uint32_t           lba,
                           uint32_t           sectors,
                           uint32_t           sectorBytes,
                           volatile uint32_t* buffer,
                           FAPI_DMA_CallbackT finishCallback );
    FAPI_DMA_RequestT* (*ATA_WriteData)(
                           FAPI_SYS_HandleT   ideHandle,
                           uint32_t           lba,
                           uint32_t           sectors,
                           uint32_t           sectorBytes,
                           volatile uint32_t* buffer,
                           FAPI_DMA_CallbackT finishCallback );
    FAPI_DMA_RequestT* (*StartMemcopy)(
                           volatile uint32_t*    target,
                           volatile uint32_t*    source,
                           uint32_t              bytes,
                           FAPI_DMA_CallbackT    finishCallback );
    FAPI_DMA_RequestT* (*StartPFlashcopy)(
                           volatile uint32_t* target,
                           volatile uint32_t* source,
                           uint32_t           bytes,
                           FAPI_DMA_CallbackT finishCallback );
    FAPI_DMA_RequestT* (*StartMemfill)(
                           volatile uint32_t*          targetPtr,
                           FAPI_DMA_MemoryFillParamsT* fillParamsPtr,
                           FAPI_DMA_CallbackT          finishCallback);
    FAPI_DMA_RequestT* (*StartSearchReplace)(
                           volatile uint32_t*             buffer,
                           uint32_t                       bytes,
                           FAPI_DMA_SearchReplaceParamsT* srParams,
                           FAPI_DMA_CallbackT             hitCallback,
                           FAPI_DMA_CallbackT             finishCallback );
    FAPI_DMA_RequestT* (*StartCcir656Capture)(
                           volatile uint32_t*              targetBuffer,
                           uint32_t                        bufferBytes,
                           uint32_t                        bufferLineInc,
                           FAPI_DMA_Ccir656CaptureParamsT* captureParams,
                           FAPI_DMA_CallbackT              syncCallback,
                           void*                           syncOptDataPtr,
                           FAPI_DMA_CallbackT              finishCallback,
                           void*                           finishOptDataPtr,
                           int32_t*                        errorCodePtr );
    FAPI_DMA_RequestT* (*StartCcir656CaptureHalfMode)(
                           volatile uint32_t*              targetBuffer,
                           uint32_t                        bufferBytes,
                           uint32_t                        bufferLineInc,
                           FAPI_DMA_Ccir656CaptureParamsT* captureParams,
                           FAPI_DMA_CallbackT              syncCallback,
                           void*                           syncOptDataPtr,
                           FAPI_DMA_CallbackT              finishCallback,
                           void*                           finishOptDataPtr,
                           int32_t*                        errorCodePtr );
    int32_t            (*StopCcir656Capture)( FAPI_DMA_RequestT* dmaRequest );
    FAPI_DMA_RequestT* (*StartAudioCapture)(
                           volatile uint32_t*            targetBuffer,
                           uint32_t                      bufferBytes,
                           FAPI_DMA_AudioCaptureParamsT* captureParams,
                           FAPI_DMA_CallbackT            finishCallback,
                           void*                         finishOptDataPtr,
                           FAPI_DMA_CallbackT            splitCallback,
                           void*                         splitOptDataPtr,
                           uint32_t                      numberOfSplit,
                           int32_t*                      errorCodePtr );
    int32_t            (*StopAudioCapture)( FAPI_DMA_RequestT* dmaRequest );
    FAPI_DMA_RequestT* (*StartCrypto)(
                           volatile uint32_t*      sourceBuffer,
                           volatile uint32_t*      targetBuffer,
                           uint32_t                bufferBytes,
                           FAPI_DMA_CryptoParamsT* cryptoParams,
                           FAPI_DMA_CryptoKeysT*   cryptoKeys,
                           FAPI_DMA_CallbackT      finishCallback );
    FAPI_DMA_RequestT* (*StartCsConversion)(
                           volatile uint32_t*            sourceBuffer,
                           volatile uint32_t*            targetBuffer,
                           uint32_t                      sourceBytes,
                           FAPI_DMA_CsConversionParamsT* csParams,
                           FAPI_DMA_CsConversionCoeffT*  csCoeff,
                           FAPI_DMA_CallbackT            finishCallback );
} FAPEX_DMA_FuncStrT;

#endif /* _FAPEX_DMA_FUNC_H */
