/*!
*************************************************************************
**
** \file    fapex_flash_func.h
**
** \brief   function pointer definiton header file 
**
** \attention  This program is provided as is. You can redistribute it and/or
** modify it. Fujitsu Semiconductor Limited accepts no responsibility or
** liability for any errors or omissions.
**
** Copyright (C) 2009-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
**
*************************************************************************
*/
#ifndef _FAPEX_FLASH_FUNC_H
#define _FAPEX_FLASH_FUNC_H

#include "fapi/drv_flash.h"

typedef struct {
    void             (*SetEnable)(uint32_t deviceType, uint32_t enableFlag);
    void             (*SetUpiIndex)( uint32_t deviceType, uint32_t upiIndex );
    int32_t          (*Init)(void);
    void             (*Exit)(void);
    FAPI_SYS_HandleT (*Open)(const FAPI_FLASH_OpenParamsT* paramsPtr, int32_t* errorCodePtr );
    int32_t          (*Close)(FAPI_SYS_HandleT handle);
    int32_t          (*EraseChip)(FAPI_SYS_HandleT handle);
    int32_t          (*EraseSector)(FAPI_SYS_HandleT handle, uint32_t sectorIndex);
    int32_t          (*ReadBytes)(FAPI_SYS_HandleT handle, uint32_t address, uint32_t byteCount, void* bufferPtr);
    int32_t          (*WriteBytes)(FAPI_SYS_HandleT handle, uint32_t address, uint32_t byteCount, const void* bufferPtr);
    int32_t          (*Flush)( FAPI_SYS_HandleT handle );
    int32_t          (*GetDeviceSize)(FAPI_SYS_HandleT handle);
    int32_t          (*GetDeviceId)(FAPI_SYS_HandleT handle);
    int32_t          (*GetManufacturerId)(FAPI_SYS_HandleT handle);
    const char_t*    (*GetDeviceName)(FAPI_SYS_HandleT handle);
    const char_t*    (*GetManufacturerName)(FAPI_SYS_HandleT handle);
    int32_t          (*GetSectorCount)(FAPI_SYS_HandleT handle);
    int32_t          (*GetSectorIndex)(FAPI_SYS_HandleT handle, uint32_t address);
    int32_t          (*GetSectorStart)(FAPI_SYS_HandleT handle, uint32_t sectorIndex);
    int32_t          (*GetSectorBytes)(FAPI_SYS_HandleT handle, uint32_t sectorIndex);
    int32_t          (*IsSectorEmpty)(FAPI_SYS_HandleT handle, uint32_t sectorIndex);
    int32_t          (*IsChipEmpty)(FAPI_SYS_HandleT handle);
    int32_t          (*IsSectorProtected)(FAPI_SYS_HandleT handle, uint32_t sectorIndex);
    int32_t          (*ProtectSector)(FAPI_SYS_HandleT handle, uint32_t sectorIndex, uint32_t flag);
} FAPEX_FLASH_FuncStrT;

#endif /* _FAPEX_FLASH_FUNC_H */
