/*!
*************************************************************************
**
** \file    fapex_hdmi_func.h
**
** \brief   function pointer definiton header file 
**
** \attention  This program is provided as is. You can redistribute it and/or
** modify it. Fujitsu Semiconductor Limited accepts no responsibility or
** liability for any errors or omissions.
**
** Copyright (C) 2009-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
**
*************************************************************************
*/
#ifndef _FAPEX_HDMI_FUNC_H
#define _FAPEX_HDMI_FUNC_H

#include "fapi/drv_hdmi.h"

typedef struct {
    int32_t          (*Init)(void);
    void             (*Exit)(void);
    FAPI_SYS_HandleT (*Open)(const FAPI_HDMI_OpenParamsT* pParams,
                             int32_t* errorCodePtr);
    int32_t          (*Close)(FAPI_SYS_HandleT handle);
    int32_t          (*GetPhysicalAddressInEDID)(FAPI_SYS_HandleT handle, uint32_t* physAddrPtr);
    int32_t          (*ReadEDID)(FAPI_SYS_HandleT handle);
    int32_t          (*LinkHDCP)(FAPI_SYS_HandleT handle);
    int32_t          (*VIDEO_Start)( FAPI_SYS_HandleT handle );
    int32_t          (*AUDIO_Start)( FAPI_SYS_HandleT handle );
    void             (*VIDEO_Stop)( FAPI_SYS_HandleT handle );
    void             (*AUDIO_Stop)( FAPI_SYS_HandleT handle );
    int32_t          (*SetVideoParameters)(FAPI_SYS_HandleT handle,
                                           FAPI_HDMI_VideoParamsT* pVParams);
    int32_t          (*GetVideoParameters)(FAPI_SYS_HandleT handle,
                                           FAPI_HDMI_VideoParamsT* pVParams);
    int32_t          (*SetAudioParameters)(FAPI_SYS_HandleT handle,
                                           FAPI_HDMI_AudioParamsT* pAParams);
    int32_t          (*GetAudioParameters)(FAPI_SYS_HandleT handle,
                                           FAPI_HDMI_AudioParamsT* pAParams);
    int32_t          (*SetAudioSampleFrequency)(FAPI_SYS_HandleT handle,
                                                uint32_t sampleFreq);
    int32_t          (*CheckHPD)(FAPI_SYS_HandleT handle );
    int32_t          (*CheckHDCPState)(FAPI_SYS_HandleT handle );
    int32_t          (*GetEDID_Info)(FAPI_SYS_HandleT handle,
                                     FAPI_HDMI_EdidInfoT* pEdidInfo);
    int32_t          (*GetEDID_Data)(FAPI_SYS_HandleT handle,
                                     uint8_t* pEdidData, uint32_t bufLen);
    int32_t          (*SetRevocationList)(FAPI_SYS_HandleT handle,
                                          uint8_t* deviceKsv,
                                          uint32_t numDevices);
    int32_t          (*UpdateAviInfoFrame)(FAPI_SYS_HandleT handle );
    int32_t          (*DynamicChangeHDCP)(FAPI_SYS_HandleT handle,
                                          uint32_t onOffFlag);

} FAPEX_HDMI_FuncStrT;

#endif /* _FAPEX_HDMI_FUNC_H */
