/*!
*************************************************************************
**
** \file    fapex_i2c_func.h
**
** \brief   function pointer definiton header file 
**
** \attention  This program is provided as is. You can redistribute it and/or
** modify it. Fujitsu Semiconductor Limited accepts no responsibility or
** liability for any errors or omissions.
**
** Copyright (C) 2009-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
**
*************************************************************************
*/
#ifndef _FAPEX_I2C_FUNC_H
#define _FAPEX_I2C_FUNC_H

#include "fapi/drv_i2c.h"

typedef struct {
    int32_t          (*Init)(void);
    void             (*Exit)(void);
    uint32_t         (*Isr)(void);
    FAPI_SYS_HandleT (*Open)(const FAPI_I2C_OpenParamsT* pParams,
                             int32_t* errorCodePtr);
    void             (*Reset)(FAPI_SYS_HandleT handle);
    int32_t          (*Close)(const FAPI_SYS_HandleT handle);
    void             (*CancelProcess)(const FAPI_SYS_HandleT handle);
    int32_t          (*Process)(const FAPI_SYS_HandleT handle,
                                uint8_t* dataPtr,
                                const uint32_t dataLength);
    int32_t          (*Read)(const FAPI_SYS_HandleT handle,
                             const uint8_t msSlaveAddress,
                             uint8_t* txDataPtr,
                             const uint32_t txDataLength,
                             uint8_t* rxDataPtr,
                             uint32_t rxDataLength);
    int32_t          (*Write)(const FAPI_SYS_HandleT handle,
                              const uint8_t msSlaveAddress,
                              uint8_t* txDataPtr,
                              const uint32_t txDataLength);
    void             (*SetOperationMode)(const uint32_t blockIndex,
                                         const FAPI_I2C_OperationModeT operationMode);
    int32_t          (*SetProtocol)(const FAPI_SYS_HandleT handle,
                                          const FAPI_I2C_ProtocolT protocol);
    void             (*SetMaxHandleCount)(const uint32_t blockIndex,
                                          const uint32_t maximum);
} FAPEX_I2C_FuncStrT;

#endif /* _FAPEX_I2C_FUNC_H */
