/*!
*************************************************************************
**
** \file    fapex_ir_func.h
**
** \brief   function pointer definiton header file 
**
** \attention  This program is provided as is. You can redistribute it and/or
** modify it. Fujitsu Semiconductor Limited accepts no responsibility or
** liability for any errors or omissions.
**
** Copyright (C) 2009-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
**
*************************************************************************
*/
#ifndef _FAPEX_IR_FUNC_H
#define _FAPEX_IR_FUNC_H

#include "fapi/drv_ir.h"
#include "fapi/drv_ir_nec.h"
#include "fapi/drv_ir_rc5.h"

typedef struct {
    int32_t           (*Init)(void);
    void              (*Exit)(void);
    uint32_t          (*Isr)(void);
    void              (*SetFrequency)(const uint32_t Hz);
//  void              (*SetProtocolMode)(const uint32_t supportMode);
    FAPI_SYS_HandleT  (*Open)(const FAPI_IR_OpenParamsT *paramsPtr,
                              int32_t* errorCodePtr);
    int32_t           (*Close)(FAPI_SYS_HandleT handle);
    int32_t           (*ReadData)(const FAPI_SYS_HandleT handle,
                                        uint32_t *dataPtr,
                                  const uint32_t length);
    int32_t           (*GetStatus)(const FAPI_SYS_HandleT handle);
    int32_t           (*GetKey)(const FAPI_SYS_HandleT handle,
                                int32_t*  deviceIdPtr,
                                uint32_t* keyValuePtr,
                                uint32_t* RcmCode,
                                uint32_t* repeatFlagPtr);
    void              (*Bsr)(uint32_t handle);
    int32_t           (*SetMapTable)(const FAPI_IR_MappingTableT *mappingTablePtr);
    int32_t           (*RemoveMapTable)(const int32_t mappingDeviceId);
    int32_t           (*SetUserProtocol)(const FAPI_IR_ProtocolParamsT *paramsPtr,
                                              const uint32_t Hz, const uint32_t userMode);
    void              (*SetMaskData)(const uint32_t id_mask, const uint32_t data_mask);
    void              (*SetWakeUpKey)(const uint32_t rawData,const uint32_t mask);

    int32_t           (*NEC_Mode)(FAPI_SYS_HandleT handle);
    int32_t           (*RC5_Mode)(FAPI_SYS_HandleT handle);
} FAPEX_IR_FuncStrT;

#endif /* _FAPEX_IR_FUNC_H */
