/*!
*************************************************************************
**
** \file    fapex_mmu_func.h
**
** \brief   function pointer definiton header file 
**
** \attention  This program is provided as is. You can redistribute it and/or
** modify it. Fujitsu Semiconductor Limited accepts no responsibility or
** liability for any errors or omissions.
**
** Copyright (C) 2009-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
**
*************************************************************************
*/
#ifndef _FAPEX_MMU_FUNC_H
#define _FAPEX_MMU_FUNC_H

#include "fapi/drv_mmu.h"

typedef struct {
    void             (*SetSegmentSizeMBytes)(uint32_t index,uint32_t sizeMBytes);
    void             (*SetSegmentStragegyMode)(uint32_t index,uint32_t strategyMode);
    void             (*SetSegmentFunctionMode)(uint32_t index,uint32_t functionMode);
    void             (*SetSegmentAlignmentBits)(uint32_t index,uint32_t alignmentBits);
    int32_t          (*Init)( void );
    void             (*Exit)( void );
    void             (*SetTlb1BaseAddr)( uint32_t tlb1BaseAddr );
    void             (*SetTlb2BaseAddr)( uint32_t tlb2BaseAddr );
    FAPI_SYS_HandleT (*Open)( FAPI_MMU_OpenParamsT* openParamsPtr, int32_t* errorCodePtr );
    int32_t          (*Close)( FAPI_SYS_HandleT handle );
    uint32_t         (*GetPhysAddr)( uint32_t virtAddr );
    uint32_t         (*GetVirtAddr)( uint32_t physAddr );
    int32_t          (*ChangeMapEntry)( FAPI_SYS_HandleT handle, uint32_t virtAddr,
                                        uint32_t physAddr, uint32_t accessFlags );
    FAPI_SYS_HandleT (*CreateSegment)(const FAPI_MMU_HeapParamsT* paramsPtr,int32_t* errorCodePtr);
    int32_t          (*ReleaseSegment)( FAPI_SYS_HandleT handle );
    int32_t          (*ResetSegment)( FAPI_SYS_HandleT handle );
    FAPI_SYS_HandleT (*GetSegment)( void* memPtr );
    int32_t          (*GetSegmentInfo)( FAPI_SYS_HandleT handle, FAPI_MMU_HeapSegmentInfoT* infoPtr );

    void*            (*Malloc)(FAPI_SYS_HandleT handle, uint32_t bytes);
    void*            (*Zalloc)(FAPI_SYS_HandleT handle, uint32_t bytes);
    void*            (*Realloc)(FAPI_SYS_HandleT handle,void* memPtr,uint32_t bytes);
    void*            (*Free)(FAPI_SYS_HandleT handle,void* memPtr);
    void             (*ReportMemory)(FAPI_SYS_HandleT handle,uint32_t dumpChunks);
} FAPEX_MMU_FuncStrT;

#endif /* _FAPEX_MMU_FUNC_H */
