/*!
*************************************************************************
**
** \file    fapex_viscale_func.h
**
** \brief   function pointer definiton header file 
**
** \attention  This program is provided as is. You can redistribute it and/or
** modify it. Fujitsu Semiconductor Limited accepts no responsibility or
** liability for any errors or omissions.
**
** Copyright (C) 2009-2010 Fujitsu Semiconductor Limited, All Rights Reserved.
**
*************************************************************************
*/
#ifndef _FAPEX_VISCALE_FUNC_H
#define _FAPEX_VISCALE_FUNC_H

#include "fapi/drv_viscale.h"

typedef struct {
    void             (*PreInit)(const uint32_t piIpcMemoryAllocationType);
    int32_t          (*Init)(void);
    void             (*Exit)(void);
    uint32_t         (*Isr)(void);
    void             (*Bsr)(uint32_t arg);
    FAPI_SYS_HandleT (*Open)(
                          const FAPI_VISCALE_OpenParamsT* paramsPtr,
                                int32_t* errorCodePtr);
    int32_t          (*Close)(FAPI_SYS_HandleT handle);
    uint32_t         (*BackgroundColour)(
                                FAPI_VISCALE_ParameterModeEnumT mode,
                                uint8_t Y, uint8_t Cr, uint8_t Cb);
    uint32_t         (*PlaneOrder)(
                                 FAPI_VISCALE_ParameterModeEnumT mode,
                                 FAPI_VISCALE_PlaneDataSelectorEnumT plane5,
                                 FAPI_VISCALE_PlaneDataSelectorEnumT plane4,
                                 FAPI_VISCALE_PlaneDataSelectorEnumT plane3,
                                 FAPI_VISCALE_PlaneDataSelectorEnumT plane2,
                                 FAPI_VISCALE_PlaneDataSelectorEnumT plane1,
                                 FAPI_VISCALE_PlaneDataSelectorEnumT plane0);
    int32_t          (*KillOutputs)(void);
    int32_t          (*SetBrightness)(FAPI_VISCALE_ParameterModeEnumT mode,
                                      int32_t level);
    int32_t          (*SetContrast)(FAPI_VISCALE_ParameterModeEnumT mode,
                                    int32_t level);
    int32_t          (*SetSaturation)(FAPI_VISCALE_ParameterModeEnumT mode,
                                      int32_t level);
    int32_t          (*Start)(FAPI_SYS_HandleT handle,
                                FAPI_VISCALE_ScalerParamsT*
                                scaleParamsPtr);
    int32_t          (*Stop)(FAPI_SYS_HandleT handle);
    int32_t          (*EnableDisable)(FAPI_SYS_HandleT handle,
                                FAPI_VISCALE_ParameterModeEnumT mode);
    uint32_t         (*Alpha)(FAPI_SYS_HandleT handle,
                                FAPI_VISCALE_ParameterModeEnumT mode,
                                uint32_t level);
    int32_t          (*Resize)(FAPI_SYS_HandleT handle ,
                                FAPI_VISCALE_ScalerParamsT*
                                scaleParamsPtr);
    int32_t          (*Hilight)(FAPI_SYS_HandleT handle,
                                FAPI_VISCALE_ParameterModeEnumT mode);
    int32_t          (*AutoPlaneFading)(FAPI_SYS_HandleT handle,
                                FAPI_VISCALE_VideoFadingModeEnumT fadingMode);
    int32_t          (*AutoPlaneFadingStepSize)(FAPI_SYS_HandleT handle,
                                             FAPI_VISCALE_ParameterModeEnumT mode,
                                                           uint32_t* stepSizePtr);
    int32_t          (*UpdateClut)(FAPI_SYS_HandleT handle,
                                uint32_t* colour);
    int32_t          (*UpdateSingleClut)(FAPI_SYS_HandleT handle,
                                uint32_t entryNumber, uint32_t entryColour);
    void             (*AutoScaler)(void* frameCropPtr,
                                FAPI_SYS_HandleT handle,
                                uint32_t suspendType);
    int32_t          (*SetScalerToHdEncoder)(FAPI_SYS_HandleT handle,
                                FAPI_SYS_HandleT viencHdHandle);
    int32_t          (*SetScalingMethod)(FAPI_SYS_HandleT handle,
                                FAPI_VISCALE_ScalingMethodEnumT scalingMethod);
    int32_t          (*SetDisplayAspectRatio)(FAPI_SYS_HandleT handle,
                                FAPI_VISCALE_AspectRatioEnumT displayAr);
    int32_t          (*SetWssState)(FAPI_SYS_HandleT handle,
                                              FAPI_VISCALE_WssStateEnumT wssState);
    int32_t          (*SetParamChangeCallback)(FAPI_SYS_HandleT handle,
                                FAPI_VISCALE_ParamChangeCallbackT
                                                     paramChangeCallbackFunction,
                                                       void* paramChangeOptData);
    int32_t          (*SetWbuCaptureMode)(FAPI_SYS_HandleT handle,
                                  FAPI_VISCALE_WbuCaptureModeEnumT wbuCaptureMode);
    int32_t          (*ClearScalerToHdEncoder)(FAPI_SYS_HandleT handle);
    int32_t          (*ClearScalingMethod)(FAPI_SYS_HandleT handle);
    int32_t          (*ClearDisplayAspectRatio)(FAPI_SYS_HandleT handle);
    int32_t          (*ClearParamChangeCallback)(FAPI_SYS_HandleT handle);
    int32_t          (*ClearWbuCaptureMode)(FAPI_SYS_HandleT handle);
    FAPI_VISCALE_SystemModeEnumT 
                     (*GetSystemType)(FAPI_SYS_HandleT handle);
    int32_t          (*GetWssSignal)(FAPI_SYS_HandleT handle,
                                                 uint32_t* wssSignalPtr);
    int32_t          (*GetScalerParams)(FAPI_SYS_HandleT handle,
                               FAPI_VISCALE_GetScalerParamsT* getScalerParamsPtr);

} FAPEX_VISCALE_FuncStrT;

#endif /* _FAPEX_VISCALE_FUNC_H */
