@echo off
rem ############################################################################
rem ############################################################################
rem ##
rem ## BurnFlash4Linux.bat -- Windows command script to burn image into Flash
rem ##
rem ## (C) Copyright 2009-2010 by Fujitsu Microelectronics Limited
rem ##
rem ## This batch file creates the flash image file for Linux environmnt
rem ## and burns it into Flash memory.
rem ##
rem ## Usage: BurnFlash4Linux [arguments...]
rem ##  The [arguments...] are accepted only once
rem ##   -boot={rom, slot}          : booter type
rem ##                                - rom  : ROM loader booting
rem ##                                - slot : 2nd level 'booter' booting
rem ##   -firm={external, embedded} : firmware registration type
rem ##                                - external : External Solution
rem ##                                - embedded : Embedded Solution
rem ##   -load=<hex value>        : kernel loading address
rem ##                                set the value whitch added size of
rem ##                                "Memory Size for Video/Audio Firmware"
rem ##                                at "meke menuconfig" and 0x20008000.
rem ##   If arguments are not set, 'slot', 'external' and '64M' are selected.
rem ##   Please refer "Booting Methods" of HTML documentation for more detail.
rem ##
rem ############################################################################
rem ############################################################################

setlocal

rem ############################################################################
rem ##
rem ## user environment parameters
rem ##
rem ############################################################################

set TOP_DIR=../..

rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // configuration file
rem //
rem ////////////////////////////////////////////////////////////////////////////

set CONFIG_FILE=%TOP_DIR%/Config.mak


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // booter image file (It is necessary if booter type is "slot")
rem //
rem ////////////////////////////////////////////////////////////////////////////

set BOOTER_FILE=fastboot.bin

rem If you can use RTOS environment, the following binary file might be able to
rem be used for BOOTER_FILE.
rem set BOOTER_FILE=%TOP_DIR%/applications/BurnFlash4Linux/fastboot/build/flimage.bin


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // slot parameters (It is necessary if booter type is "slot")
rem //
rem ////////////////////////////////////////////////////////////////////////////

set BOOTER_SLOT_NAME="fapex_Linux_demo"
set BOOTER_SLOT_0=1
set BOOTER_SLOT_1=2
set BOOTER_IMGVENDOR="fmc"
set BOOTER_SLOT_VERSION=1
set BOOTER_SLOT_COMPRESS=off
set BOOTER_SLOT_OFFSETS=0x00140000 0x00800000

set BOOTER_SLOT_INDEX=%BOOTER_SLOT_0%

for /f "tokens=1,2 delims= " %%a in ("%BOOTER_SLOT_OFFSETS%") do (
    if "%BOOTER_SLOT_INDEX%"=="%BOOTER_SLOT_0%" (
        set BOOTER_SLOT_OFFSET=%%a
    ) else (
        set BOOTER_SLOT_OFFSET=%%b
    )
)
:SLOT_PARAM_END


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // Linux kernel image file
rem //
rem ////////////////////////////////////////////////////////////////////////////

rem Please copy the Kernel image file "${KERNEL_TOP}/arch/arm/boot/Image or zImage"
rem to this folder.

set KIMAGE_FILE=Image
rem set KIMAGE_FILE=X:EmbeddedLinuxKernel-v16.1-mb86hxx/arch/arm/boot/Image

rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // Linux kernel loading address (BUF area size + 0x20008000)
rem //
rem ////////////////////////////////////////////////////////////////////////////

set KERNEL_LOAD_ADDRESS=0x22008000

rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // USB port parameter
rem //
rem ////////////////////////////////////////////////////////////////////////////

set USB_PORT=-port=USB0:DEBUGPORT_JTAGPORT:SYSFREQ=396000000:STARTBAUD=38400:BAUDRATE=1152000:RAMBASE=0x02004000


rem ############################################################################
rem ##
rem ## common parameters
rem ##
rem ############################################################################

rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // parameter files
rem //
rem ////////////////////////////////////////////////////////////////////////////

set RVS_FILE=%TOP_DIR%/env/execute/ddr_init_system1gb_video1gb.rvs
set WRITER_FILE=%TOP_DIR%/env/bin/M-arm11-realview/flasher_61.axf
set RELEASE_FILE=%TOP_DIR%/env/make/Makefile.release


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // output files
rem //
rem ////////////////////////////////////////////////////////////////////////////

set IMAGE_FILE=flimage.bin
set SLOT_IMAGE_FILE=bslotimg.bin


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // execution program directory
rem //
rem ////////////////////////////////////////////////////////////////////////////

set EXE_DIR=%TOP_DIR%/env/bin/M-win32

set PATH_SAVE=%PATH%
set PATH=%EXE_DIR%;%PATH%


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // firmware directory
rem //
rem ////////////////////////////////////////////////////////////////////////////

set AFIRM_DIR=%TOP_DIR%/subsystems/audio/firmware/build/MB86H61/firmware
set VFIRM_DIR=%TOP_DIR%/subsystems/video/firmware/build/MB86H61/firmware


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // flash parameters
rem //
rem ////////////////////////////////////////////////////////////////////////////

for /f "tokens=1,3" %%a in (%CONFIG_FILE%) do (
   if %%a == FLASHTYPE          set FLASHTYPE=%%b
   if %%a == FLASHBASE          set FLASHBASE=%%b
   if %%a == FLASHER_INFO_BASE  set FLASHER_INFO_BASE=%%b
   if %%a == FLASHER_IMAGE_BASE set FLASHER_IMAGE_BASE=%%b
)


rem ############################################################################
rem ##
rem ## check arguments
rem ##
rem ############################################################################

set BOOT_TYPE=slot
set FIRM_TYPE=external

if "%1"=="" goto SETTING_END

:ARG_SETTING
rem ============================================================================
rem echo %1 %2
if "%1"=="-boot" goto BOOT_SETTING
if "%1"=="-firm" goto FIRM_SETTING
if "%1"=="-load" goto LOAD_SETTING
goto SETTING_ERROR


:BOOT_SETTING
rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // booting type setting
rem //
rem ////////////////////////////////////////////////////////////////////////////

if "%2"=="rom"  goto BOOT_SETTING_END
if "%2"=="slot" goto BOOT_SETTING_END
goto SETTING_ERROR

:BOOT_SETTING_END
set BOOT_TYPE=%2
shift
shift
if "%1"=="" goto SETTING_END
goto ARG_SETTING


:FIRM_SETTING
rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // firmware registration type setting
rem //
rem ////////////////////////////////////////////////////////////////////////////

if "%2"=="external" goto FIRM_SETTING_END
if "%2"=="embedded" goto FIRM_SETTING_END
goto SETTING_ERROR

:FIRM_SETTING_END
set FIRM_TYPE=%2
shift
shift
if "%1"=="" goto SETTING_END
goto ARG_SETTING


:LOAD_SETTING
rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // Linux kernel loading address (BUF area size + 0x20008000)
rem //
rem ////////////////////////////////////////////////////////////////////////////

set TEMP=%2
if %TEMP:~0,2% == 0x goto LOAD_SETTING_END
goto SETTING_ERROR

:LOAD_SETTING_END
set KERNEL_LOAD_ADDRESS=%2
shift
shift
if "%1"=="" goto SETTING_END
goto ARG_SETTING


:SETTING_ERROR
rem ============================================================================
echo;
echo Usage: BurnFlash4Linux [arguments...]
echo  The [arguments...] are accepted only once
echo   -boot={rom, slot}          : booter type
echo                                - rom  : ROM loader booting
echo                                - slot : 2nd level 'booter' booting (*)
echo   -firm={external, embedded} : firmware registration type
echo                                - external : External Solution (*)
echo                                - embedded : Embedded Solution
echo   -load=^<hex value^>        : kernel loading address
echo                                set the value whitch added size of "Memory Size for Video/Audio Firmware"
echo                                at "meke menuconfig" and 0x20008000. (* 0x22008000)
echo  (*) ... default setting
echo;
goto END

:SETTING_END


rem ############################################################################
rem ##
rem ## clean flash image file
rem ##
rem ############################################################################

if "%BOOT_TYPE%"=="slot" (
    rm.exe -rf %SLOT_IMAGE_FILE%
) else (
    rm.exe -rf %IMAGE_FILE%
)


rem ############################################################################
rem ##
rem ## generate flash image file
rem ##
rem ############################################################################

rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // setup firmware version
rem //
rem ////////////////////////////////////////////////////////////////////////////

for /f "tokens=1,3" %%a in (%RELEASE_FILE%) do (
   if %%a == RELEASE_AUDIO_DECODER_VERSION set ADEC0_VERSION=%%b
   if %%a == RELEASE_VIDEC_MPEG_0_VERSION  set VDEC_M0_VERSION=%%b
   if %%a == RELEASE_VIDEC_MPEG_1_VERSION  set VDEC_M1_VERSION=%%b
   if %%a == RELEASE_VIDEC_MPEG_2_VERSION  set VDEC_M2_VERSION=%%b
   if %%a == RELEASE_VIDEC_MPEG_3_VERSION  set VDEC_M3_VERSION=%%b
   if %%a == RELEASE_VIDEC_MPEG_4_VERSION  set VDEC_M4_VERSION=%%b
   if %%a == RELEASE_VIDEC_H264_0_VERSION  set VDEC_H0_VERSION=%%b
   if %%a == RELEASE_VIDEC_H264_1_VERSION  set VDEC_H1_VERSION=%%b
   if %%a == RELEASE_VIDEC_H264_2_VERSION  set VDEC_H2_VERSION=%%b
   if %%a == RELEASE_VIDEC_AVS_0_VERSION   set VDEC_A0_VERSION=%%b
   if %%a == RELEASE_VIDEC_AVS_1_VERSION   set VDEC_A1_VERSION=%%b
   if %%a == RELEASE_VIDEC_AVS_2_VERSION   set VDEC_A2_VERSION=%%b
   if %%a == RELEASE_VIDEC_AVS_3_VERSION   set VDEC_A3_VERSION=%%b
   if %%a == RELEASE_VIDEC_AVS_4_VERSION   set VDEC_A4_VERSION=%%b
   if %%a == RELEASE_VIDEC_VC1_0_VERSION   set VDEC_V0_VERSION=%%b
   if %%a == RELEASE_VIDEC_VC1_1_VERSION   set VDEC_V1_VERSION=%%b
   if %%a == RELEASE_VIDEC_VC1_2_VERSION   set VDEC_V2_VERSION=%%b
   if %%a == RELEASE_VIDEC_VC1_3_VERSION   set VDEC_V3_VERSION=%%b
   if %%a == RELEASE_VIDEC_VC1_4_VERSION   set VDEC_V4_VERSION=%%b
)


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // Audio firmware parameter
rem //
rem ////////////////////////////////////////////////////////////////////////////

set AFIRM_PARAM=-s3hex=%AFIRM_DIR%/audio_decoder.s3hex@ARC,,,EXEC,,,%ADEC0_VERSION%


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // Video firmware parameter (refer subsystems\video\libfirmware\Makefile)
rem //
rem ////////////////////////////////////////////////////////////////////////////

set FW_BINFILE_VIDEO_MPEG_JUMP=%VFIRM_DIR%/dec_mpeg0.bin
set FW_BINFILE_VIDEO_MPEG_CODE=%VFIRM_DIR%/dec_mpeg1.bin
set FW_BINFILE_VIDEO_MPEG_CTRL=%VFIRM_DIR%/dec_mpeg2.bin
set FW_BINFILE_VIDEO_MPEG_PARS=%VFIRM_DIR%/dec_mpeg_parse.bin
set FW_BINFILE_VIDEO_MPEG_VLDC=%VFIRM_DIR%/dec_mpeg_vld.bin
set FW_BINFILE_VIDEO_H264_JUMP=%VFIRM_DIR%/dec_h2640.bin
set FW_BINFILE_VIDEO_H264_CODE=%VFIRM_DIR%/dec_h2641.bin
set FW_BINFILE_VIDEO_H264_CTRL=%VFIRM_DIR%/dec_h2642.bin
set FW_BINFILE_VIDEO_AVS_JUMP=%VFIRM_DIR%/dec_avs0.bin
set FW_BINFILE_VIDEO_AVS_CODE=%VFIRM_DIR%/dec_avs1.bin
set FW_BINFILE_VIDEO_AVS_CTRL=%VFIRM_DIR%/dec_avs2.bin
set FW_BINFILE_VIDEO_AVS_PARS=%VFIRM_DIR%/dec_avs_parse.bin
set FW_BINFILE_VIDEO_AVS_VLDC=%VFIRM_DIR%/dec_avs_vld.bin
set FW_BINFILE_VIDEO_VC1_JUMP=%VFIRM_DIR%/dec_vc10.bin
set FW_BINFILE_VIDEO_VC1_CODE=%VFIRM_DIR%/dec_vc11.bin
set FW_BINFILE_VIDEO_VC1_CTRL=%VFIRM_DIR%/dec_vc12.bin
set FW_BINFILE_VIDEO_VC1_PARS=%VFIRM_DIR%/dec_vc1_parse.bin
set FW_BINFILE_VIDEO_VC1_VLDC=%VFIRM_DIR%/dec_vc1_vld.bin

set VFIRM_M0_PARAM=-bin=%FW_BINFILE_VIDEO_MPEG_JUMP%@RAW,0x60040000,,NONE,SWAP,,%VDEC_M0_VERSION%
set VFIRM_M1_PARAM=-bin=%FW_BINFILE_VIDEO_MPEG_CODE%@RAW,0x60080000,,NONE,SWAP,,%VDEC_M1_VERSION%
set VFIRM_M2_PARAM=-bin=%FW_BINFILE_VIDEO_MPEG_CTRL%@RAW,0x60097c00,,NONE,SWAP,,%VDEC_M2_VERSION%
set VFIRM_M3_PARAM=-bin=%FW_BINFILE_VIDEO_MPEG_PARS%@RAW,0xFFFFFFFF,,NONE,SWAP,,%VDEC_M3_VERSION%
set VFIRM_M4_PARAM=-bin=%FW_BINFILE_VIDEO_MPEG_VLDC%@RAW,0xFFFFFFFF,,NONE,SWAP,,%VDEC_M4_VERSION%
set VFIRM_H0_PARAM=-bin=%FW_BINFILE_VIDEO_H264_JUMP%@RAW,0x60049000,,NONE,SWAP,,%VDEC_H0_VERSION%
set VFIRM_H1_PARAM=-bin=%FW_BINFILE_VIDEO_H264_CODE%@RAW,0x60080000,,NONE,SWAP,,%VDEC_H1_VERSION%
set VFIRM_H2_PARAM=-bin=%FW_BINFILE_VIDEO_H264_CTRL%@RAW,0x60097c00,,NONE,SWAP,,%VDEC_H2_VERSION%
set VFIRM_A0_PARAM=-bin=%FW_BINFILE_VIDEO_AVS_JUMP%@RAW,0x60040000,,NONE,SWAP,,%VDEC_A0_VERSION%
set VFIRM_A1_PARAM=-bin=%FW_BINFILE_VIDEO_AVS_CODE%@RAW,0x60080000,,NONE,SWAP,,%VDEC_A1_VERSION%
set VFIRM_A2_PARAM=-bin=%FW_BINFILE_VIDEO_AVS_CTRL%@RAW,0x60097c00,,NONE,SWAP,,%VDEC_A2_VERSION%
set VFIRM_A3_PARAM=-bin=%FW_BINFILE_VIDEO_AVS_PARS%@RAW,0xFFFFFFFF,,NONE,SWAP,,%VDEC_A3_VERSION%
set VFIRM_A4_PARAM=-bin=%FW_BINFILE_VIDEO_AVS_VLDC%@RAW,0xFFFFFFFF,,NONE,SWAP,,%VDEC_A4_VERSION%
set VFIRM_V0_PARAM=-bin=%FW_BINFILE_VIDEO_VC1_JUMP%@RAW,0x60040000,,NONE,SWAP,,%VDEC_V0_VERSION%
set VFIRM_V1_PARAM=-bin=%FW_BINFILE_VIDEO_VC1_CODE%@RAW,0x60080000,,NONE,SWAP,,%VDEC_V1_VERSION%
set VFIRM_V2_PARAM=-bin=%FW_BINFILE_VIDEO_VC1_CTRL%@RAW,0x60097c00,,NONE,SWAP,,%VDEC_V2_VERSION%
set VFIRM_V3_PARAM=-bin=%FW_BINFILE_VIDEO_VC1_PARS%@RAW,0xFFFFFFFF,,NONE,SWAP,,%VDEC_V3_VERSION%
set VFIRM_V4_PARAM=-bin=%FW_BINFILE_VIDEO_VC1_VLDC%@RAW,0xFFFFFFFF,,NONE,SWAP,,%VDEC_V4_VERSION%

set VFIRM_PARAM=%VFIRM_M0_PARAM% %VFIRM_M1_PARAM% %VFIRM_M2_PARAM% %VFIRM_M3_PARAM% %VFIRM_M4_PARAM% %VFIRM_H0_PARAM% %VFIRM_H1_PARAM% %VFIRM_H2_PARAM%
rem set VFIRM_PARAM=%VFIRM_M0_PARAM% %VFIRM_M1_PARAM% %VFIRM_M2_PARAM% %VFIRM_H0_PARAM% %VFIRM_H1_PARAM% %VFIRM_H2_PARAM%


rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // Linux kernel Image parameter
rem //
rem ////////////////////////////////////////////////////////////////////////////

set LINUX_IMG_PARAM=-bin=%KIMAGE_FILE%@ARM,%KERNEL_LOAD_ADDRESS%,%KERNEL_LOAD_ADDRESS%,EXEC

rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // Image binary data
rem //
rem ////////////////////////////////////////////////////////////////////////////

if "%FIRM_TYPE%"=="embedded" (
    set BIN_PARAMS=%LINUX_IMG_PARAM%
) else (
    set BIN_PARAMS=%AFIRM_PARAM% %VFIRM_PARAM% %LINUX_IMG_PARAM%
)


rem ============================================================================
if "%BOOT_TYPE%"=="slot" goto GENERATE_SLOT_IMAGE
    rem ////////////////////////////////////////////////////////////////////////
    rem //
    rem // generate flash image file for ROM boot
    rem //
    rem ////////////////////////////////////////////////////////////////////////

    set FLIMGEN_FLAGS=-verbose -decoder=MB86H61 -flashtype=%FLASHTYPE% -out=%IMAGE_FILE%

    flimgen.exe %FLIMGEN_FLAGS% -reg=%RVS_FILE% %BIN_PARAMS%

    goto GENERATE_IMAGE_END

:GENERATE_SLOT_IMAGE
    rem ////////////////////////////////////////////////////////////////////////
    rem //
    rem // generate flash image file for slot boot
    rem //
    rem ////////////////////////////////////////////////////////////////////////

    set BIMGEN_FLAGS=-verbose -compress=%BOOTER_SLOT_COMPRESS% -name=%BOOTER_SLOT_NAME% -out=%SLOT_IMAGE_FILE% -version=%BOOTER_SLOT_VERSION% -reg=%RVS_FILE%

    bimgen.exe %BIMGEN_FLAGS% %BIN_PARAMS%

:GENERATE_IMAGE_END


rem ############################################################################
rem ##
rem ## burn flash image file
rem ##
rem ############################################################################

rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // setup burn flash parameters
rem //
rem ////////////////////////////////////////////////////////////////////////////

set USB_DLL=%EXE_DIR%/usbdbgport.dll

set FLASHER_FLAGS=-verbose -flash -flashtype=%FLASHTYPE% -flashbase=%FLASHBASE% -flashinfo=%FLASHER_INFO_BASE%
set BURN_PARAMS=-dll=%USB_DLL% %USB_PORT% -elf=%WRITER_FILE%  
rem -wait=polling -rblmsg


echo;
echo Please push the reset button SW1(H61-DK) Boot until starting count-up!
color 70
sleep 1
color
pause

if "%BOOT_TYPE%"=="slot" goto SETUP_SLOT_PARAMS
    set FLASH_OFFSET=0x0
    set IMAGE_PARAM=-bin=%IMAGE_FILE%@RAW,%FLASHER_IMAGE_BASE%,,LOAD

    goto BURN_FLASH_IMAGE

:SETUP_SLOT_PARAMS
    rem ////////////////////////////////////////////////////////////////////////
    rem //
    rem // burn flash booter file
    rem //
    rem ////////////////////////////////////////////////////////////////////////

    execute.exe %FLASHER_FLAGS% -flashoffset=0x0 -gpio-rx=2 -gpio-tx=3 %BURN_PARAMS% -reg=%RVS_FILE% -bin=%BOOTER_FILE%@RAW,%FLASHER_IMAGE_BASE%,,LOAD

    set FLASH_OFFSET=%BOOTER_SLOT_OFFSET%
    set IMAGE_PARAM=-bin=%SLOT_IMAGE_FILE%@RAW,%FLASHER_IMAGE_BASE%,,LOAD
    echo;
    echo Please push the reset button SW1(H61-DK) Kernel until starting count-up again!
    color 70
    sleep 1
    color
    pause

:BURN_FLASH_IMAGE
rem ////////////////////////////////////////////////////////////////////////////
rem //
rem // burn flash image file
rem //
rem ////////////////////////////////////////////////////////////////////////////

execute.exe %FLASHER_FLAGS% -flashoffset=%FLASH_OFFSET% -gpio-rx=2 -gpio-tx=3 %BURN_PARAMS% -reg=%RVS_FILE% %IMAGE_PARAM%

:END
endlocal

