/*!
*******************************************************************************
**
** \file        ./applications/fastboot/src/app.c
** 
** \version     $Id$
**
** \author      
**
** \brief       Bootloader application 'fastboot'
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU MICROELECTRONICS
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2008-2010 by Fujitsu Microelectronics Europe GmbH
** (C) Copyright 2009-2010 by Fujitsu Microelectronics Limited
*******************************************************************************
*/
#define DEBUG 1

/*---------------------------------------------------------------------------*/
/* FAPI Build configurations                                                 */
/*---------------------------------------------------------------------------*/
//#define FAPI_SYS_INIT_REG_ADC           0
#define FAPI_SYS_INIT_REG_AO            0
#define FAPI_SYS_INIT_REG_BM            0
#define FAPI_SYS_INIT_REG_CLKPWR        0
#define FAPI_SYS_INIT_REG_DMA           1 /* USE DMA */
#define FAPI_SYS_INIT_REG_ETH           0
#define FAPI_SYS_INIT_REG_FPC           0
#define FAPI_SYS_INIT_REG_GPIO          1
#define FAPI_SYS_INIT_REG_I2C           0
#define FAPI_SYS_INIT_REG_ICC           0
#define FAPI_SYS_INIT_REG_IR            0
#define FAPI_SYS_INIT_REG_SFLASH        0//1
#define FAPI_SYS_INIT_REG_SSP           0
#define FAPI_SYS_INIT_REG_TSD           0
#define FAPI_SYS_INIT_REG_UART          1
#define FAPI_SYS_INIT_REG_UPI           1
#define FAPI_SYS_INIT_REG_VE            0
#define FAPI_SYS_INIT_REG_VE_HD         0
#define FAPI_SYS_INIT_REG_VIC           0
#define FAPI_SYS_INIT_REG_VO            0

#define FAPI_SYS_INIT_MMU               1
#define FAPI_SYS_INIT_GPIO              1
#define FAPI_SYS_INIT_AUDEC             0
#define FAPI_SYS_INIT_AUOUT             0
#define FAPI_SYS_INIT_BM                0
#define FAPI_SYS_INIT_CLKPWR            0
#define FAPI_SYS_INIT_DESC              0
#define FAPI_SYS_INIT_DMA               1 /* USE DMA */
#define FAPI_SYS_INIT_ETH               0
#define FAPI_SYS_INIT_FPC               0
#define FAPI_SYS_INIT_FLASH             1
#define FAPI_SYS_INIT_GPREG             1  /* Must be 1 ,0*/
#define FAPI_SYS_INIT_HDMI              0
#define FAPI_SYS_INIT_I2C               0
#define FAPI_SYS_INIT_ICC               0
#define FAPI_SYS_INIT_INTR              1
#define FAPI_SYS_INIT_IR                0
#define FAPI_SYS_INIT_SSP               0
#define FAPI_SYS_INIT_SYNC              0
#define FAPI_SYS_INIT_TIMER             1 /*0*/
#define FAPI_SYS_INIT_TSD               0
#define FAPI_SYS_INIT_TTX               0
#define FAPI_SYS_INIT_UART              1
#define FAPI_SYS_INIT_UPI               1
#define FAPI_SYS_INIT_USB               0
#define FAPI_SYS_INIT_VIDEC             0
#define FAPI_SYS_INIT_VIENC             0
#define FAPI_SYS_INIT_VISCALE           0

#define FAPI_SYS_INIT_BOOT              0         /* add */

/*---------------------------------------------------------------------------*/
/* include files                                                             */
/*---------------------------------------------------------------------------*/

#include <fapi/sys_services.h>
#include <fapi/sys_init.h>
#include <boot/cpu_control.h>
#include "fastboot.h"
#include "app.h"

/*---------------------------------------------------------------------------*/
/* internel data                                                             */
/*---------------------------------------------------------------------------*/
uint32_t flashType = FAPI_FLASH_DEVICE_PFLASH;

#if APP_PROFILE
/*---------------------------------------------------------------------------*/
/* Profile sub-routines                                                       */
/*---------------------------------------------------------------------------*/
uint32_t bsSampleNum;
APP_ProfileT bsProfData[APP_SAMPLE_NUM];
    
/*lint -save -e534 */
static void StartProfile(void)
{
    bsSampleNum=0;
    memset( bsProfData, 0, sizeof(APP_ProfileT)*APP_SAMPLE_NUM );
    
    /* Set start point time */
    SetProfile(0);
}

static void SetProfile(int32_t nPoint)
{
    if(bsSampleNum >= APP_SAMPLE_NUM){
        return;
    }
        
    bsProfData[bsSampleNum].nPoint = nPoint;
    bsProfData[bsSampleNum].nTime  = FAPI_TIMER_GetTimeStamp( FAPI_TIMER_RESOLUTION_1_MSEC );
    bsSampleNum++;
}

static void PrintProfile(void)
{
    uint32_t i;
    
    for(i=1; i< bsSampleNum; i++){
        FAPI_SYS_PRINT_MSG( " [%u] %lld msecs \n", 
                                    bsProfData[i].nPoint,
                                    bsProfData[i].nTime - bsProfData[i-1].nTime);
    }
}
/*lint -restore */
#define APP_PROFILE_FUNC(point)  {SetProfile(point);} 

#else
    #define APP_PROFILE_FUNC(point) {}

#endif /* APP_PROFILE */

/*!
*******************************************************************************
** 
** \brief  Cleanup allocated area.
** 
** \param  *pInflateArea[nNum]   pointer to cleanup area.
** \param  nNum                  area num.
** 
** \return -
******************************************************************************/
static void CleanupDMAInflateArea(void**pInflateArea, uint32_t nNum)
{
    uint32_t i;
    for(i=0;i<nNum;i++){
        if( pInflateArea[i] != NULL){
            FAPI_SYS_FREE(pInflateArea[i]);
            pInflateArea[i]=NULL;
        }
    }
}


/*!
*******************************************************************************
** 
** \brief  Check image alignment and setup DMA/PIO Image List. Regist DMA Transfer.
** 
** \param  pHandle       Fast Bootloader handle.
** \param  nImgHdrAddr   Image header address.
** \param  nImgNum       Image num included in the slot.
** \param  pImgList      DMA/PIO Image List
** \param  bCmp          Image compression flag
** \param  ppInflateArea Pointer to decompression work area
** \param  *pJumAdr      Pointer to ARM/ARC execution address data struct
** 
** \return 
** -    0                 DMA Transfered image num.
** -   -1                 If error happend.
******************************************************************************/
static int32_t CheckImages
            (FBOOT_HandleT* pHdl, uint32_t nImgHdrAddr, uint32_t nImgNum, APP_ImageList* pImgList, 
                uint32_t bCmp, void **pInflateArea, APP_JMPAdrT* pJmpAdr)
{
    uint32_t  i;
    uint32_t nImageAddr;
    BOOTER_FlashImageHeaderT sImageHeader;
    APP_TxImageT *pDImgList, *pPImgList;
    int32_t retVal = 0;
    uint32_t nNumDMA=0;
    uint32_t nNumPIO=0;
    uint32_t irq_flags = 0;
    
    /* parameter check */
    if( ((pHdl==NULL) || (nImgHdrAddr==0) || (pImgList==NULL)) 
            || ((bCmp==FBOOT_TRUE) && (pInflateArea==NULL)) ){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_MSG("CopyDmaImage. Parameter Error\n");
        /*lint -restore */
        return -1;
    }
    
    pDImgList=pImgList->pDMAList;
    pPImgList=pImgList->pPIOList;
    memset( &sImageHeader, 0, sizeof(BOOTER_FlashImageHeaderT) );
    
    /* Mask interrupts */
    irq_flags = FAPI_SYS_DISABLE_IRQ;
        
    /* Image Check loop */
    for ( i = 0; i < nImgNum; i++ )
    {
        /* read image headers from flash to #imageHeader */
        (void)FAPI_FLASH_ReadBytes( pHdl->hFlashHandle, nImgHdrAddr,
                                    sizeof(BOOTER_FlashImageHeaderT), &sImageHeader);

        #if (APP_LOAD_ARMEXE_ONLY)
        /* FAPI booter loads application(ARM + EXEC) only.
           Application should load other images. */
        if(
            ((sImageHeader.imageControl & (uint32_t)BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
                    ==BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
            && (( sImageHeader.imageControl & (uint32_t)BOOT_FLASH_IMAGE_CONTROL_TYPE_ARM)!=0)
            && (sImageHeader.targetAddr !=0)  
            && (sImageHeader.imageSize !=0)
        )
        #else /* (APP_LOAD_ARMEXE_ONLY) */
        /* fast booter loads LOAD, EXECUTE images. */
        if(
            (((sImageHeader.imageControl & (uint32_t)BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
                    ==BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
                || ((sImageHeader.imageControl & (uint32_t)BOOT_FLASH_IMAGE_CONTROL_MODE_LOAD)
                        ==BOOT_FLASH_IMAGE_CONTROL_MODE_LOAD))
            && (sImageHeader.targetAddr !=0)  
            && (sImageHeader.imageSize !=0)
        )
        #endif /* (APP_LOAD_ARMEXE_ONLY) */
        {
            /* Skip already ARM Execute Image has been loaded. */
            if(
                ((sImageHeader.imageControl & (uint32_t)BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
                        ==BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
                && (( sImageHeader.imageControl & (uint32_t)BOOT_FLASH_IMAGE_CONTROL_TYPE_ARM)!=0)
                && (pJmpAdr->nAdrARM != 0)   
            ){
                /* go to next image */
                nImgHdrAddr 
                    += ( sizeof(BOOTER_FlashImageHeaderT) 
                                + ((sImageHeader.imageSize) * sizeof(uint32_t)));
                continue;
            }

            /* Skip already ARC Execute Image has been loaded. */
            if(
                ((sImageHeader.imageControl & (uint32_t)BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
                        ==BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
                && (( sImageHeader.imageControl & (uint32_t)BOOT_FLASH_IMAGE_CONTROL_TYPE_ARC)!=0)
                && (pJmpAdr->nAdrARC != 0)   
            ){
                /* go to next image */
                nImgHdrAddr 
                    += ( sizeof(BOOTER_FlashImageHeaderT) 
                                + ((sImageHeader.imageSize) * sizeof(uint32_t)));
                continue;
            }

            /* Alignment check */
            nImageAddr = nImgHdrAddr + sizeof(BOOTER_FlashImageHeaderT);

			if (flashType == FAPI_FLASH_DEVICE_PFLASH)
			{
	            /* If DMA available case. */
	            if( ((nImageAddr & 0x3)==0) && ((sImageHeader.targetAddr & 0x3)==0)  
	                    && (((sImageHeader.imageSize * sizeof(uint32_t)) & 0x3)==0) )
	            {
	                /* Aligned. So regist DMA Transmission List */
	                pDImgList[nNumDMA].pSrcAddr  = (void*)nImageAddr;
	                pDImgList[nNumDMA].pDstAddr  = (void*)sImageHeader.targetAddr;
	                pDImgList[nNumDMA].nSize         = sImageHeader.imageSize * sizeof(uint32_t);
	                pDImgList[nNumDMA].nUncompSize  = sImageHeader.imageUncomprSize * sizeof(uint32_t);
	                
	                /* Allocate infrate area */
	                if(bCmp == FBOOT_TRUE){
	                    pInflateArea[nNumDMA] = FAPI_SYS_MALLOC_UNCACHED(pDImgList[nNumDMA].nSize);
	                    if(pInflateArea[nNumDMA] == NULL){
	                        /*lint -save -e534 */
	                        FAPI_SYS_PRINT_MSG("Fail to allocate Infrate Area for DMA\n");
	                        /*lint -restore */
	                        retVal = -1;
	                        break;
	                    }
	                    pDImgList[nNumDMA].pInflateAddr = pInflateArea[nNumDMA];
	                }   
	                else{
	                    pDImgList[nNumDMA].pInflateAddr = pDImgList[nNumDMA].pDstAddr;
	                }
	                /* Request DMA Transmission */
	                retVal = FBOOT_ExecuteBurstTransmission(pHdl, 
	                                                pDImgList[nNumDMA].pSrcAddr, 
	                                                pDImgList[nNumDMA].pInflateAddr,
	                                                pDImgList[nNumDMA].nSize);
	                if(retVal < 0){
	                    /*lint -save -e534 */
	                    FAPI_SYS_PRINT_MSG("Fail to start DMA Transfer\n");
	                    /*lint -restore */
	                    retVal = -1;
	                    break;
	                }
	                else{
	                    pDImgList[nNumDMA].nTxID = (uint32_t)retVal;
	                    nNumDMA++;
	                }
	            }
			}
	        else
            {
                /* Unaligned. So regist PIO Transmission */
                //pPImgList[nNumPIO].nTxID         = 0;
                pPImgList[nNumPIO].pSrcAddr  = (void*)nImageAddr;
                pPImgList[nNumPIO].pDstAddr  = (void*)sImageHeader.targetAddr;
                pPImgList[nNumPIO].nSize         = sImageHeader.imageSize * sizeof(uint32_t);
                pPImgList[nNumPIO].nUncompSize  = sImageHeader.imageUncomprSize * sizeof(uint32_t);

                nNumPIO++;
            }

            /* Set Execution Flag */
            if( (sImageHeader.imageControl & BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE)
                    ==BOOT_FLASH_IMAGE_CONTROL_MODE_EXECUTE){
                /* ARM */
                if( ((sImageHeader.imageControl & BOOT_FLASH_IMAGE_CONTROL_TYPE_ARM)!=0)
                        && (pJmpAdr->nAdrARM == 0) ){
                    pJmpAdr->nAdrARM = sImageHeader.jumpAddr;
                }
                
                /* ARC */
                if( ((sImageHeader.imageControl & BOOT_FLASH_IMAGE_CONTROL_TYPE_ARC)!=0)
                        && (pJmpAdr->nAdrARC == 0) ){
                    pJmpAdr->nAdrARC = sImageHeader.jumpAddr;
                }
            }
        }
        /* go to next image */
        nImgHdrAddr 
            += ( sizeof(BOOTER_FlashImageHeaderT) + ((sImageHeader.imageSize) * sizeof(uint32_t)));
        
        
    } /* End of Image Check loop */
    
    /* Enable Interrupts */
    FAPI_SYS_ENABLE_IRQ( irq_flags );
    
    /* Check error status */
    if(retVal<0){
        /* Cancel DMA Transmission */
        (void)FBOOT_CleanupBurstTransmission(pHdl);
        
        /* Free Inflate area */
        if(bCmp == FBOOT_TRUE){
            CleanupDMAInflateArea(pInflateArea, nNumDMA);
        }
        retVal = -1;
    }else{
        pImgList->nNumDMA = nNumDMA;
        pImgList->nNumPIO = nNumPIO;
        retVal = 0;
    }

    return retVal;
}

/*!
*******************************************************************************
** 
** \brief  Transfer unaligned image by PIO.
** 
** \param  pHandle       Fast Bootloader handle.
** \param  pPIOTxList    PIO Image List
** \param  nNum          Image num
** \param  bCmp          Image compression flag
** 
** \return 
** -  0                  Success.
** -  -1                 If error happend.
******************************************************************************/

static int32_t CopyUnalignedImages
        (FBOOT_HandleT* pHandle,APP_TxImageT* pPIOTxList,uint32_t nNum, uint32_t bCmp)
{
    uint32_t  i;
    uint32_t nDecompSize;
    void *pInflateAreaPIO=NULL;
    int32_t retVal = 0;
    
    if( (pHandle==NULL)|| (pPIOTxList==NULL) ){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_MSG("CopyPioImage. Parameter Error\n");
        /*lint -restore */
        return -1;
    }
    
    for(i=0; i < nNum; i++){
        if(bCmp == FBOOT_TRUE){
            pInflateAreaPIO = FAPI_SYS_MALLOC_CACHED(pPIOTxList[i].nSize);
            if(pInflateAreaPIO == NULL){
                /*lint -save -e534 */
                FAPI_SYS_PRINT_MSG("Fail to allocate Infrate Area for PIO\n");
                /*lint -restore */
                retVal = -1;
                break;
            }
        }   
        else{
            pInflateAreaPIO = pPIOTxList[i].pDstAddr;
        }   

        /* PIO Transmission */
        retVal = FBOOT_ExecutePioTransmission (pHandle, 
                                        pPIOTxList[i].pSrcAddr, 
                                        pInflateAreaPIO,
                                        pPIOTxList[i].nSize);
        if(retVal != FAPI_OK){
            /*lint -save -e534 */
            FAPI_SYS_PRINT_MSG("Fail PIO Transmission\n");
            /*lint -restore */
            FAPI_SYS_FREE(pInflateAreaPIO);
            pInflateAreaPIO = NULL;
            retVal = -1;
            break;
        }
        
        /* Decompress Image */
        if(bCmp == FBOOT_TRUE){
            nDecompSize = FBOOT_Inflate( pInflateAreaPIO,
                                pPIOTxList[i].nSize, 
                                pPIOTxList[i].pDstAddr,
                                pPIOTxList[i].nUncompSize);
            /* Free temporary area */
            FAPI_SYS_FREE(pInflateAreaPIO);
            pInflateAreaPIO = NULL;
            
            /* Check decompressed size */
            if(nDecompSize != pPIOTxList[i].nUncompSize){
                /*lint -save -e534 */
                FAPI_SYS_PRINT_MSG("(PIO)Uncompressed size is not correct\n");
                /*lint -restore */
                retVal = -1;
                break;
            }
            
        }
        
    }

    return retVal;
}


/*!
*******************************************************************************
** 
** \brief  Initialize drivers.
** 
** \param  pHandle       Fast Bootloader handle.
** 
** \return 
** -  0                  Success.
** -  -1                 If error happend.
******************************************************************************/
static int32_t InitDrivers(FBOOT_HandleT* pHandle)
{
    FAPI_FLASH_OpenParamsT  sFlashOpt;
    int32_t retVal;
    
    if(pHandle==NULL){
        return -1;
    }
    
    /* clear structures */
    memset(&sFlashOpt, 0, sizeof(sFlashOpt));
    
    
    /* open the FLASH drvier to access parallel flash */
    sFlashOpt.version = FAPI_FLASH_VERSION;
    sFlashOpt.deviceType = FAPI_FLASH_DEVICE_PFLASH;
    sFlashOpt.firstSector = 0;
    sFlashOpt.sectorCount = 0;
    sFlashOpt.fsysFlag = 0;
    sFlashOpt.forceFlush = 0;
    
    FAPI_SYS_PRINT_MSG( "FAPI_FLASH_Open: PFlash->first");
    pHandle->hFlashHandle = FAPI_FLASH_Open( &sFlashOpt, NULL );

    /* Added by Anthony Wang for support sflash as well */
    if (pHandle->hFlashHandle == 0)
    {
	    FAPI_SYS_PRINT_MSG( ":Error!\r\n");
	    sFlashOpt.deviceType = FAPI_FLASH_DEVICE_SFLASH;
	    FAPI_SYS_PRINT_MSG( "FAPI_FLASH_Open: SFlash->Second");
	    pHandle->hFlashHandle = FAPI_FLASH_Open( &sFlashOpt, NULL );
	    if (pHandle->hFlashHandle != 0)
	    {
            FAPI_SYS_PRINT_MSG("\nBoot from SFlash.\n");
            flashType = FAPI_FLASH_DEVICE_SFLASH;
	        retVal = 0;
        }
        else
        {
		    FAPI_SYS_PRINT_MSG( ":Error!\r\n");
		    retVal = -1;
		}
    }
    else
    {
        FAPI_SYS_PRINT_MSG("Boot from PFlash.\n");

	    /* Initialize DMA Driver */
	    retVal=FBOOT_InitBurstTransmission (pHandle);
	    
	    /* Check driver initialization status. */
	    if(((pHandle->hFlashHandle)==0 ) || (retVal != FAPI_OK) ){
	        /*lint -save -e534 */
	        FAPI_SYS_PRINT_MSG("Driver Initialization Failed\n");
	        /*lint -restore */
	        retVal = -1;
	        
	        /* Cleanup Driver handles*/
	        if( (pHandle->hFlashHandle)==0 ){
	            (void)FAPI_FLASH_Close(pHandle->hFlashHandle);
	        }
	    }
	    else{
	        retVal = 0;
	    }
    }

    return retVal;
}

/*!
*******************************************************************************
** 
** \brief  Fastboot application main.
** 
** \param  -
** 
** \return 
** -  0                  Success.
** -  -1                 If error happend.
******************************************************************************/
int32_t fboot_main( void )
{
    int32_t retVal = 0;
    FBOOT_HandleT*  pHandle;
    BOOTER_FlashSlotHeaderT  sSlotHeader;
    uint32_t nSlotHeaderAddr;
    void *pInflateAreaDMA[FBOOT_IMAGE_NUM_MAX]; 
    uint32_t bCompress;
    uint32_t nDecompSize;
    uint32_t nRemainDMA;
    uint32_t nImageNum;
    
    APP_TxImageT sDmaImgList[FBOOT_IMAGE_NUM_MAX];
    APP_TxImageT sPioImgList[FBOOT_IMAGE_NUM_MAX];
    APP_ImageList sImgList;
    APP_JMPAdrT sJmpAdr;
#if APP_PROFILE
    uint32_t i;
#endif /* APP_PROFILE */


#if !(APP_ENABLE_LINUX_EXT)
    CPU_BootFunctionT   pRomPollingFunction;
    BOOT_ROM_SymbolT*   pRomSymbol       = (BOOT_ROM_SymbolT*)BOOT_ROM_SYMBOLS_STRUCTURE_ADDRESS;
#endif /* !(APP_ENABLE_LINUX_EXT) */

    volatile uint32_t*  pArcControl_stat = (volatile uint32_t*)BOOT_CPU_CONTROL_ARC_STATUS;
    volatile uint32_t*  pArcControl_jump = (volatile uint32_t*)BOOT_CPU_CONTROL_ARC_JUMP;
    volatile uint32_t*  pArmControl_stat = (volatile uint32_t*)BOOT_CPU_CONTROL_ARM_STATUS;
    volatile uint32_t*  pArmControl_jump = (volatile uint32_t*)BOOT_CPU_CONTROL_ARM_JUMP;


#if APP_PROFILE
    StartProfile();
#endif /* APP_PROFILE */
    
    /* Initialize hHandle area */
    pHandle = FAPI_SYS_MALLOC_CACHED(sizeof(FBOOT_HandleT));
    if(pHandle==NULL){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_MSG( "Failed to allocate handle area\r\n" );
        /*lint -restore */
        return -1;
    }
    memset(pHandle, 0, sizeof(FBOOT_HandleT));
    memset(&sImgList, 0, sizeof(APP_ImageList));
    memset(sDmaImgList, 0, sizeof(APP_TxImageT)*FBOOT_IMAGE_NUM_MAX);
    memset(sPioImgList, 0, sizeof(APP_TxImageT)*FBOOT_IMAGE_NUM_MAX);
    memset(&sSlotHeader, 0, sizeof(sSlotHeader));
    memset(&sJmpAdr, 0, sizeof(APP_JMPAdrT));
    
    sImgList.pDMAList=sDmaImgList;
    sImgList.pPIOList=sPioImgList;

    /*lint -save -e534 */
    FAPI_SYS_PRINT_MSG( "\033[2J\033[0;0H\r\n" ); /* clear screen and position cursor */
    FAPI_SYS_PRINT_MSG( "===============================================================\r\n" );
    FAPI_SYS_PRINT_MSG( "Fast booter application\r\n" );
    FAPI_SYS_PRINT_MSG( "Target: MB86H61\r\n" );
    FAPI_SYS_PRINT_MSG( "(c) Copyright 2008-2010 by Fujitsu Microelectronics Europe\r\n" );
    FAPI_SYS_PRINT_MSG( "(c) Copyright 2009-2010 by Fujitsu Microelectronics Limited\r\n" );
    FAPI_SYS_PRINT_MSG( "===============================================================\r\n" );
    /*lint -restore */

    /* ---- Initialize drivers ------------------------- */
    retVal = InitDrivers(pHandle);
    if ( retVal != FAPI_OK ){
	    FAPI_SYS_PRINT_MSG( "Initial driver Error: 0x%x\r\n" ,retVal);
        FAPI_SYS_FREE(pHandle);
        return -1;
    }

    APP_PROFILE_FUNC(1);

    /* ---- select an valid slot for loading ------------------------- */

    nSlotHeaderAddr = (uint32_t)FBOOT_ChooseSlot(pHandle->hFlashHandle, &sSlotHeader);
    if( (sSlotHeader.signature != BOOTER_SLOTHDR_SIG) || (nSlotHeaderAddr==0)){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_MSG("No valid slot \n");
        /*lint -restore */
		if (flashType == FAPI_FLASH_DEVICE_PFLASH)
		{
        	(void)FBOOT_DeinitBurstTransmission (pHandle);
        }
        
        (void)FAPI_FLASH_Close(pHandle->hFlashHandle);
        FAPI_SYS_FREE(pHandle);
        return -1;
    }

    APP_PROFILE_FUNC(2);

    /*lint -save -e534 */
    FAPI_SYS_PRINT_DEBUG(1, "Slot at 0x%08x is selected.\r\n",nSlotHeaderAddr );
    /*lint -restore */
    
    if(sSlotHeader.imageFormat == (uint32_t)BOOTER_COMPRESS_MODE_OFF){
        bCompress = FBOOT_FALSE;
        
    }else{
        bCompress = FBOOT_TRUE;
    
    }

    APP_PROFILE_FUNC(3);

    if(sSlotHeader.imageCount <= FBOOT_IMAGE_NUM_MAX){
        nImageNum = sSlotHeader.imageCount;
    }else{
        nImageNum = FBOOT_IMAGE_NUM_MAX;
    }
    retVal= CheckImages(pHandle,
                (nSlotHeaderAddr + sizeof(BOOTER_FlashSlotHeaderT)), /* Image Header addr*/
                nImageNum, 
                &sImgList, 
                bCompress, 
                pInflateAreaDMA,
                &sJmpAdr);
    
    if( retVal < 0){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_MSG("Failed to DMA transmission\n");
        /*lint -restore */

        /* Close drivers */
		if (flashType == FAPI_FLASH_DEVICE_PFLASH)
		{
	        (void)FBOOT_DeinitBurstTransmission (pHandle);
	    }
        (void)FAPI_FLASH_Close(pHandle->hFlashHandle);
        FAPI_SYS_FREE(pHandle);
        return -1;
    }else{
        /*lint -save -e534 */
        FAPI_SYS_PRINT_DEBUG(1, "DMA transfer: %d  PIO transfer: %d \r\n"
                                    , sImgList.nNumDMA, sImgList.nNumPIO);
        /*lint -restore */
    }
    
    APP_PROFILE_FUNC(4);
    
    /* Load unaligned image by PIO */
    retVal=CopyUnalignedImages(pHandle,sPioImgList, sImgList.nNumPIO , bCompress);

    APP_PROFILE_FUNC(5);

    if(retVal != FAPI_OK){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_MSG("Failed to PIO transmission\n");
        /*lint -restore */
        /* Cancel DMA Transmission */
		if (flashType == FAPI_FLASH_DEVICE_PFLASH)
		{
	        (void)FBOOT_CleanupBurstTransmission(pHandle);
	    
	        if(bCompress == FBOOT_TRUE){
    	        CleanupDMAInflateArea(pInflateAreaDMA, sImgList.nNumDMA);
    	    }
	        /* Close drivers */
	        (void)FBOOT_DeinitBurstTransmission (pHandle);
	    }

        (void)FAPI_FLASH_Close(pHandle->hFlashHandle);
        FAPI_SYS_FREE(pHandle);
        return -1;
    }
    /*lint -save -e534 */
    if(sImgList.nNumPIO){
        FAPI_SYS_PRINT_DEBUG(1, "PIO transfer completed\r\n");
    }
    /*lint -restore */

    /* Wait DMA Transmission Complete */
    nRemainDMA = sImgList.nNumDMA;
    
    while( nRemainDMA != 0){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_DEBUG(1, "DMA transfer waiting. Remain:%d\r\n", nRemainDMA);
        /*lint -restore */

		if (flashType == FAPI_FLASH_DEVICE_PFLASH)
		{
	        retVal = FBOOT_WaitBurstTransmission (pHandle);
        	if(retVal < 0){
            	/*lint -save -e534 */
            	FAPI_SYS_PRINT_MSG("Failed in DMA transmission\n");
            	/*lint -restore */
            
            	/* Cancel DMA Transmission */
            	(void)FBOOT_CleanupBurstTransmission(pHandle);
            	if(bCompress == FBOOT_TRUE){
                	CleanupDMAInflateArea(pInflateAreaDMA, sImgList.nNumDMA);
            	}
            	break;
            }
        }
        
#if 0
        FAPI_SYS_FlushDataCache( (uint32_t)sDmaImgList[retVal].pInflateAddr,
                                    sDmaImgList[retVal].nSize);

#endif
        
        /*lint -save -e534 */
        FAPI_SYS_PRINT_DEBUG(1, "DMA Transfer Image # %d completed. \r\n", retVal);
        
        /*lint -restore */
        
        /* Decompress Image */
        if(bCompress == FBOOT_TRUE){
            APP_PROFILE_FUNC(((retVal+1)*10) + 0);
            
            nDecompSize = FBOOT_Inflate( sDmaImgList[retVal].pInflateAddr,
                                sDmaImgList[retVal].nSize, 
                                sDmaImgList[retVal].pDstAddr,
                                sDmaImgList[retVal].nUncompSize);
            /* Free temporary area */
            FAPI_SYS_FREE(pInflateAreaDMA[retVal]);
            pInflateAreaDMA[retVal]=NULL;
            
            APP_PROFILE_FUNC(((retVal+1)*10) + 1);

            /* Check decompressed size */
            if(nDecompSize != sDmaImgList[retVal].nUncompSize){
                /*lint -save -e534 */
                FAPI_SYS_PRINT_MSG("(DMA)Uncompressed size is not correct\n");
                /*lint -restore */
        
                retVal = -1;
                break;
            }
        }
        nRemainDMA--;
    }

    /* Close DMA Driver */
	if (flashType == FAPI_FLASH_DEVICE_PFLASH)
	{
	    (void)FBOOT_DeinitBurstTransmission (pHandle);
	}

    /* FLASH handle no longer needed */
    (void)FAPI_FLASH_Close(pHandle->hFlashHandle);
    pHandle->hFlashHandle=0;
    
    FAPI_SYS_FREE(pHandle);

    /* If error has happened at this time, do not execute image */
    if(retVal < 0){
        return -1;
    }

#if APP_PROFILE
    SetProfile(99);
    
    /*lint -save -e534 */
    if(sImgList.nNumDMA){
        for(i=0;i< (sImgList.nNumDMA); i++){
            FAPI_SYS_PRINT_MSG( "Transfered(DMA) %u Byte\n", sDmaImgList[i].nSize);
        }
    }
    if(sImgList.nNumPIO){
        for(i=0;i< (sImgList.nNumPIO);i++){
            FAPI_SYS_PRINT_MSG( "Transfered(PIO) %u Byte\n", sPioImgList[i].nSize);
        }
    }

    PrintProfile();
    /*lint -restore */
#endif /* APP_PROFILE */


    /* request execute the image in ARC */
    if((sJmpAdr.nAdrARC !=0) && ((sJmpAdr.nAdrARC & 0x3) ==0) ){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_DEBUG(1, "ARC jumps to 0x%08x. \r\n", sJmpAdr.nAdrARC);
        /*lint -restore */
        *pArcControl_jump = sJmpAdr.nAdrARC; 
        *pArcControl_stat = BOOT_CPU_CONTROL_REQUEST_EXECUTE;
    }
    /* request execute the image in ARM */
    if((sJmpAdr.nAdrARM !=0) && ((sJmpAdr.nAdrARM & 0x3) ==0)){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_DEBUG(1, "ARM jumps to 0x%08x. \r\n", sJmpAdr.nAdrARM);
        /*lint -restore */
        *pArmControl_jump = sJmpAdr.nAdrARM;
        *pArmControl_stat = BOOT_CPU_CONTROL_REQUEST_EXECUTE;
        
        /* set the CPU to polling mode and perform the cpu execution request */
        
        #if !(APP_ENABLE_LINUX_EXT)
        pRomPollingFunction = (CPU_BootFunctionT)(pRomSymbol->arm1176_control_mode_handler);
        pRomPollingFunction();
        #else
        ARM1176_set_and_go(0, APP_LINUX_MACHINE_ID , 0, sJmpAdr.nAdrARM);
        #endif /* !(APP_ENABLE_LINUX_EXT) */
    }
    
    /*lint -save -e534 */
    FAPI_SYS_PRINT_DEBUG(1, "Fast Boot loader finished. (No ARM execution image.). \r\n");
    /*lint -restore */

    return 0;
}



int32_t main( void )
{
    int32_t retVal = 0;
    uint32_t ret;
    uint32_t      bsStack[APP_THREAD_STACK_SIZE];
    RTOS_ThreadT  bsMainThread;

    // perform some pre-init setup
    FAPI_GPIO_SetPinFunction( 2, FAPI_GPIO_IN_UART0_DATA );
    FAPI_GPIO_SetPinFunction( 3, FAPI_GPIO_OUT_UART0_DATA );
    FAPI_UART_SetBaudrate( 0, FAPI_UART_BAUDRATE_115200 );
    FAPI_UART_SetDataBits( 0, FAPI_UART_DATABITS_8 );
    FAPI_UART_SetStopBits( 0, FAPI_UART_STOPBITS_1 );
    FAPI_UART_SetParity( 0, FAPI_UART_PARITY_NONE );
    FAPI_UART_SetFlowControl( 0, FAPI_UART_FLOWCTRL_NONE );
    FAPI_UART_SetStdioBlockIndex( 0 );

    /* Initialise RTOS and FAPI */
    RTOS_InitServices();
    retVal = FAPI_SYS_Init();
    if ( retVal != FAPI_OK ){
        return( retVal );
    }

    ret = RTOS_InitKernel();
    if( ret != 1 ){
        /*lint -save -e534 */
        FAPI_SYS_PRINT_MSG( "RTOS_InitKernel failed.\n");
        /*lint -restore */
        return -1;
    }

    bsMainThread = RTOS_CreateThread(
                                bsStack, 
                                APP_THREAD_STACK_SIZE,  
                                APP_THREAD_PRIORITY, 
                                (RTOS_ThreadFunctionT) fboot_main, 
                                NULL, 
                                NULL, 
                                APP_THREAD_NAME);
    if( !bsMainThread )
    {
        /*lint -save -e534 */
        FAPI_SYS_PRINT_MSG( "Creating start thread  failed.\n");
        /*lint -restore */
        return -1;
    }

    RTOS_StartMultitasking();
    
    return( 0 );
}
