/*!
*******************************************************************************
**
** \file        ./applications/fastboot/src/app.h
** 
** \version     $Id$
**
** \author      
**
** \brief       Bootloader application 'fastboot'
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU MICROELECTRONICS
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2008-2010 by Fujitsu Microelectronics Europe GmbH
** (C) Copyright 2009-2010 by Fujitsu Microelectronics Limited
*******************************************************************************
*/

#ifndef __APP_H__
#define __APP_H__


/*---------------------------------------------------------------------------*/
/* Firm Load Option                                                          */
/*                                                                           */
/* RTOS booter only loads ARM execution image. This is compatibility option  */
/*   with RTOS booter.                                                       */
/*---------------------------------------------------------------------------*/
#define  APP_LOAD_ARMEXE_ONLY       0


/*---------------------------------------------------------------------------*/
/* Linux Specific Option                                                     */
/*                                                                           */
/* For linux execution,  it's necessarry to set MachineID on R1 register.    */
/* If this option specified, R1 register will be set before executing ARM    */
/*   execution image.                                                        */
/*---------------------------------------------------------------------------*/
#define  APP_ENABLE_LINUX_EXT       1

/* MachineID for MB86H60 */
#define APP_LINUX_MACHINE_ID      0x534


/*---------------------------------------------------------------------------*/
/* Application thread settings                                               */
/*---------------------------------------------------------------------------*/
/* no need to change */
#define APP_THREAD_NAME "fboot" 
#define APP_THREAD_STACK_SIZE       (0x1000)
#define APP_THREAD_PRIORITY         (RTOS_PRIORITY_MEDIUM)


/*---------------------------------------------------------------------------*/
/* Profiling option                                                          */
/*---------------------------------------------------------------------------*/

/* If you want to profile boot time, set this value 1. */
#define APP_PROFILE      0

/* Time sampling point num. if you add sampling point, adjust this value.*/
#define APP_SAMPLE_NUM   20

/*!
****************************************************************************
** \brief Image data structure which is application uses.
**
** This data structure is used on fast boot application to manage image.
****************************************************************************
*/
typedef struct
{
    uint32_t    nTxID;
    void *      pSrcAddr;
    void *      pDstAddr;
    uint32_t    nSize;
    uint32_t    nCompressed; //for extension.
    uint32_t    nUncompSize; 
    void *      pInflateAddr;
}
APP_TxImageT;

/*!
****************************************************************************
** \brief Image management data structure which is application uses.
**
** This data structure is used on fast boot application to manage image.
****************************************************************************
*/

typedef struct
{
    uint32_t nNumDMA;
    uint32_t nNumPIO;
    APP_TxImageT *pDMAList;
    APP_TxImageT *pPIOList;
}
APP_ImageList;


/*!
****************************************************************************
** \brief ARM/ARC execute address data structure which is application uses.
**
** This data structure is used on fast boot application to manage image.
****************************************************************************
*/
typedef struct
{
    uint32_t nAdrARM;
    uint32_t nAdrARC;
}
APP_JMPAdrT;




/*!
****************************************************************************
** \brief Boot time profiling data structure.
**
** This data structure is used if APP_PROFILE option is enabled.
****************************************************************************
*/
typedef struct
{
    int32_t nPoint;
    uint64_t nTime;
}
APP_ProfileT;
    
/*!
****************************************************************************
** \brief CPU_BootFunction type.
**
** This type will be used on executing EXEC image.
****************************************************************************
*/
typedef void(*CPU_BootFunctionT)(void);


/*---------------------------------------------------------------------------*/
/* Functions of fastboot application                                         */
/*---------------------------------------------------------------------------*/
static int32_t CheckImages(FBOOT_HandleT* , uint32_t, uint32_t, APP_ImageList*, uint32_t, void **, APP_JMPAdrT*);
static int32_t CopyUnalignedImages(FBOOT_HandleT*,APP_TxImageT*,uint32_t, uint32_t);
static int32_t InitDrivers(FBOOT_HandleT* pHandle);
static void CleanupDMAInflateArea(void**, uint32_t);
void * ConvertToAbsFlashAddr(FAPI_SYS_HandleT, uint32_t);
int32_t fboot_main(void);

#if (APP_PROFILE)
static void StartProfile(void);
static void SetProfile(int32_t);
static void PrintProfile(void);
#endif /* APP_PROFILE */

#if (APP_ENABLE_LINUX_EXT)
extern void ARM1176_set_and_go(uint32_t, uint32_t, uint32_t, uint32_t);
#endif /* APP_ENABLE_LINUX_EXT */

#endif /* __APP_H__ */
