/*!
*******************************************************************************
**
** \file        ./applications/fastboot/src/fastboot_zlib.c
** 
** \version     $Id$
**
** \author      
**
** \brief       decompress code using zlib
**
** \attention   THIS SAMPLE CODE IS PROVIDED AS IS. FUJITSU MICROELECTRONICS
**              ACCEPTS NO RESPONSIBILITY OR LIABILITY FOR ANY ERRORS OR
**              OMMISSIONS.
**
** (C) Copyright 2008-2010 by Fujitsu Microelectronics Europe GmbH
** (C) Copyright 2009-2010 by Fujitsu Microelectronics Limited
*******************************************************************************
*/
#include <fapi/sys_services.h>
#include <compression/zlib.h>
#include "fastboot_zlib.h"
/*!
*******************************************************************************
** 
** \brief  Decompress data. 
** 
** \param  pComp          Compressed Data address
** \param  sCompLen       Compressed Data Size
** \param  pUnComp        Address for uncompressed area
** \param  sUnCompLen     Uncompressed data size
** 
** \return Uncompressed size
******************************************************************************/
uint32_t FBOOT_Inflate( uint8_t* pComp, uint32_t sCompLen, 
                          uint8_t* pUnComp, uint32_t sUnCompLen )
{
    int err;
    z_stream d_stream;
    
    d_stream.zalloc    = (alloc_func)0;
    d_stream.zfree     = (free_func)0;
    d_stream.opaque    = (voidpf)0;
    d_stream.next_in   = pComp;
    d_stream.avail_in  = sCompLen;
    d_stream.next_out  = pUnComp;
    d_stream.avail_out = sUnCompLen;
    
    err = inflateInit(&d_stream);
    if ( err != Z_OK ){ 
        return 0;
    }
    
    while ( (d_stream.total_out < sUnCompLen) &&
            (d_stream.total_in  < sCompLen) )
    {
        err = inflate(&d_stream, Z_NO_FLUSH);
        if (err == Z_STREAM_END){ 
            break;
        }
        if ( err != Z_OK ){
             return 0;
        }
    }
    
    err = inflateEnd(&d_stream);
    if ( err != Z_OK ) {
        return 0;
    }
    return d_stream.total_out;
}

